/******************************************************************

"Copyright  1994 by the Massachusetts  Institute  of  Technology.  All
rights reserved.

Developed by John Y. A. Wang at the  Media Laboratory, MIT, Cambridge,
Massachusetts, with  support  from Secom Ltd., Goldstar Ltd.,  and the
Television of Tomorrow Program.

For use by authorized users only.

This distribution is approved  by Nicholas Negroponte, Director of the
Media Laboratory, MIT.

Permission to use, copy, or modify this software and its documentation
for educational and research purposes only and without  fee is  hereby
granted, provided that this copyright notice and the original authors'
names  appear  on  all  copies  and   supporting   documentation.   If
individual  files  are  separated  from  this  distribution  directory
structure, this copyright notice must be included.  For any other uses
of  this software, in  original or  modified form, including  but  not
limited to distribution in whole or in part, specific prior permission
must be  obtained  from  MIT.   These  programs  shall  not  be  used,
rewritten,  or  adapted  as  the  basis  of a  commercial software  or
hardware  product without  first  obtaining  appropriate licenses from
MIT.  MIT.  makes no  representations about  the  suitability of  this
software for any purpose.  It is  provided "as is" without  express or
implied warranty.

******************************************************************/


/***************************************************************
*** File: affine-example.c
*** Author: John Y. A. Wang
*** Date: 1993.
*** Copyright 1993, Massachusetts Institute of Technology, All Rights Reserved.
***
*** Description: test affine source library
***   program does the following:
***      1) makes a test image consisting of sum of x-sine grating and
***         y-sine grating producing a 2-D sine grating.
***      2) prompts user for affine motion paramters. A set of example
***         parameters are show below.
***      3) calls warping routine to warp the test image.
***      4) performs affine motion estimation on the test image and
***         the warped image.
***
***   also checks to make sure other warp code is in library
***************************************************************/

static  char *copyright="\
\n\n***********************************************\n\
Author: John Y. A. Wang.\n\
The MIT Media Lab.\n\
Copyright (c) 1993, All Rights Reserved.\n\
***********************************************\n\n";

/***************************

  compile and link this program a follows:

  cc affine-example.c -o affine-example libdynamo.a -lm

**************************/

#ifdef NEVER
/**************************************************/
/**** a set of test values for motion estimation **/

affine motion:
  Vy = ay + by*x + cy*y;
  Vx = ax + bx*x + cx*y;

origin of image coordinates is at the center of the image.
  cx = xdim/2;   xcenter
  cy = ydim/2;   ycenter
  
 motion:        ay   by   cy    ax   bx   cx 
-------------------------------------------
 no motion:    0.0  0.0  0.0   0.0  0.0  0.0
 translation: -2.7  0.0  0.0   1.5  0.0  0.0
 rotation:     0.0  0.1  0.0   0.0  0.0 -0.1
 zoom in:      0.0  0.0  0.1   0.0  0.1  0.0

WARNING:

If the parameters by, cy, bx, and cx are large, the image will
undergo large motions at the image boundaries. When these motions
are large compared to the spatial frequency of the image, aliasing
will occur and motion estimation will not produce accurate results.

/**************************************************/
#endif

#include "motion.h"

/* global affine estimation code */
void internal_coarse_fine_affine_estimator(/*im1, im2, ap, init_ap,
					     xdim, ydim,level,method,
				      xstart,ystart,xend,yend,ttemp,tempsize,
				      f_array, mask, masktemp,
				      limit */
				      );
/* affine warper given affine motion parameters */
void bicubic_affine_warp(/*im, res, ax, bx, cx, ay, by, cy, ydim, xdim*/);

/* image warper given arbitrary motion fields */
void bicubic_warp(/*im, res, wy, wx, ydim, xdim*/);

main()
{
  /* test affine estimator on 100x100 image*/
  {
    float *im1, *im2;
    float ap[6];
    static float init_ap[6]={0.0, 0.0, 0.0, 0.0, 0.0, 0.0};
    int xdim,ydim;
    float *ttemp=NULL;
    int tempsize=0;
    int xstart,ystart,xend,yend;
    int level;
    static float method[3]={1.0, 1.0, 1.0};    
    static float f_array[6]={1.0, 1.0, 1.0, 1.0, 1.0, 1.0} ;   /* factor */
    float *mask=NULL, *masktemp=NULL;
    int limit=1;
    
    float  ay, by, cy, ax, bx, cx;

    xdim=100; ydim=100; level=2; xstart=0; ystart=0; xend=99; yend=99;

    printf("\n");
    printf("affine-example: this program creates a synthetic sine grating\n");
    printf("   to test the affine motion estimation rountines. The user is\n");
    printf("   is asked to enter a set of 6 affine parameters. The program\n");
    printf("   transforms the sine grating according to these parameters \n");
    printf("   to create a new \"warped\" image. Affine parameters are then\n");
    printf("   estimated by the affine motion estimation routines. The \n");
    printf("   estimated parameters are displayed.\n\n");
    printf("   For more information on affine transformation and affine\n");
    printf("   parameters, see affine-example.c and the DYNAMO User\n");
    printf("   Documentation. When the parameters results in a large\n");
    printf("   transformation, the estimator will not be able to correctly\n");
    printf("   estimate the parameters.");
    printf("\n\n");

    /* allocate input image memory */
    if((im1=(float *)malloc(sizeof(float)*xdim*ydim))==NULL) {
      fprintf(stderr,"cannot allocate memory for input.\n"); return;}
    if((im2=(float *)malloc(sizeof(float)*xdim*ydim))==NULL) {
      fprintf(stderr,"cannot allocate memory for input.\n"); return;}

    
    /* allocate working memory for affine estimation. requires 5x
     * of input image size. If memory not allocated memory will be
     * allocated internally by affine routine.
     */
    tempsize=xdim*ydim*5;
    if((ttemp=(float *)malloc(sizeof(float)*tempsize))==NULL) {
      fprintf(stderr,"cannot allocate memory for ttemp.\n"); return;}
    
    make_test_image(im1,xdim,ydim);  /* generate a test image */

    while(1) {
      /* prompt user for input: see examples above */
      fprintf(stdout,"\nEnter ay, by, cy, ax, bx, cx >> ");
      fscanf(stdin,"%f%f%f%f%f%f",&ay, &by, &cy, &ax, &bx, &cx);

      /* clear im2 then warp im1 to im2 */
      {	int i,imsize=xdim*ydim;	for(i=0;i<imsize;im2[i]=0.0,i++); } 
      bicubic_affine_warp(im1, im2, ax, bx, cx, ay, by, cy, ydim, xdim);

      /* display images. Use whatever image displayer you have */
      display_float_image(im1,xdim,ydim,"im1");
      display_float_image(im2,xdim,ydim,"im2");

      /* perform the coarse-to-fine affine estimation */
      internal_coarse_fine_affine_estimator(im1, im2, ap, init_ap,
					    xdim, ydim,level,method,
					    xstart,ystart,xend,yend,
					    ttemp,tempsize,
					    f_array, mask, masktemp,
					    limit
					    );
      /* show estimation results */
      fprintf(stdout,"Estimation result: %6.2f %7.4f %7.4f  %6.2f %7.4f %7.4f\n",
	      ap[0],ap[1],ap[2],ap[3],ap[4],ap[5]);
    }
  }

  {
    float *im, *res;    
    float *wy, *wx;
    int xdim, ydim;

    bicubic_warp(im, res, wy, wx, ydim, xdim);
  }

  {
    float *im, *res;     /*  *wx, *wy; */
    float ax, bx, cx, ay, by, cy;
    int xdim, ydim;

    bicubic_affine_warp(im, res, ax, bx, cx, ay, by, cy, ydim, xdim);
  }
}


#define PI 3.1415
double sin();

/* test image consists of sin(x)+sin(y), ie, a 2-d grating */
make_test_image(im,xdim,ydim)
  float *im;
  int xdim,ydim;
{
  int x,y,i;

  for(y=0,i=0;y<ydim;y++) {
    for(x=0;x<xdim;x++,i++) {
      im[i]=(float)sin((double)((PI/5)*x)); 
      im[i]+=(float)sin((double)((PI/5)*y)); 
    }
  }
  rescale_float(im,xdim*ydim,(float)-2.0,(float)4.0);
}

 

/********************************************************/
/***************** Extra stuff **************************/

/* rescale image to go from 0-255 for easy display */
rescale_float(im,imsize,pedestal,scale)
  float *im,pedestal,scale;
  int imsize;
{
  int i;
  float min,max;
  
/*  fprintf(stderr,"pedestal=%f scale=%f\n",pedestal,scale); */
  minmax(im,&min,&max,imsize);
/*  fprintf(stderr,"min=%f max=%f\n",min,max); */
  scale= 255.0/scale;
  for(i=0;i<imsize;i++) {
    im[i]=(im[i]-pedestal)*scale;
    if(im[i]<0) im[i]=0.0;
    else if(im[i]>255.0) im[i]=255.0;
  }
}

/*************************************************************************/
/* converts a float to a byte assumes pedestal=0 scale=1 */
void convert_float_to_byte(image,array,imsize)
  float *image;
  unsigned char *array;
  int imsize;
{
  int i;
  float im;

  for(i=0;i<imsize;i++) {
/*    array[i] = (unsigned char)(image[i]); */
    im=image[i];
    if(im>255)
      array[i]=(unsigned char)255;
    else if(im<0)
      array[i]=(unsigned char)0;
    else
      array[i]=(unsigned char)((int)im);
  }
}

/* find min and max */
minmax(im,minimum,maximum,imsize)
  float *im,*minimum,*maximum;
  int imsize;
{
  int i;
  float min,max;
  for(i=0,min=im[0],max=im[0];i<imsize;i++) {
    if(im[i]<min) min=im[i]; else if(im[i]>max) max=im[i];
  }
  *minimum=min; *maximum=max;
}

/* displays an image float array for debugging */
int display_float_image(image,xdim,ydim,title)
  float *image;
  int xdim,ydim;
  char *title;
{

#ifdef NEVER

  FILE *pid;
  float zoom=1;
  float gamma=1.0;
  unsigned char *array;
  int imsize=xdim*ydim;
  if((array=(unsigned char *) malloc((unsigned)imsize))==NULL) {
    printf("display_float_image; cannot display image %d\n");
    return(-1); }
  pid=OpenGrayScreen(ydim,xdim,title,gamma,zoom,1);
  convert_float_to_byte(image,array,imsize);
  DumpBufToScreen(array,ydim,xdim,pid,imsize);
  Close_Screen(pid);
  free((char *)array);

#endif

  return(0);
}



