/******************************************************************

"Copyright  1994 by the Massachusetts  Institute  of  Technology.  All
rights reserved.

Developed by John Y. A. Wang at the  Media Laboratory, MIT, Cambridge,
Massachusetts, with  support  from Secom Ltd., Goldstar Ltd.,  and the
Television of Tomorrow Program.

For use by authorized users only.

This distribution is approved  by Nicholas Negroponte, Director of the
Media Laboratory, MIT.

Permission to use, copy, or modify this software and its documentation
for educational and research purposes only and without  fee is  hereby
granted, provided that this copyright notice and the original authors'
names  appear  on  all  copies  and   supporting   documentation.   If
individual  files  are  separated  from  this  distribution  directory
structure, this copyright notice must be included.  For any other uses
of  this software, in  original or  modified form, including  but  not
limited to distribution in whole or in part, specific prior permission
must be  obtained  from  MIT.   These  programs  shall  not  be  used,
rewritten,  or  adapted  as  the  basis  of a  commercial software  or
hardware  product without  first  obtaining  appropriate licenses from
MIT.  MIT.  makes no  representations about  the  suitability of  this
software for any purpose.  It is  provided "as is" without  express or
implied warranty.

******************************************************************/


/***************************************************************
*** File: affine-motion-cf.c
*** Author: John Y. A. Wang
*** Date: November 18, 1992.
*** Copyright 1992, Massachusetts Institute of Technology, All Rights Reserved.
***
*** Description: coarse to fine affine motion estimation
***************************************************************/

static  char *copyright="\
\n\n***********************************************\n\
File: affine-mot-cf.c\n\
Author: John Y. A. Wang.\n\
Copyright 1992 by The MIT Media Lab. All Rights Reserved.\n\
***********************************************\n\n";

#include "motion.h"
#include "pyramid.h"

/*
  cc -c -O4 -fsingle affine-motion-cf.c
*/

#ifdef lint
/*
  lint affine-motion-cf.c
*/  
#define register
#endif

static int not_zero(/*params*/);
static void expand_affine_params(/*p1,pres*/);

/* init_ap:  the initial guess  can use zeros
 * level:    level of the pyramid to begin analysis. use 2.
 * method:   no. of iterations at each level of pyr.
             Example:
	     method[3]={1,2,3};
               iterate 3 time on level 2, twice on lev. 1, once on level 0.
	     method[3]={3,1,1};
               iterate once on lev. 2, once on lev. 1, 3 times on level 0.
 * f_array:  mask the parameters when updating estimates. use {1,1,1,1,1,1}
             Example {1,0,0,1,0 0} if you just want translation.
	     
 *masktemp: temporary array size = size of mask array;
 */


extern void internal_coarse_fine_affine_estimator
  (im1, im2, ap, init_ap, xdim, ydim,level,method,
   xstart,ystart,xend,yend,ttemp,tempsize,
   f_array, mask, masktemp,
   limit
   )
  float *im1, *im2;         /* input images */
  float *ap,*init_ap;       /* parameters. array of 6 floats */
  int xdim,ydim;            /* image dimensions */
  float *ttemp;             /* temp memory for calculations */
  int tempsize;             /* size of input temp memory */
  int xstart,ystart,xend,yend;   /* analysis region */
  int level;                /* pyramid level to begin analysis */
  float *method;            /* number of iterations at each level of pyramid */
  float *f_array;           /* parameter update factor */
  float *mask, *masktemp;   /* use for analysis within the masked region */
/*  float *tempdbg1,*tempdbg2; */
  int limit;                /* limit warping only within window. For speed */

{
  Pyramid pyr1;        /* pyr of 1st image */
  Pyramid pyr2;        /* pyr of 2nd image */
  Pyramid pyrwarp;     /* tmp space for warped images */
  Pyramid pyrmask;     /* tmp space for mask images */
  int xsize, ysize,i,j;
  int im_size = xdim*ydim;
  float *temp;
  float p1[6],p2[6];
  int temp_image = 5;   /* requires 5 temporary images */
  int allocate = 0;
  int factor,xs,ys,xe,ye;
  static float g_filt[5]={0.0625, 0.25, 0.375, 0.25, 0.0625}; 

/*
  fprintf(stderr,"internal_coarse_fine_affine_estimator:\n");
*/
  
  if(tempsize < im_size*temp_image) {
    allocate = 1;
    fprintf(stderr,"\naffine-mot-cf: allocate temporary image memory\n");
    tempsize=im_size*temp_image;
    if((temp = (float *) (malloc(tempsize*sizeof(float)))) == NULL) {
      fprintf(stderr,"\nmalloc: cannot allocate memory for temporary image\n");
      return;
    }
  }
  else
    temp = ttemp;

  /* use last temp image for the pyramid level>0. This is safe because temp
     space used by the flow routine will use only (5* 1/4 * im_size) for
     level>0, since images on level 1 are 1/4 size.
  */

/*   fprintf(stderr,"\nmaking pyramids.... ");  */
  make_pyramid(im1,level,xdim,ydim,&pyr1,(temp+4*im_size),temp,1,g_filt,5);
  make_pyramid(im2,level,xdim,ydim,&pyr2,(temp+(int)(4.5*(float)im_size)),
	       temp,1,g_filt,5);

  if ((mask != NULL) && (masktemp != NULL))
    make_pyramid(mask,level,xdim,ydim,&pyrmask,masktemp,temp,1,g_filt,5); 
  make_pyramid((temp+4*im_size),level,xdim,ydim,&pyrwarp,(temp+3*im_size),
	       temp,0,g_filt,5); 
  
/*  fprintf(stderr,"\n cf:  (%d,%d,%d,%d) :",xstart,xend,ystart,yend);  */
  
  for(i=0,factor=1;i<level;i++)
    factor*=2;
 
/*   fprintf(stderr,"\nfactor=%d\n",factor); */

  p1[0]=init_ap[0]/(float)factor; p1[1]=init_ap[1]; p1[2]=init_ap[2];
  p1[3]=init_ap[3]/(float)factor; p1[4]=init_ap[4]; p1[5]=init_ap[5];

  for(i=level; i>=0; i--) {
    xsize=pyr1.level[i].xdim;
    ysize=pyr1.level[i].ydim;
    xs = (int)(xstart/factor);  xe = (int)(xend/factor);
    ys = (int)(ystart/factor);  ye = (int)(yend/factor);
/*
    fprintf(stderr,"\n level %d (%d,%d) (%d,%d,%d,%d) :",
            i,xsize,ysize,xstart,xend,ystart,yend); 
*/
    for(j=0;j<method[i];j++) {
      if(not_zero(p1) == 1) {
	fill_im(pyrwarp.level[i].array,0.0,xsize*ysize);
	if (limit==1) {
	  bicubic_affine_warp_limit(pyr1.level[i].array,
				    pyrwarp.level[i].array,
				    p1[3],p1[4],p1[5],p1[0],p1[1],p1[2],
				    ysize, xsize,
				    (xs-5),(ys-5),(xe+5),(ye+5));
	}
	else {
	  bicubic_affine_warp(pyr1.level[i].array, pyrwarp.level[i].array,
			      p1[3],p1[4],p1[5],p1[0],p1[1],p1[2],
			      ysize, xsize);
	}
      }
      else {
	copy_im(pyr1.level[i].array, pyrwarp.level[i].array,xsize*ysize);
      }

      internal_affine_estimator_global(pyrwarp.level[i].array,
				       pyr2.level[i].array, p2,
				       xsize,ysize,xs,ys,xe,ye,
				       temp, tempsize,
				       (mask!=NULL? pyrmask.level[i].array : 0));
      /* update the affine params */
      internal_update_affine_params_from(p1, p2, p1, f_array);
    }
    if(i>0) { expand_affine_params(p1,p1); }
    factor /= 2;
  }

  /* copy to result array */
  for(i=0;i<6;i++) {ap[i]=p1[i];}

  if (allocate) {free(temp);}
}

static int not_zero(params)
  float *params;
{
  int i;
  for(i=0;i<6;i++) {
    if(params[i] != 0.0)
      return(1);
  }
  return(0);
}

static void expand_affine_params(p1,pres)
  float *p1,*pres;
{
  pres[0] = 2*p1[0];  pres[1] = p1[1];  pres[2] = p1[2];
  pres[3] = 2*p1[3];  pres[4] = p1[4];  pres[5] = p1[5];
}

