/******************************************************************

"Copyright  1994 by the Massachusetts  Institute  of  Technology.  All
rights reserved.

Developed by John Y. A. Wang at the  Media Laboratory, MIT, Cambridge,
Massachusetts, with  support  from Secom Ltd., Goldstar Ltd.,  and the
Television of Tomorrow Program.

For use by authorized users only.

This distribution is approved  by Nicholas Negroponte, Director of the
Media Laboratory, MIT.

Permission to use, copy, or modify this software and its documentation
for educational and research purposes only and without  fee is  hereby
granted, provided that this copyright notice and the original authors'
names  appear  on  all  copies  and   supporting   documentation.   If
individual  files  are  separated  from  this  distribution  directory
structure, this copyright notice must be included.  For any other uses
of  this software, in  original or  modified form, including  but  not
limited to distribution in whole or in part, specific prior permission
must be  obtained  from  MIT.   These  programs  shall  not  be  used,
rewritten,  or  adapted  as  the  basis  of a  commercial software  or
hardware  product without  first  obtaining  appropriate licenses from
MIT.  MIT.  makes no  representations about  the  suitability of  this
software for any purpose.  It is  provided "as is" without  express or
implied warranty.

******************************************************************/


/******************************************************************************
  Author: John Y. A. Wang
  Date  : Sept. 30 1991
  Vision and Modelling Group: MIT Media Lab
  Copyright 1991, Masschusetts Institute of Technology,
  All Rights Reserved

  Modified 6/28/94: optimize summation within analysis window. jyawang

****************************************************************************/  

static  char *copyright="\
\n\n***********************************************\n\
File: affine-motion.c\n\
Author: John Y. A. Wang.\n\
The MIT Media Lab.\n\
Copyright 1991, All Rights Reserved.\n\
***********************************************\n\n";

#include "motion.h"

/*
  cc -c -O4 -fsingle  affine-motion.c
*/


/*
  needs internal_pinverse() code.
  ~jyawang/obvius/stuff-2.2/inverse/....
*/

/*-- Finds the pseudo-inverse of mat and puts it into imat --*/
/* -- n is number of rows, m is number of columns in mat -- */

#ifdef lint
#define register
#endif

static float sum_array
  (/*_ float *array, int xdim, int ydim,
     int xstart, int ystart, int xend, int yend, float *mask _*/);
static float sum_mul_xramp
  (/*_ float *array, int xdim, int ydim,
     int xstart, int ystart, int xend, int yend, float *mask _*/);
static float sum_mul_x2ramp
  (/*_ float *array, int xdim, int ydim,
     int xstart, int ystart, int xend, int yend, float *mask _*/);
static float sum_mul_yramp
  (/*_ float *array, int xdim, int ydim,
     int xstart, int ystart, int xend, int yend, float *mask _*/);
static float sum_mul_y2ramp
  (/*_ float *array, int xdim, int ydim,
     int xstart, int ystart, int xend, int yend, float *mask _*/);
static float sum_mul_xyramp
  (/*_ float *array, int xdim, int ydim,
     int xstart, int ystart, int xend, int yend, float *mask _*/);

/******************************************************************************
  determins the 6 affine parameters describing motion in the two images
  im1 and im2.
  im1,im2: float arry of the images
  ap: an array of 6 floats ; [ay by cy ax bx cx]
  xstart,ystart: upper left hand corner of region to do ananlysis
  xend, yend:    lower rigth hand corner of region to do ananlysis
  ttemp, tempsize: temporary buffer for intermidiate calculations. Tempsize
  is the temp buffer size and should be greater than 5*xdim*ydim*sizeof(float).
  if tempsize is too small new memory is allocated and freed when completed.

  Algorithm is modified from that described  by James Bergen using a gradient
  least squares estimation.
  Reference: J. Bergen and P. Burt, "Computing two  motions from three frames",
             In Proc. 3rd International Conference on Computer Vision", 
	     pages 27-32, Osaka, Japan, 1990.

  Affine motion is described as follows:
    Vx = ax + bx*x + cx*y
    Vy = ay + by*x + cy*y

  Image model is as follows:
    I1(x,y) = I2(x-Vx, y-Vy)

  Linear least square:

    E^2 = I2(x,y)-I1(x,y) - Vx*Ix - Vy*Iy

  Solving system of equations (normal equations):
    phi(x,y)^T=[Ix(x,y)^2  x*Ix(x,y)^2   y*Ix(x,y)^2
                Iy(x,y)^2  x*Iy(x,y)^2   y*Iy(x,y)^2]
    theta_hat^T=[ax bx cx ay by cy]
    It(x,y) = phi^T . theta_hat + E^2
    theta_hat = {sum(phi . phi^T)}^-1 . phi . It(x,y)

    Summation window taken over a weighted gaussian window implemeted
    with a separable convolution algorithm.

*******************************************************************************/  

extern void internal_affine_estimator_global
  (im1, im2, ap, xdim, ydim,
   xstart, ystart, xend, yend, ttemp, tempsize, mask)
  float *im1, *im2;
  float *ap;
  register int xdim,ydim;
  float *ttemp;
  int tempsize;
  int xstart,ystart,xend,yend;
  float *mask;
{
  register int im_size = ydim*xdim;

  
  register float m[36], im[36];  /* matrices */
  register float v[6];           /* output vector */

  static float d_filt[2]  = {-1.0, 1.0};  /* deriv filter */
  static float s_filt[2]  = {0.5,  0.5};  /* sum filt */
  static float s2_filt[2] = {1.0,  1.0};  /* twice sum filt */

  /* temp image */
  register float *temp, *temp1;
  register float *ix,*iy,*it;
  int temp_image = 5;
  int allocate = 0;

/*  
  fprintf(stderr,"\naff_mot_dbg: (%d,%d)(%d %d %d %d)\n",
	  xdim,ydim,ystart,xstart,yend,xend);
*/
  
  /*************************************************/  
  if(tempsize < im_size*temp_image) {
    allocate = 1;
    fprintf(stderr,"\naffine-motion: allocate temporary image memory\n");
    tempsize=im_size*temp_image;
    if((temp = (float *)(malloc(tempsize*sizeof(float)))) == NULL) {
      fprintf(stderr,"\nmalloc: cannot allocate memory for temporary image\n");
      return;
    }
  }
  else
    temp = ttemp;

  temp1 = temp + im_size;
  ix    = temp + 2*im_size;
  iy    = temp + 3*im_size;
  it    = temp + 4*im_size;

  /********************************************************************/
  add_im(im1,im2,temp1,im_size);

#if DBG_DISPLAY  /* display the analysis region */
  {
    int x,y,ypos,pos; 
    
    xend = ((xend<xdim)? xend : xdim);  yend = ((yend<ydim)? yend : ydim);
    
    zero_im(temp,xdim*ydim);
    if(mask==NULL) {
      for(y=ystart,ypos=y*xdim; y<yend; y++,ypos+=xdim)
	for(x=xstart,pos=ypos+x; x<xend; x++,pos++)
	  temp[pos]=temp1[pos]/2.0; }
    else {
      for(y=ystart,ypos=y*xdim; y<yend; y++,ypos+=xdim)
	for(x=xstart,pos=ypos+x; x<xend; x++,pos++)
	  temp[pos]=temp1[pos]*mask[pos]/2.0; }
    display_float_image(temp,xdim,ydim,"temp");
  }
#endif

  /* x-derivative */
  apply_sep_filter_reflect(temp1,ix,temp,xdim,ydim,d_filt,s_filt,2,2,1,1);
  /* y-derivative */
  apply_sep_filter_reflect(temp1,iy,temp,xdim,ydim,s_filt,d_filt,2,2,1,1);

  sub_im(im1,im2,it,im_size);
  apply_sep_filter_reflect(it,it,temp,xdim,ydim,s2_filt,s_filt,2,2,1,1);

  /********** temp imges *********/
  mul_im (ix, ix, temp, im_size);
    
  m[7]  = sum_mul_x2ramp(temp,xdim,ydim,
		    xstart,ystart,xend,yend,mask); /* x2*Ix2[i]*/
  m[14] = sum_mul_y2ramp(temp,xdim,ydim,
		    xstart,ystart,xend,yend,mask); /* y2*Ix2 */
  m[8]  = sum_mul_xyramp(temp,xdim,ydim,
		    xstart,ystart,xend,yend,mask); /* xy*Ix2 */
  m[1]  = sum_mul_xramp(temp,xdim,ydim,
		    xstart,ystart,xend,yend,mask);  /* x*Ix2 */
  m[2]  = sum_mul_yramp(temp,xdim,ydim,
		    xstart,ystart,xend,yend,mask);  /* y*Ix2 */
  m[0]  = sum_array(temp,xdim,ydim,xstart,ystart,xend,yend,mask); /* Ix2 */

  /********* IxIy term *********/
  mul_im (ix, iy, temp, im_size);   /* temp = ixy */
  m[10] = sum_mul_x2ramp(temp,xdim,ydim,
		    xstart,ystart,xend,yend,mask); /* x2*IxIy */
  m[17] = sum_mul_y2ramp(temp,xdim,ydim,
		    xstart,ystart,xend,yend,mask); /* y2*IxIy */
  m[11] = sum_mul_xyramp(temp,xdim,ydim,
		    xstart,ystart,xend,yend,mask); /* xy*IxIy */
  m[4]  = sum_mul_xramp(temp,xdim,ydim,
		    xstart,ystart,xend,yend,mask);  /* x*IxIy */
  m[5]  = sum_mul_yramp(temp,xdim,ydim,
		    xstart,ystart,xend,yend,mask);  /* y*IxIy */
  m[3]  = sum_array(temp,xdim,ydim,xstart,ystart,xend,yend,mask); /* IxIy */
    
  /********** Iy2 term **********/
    
  mul_im (iy, iy, temp, im_size);   /* temp = iy2 */
  m[28] = sum_mul_x2ramp(temp,xdim,ydim,
		    xstart,ystart,xend,yend,mask); /* x2*Iy2 */
  m[35] = sum_mul_y2ramp(temp,xdim,ydim,
		    xstart,ystart,xend,yend,mask); /* y2*Iy2 */
  m[29] = sum_mul_xyramp(temp,xdim,ydim,
		    xstart,ystart,xend,yend,mask); /* xy*Iy2 */
  m[22] = sum_mul_xramp(temp,xdim,ydim,
		    xstart,ystart,xend,yend,mask);  /* x*Iy2 */
  m[23] = sum_mul_yramp(temp,xdim,ydim,
		    xstart,ystart,xend,yend,mask);  /* y*Iy2 */
  m[21] = sum_array(temp,xdim,ydim,xstart,ystart,xend,yend,mask); /* Iy2 */
    
  /************ IxIt term ***********/
  mul_im (ix, it, temp, im_size);   /* temp = it2 */
  v[1]  = sum_mul_xramp(temp,xdim,ydim,
		    xstart,ystart,xend,yend,mask); /* -x*IxIt */
  v[2]  = sum_mul_yramp(temp,xdim,ydim,
		    xstart,ystart,xend,yend,mask); /* -y*IxIt */
  v[0]  = sum_array(temp,xdim,ydim,xstart,ystart,xend,yend,mask); /* -IxIt */
    
  /************ IxIt term ***********/
  mul_im (iy, it, temp, im_size);   /* temp = it2 */
  v[4]  = sum_mul_xramp(temp, xdim,ydim,
		    xstart,ystart,xend,yend,mask); /* -x*IyIt */
  v[5]  = sum_mul_yramp(temp,xdim,ydim,
		    xstart,ystart,xend,yend,mask); /* -y*IyIt */
  v[3]  = sum_array(temp,xdim,ydim,xstart,ystart,xend,yend,mask); /* -IyIt  */

  /******** symmetric bindings ********/
  /* Ix2 terms */
  m[13]=m[8];
  m[6] =m[1];
  m[12]=m[2];
    
  /* Iy2 terms */
  m[34]=m[29];
  m[27] =m[22];
  m[33]=m[23];
    
  /*IxIy terms */
  m[18]=m[3];
  m[19]=m[9]=m[24]=m[4];
  m[20]=m[15]=m[30]=m[5];
  m[26]=m[16]=m[31]=m[11];
  m[25]=m[10];
  m[32]=m[17];

#ifdef NOISE_STABLIZE 
  {
    /* bias diagonal with noise assumptions for stability */
    int x,y,ypos,pos; 
    float noise=1;
    double nz=0;
    
    xend = ((xend<xdim)? xend : xdim);
    yend = ((yend<ydim)? yend : ydim);
    
    if(mask==NULL)  nz=(xend-xstart)*(yend-ystart)*noise;
    else {
      for(y=ystart,ypos=y*xdim; y<yend; y++,ypos+=xdim)
	for(x=xstart,pos=ypos+x; x<xend; x++,pos++)
	  nz+=noise*mask[pos];
    }
    m[0] =m[0] +nz;
    m[7] =m[7] +nz/0.01;
    m[14]=m[14]+nz/0.01;
    m[21]=m[21]+nz;
    m[28]=m[28]+nz/0.01;
    m[35]=m[35]+nz/0.01;
  }
#endif
    
  /* inverse-matrix */
/*  fprintf(stderr,"calling pinverse....");  */
  internal_pinverse (m, im, 6, 6);
/*  fprintf(stderr,"return from pinverse\n"); */

/*  fprintf(stderr,"/n(%4.4f %4.4f %4.4f %4.4f %4.4f %4.4f)\n", ap[0],ap[1],ap[2],ap[3],ap[4],ap[4]); */
    
    /* mul matrix    [im]*[v] */
  ap[3] = im[0]  * v[0] + im[1]  * v[1] + im[2]  * v[2] +
          im[3]  * v[3] + im[4]  * v[4] + im[5]  * v[5];
  ap[4] = im[6]  * v[0] + im[7]  * v[1] + im[8]  * v[2] + 
          im[9]  * v[3] + im[10] * v[4] + im[11] * v[5];
  ap[5] = im[12] * v[0] + im[13] * v[1] + im[14] * v[2] +
          im[15] * v[3] + im[16] * v[4] + im[17] * v[5];
  ap[0] = im[18] * v[0] + im[19] * v[1] + im[20] * v[2] +
          im[21] * v[3] + im[22] * v[4] + im[23] * v[5];
  ap[1] = im[24] * v[0] + im[25] * v[1] + im[26] * v[2] +
          im[27] * v[3] + im[28] * v[4] + im[29] * v[5];
  ap[2] = im[30] * v[0] + im[31] * v[1] + im[32] * v[2] +
          im[33] * v[3] + im[34] * v[4] + im[35] * v[5];

/*  fprintf(stderr,"\naf-est: (%4.8f %4.8f %4.8f %4.8f %4.8f %4.8f)\n",
	 ap[0],ap[1],ap[2],ap[3],ap[4],ap[5]);    */

  if (allocate)
    free(temp);
}

/* sum all the points in the array bounded by xstart,ystart,xend,yend; */
static float sum_array (array,xdim,ydim,xstart,ystart,xend,yend,mask)
  register float *array;
  register int xdim,ydim,xstart,ystart,xend,yend;
  float *mask;
{
  register double sum=0;
  register int x,y;
  register int ymax,pos;

  ymax = ((yend<ydim)? yend :ydim)*xdim;

  if(mask==NULL) {
/*    fprintf(stderr,"null "); */
    for(y=ystart*xdim;y<ymax;y+=xdim)
      for(x=xstart,pos=y+x;x<xend;x++,pos++)
	sum += (double)array[pos];
    return((float) sum);
  }
  else {
    for(y=ystart*xdim;y<ymax;y+=xdim)
      for(x=xstart,pos=y+x;x<xend;x++,pos++)
	sum += (double)array[pos]*mask[pos];
    return((float) sum);
  }
}

/* multiply array by x. center at center of image */
static float sum_mul_xramp (array,xdim,ydim,
			    xstart,ystart,xend,yend,mask)
  register float  *array,*mask;
  int xstart,ystart,xend,yend;
  register int xdim,ydim;
{
  double sum=0;
  register float x,y,xc;
  register int pos,ymax; 

  xend = ((xend<xdim)? xend : xdim);
  yend = ((yend<ydim)? yend : ydim);

  ymax = yend*xdim;

  xc = (float)(((float )(xdim-1))/2.0);

  if(mask==NULL) {
    for(y=ystart*xdim; y<ymax; y+=xdim)
      for(x=xstart,pos=y+x; x<xend; x++,pos++)
	sum+=(x-xc)*array[pos];
  }
  else {
    for(y=ystart*xdim; y<ymax; y+=xdim)
      for(x=xstart,pos=y+x; x<xend; x++,pos++)
	sum+=(x-xc)*array[pos]*mask[pos];
  }
  return((float)sum);
}


/* multiply array by x-squared. center at center of image */
static float sum_mul_x2ramp (array,xdim,ydim,
			     xstart,ystart,xend,yend,mask)
  register float  *array, *mask;	
  int xstart,ystart,xend,yend;
  register int xdim,ydim;
{
  double sum=0;
  register float x,y,xc,z;
  register int pos,ymax; 

  xend = ((xend<xdim)? xend : xdim);
  yend = ((yend<ydim)? yend : ydim);

  ymax = yend*xdim;

  xc = (float)(((float )(xdim-1))/2.0);

  if(mask==NULL) {
    for(y=ystart*xdim; y<ymax; y+=xdim)
      for(x=xstart,pos=y+x; x<xend; x++,pos++) {
	z=(x-xc); sum+=z*z*array[pos]; }
  }
  else {
    for(y=ystart*xdim; y<ymax; y+=xdim)
      for(x=xstart,pos=y+x; x<xend; x++,pos++) {
	z=(x-xc); sum+=z*z*array[pos]*mask[pos]; }
  }
  return((float)sum);
}

/* multiply array by y. center at center of image */
static float sum_mul_yramp (array,xdim,ydim,
			     xstart,ystart,xend,yend,mask)
  register float  *array, *mask;	
  int xstart,ystart,xend,yend;
  register int xdim,ydim;
{
  register float z,yc;
  register int x,y,ypos,pos; 
  register double sum=0;

  xend = ((xend<xdim)? xend : xdim);
  yend = ((yend<ydim)? yend : ydim);

  yc = (float)(((float)(ydim-1))/2.0);

  if(mask==NULL) {
    for(y=ystart,ypos=y*xdim; y<yend; y++,ypos+=xdim) {
      z=(float)(y-yc);
      for(x=xstart,pos=ypos+x; x<xend; x++,pos++) 
	sum+=z*array[pos];
    }
  }
  else {
    for(y=ystart,ypos=y*xdim; y<yend; y++,ypos+=xdim) {
      z=(float)(y-yc);
      for(x=xstart,pos=ypos+x; x<xend; x++,pos++) 
	sum+=z*array[pos]*mask[pos];
    }
  }
  return((float)sum);
}


/* multiply array by y-squared. center at center of image */
static float sum_mul_y2ramp (array,xdim,ydim,
			     xstart,ystart,xend,yend,mask)
  register float  *array, *mask;	
  int xstart,ystart,xend,yend;
  register int xdim,ydim;
{
  register float z,yc;
  register int x,y,ypos,pos; 
  register double sum=0;

  xend = ((xend<xdim)? xend : xdim);
  yend = ((yend<ydim)? yend : ydim);

  yc = (float)(((float)(ydim-1))/2.0);

  if(mask==NULL) {
    for(y=ystart,ypos=y*xdim; y<yend; y++,ypos+=xdim) {
      z=(float)(y-yc); z=z*z;
      for(x=xstart,pos=ypos+x; x<xend; x++,pos++) 
	sum+=z*array[pos];
    }
  }
  else {
    for(y=ystart,ypos=y*xdim; y<yend; y++,ypos+=xdim) {
      z=(float)(y-yc); z=z*z;
      for(x=xstart,pos=ypos+x; x<xend; x++,pos++) 
	sum+=z*array[pos]*mask[pos];
    }
  }
  return((float)sum);
}


/* multiply array by x*y. center at center of image */
static float sum_mul_xyramp (array,xdim,ydim,
			     xstart,ystart,xend,yend,mask)
  register float  *array, *mask;	
  int xstart,ystart,xend,yend;
  register int xdim,ydim;
{
  register float z,yc,xc;
  register int x,y,ypos,pos; 
  register double sum=0;

  xend = ((xend<xdim)? xend : xdim);
  yend = ((yend<ydim)? yend : ydim);

  xc = (float)(((float)(xdim-1))/2.0);
  yc = (float)(((float)(ydim-1))/2.0);

  if(mask==NULL) {
    for(y=ystart,ypos=y*xdim; y<yend; y++,ypos+=xdim) {
      z=(float)(y-yc);
      for(x=xstart,pos=ypos+x; x<xend; x++,pos++)
	sum+=z*(x-xc)*array[pos];
    }
  }
  else {
    for(y=ystart,ypos=y*xdim; y<yend; y++,ypos+=xdim) {
      z=(float)(y-yc);
      for(x=xstart,pos=ypos+x; x<xend; x++,pos++)
	sum+=z*(x-xc)*array[pos]*mask[pos];
    }
  }
  return((float)sum);
}
