/******************************************************************

"Copyright  1994 by the Massachusetts  Institute  of  Technology.  All
rights reserved.

Developed by John Y. A. Wang at the  Media Laboratory, MIT, Cambridge,
Massachusetts, with  support  from Secom Ltd., Goldstar Ltd.,  and the
Television of Tomorrow Program.

For use by authorized users only.

This distribution is approved  by Nicholas Negroponte, Director of the
Media Laboratory, MIT.

Permission to use, copy, or modify this software and its documentation
for educational and research purposes only and without  fee is  hereby
granted, provided that this copyright notice and the original authors'
names  appear  on  all  copies  and   supporting   documentation.   If
individual  files  are  separated  from  this  distribution  directory
structure, this copyright notice must be included.  For any other uses
of  this software, in  original or  modified form, including  but  not
limited to distribution in whole or in part, specific prior permission
must be  obtained  from  MIT.   These  programs  shall  not  be  used,
rewritten,  or  adapted  as  the  basis  of a  commercial software  or
hardware  product without  first  obtaining  appropriate licenses from
MIT.  MIT.  makes no  representations about  the  suitability of  this
software for any purpose.  It is  provided "as is" without  express or
implied warranty.

******************************************************************/


/***************************************************************
*** File: affine-src.c
*** Author: John Y. A. Wang
*** Date: January 15, 1993
*** Copyright 1993, Massachusetts Institute of Technology, All Rights Reserved.
***
*** Description: affine motion program from command line
***************************************************************/

/*****************************************************************/
#include "dynamo.h"
/*****************************************************************/

/********** IMPORTANT:  put copyright label in all the files ***********/
static  char *copyright="\
\n\n***********************************************\n\
Developed by John Y. A. Wang at the Media Laboratory, MIT, Cambridge,Massachusetts\n\
Copyright  1993 by the Massachusetts  Institute  of  Technology.  All rights reserved.\n\
***********************************************\n\n";


/**************************************************/
display_affine_motion_options(image_mgr)
  Image_Manager *image_mgr;
{
  Affine_Analysis *an;
  Image_Cache *cache;
  int i;

  if((cache=ask_name_get_cache(image_mgr))==NULL)
    return(0);
  
  an= &cache->affine_analysis;
  
  printf("Using the following options:\n");

  printf("\tLevel: %d\n",an->level);

  printf("\tMethod: "); printf("(");
  for(i=0;i<=an->level;i++){
    printf("%d ",(int)an->method[i]);
  }
  printf(")\n");
  printf("\tCrop region: (%d %d %d %d)\n",an->xstart,an->ystart,an->xend,an->yend);
  printf("\tStart frame: %d\n",an->start);
  printf("\tEnd frame: %d\n",an->end);
  printf("\tTime step between analysis: %d\n",an->tstep);
  return(0);
}

/*****************************************************************/
/**************************************************/
int change_affine_analysis_options(cache)
  Image_Cache *cache;
{
  extern char *args;

  Affine_Analysis *an=&(cache->affine_analysis);

  int i;

  printf("\tLevel: (%d) : ",an->level);
  readline(args,ARGSIZE,stdin); sscanf(args,"%d",&an->level);

  printf("\tMethod: (");
  for(i=0;i<=an->level;i++){ printf("%d ",(int)an->method[i]); }
  printf(") : ");
  readline(args,ARGSIZE,stdin);
  sscanf(args,"%f %f %f %f %f %f %f %f %f %f",
	 &an->method[0],&an->method[1],&an->method[2],&an->method[3],
	 &an->method[4],&an->method[5],&an->method[6],&an->method[7],
	 &an->method[8],&an->method[9]);
  
  printf("\tUpdate Factor: (%4.2f %4.2f %4.2f %4.2f %4.2f %4.2f) : ",
	 an->factor[0],an->factor[1],an->factor[2],
	 an->factor[3],an->factor[4],an->factor[5]);fflush(stdout);
  readline(args,ARGSIZE,stdin);
  sscanf(args,"%f %f %f %f %f %f",
	 &an->factor[0],&an->factor[1],&an->factor[2],
	 &an->factor[3],&an->factor[4],&an->factor[5]);

  printf("\tCrop region (xs ys xe ye): (%d %d %d %d) : ",
	 an->xstart,an->ystart,an->xend,an->yend);
  readline(args,ARGSIZE,stdin);
  sscanf(args,"%d %d %d %d",&an->xstart,&an->ystart,&an->xend,&an->yend);

  printf("\tStart frame: (%d) : ",an->start);
  readline(args,ARGSIZE,stdin); sscanf(args,"%d",&an->start);

  printf("\tEnd frame: (%d) : ",an->end);
  readline(args,ARGSIZE,stdin); sscanf(args,"%d",&an->end);

  printf("\tTime step between analysis: (%d) : ",an->tstep);
  readline(args,ARGSIZE,stdin); sscanf(args,"%d",&an->tstep);

  if(an->end>(cache->ext_start+cache->tdim-1 - an->tstep)) {
    an->end = cache->ext_start + (cache->tdim-1) - an->tstep; }
/*
  an->end=(((an->end)<=(tdim-(an->tstep)-1)) ?
	   (an->end) : (tdim-(an->tstep)-1));
*/
  return(0);
}

/*****************************************************************/
/*****************************************************************/
save_param_array(image_mgr)
  Image_Manager *image_mgr;
{
  FILE *fp;
  int i;

  Image_Cache *cache;
  Affine_Analysis *an;
  float *param;
  char *filename;
  
  float *cur_param;
  char *progn="save_param_array";

  if((cache=ask_name_get_cache(image_mgr))==NULL)
    return(0);

  an= &cache->affine_analysis;
  param=cache->param;
  filename=cache->paramfilename;
  
  if((fp=open_file(filename,"w"))==NULL) {
    fclose(fp);
    return(-1);;
  }
  printf("writing to %s ......\n", filename);
  fprintf(fp,";;; Affine parameters generated by %s:\n", progn);
  fprintf(fp,";;; Start=%d : End=%d : tstep=%d\n", an->start,an->end,an->tstep);
  fprintf(fp,";;; Level=%d \n", an->level);

  fprintf(fp,";;; Method: (");
  for(i=0;i<=an->level;i++){fprintf(fp,"%d ",(int)an->method[i]);}
  fprintf(fp,")\n");

  fprintf(fp,";;; Crop region: (%d %d %d %d)\n",an->xstart,an->ystart,an->xend,an->yend);
  
  fprintf(fp,"(setq param-list '(\n");

  for(i=an->start;i<=an->end;i++) {
    cur_param=param+i*PARAM_WIDTH;
    fprintf(fp,"\t(%10.4f %10.6f %10.6f %10.4f %10.6f %10.6f)\n",
	   cur_param[0],cur_param[1],cur_param[2],cur_param[3],cur_param[4],cur_param[5]);
  }
  fprintf(fp,"\t)\n"); /* end list */
  fprintf(fp,")\n"); /* end setq */
  fclose(fp);
  printf("parameters written.\n");
  return(0);
}

/*****************************************************************/
/*****************************************************************/
display_affine_params(image_mgr)
  Image_Manager *image_mgr;
{
  int i;
  float *arr;

  Image_Cache *cache;
  Affine_Analysis *an;
  float *param;

  if((cache=ask_name_get_cache(image_mgr))==NULL)
    return(0);

  an= &cache->affine_analysis;
  param=cache->param;
  
  for(i=an->start;i<=an->end;i++) {
    arr=param+i*PARAM_WIDTH;
    printf("%3d:  (%8.2f %8.4f %8.4f %8.2f %8.4f %8.4f)\n", i,
	   arr[0],arr[1],arr[2],arr[3],arr[4],arr[5]);
  }
  fflush(stdout);
  return(0);
}

/*****************************************************************/
/*****************************************************************/
load_affine_params(image_mgr)
  Image_Manager *image_mgr;
{
  extern char *args;

  float *param;
  int start,end;
  char *filename;

  int i;
  int offset=0;
  float *arr;
  char *buf=image_mgr->cbuf1;
  char *buf2=image_mgr->cbuf2;
  FILE *fp;
  Image_Cache *cache;
  
  if((cache=ask_name_get_cache(image_mgr))==NULL)
    return(0);

  param=cache->param;
  filename=cache->paramfilename;
  start=cache->affine_analysis.start;
  end=cache->affine_analysis.end;
  
  if((fp=open_file(filename,"r"))==NULL) {
    return(-1);
  }
    
  /************ parse comments *****************************/
  for(;;) {
    if(fgets(buf,100,fp)==NULL) {
      printf("********EOF********\n");	
      break;
    }
    if(strncmp(buf,";",1) != 0)
      break;
    else
      printf("%s",buf); /* print comment. may be useful */
  }
  
  /************ parse setq ****************************/
  for(;;) {
    if(strncmp(buf,"(setq param-list",15)) { 
      break;
    }
    else
      if(fgets(buf,100,fp)==NULL) {
	printf("********EOF********\n");	
	break;
      }
  }

  printf("Enter Start: (%d) : ",start);
  readline(args,ARGSIZE,stdin);
  sscanf(args,"%d",&start);

  printf("Enter End: (%d) : ",end);
  readline(args,ARGSIZE,stdin);
  sscanf(args,"%d",&end);

  offset=start;

  printf("Enter Offset: (%d) : ",offset);
  readline(args,ARGSIZE,stdin);
  sscanf(args,"%d",&offset);

  /************ parse data tokens ****************************/
    for(i=0;i<=end;i++) {
      if(i>=start)
	arr=param+i*PARAM_WIDTH;
      else
	arr=param;
      fprintf(stderr,"buf=%s\n",buf);
      if(sscanf(buf,"%s %f %f %f %f %f %f %s",
		buf2,&arr[0],&arr[1],&arr[2],&arr[3],&arr[4],&arr[5],buf2)
	 !=8) {
	printf("*i=%d,Incorrect format or EOF\nbuf=%s",i,buf);
	break;
      }
      else
	printf("%8d: %8.2f %8.4f %8.4f %8.2f %8.4f %8.4f\n",
	       i,arr[0],arr[1],arr[2],arr[3],arr[4],arr[5]);

      if(fgets(buf,100,fp)==NULL) {
	printf("********EOF********\n");
	break;
      }
    }
  fflush(stdout);
  fclose(fp);
  return(0);
}

/*****************************************************************/
/*****************************************************************/
int run_global_affine_motion_analysis(image_mgr)
  Image_Manager *image_mgr;
{
  extern char *args;

  float *im2;
  float *ttemp=NULL,*im1=NULL;  /* these are malloc-ed here */
  int ttempsize;  
  float *cur_param;
  float *temp;
  int i;
  Affine_Analysis *an;
  Image_Cache *cache;
  int imsize;

  float xs,ys,xe,ye;
  int xs_i,ys_i,xe_i,ye_i;
  int track;
/*   float xmid,ymid; */
/*  float inv_param[6];*/
  
  char *yn=image_mgr->yn;

  int limit;

  FILE *pid=NULL;  /* for displaying tracked window */


  /**************** Analysis *******************************/
  printf("Running affine analysis ...........\n");

 if((cache=ask_name_get_cache(image_mgr))==NULL)
   return(0);

  imsize=cache->imsize;
  ttempsize=5*imsize;
  an= &(cache->affine_analysis);

  ttemp=NULL;im1=NULL;
  if((ttemp=(float *)malloc((unsigned)ttempsize*sizeof(float)))==NULL) {
    printf("cannot allocate memory for affine analysis (ttemp)\n");
    free((char *)ttemp);
    return(-1);
  }
  if((im1=(float *)malloc((unsigned)imsize*sizeof(float)))==NULL) {
    printf("cannot allocate memory for affine analysis (im1)\n");
    free((char *)ttemp);free((char *)im1);
    return(-1);
  }

  change_affine_analysis_options(cache);
  if(an->start<0) {
    printf("Incorrect start specification....\n"); an->start=0; return(-1);}
  if(an->end<0) {
    printf("Incorrect end specification....\n"); an->end=cache->tdim;
    return(-1);}

  limit=1; sprintf(yn,"y");
  printf("\tLimit window? (y/n) (%s) >> ", yn);
  readline(args,ARGSIZE,stdin);  sscanf(args,"%s",yn);
  if(strncmp(yn,"n",1)==0) { limit=0; }  /* no */

  track=1; sprintf(yn,"y");
  printf("\tTrack analysis window? (y/n) (%s) >> ",yn);
  readline(args,ARGSIZE,stdin);  sscanf(args,"%s",yn);
  if(strncmp(yn,"n",1)==0) { track=0; }  /* no */

  xs=(float)an->xstart; xe=(float)an->xend;
  ys=(float)an->ystart; ye=(float)an->yend;
  xs_i=(int)xs; xe_i=(int)xe; ys_i=(int)ys; ye_i=(int)ye;

  for(i=an->start;i<=an->end;i++) {
    printf("\nAnalysis on frames %d and %d .... \n",i,i+an->tstep);
    cur_param=(cache->param)+(i*PARAM_WIDTH);
    if((temp=get_image(cache,i))==NULL) {
      printf("Analysis: cannot get frame %d\n",i);
      return(-1);
    }
    /* save im1 into static space in case cache changes while getting im2 */
    copy_im(temp,im1,cache->imsize);  
    fill_im(ttemp,(float)0,ttempsize);
    if((im2=get_image(cache,(i+an->tstep)))==NULL){
      printf("Analysis: cannot get frame %d\n",i);
      break;
    }

/*    printf("xs=%4.1f, ys=%4.1f, xe=%4.1f, ye=%4.1f\n",xs,ys,xe,ye); */
    printf("xs_i=%d ys_i=%d xe_i=%d ye_i=%d\n",xs_i,ys_i,xe_i,ye_i);
    internal_coarse_fine_affine_estimator(im1,im2,cur_param,an->param_init,
					  cache->xdim,cache->ydim,
					  an->level,an->method,
					  xs_i,ys_i,xe_i,ye_i,
					  ttemp,ttempsize,
					  an->factor,NULL,NULL,
					  limit);

    printf("\t parameters: (%4.1f %8.3f %8.3f  %4.1f %8.3f %8.3f)\n",
	   cur_param[0],cur_param[1],cur_param[2],
	   cur_param[3],cur_param[4],cur_param[5]);

    if(track==1) { /* currrently only handles translation cannot */
      if(i==an->start) {  /* first time around. open window for display */
	pid=OpenGrayScreen(cache->ydim,cache->xdim,"tracked window",
			   cache->gamma,(float)1.0,(an->end-an->start+1));
	if(pid!=NULL) { 
	  convert_float_to_byte(im1,cache->array,cache->imsize);
	  draw_window_bounds(cache->array,xs_i,ys_i,xe_i,ye_i,
			     cache->xdim,cache->ydim);
	  display_number_on_image(cache->array,cache->xdim,cache->ydim,
				  (float)i,(int)5,(int)5,(int)255,(int)0,(int)4);
	  DumpBufToScreen(cache->array,cache->ydim,cache->xdim,pid,imsize);
	}
      }

      set_new_bounds(cur_param,(float *)&xs,(float *)&ys,
		     (float *)&xe,(float *)&ye,cache->xdim,cache->ydim);

      xs_i=(int)(xs+0.5); xe_i=(int)(xe+.5);
      ys_i=(int)(ys+0.5); ye_i=(int)(ye+.5);
      if(xs_i<5) xs_i=(int)5; if(xe_i>cache->xdim-5) xe_i=cache->xdim-6;
      if(ys_i<5) ys_i=(int)5; if(ye_i>cache->ydim-5) ye_i=cache->ydim-6;
      if(pid!=NULL) {
	convert_float_to_byte(im2,cache->array,cache->imsize);
	draw_window_bounds(cache->array,xs_i,ys_i,xe_i,ye_i,
			   cache->xdim,cache->ydim);
	display_number_on_image(cache->array,cache->xdim,cache->ydim,
			    (float) i,(int)5,(int)5,(int)255,(int)0,(int)4);
	DumpBufToScreen(cache->array,cache->ydim,cache->xdim,pid,imsize);
/*	fprintf(stderr,"done display...\n"); */
      }
    }
  }
  if(track==1 && pid!=NULL) {
/*     fprintf(stderr,"closing window...\n"); */
    Close_Screen(pid);
  }

/*  fprintf(stderr,"free memory...\n");*/
  free((char *)ttemp); ttemp=NULL;
  free((char *)im1); im1=NULL;
  fprintf(stderr,"affine analysis: done\n");
  return(0);
}


/*****************************************************************/
/*****************************************************************/
/*  (list 0 0 (- 1 (/ factor)) 0 (- 1 (/ factor)) 0)) */
zoom_param(res,xzoom,yzoom)
  float *res;
  float xzoom,yzoom;
{
  if(xzoom==0 || yzoom==0) {
    printf("Warning: zoom factors are zero. Setting zoom to 1.\n");
    res[0]=0; res[1]=0; res[2]=0;
    res[3]=0; res[2]=0; res[5]=0;
  }
  else {
    res[0]=0; res[1]=0; res[2]=1-(1/yzoom);
    res[3]=0; res[4]=1-(1/xzoom); res[5]=0;
  }
}  


/********************************************************************************/
int draw_window_bounds(array,xs,ys,xe,ye,xdim,ydim)
  unsigned char *array;
  int xs,ys,xe,ye,xdim,ydim;
{
  drawline(array,xdim,ydim,xs,ys,xs,ye,(int)255);  /* left */
  drawline(array,xdim,ydim,xe,ys,xe,ye,(int)255);  /* right */
  drawline(array,xdim,ydim,xs,ys,xe,ys,(int)255);  /* top */
  drawline(array,xdim,ydim,xs,ye,xe,ye,(int)255);  /* bottom */
}

int show_window_bounds_cache(cache)
  Image_Cache *cache;
{
  unsigned char *array=cache->array;
  int xdim=cache->xdim, ydim=cache->ydim;
  int imsize=xdim*ydim;
  float *im;
  Affine_Analysis *an = &(cache->affine_analysis);
  FILE *pid;
  float *cur_param;
  int i,xs_i,ys_i,xe_i,ye_i;
  float xs,ys,xe,ye;


  if((pid=OpenGrayScreen(ydim,xdim,"tracked window",cache->gamma,(float)1.0,
		     (an->end-an->start+1)))==NULL) {
    return(0);}

  xs=(float)an->xstart; xe=(float)an->xend;
  ys=(float)an->ystart; ye=(float)an->yend;

  for(i=an->start; i<=an->end;i++) {
    if((im=get_image(cache,(i+an->tstep)))==NULL){
      break;
    }
    convert_float_to_byte(im,array,imsize);

    xs_i=(int)(xs+0.5); xe_i=(int)(xe+.5);
    ys_i=(int)(ys+0.5); ye_i=(int)(ye+.5);
    if(xs_i<0) xs_i=(int)0; if(xe_i>cache->xdim) xe_i=cache->xdim-1;
    if(ys_i<0.0) ys_i=(int)0; if(ye_i>cache->ydim) ye_i=cache->ydim-1;
/*    printf("xs=%4.1f, ys=%4.1f, xe=%4.1f, ye=%4.1f\n",xs,ys,xe,ye); */
    printf("xs_i=%d ys_i=%d xe_i=%d ye_i=%d\n",xs_i,ys_i,xe_i,ye_i);
    
    drawline(array,xdim,ydim,xs_i,ys_i,xs_i,ye_i,(int)255); /*left*/
    drawline(array,xdim,ydim,xe_i,ys_i,xe_i,ye_i,(int)255); /*right*/
    drawline(array,xdim,ydim,xs_i,ys_i,xe_i,ys_i,(int)255); /*top*/
    drawline(array,xdim,ydim,xs_i,ye_i,xe_i,ye_i,(int)255); /*bot*/

    cur_param=(cache->param)+(i*PARAM_WIDTH);
    set_new_bounds(cur_param,(float *)&xs,(float *)&ys,(float*)&xe,(float *)&ye,
		   xdim,ydim);
    DumpBufToScreen(array,ydim,xdim,pid,imsize);
  }
  Close_Screen(pid);
  return(0);
}



int set_new_bounds(cur_param,xstart,ystart,xend,yend,xdim,ydim) 
  float *cur_param;
  float *xstart,*ystart,*xend,*yend;
  int xdim,ydim;
{
  float dx,dy;
  float inv_param[6];
  int i;
  float xs,ys,xe,ye;
  float x[4],y[4];
  float xc=(float)(xdim/2.0), yc=(float)(ydim/2.0);

  internal_invert_affine_params_from(cur_param,inv_param);

/*
  printf("set: xstart=%4.1f ystart=%4.1f xend=%4.1f yend=%4.1f\n",
	 *xstart,*ystart,*xend,*yend);
*/
  xs=(float)(*xstart-xc); ys=(float)(*ystart-yc);
  xe=(float)(*xend-xc); ye=(float)(*yend-yc);

/*  printf("set: xs=%4.1f ys=%4.1f xe=%4.1f ye=%4.1f\n", xs,ys,xe,ye);  */
  x[0]=xs;y[0]=ys; /* up left */
  x[1]=xe;y[1]=ys; /* up right */
  x[2]=xs;y[2]=ye; /* bot left */
  x[3]=xe;y[3]=ye; /* bot right */

  for(i=0;i<4;i++) {
    dy=((inv_param[0]+x[i]*inv_param[1]+y[i]*inv_param[2]));
    dx=((inv_param[3]+x[i]*inv_param[4]+y[i]*inv_param[5]));
    x[i]=x[i]-dx; y[i]=y[i]-dy;
/*    printf("%4.1f: x=%4.1f y=%4.1f dx=%4.1f dy=%4.1f\n",i,x[i],y[i],dx,dy);*/
  }

  xs=x[0];xe=x[0];ys=y[0];ye=y[0];
  for(i=0;i<4;i++) {
/*    printf("%d: x=%4.1f y=%4.1f\n",i,x[i],y[i]); */
    if(x[i]<xs) xs=x[i];
    else if(x[i]>xe) xe=x[i];
    if(y[i]<ys) ys=y[i];
    else if(y[i]>ye) ye=y[i];
  }
  *xstart=xs+xc; *ystart=ys+yc; *xend=xe+xc; *yend=ye+yc;
  return(0);
}

