/******************************************************************

"Copyright  1994 by the Massachusetts  Institute  of  Technology.  All
rights reserved.

Developed by John Y. A. Wang at the  Media Laboratory, MIT, Cambridge,
Massachusetts, with  support  from Secom Ltd., Goldstar Ltd.,  and the
Television of Tomorrow Program.

For use by authorized users only.

This distribution is approved  by Nicholas Negroponte, Director of the
Media Laboratory, MIT.

Permission to use, copy, or modify this software and its documentation
for educational and research purposes only and without  fee is  hereby
granted, provided that this copyright notice and the original authors'
names  appear  on  all  copies  and   supporting   documentation.   If
individual  files  are  separated  from  this  distribution  directory
structure, this copyright notice must be included.  For any other uses
of  this software, in  original or  modified form, including  but  not
limited to distribution in whole or in part, specific prior permission
must be  obtained  from  MIT.   These  programs  shall  not  be  used,
rewritten,  or  adapted  as  the  basis  of a  commercial software  or
hardware  product without  first  obtaining  appropriate licenses from
MIT.  MIT.  makes no  representations about  the  suitability of  this
software for any purpose.  It is  provided "as is" without  express or
implied warranty.

******************************************************************/


/***************************************************************
*** File: affine-util.c
*** Author: John Y. A. Wang
*** Date: January 19, 1993.
*** Copyright ****, Massachusetts Institute of Technology, All Rights Reserved.
***
*** Description: affine utilities.
***************************************************************/

static  char *copyright="\
\n\n***********************************************\n\
File: affine-util.c\n\
Author: John Y. A. Wang.\n\
The MIT Media Lab.\n\
Copyright (c) 1993, All Rights Reserved.\n\
***********************************************\n\n";

#include<stdio.h>
#include "motion.h"
#include "math.h"
/*
  cc -c -O4 -fsingle  affine-util.c
*/

#ifdef lint
/*
  lint affine-util.c
*/
#define register
#endif

static void print_aff(/*_ float *pr _*/);

/* finds composite affine param  of  transformation with p1 followed by p2 */
/* destructive modification allowed */
extern void internal_update_affine_params_from(p1, p2, pr, factor)
  float *p1,*p2,*pr,*factor;
{
  float ax1,bx1,cx1,ay1,by1,cy1;
  float ax2,bx2,cx2,ay2,by2,cy2;
  float pr2[6];
  int i;
/*  
  fprintf(stderr,"update\n");fflush(stderr);
  for(i=0;i<6;i++)
    fprintf(stderr,"%6.4f ",p1[i]);
  fprintf(stderr,"\n");
  for(i=0;i<6;i++)
    fprintf(stderr,"%6.4f ",p2[i]);
  fprintf(stderr,"\n");
  fflush(stderr);
*/  
  ay1=p1[0];  by1=p1[1];  cy1=p1[2];  ax1=p1[3];  bx1=p1[4];  cx1=p1[5];
  ay2=p2[0];  by2=p2[1];  cy2=p2[2];  ax2=p2[3];  bx2=p2[4];  cx2=p2[5];

  pr2[0] = ay1 + ay2 - cy1*ay2 - by1*ax2; /* ay */
  pr2[3] = ax1 + ax2 - cx1*ay2 - bx1*ax2; /* ax */
  
  pr2[2] = cy1 + cy2 - cy1*cy2 - by1*cx2; /* cy */
  pr2[5] = cx1 + cx2 - cx1*cy2 - bx1*cx2; /* cx */
  
  pr2[1] = by1 + by2 - cy1*by2 - by1*bx2; /* by */
  pr2[4] = bx1 + bx2 - cx1*by2 - bx1*bx2; /* bx */

  if(factor!=NULL) { for(i=0;i<6;i++) { pr[i] = pr2[i]*factor[i];} }
  else { for(i=0;i<6;i++) { pr[i] = pr2[i];} }

/*
  fprintf(stderr,"(%4.4f %4.4f %4.4f %4.4f %4.4f %4.4f)\n",
	 pr[0],pr[1],pr[2],pr[3],pr[4],pr[5]);fflush(stderr);
*/
}


/* inverts the affine param p1 and  puts it in pr */
/* destructive modification allowed */
extern void internal_invert_affine_params_from(p1, pr)
  float *p1,*pr;
{
  float det;
  
  float ay=p1[0];
  float by=p1[1];
  float cy=p1[2];
  float ax=p1[3];
  float bx=p1[4];
  float cx=p1[5];

  float binv[2][2];
  int i;

  det=(cy-1)*(bx-1) - cx*by;
/*
  fprintf(stderr,"det=%f\n",det); fflush(stderr);
  fprintf(stderr,"input: (%6.2f %6.4f %6.4f   %6.2f %6.4f %6.4f)\n",ay,by,cy,ax,bx,cx); fflush(stderr);
*/
  if(det==0) {
    fprintf(stderr,"internal_invert_affine_params_from: warning; determinant=0, returning negative param.\n");
    fflush(stderr);
    for(i=0;i<6;i++)
      pr[i]= -p1[i];
    return;
  }
  
  binv[0][0]=(bx-1)/det;
  binv[0][1]= -by/det;
  binv[1][0]= -cx/det;
  binv[1][1]=(cy-1)/det;
/*
  fprintf(stderr,"%6.4f %6.4f\n",binv[0][0],binv[0][1]); fflush(stderr);
  fprintf(stderr,"%6.4f %6.4f\n",binv[1][0],binv[1][1]); fflush(stderr);
*/
  pr[0]=binv[0][0]*ay + binv[0][1]*ax;
  pr[3]=binv[1][0]*ay + binv[1][1]*ax;

  pr[2]=binv[0][0]*cy+ binv[0][1]*cx;
  pr[1]=binv[0][0]*by + binv[0][1]*bx;

  pr[5]=binv[1][0]*cy+ binv[1][1]*cx;
  pr[4]=binv[1][0]*by + binv[1][1]*bx;
/*  
  fprintf(stderr,"output: (%6.2f %6.4f %6.4f   %6.2f %6.4f %6.4f)\n",pr[0],pr[1],pr[2],pr[3],pr[4],pr[5]);
  fflush(stderr);
*/
}


/** finds the composite aff param between from->to and puts it in pr. **/
/** destructive modification will result in unexpected results **/
extern void internal_update_affine_params_array
  (param_array, array_width, to, from, pr,
   factor_in,pre_factor_in)
  float *param_array;
  float *pr;
  int array_width, from, to;
  float *factor_in;
  float *pre_factor_in;
{
  float cur[6];
  int i,j;
  static float f1[6]={1.0,1.0,1.0,1.0,1.0,1.0};
  float *factor=factor_in;
  float *pre_factor=pre_factor_in;
  
/*  fprintf(stderr,"factor=%d pre_factor=%d f1=%d\n",factor,pre_factor,f1); */
  
  if (factor_in==NULL) { factor=f1; /*fprintf(stderr,"factor_in==NULL");fflush(stderr);*/}
  if (pre_factor_in==NULL) { pre_factor=f1; /*fprintf(stderr,"pre_factor_in==NULL");fflush(stderr);*/}
/*
  fprintf(stderr,"factor=%d pre_factor=%d f1=%d\n",factor,pre_factor,f1);
*/      
  for(j=0;j<6;j++)
    pr[j]=0.0;
  for(j=6,i=from*array_width+6;j<array_width;j++,i++)
    pr[j]=param_array[i];
  
/*  fprintf(stderr,"to=%d,from=%d,width=%d\n",to,from,array_width); fflush(stderr); */

  if(to>from) {
    for(i=from;i<to;i++) {
      for(j=0;j<6;j++) {cur[j]=pre_factor[j]*param_array[i*array_width + j];}
      internal_update_affine_params_from(pr,cur,pr,factor);
    }
  }
  else {
    for(i=to;i<from;i++) {
      for(j=0;j<6;j++) {cur[j]=pre_factor[j]*param_array[i*array_width + j];}
      internal_update_affine_params_from(pr,cur,pr,factor);
    }
    internal_invert_affine_params_from(pr,pr);
  }
/*
  fprintf(stderr,"result\n");
  print_aff(pr);
  fflush(stderr);
*/
}

/** takes an array of affine params and finds the composite params to center_frame.
 ** returns -1 on error and 0 when complete. */
/** NOTE: distructive modification will result in unexpected results. **/
extern int internal_make_composite_aff_array
  (aff_array,res_array,arraywidth,center_frame,nframes,invert)
  float *aff_array,*res_array;
  int arraywidth,center_frame,nframes;
  int invert;
{
  int frame,to,from;

/*  fprintf(stderr,"internal_make_composite_aff_array: invert=%d\n",invert);*/
  if(arraywidth>10) {
    fprintf(stderr,"ERROR *** not enough temp storage in composite.\n");
    return(-1);
  }

  if(aff_array==res_array) {
    fprintf(stderr,"\nWarning: destructive modification in affine composition caused incorrect results.....\n\n"); return(-1);
  }

  for(frame=0;frame<nframes;frame++) {
    if(invert==1) { to=frame; from=center_frame; }
    else { from=frame; to=center_frame; }
/*    fprintf(stderr,"from=%d,to=%d\n",from,to);fflush(stderr); */
    internal_update_affine_params_array(aff_array,arraywidth,to,from,
					&res_array[frame*arraywidth],
					(float *)NULL,(float *)NULL
					);
  }
  return(0);
}

/**** updates an affine parameter array with the following result:
  ***       pre_param -> param_array -> pos_param
  *** destructive modification allowed **/
extern void internal_update_aff_array(aff_array,res_array,arraywidth,
			       pre_param,post_param,nframes)
  float *aff_array,*res_array,*pre_param,*post_param;
  int arraywidth,nframes;
{
  int i,j;
  float temp[6];
  static float zero[6]={0.0,0.0,0.0,0.0,0.0,0.0};
  float *res,*cur;

  if(pre_param==NULL) { pre_param=zero;}
  if(post_param==NULL) { post_param=zero;}
  
  for(i=0;i<nframes;i++) {
    cur=aff_array+i*arraywidth;  res=res_array+i*arraywidth;
    internal_update_affine_params_from(pre_param,cur,temp,NULL);
    internal_update_affine_params_from(temp,post_param,temp,NULL);
    for(j=0;j<6;j++) { res[j]=temp[j]; }
    for(j=6;j<arraywidth;j++) { res[j]=cur[j]; }
  }
}

/** zoom the param with factors. result is xfrm with param
 ** followed by xy-zoom. returns results in param.
 ** destructive modification allowed. */
extern void zoom_affine_param(param,yfactor,xfactor)
  float *param,yfactor,xfactor;
{
  float zoom[6];

  if(yfactor==0.0 || xfactor==0.0) {
    fprintf(stderr,"ERROR: zoom factors cannot be zero: yfactor=%f xfactor=%f\n",
	    yfactor,xfactor);return;
  }
  zoom[0]=0.0;zoom[1]=0.0;zoom[2]=(1.0-1.0/yfactor);
  zoom[3]=0.0;zoom[4]=(1.0-1.0/xfactor);zoom[5]=0.0;
  internal_update_affine_params_from(param,zoom,param,(float *) NULL);
}
  

/** rotate affine parameters by angle in radians */
extern rotate_affine_param (param, angle)
  float *param,angle;
{
  float rotate[6];
  float sa=(float)sin((double)angle);
  float ca=(float)cos((double)angle);

  rotate[0]=0.0; rotate[1]=-sa;    rotate[2]=(1-ca);
  rotate[3]=0.0; rotate[4]=(1-ca); rotate[5]=sa;
/*
  fprintf(stderr,"rotate: %3.1f %6.4f %6.4f   %3.1f %6.4f %6.4f\n",
	  rotate[0],rotate[1],rotate[2],rotate[3],rotate[4],rotate[5]);
*/
  internal_update_affine_params_from(param,rotate,param,(float *) NULL);
}

static void print_aff(pr)
  float *pr;
{
  int i;
  
  for(i=0;i<6;i++)
    fprintf(stderr,"%6.4f ",pr[i]);
  fprintf(stderr,"\n");
  fflush(stderr);
}


