/******************************************************************

"Copyright  1994 by the Massachusetts  Institute  of  Technology.  All
rights reserved.

Developed by John Y. A. Wang at the  Media Laboratory, MIT, Cambridge,
Massachusetts, with  support  from Secom Ltd., Goldstar Ltd.,  and the
Television of Tomorrow Program.

For use by authorized users only.

This distribution is approved  by Nicholas Negroponte, Director of the
Media Laboratory, MIT.

Permission to use, copy, or modify this software and its documentation
for educational and research purposes only and without  fee is  hereby
granted, provided that this copyright notice and the original authors'
names  appear  on  all  copies  and   supporting   documentation.   If
individual  files  are  separated  from  this  distribution  directory
structure, this copyright notice must be included.  For any other uses
of  this software, in  original or  modified form, including  but  not
limited to distribution in whole or in part, specific prior permission
must be  obtained  from  MIT.   These  programs  shall  not  be  used,
rewritten,  or  adapted  as  the  basis  of a  commercial software  or
hardware  product without  first  obtaining  appropriate licenses from
MIT.  MIT.  makes no  representations about  the  suitability of  this
software for any purpose.  It is  provided "as is" without  express or
implied warranty.

******************************************************************/


/***************************************************************
*** File: binops.c
*** Author: John Y. A. Wang
*** Date: January 15, 1993
*** Copyright 1993, Massachusetts Institute of Technology, All Rights Reserved.
***
*** Description: binops routines
***************************************************************/

/*****************************************************************/
#include "dynamo.h"
/*****************************************************************/

/********** IMPORTANT:  put copyright label in all the files ***********/
static  char *copyright="\
\n\n***********************************************\n\
Developed by John Y. A. Wang at the Media Laboratory, MIT, Cambridge,Massachusetts\n\
Copyright  1993 by the Massachusetts  Institute  of  Technology.  All rights reserved.\n\
***********************************************\n\n";


/*
  cc -O4 -fsingle -c binops.c 
  SUN:  cc -O4 -fsingle -c binops.c
  
  lint binops.c
*/
/*****************************************************************/

binary_operate_images(image_mgr,op_type)
  Image_Manager *image_mgr;
  int op_type;
  
{
  int imsize;
  Image_Cache *cache1;
  Image_Cache *cache2;
  Image_Cache *cache_res;
  float *res;
  int start1, start2, end1,end2;
  int i1,i2,r;
  int length;
  float *im1,*im2;

  extern char *args;
/*  char *yn=image_mgr->yn; */
  char *image_name=image_mgr->image_name;
  
  printf("Binary operate two images\n");
  
  printf("Image 1 : ");
  if((cache1=ask_name_get_cache(image_mgr))==NULL)
    return(0);

  start1=cache1->ext_start; end1=cache1->ext_start+cache1->tdim-1;
  printf("Start Frame of Image 1 : (%d %d) >> ", start1,end1);
  readline(args,ARGSIZE,stdin);;
  sscanf(args,"%d %d",&start1,&end1);

  printf("Image 2 : ");
  if((cache2=ask_name_get_cache(image_mgr))==NULL)
    return(0);

  start2=cache2->ext_start; end2=cache2->ext_start+cache2->tdim-1;
  printf("Start and End Frames of Image 2 : (%d %d) >> ", start2,end2);
  readline(args,ARGSIZE,stdin);;
  sscanf(args,"%d %d",&start2,&end2);

  i1=end1-start1+1; i2=end1-start1+1;
  length=((i1>i2)?i1:i2);

  if(check_image_size(cache1,cache2)!=0) {
    printf("images for binary operation have different sizes\n");
    return(-1);
    }
  
  sprintf(image_name,"binops-res");
  printf("Result Image : ");
  if((cache_res=make_internal_image(image_mgr,cache1->xdim,cache1->ydim,length))==NULL) {
    printf("Cannot allocate memory for result image\n"); 
    return(-1);
  }

  imsize=cache_res->imsize;
  
  printf("start1=%d,end1=%d : start1=%d,end1=%d \n", start1,end1,start2,end2);
  for(i1=start1,i2=start2,r=0; (i1<=end1) || (i2<=end2); i1++,i2++,r++) {
    if(i1>end1) i1=end1;
    if(i2>end2) i2=end2;
/*    printf("(%d;%d;%d) ",i1,i2,r);*/
    printf("%d ",i1);fflush(stdout);
    if((im1=get_image(cache1,i1))==NULL) {
      printf("subtract: cannot get frame %d for Image 1\n",i1);
      return(-1);
    }
    if((im2=get_image(cache2,i2))==NULL) {
      printf("subtract: cannot get frame %d for Image 2\n",i2);
      return(-1);
    }
    if((res=get_image(cache_res,r))==NULL) {
      printf("subtract: cannot get frame %d for Result\n",r);
      return(-1);
    }
    switch(op_type) {
    case BIN_ADD_OPT:
      add_im(im1,im2,res,imsize); break;
    case BIN_SUB_OPT:
      sub_im(im1,im2,res,imsize); break;
    case BIN_MUL_OPT:
      mul_im(im1,im2,res,imsize); break;
    case BIN_DIV_OPT:
      div_im(im1,im2,res,imsize); break;
    default:
      printf("Operation type (%d) currently not supported\n",op_type);
      break;
    }
  }
  return(0);
}  

