/******************************************************************

"Copyright  1994 by the Massachusetts  Institute  of  Technology.  All
rights reserved.

Developed by John Y. A. Wang at the  Media Laboratory, MIT, Cambridge,
Massachusetts, with  support  from Secom Ltd., Goldstar Ltd.,  and the
Television of Tomorrow Program.

For use by authorized users only.

This distribution is approved  by Nicholas Negroponte, Director of the
Media Laboratory, MIT.

Permission to use, copy, or modify this software and its documentation
for educational and research purposes only and without  fee is  hereby
granted, provided that this copyright notice and the original authors'
names  appear  on  all  copies  and   supporting   documentation.   If
individual  files  are  separated  from  this  distribution  directory
structure, this copyright notice must be included.  For any other uses
of  this software, in  original or  modified form, including  but  not
limited to distribution in whole or in part, specific prior permission
must be  obtained  from  MIT.   These  programs  shall  not  be  used,
rewritten,  or  adapted  as  the  basis  of a  commercial software  or
hardware  product without  first  obtaining  appropriate licenses from
MIT.  MIT.  makes no  representations about  the  suitability of  this
software for any purpose.  It is  provided "as is" without  express or
implied warranty.

******************************************************************/


/***************************************************************
*** File: cache.c
*** Author: John Y. A. Wang
*** Date: January 15, 1993
*** Copyright 1993, Massachusetts Institute of Technology, All Rights Reserved.
***
*** Description: affine motion program from command line
***************************************************************/

/*****************************************************************************/
#include "dynamo.h"
/*****************************************************************************/

/********** IMPORTANT:  put copyright label in all the files ***********/
static  char *copyright="\
\n\n***********************************************\n\
Developed by John Y. A. Wang at the Media Laboratory, MIT, Cambridge,Massachusetts\n\
Copyright  1993 by the Massachusetts  Institute  of  Technology.  All rights reserved.\n\
***********************************************\n\n";


/**************  Macros, DEFS  *****************/
#define CACHE_NFRAMES   5

#define CACHE_NOT_MALLOC (-1)
#define CACHE_NOT_VALID 0
#define CACHE_VALID 1

#define IS_CACHE_VALID(cache) ((cache)->status==CACHE_VALID)
#define IS_CACHE_NOT_MALLOC(cache) ((cache)->status==CACHE_NOT_MALLOC)

#define SET_CACHE_VALID(cache) ((cache)->status=CACHE_VALID)
#define SET_CACHE_NOT_VALID(cache) ((cache)->status=CACHE_NOT_VALID)

#define LOAD_CACHE_EXIT(cache) {\
       printf("error in loading images into cache......\n");\
       if(fp!=NULL) {fclose(fp);} \
       SET_CACHE_NOT_VALID(cache); return(-1); }
			   
#define MALLOC_CACHE_EXIT(cache) {\
       printf("error in allocating cache memory......\n");\
       SET_CACHE_NOT_VALID(cache); return(-1); }


/*******************************************************/
int save_image(image_mgr)
  Image_Manager *image_mgr;
{
  extern char *args;

  char *dirname=image_mgr->cbuf1;
  char *save_filename=image_mgr->cbuf2;
  char *yn=image_mgr->yn;
  FILE *fp=NULL;

  Image_Cache *cache;

  int start=0, end=0;
  int ndigits=4;
  float d_clipmin,d_clipmax;

  unsigned char *array;
  int axdim, aydim, asize, num;
  int xs,ys,i;

  float pedestal,scale;

  char *progn="save_image";

  float *image;
  /*********************/
  if((cache=ask_name_get_cache(image_mgr))==NULL)
    return(-1);

  start=cache->ext_start;
  end=cache->tdim-1;

  printf("\tEnter Start: (%d) : ", start);
  readline(args,ARGSIZE,stdin);
  if(strlen(args)!=0)
    sscanf(args,"%d",&start);
  
  printf("\tEnter End: (%d) : ", end);
  readline(args,ARGSIZE,stdin);
  if(strlen(args)!=0)
    sscanf(args,"%d",&end);

  d_clipmin=0;d_clipmax=255;
  printf("\tEnter Min Max: (%6.2f %6.2f) : ", d_clipmin,d_clipmax);
  readline(args,ARGSIZE,stdin);
  sscanf(args,"%f %f",&d_clipmin,&d_clipmax);

  xs=0;axdim=cache->xdim;
  ys=0;aydim=cache->ydim;

  printf("\tCrop Region (xs ys xdim ydim) : (%d %d %d %d) : ",xs,ys,axdim,aydim);
  readline(args,ARGSIZE,stdin);
  if(strlen(args)!=0) { sscanf(args,"%d %d %d %d", &xs,&ys,&axdim,&aydim); }
  sprintf(yn,"n");
  printf("\tDisplay numbers on image? (%s) : ", yn);
  readline(args,ARGSIZE,stdin);
  if(strlen(args)!=0) { sscanf(args,"%s",yn);}

  /*********************/
  if(start<cache->ext_start) start=cache->ext_start;
  if(end>=(cache->tdim)) { end=(cache->tdim)-1; }
  
  sprintf(args,"%d",cache->tdim); ndigits=strlen(args);
/*  printf("ndigits=%d\n",ndigits); */

  sprintf(dirname,"%s-%s",cache->savedirname,"save");
  if(save_dir(dirname)==0) {
    sprintf(save_filename,"%s/data",dirname);
    fp=NULL;
    if((fp=open_file(save_filename,"w"))==NULL) {
      printf("%s: cannot open file %s. Not saving \"%s\".\n",
	     progn, save_filename, cache->image_name); return(-1); }
  }
  else { return(-1); }
  
  pedestal=d_clipmin;
  scale=(float)(255/(d_clipmax-d_clipmin));

  asize=axdim*aydim;

  if((array=(unsigned char *)malloc((unsigned) asize))==NULL){
    printf("%s: malloc cannot allocate %d bytes for array buffer \n",
	   progn,asize);  free((char *)array);  return(-1);
  }
  
  /********** image loop **********/
  for(i=start;i<=end;i++) {
    printf("%d ",i); fflush(stdout);
    if((image=get_image(cache,i))==NULL) {
      printf("Save: cannot find frame %d.... \n",i);
      fclose(fp); free((char *)array); return(-1);
    }

    convert_subfloat_to_byte(image,cache->xdim,cache->ydim,
			     xs,ys,xs+axdim-1,ys+aydim-1,
			     scale,pedestal,
			     array,axdim,aydim);
    if(strncmp(yn,"y",1)==0)
       display_number_on_image(array,axdim,aydim,
			       (float) i,(int)5,(int)5,
			       (int)255,(int)0,(int)ndigits);
    if((num=fwrite((char *)array,1,asize,fp))!=asize) {
      printf("%s: frame %d, error written %d bytes of %d bytes\n",
	    progn,i,num,asize); free((char *)array); return(-1);
    }
  }

  sprintf(args,"image from %s. start=%d, end=%d, crop(xs ys xdim ydim)=(%d %d %d %d)",
	  cache->filename,start,end,xs+cache->xcstart,ys+cache->ycstart,axdim,aydim);
  sprintf(save_filename,"%s/descriptor",dirname);
  save_descriptor(save_filename,axdim,aydim,(int)(end-start+1),(int) 1,(int)1,
		  (float)(1/scale),(float)pedestal,args);
  fclose(fp); free((char *) array);
  sprintf(cache->savedirname,"%s",dirname);
  return(0);
}

/*****************************************************************/
/* clear all caches and start from fresh */
int clear_all_memory(image_mgr)
  Image_Manager *image_mgr;
  
{
  char *yn=image_mgr->yn;
  extern char *args;

  sprintf(yn,"n");
  printf("\tProceed to clear all memory? (%s) : ", yn);
  readline(args,ARGSIZE,stdin);
  if(strlen(args)!=0) { sscanf(args,"%s",yn); }
  
  if(strncmp(yn,"y",1)==0) { init_image_manager(image_mgr); }
  return(0);
}

/*********************************/
int display_cache(cache)
  Image_Cache *cache;
{
  printf("\n\tCACHE STATE\n");
  printf("start=%d,end=%d,nframes=%d,status=%d,prev_frame=%d\n",
	 cache->start,cache->end,cache->nframes,cache->status,cache->prev_frame);
  printf("xdim=%d,ydim=%d,tdim=%d,imsize=%d,filetype=%d\n",
	 cache->xdim,cache->ydim,cache->tdim,cache->imsize,cache->filetype);
  printf("file_xdim=%d,file_ydim=%d file_imsize=%d, xcstart=%d, ycstart=%d\n",
	 cache->file_xdim,cache->file_ydim,cache->file_imsize,
	 cache->xcstart,cache->ycstart);
  printf("filename=%s\n",cache->filename);
  printf("datfilefile_ext=%s\n",cache->datafile_ext);
  printf("images=%d,array=%d,param=%d\n",
	 (int)cache->images,(int)cache->array,(int)cache->param);
  printf("\n\n");
  return(0);
}

/*******************************************************/
/*******************************************************/
int display_image_list(image_mgr)
  Image_Manager *image_mgr;
{
  int i,n=0;
  Image_Cache *cache;
  char *progn="dislay_image_list";
  
/*  printf("%s: ...\n",progn); */

  printf("\n\n\tMemory Status\n");
  printf(    "\t___________________\n\n");
  for(i=0,n=0;i<(image_mgr->n_images);i++) {
    cache=image_mgr->image_list[i];
    if(IS_CACHE_VALID(cache)) {
      printf("%2d:: \"%s\": xdim=%4d, ydim=%4d, tdim=%d, cache_size=%d\n",
	     i,cache->image_name,cache->xdim,cache->ydim,cache->tdim,
	     cache->nframes);
      printf("\tdata=%s\n", cache->filename);
/*      display_cache(cache); */
      printf("\n");
      n++;
    }
  }
  if(n==0) { printf("******* Currently no data in memory *******\n"); }
  printf(    "\t___________________\n\n");
  return(0);
}

/*******************************************************/
int load_image(image_mgr)
  Image_Manager *image_mgr;
{
  extern char *args;
  extern int untitled_num;
  
  Image_Cache *cache;
  char *progn="load_image";
  int i;
  char *yn=image_mgr->yn;
  char *image_name=image_mgr->image_name;
  
/*  printf("%s: ...\n",progn); */

  cache=NULL;

#define UNTITLED_NAME "untitled"
  sprintf(image_name,"%s_%d",UNTITLED_NAME,untitled_num);
  if((cache=ask_name_get_cache(image_mgr))!=NULL) {
    sprintf(yn,"n");
    printf("Image \"%s \"currently exists. Flush old results(y/n)? (%s) >> ",image_name,yn);
    readline(args,ARGSIZE,stdin);
    sscanf(args,"%s",yn);
      
    if(strncmp(yn,"y",1)!=0) {
      printf("%s: Ignoring load-image command \n",progn); return(0); }
    else {
      printf("%s: replacing existing image with new image\n",progn);
      free_cache_memory(cache); }
  }
  else { /* new image name */
    printf("%s: Creating  a new image \"%s\".\n",progn,image_name);
  }
  
  if(cache==NULL){
    for(i=0;i<MAXIMAGES;i++) {
      if(IS_CACHE_NOT_MALLOC(image_mgr->image_list[i])) {
	cache=image_mgr->image_list[i];	image_mgr->n_images++;
/*	printf("%s: using and unused image slot\n",progn); */
	break;
      }}
  }

  if(cache==NULL) {
    printf("%s: Error!! Cannot insert image cache into image_mgr\n",progn);
    return(-1);
  }

/*  printf("%s:image_name=\"%s\", strlen=%d\n",progn,image_name,strlen(image_name)); */
  
  printf("(%d)datafile or (%d)dataset format? : (%d)>> ",
	 (int)DATAFILE, (int)DATASET,cache->filetype);
  readline(args,ARGSIZE,stdin); sscanf(args,"%d",&cache->filetype);


  switch(cache->filetype) {
  case DATAFILE: case DATASET:
    break;
  default:
    printf("%s: datatype %d not supported\n",progn,cache->filetype);
    return(-1);
  }
  
  printf("Enter data file: (%s) >> ",cache->filename);
  readline(args,ARGSIZE,stdin);
  if(strlen(args)!=0) { sscanf(args,"%s",cache->filename); }
  printf("%s\n", cache->filename);
  if(strlen(cache->filename)==0) { init_cache(cache);  return(-1); }

  if(stat(cache->filename,&cache->file_stat)==0) {
    switch(cache->filetype) {
    case DATAFILE:
      printf("datafile: ");
      if(!(S_ISDIR(cache->file_stat.st_mode))) {
	printf("%s: %s is not a valid datafile directory.\n", progn,cache->filename);
	init_cache(cache); return(-1); }

      printf("Enter File Extension Name: (%s) >> ",cache->datafile_ext);
      readline(args,ARGSIZE,stdin); sscanf(args,"%s",cache->datafile_ext);
      printf("%s\n",cache->datafile_ext);
      
      printf("Enter Start Frame of Extension: (%d) >> ",cache->ext_start);
      readline(args,ARGSIZE,stdin); sscanf(args,"%d",&cache->ext_start);
      printf("%d\n",cache->ext_start);

      break;
    case DATASET:
      printf("dataset: ");
      if((S_ISREG(cache->file_stat.st_mode))==0) {
	printf("%s: %s is not a valid dataset raw file.\n", progn,cache->filename);
	init_cache(cache); return(-1); }
      printf("main: dataset has %d bytes\n", cache->file_stat.st_size);
    } 
  }
  else {
    printf("%s: cannot find %s\n",progn,cache->filename);
    init_cache(cache); return(-1); }
  
  sprintf(cache->image_name,"%s",image_name);  /* copy name into cache */
  
  printf("Enter xdim: (%d) >> ",cache->file_xdim);
  readline(args,ARGSIZE,stdin); sscanf(args,"%d",&cache->file_xdim);
  printf("%d\n",cache->file_xdim);

  printf("Enter ydim: (%d) >> ",cache->file_ydim);
  readline(args,ARGSIZE,stdin); sscanf(args,"%d",&cache->file_ydim);
  printf("%d\n",cache->file_ydim);

  cache->xdim=cache->file_xdim;
  cache->ydim=cache->file_ydim;

  printf("Enter image header size: (%d) >> ",cache->headersize);
  readline(args,ARGSIZE,stdin); sscanf(args,"%d",&cache->headersize);
  printf("%d\n",cache->headersize);
  
  printf("Enter crop input (xstart ystart xdim ydim) : (%d %d %d %d) >> ",
	 cache->xcstart,cache->ycstart,cache->xdim,cache->ydim);
  readline(args,ARGSIZE,stdin);
  sscanf(args,"%d %d %d %d",&cache->xcstart,&cache->ycstart,&cache->xdim,&cache->ydim);
  printf("%d %d %d %d\n",cache->xcstart,cache->ycstart,cache->xdim,cache->ydim);

  cache->imsize=(cache->xdim)*(cache->ydim);
  cache->file_imsize=(cache->file_xdim)*(cache->file_ydim);

  set_image_tdim(cache);
  
  if(cache->tdim<=0) {
    printf("\nInvalid dimensions (tdim=%d) for the sequence. No data.:\n %s\n\n"
	   ,cache->tdim,cache->filename);
    init_cache(cache); return(-1);
  }
  
  cache->nframes=((cache->tdim<cache->nframes)? cache->tdim : cache->nframes);
  printf("Enter number of frames for cache (%d) >> ", cache->nframes);
  readline(args,ARGSIZE,stdin);
  sscanf(args,"%d",&cache->nframes);

  cache->nframes=((cache->tdim<cache->nframes)? cache->tdim : cache->nframes);
  printf("Setting cache size to %d frames\n",cache->nframes);

  if(IS_CACHE_NOT_MALLOC(cache)) {
    if(malloc_cache(cache)== -1) {
      printf("Error: in allocating cache\n"); init_cache(cache); return(-1); }
  }
  else {
    printf("error: in caching. cache should be freed\n"); exit(-1);
  }

/*  display_cache(cache); */
  
  /* bypass the load_cache now. Let the get_image determine which frames
   * to grab from memory */

  if((cache->start >= 0) && (cache->end >= 0 )) {
    fprintf(stderr,"cache not properly initialized in start=%d end=%d \n",
	    cache->start,cache->end);
    cache->start=(int)(-99); cache->end=(int)(-100); }

  sprintf(cache->paramfilename, "%s.param", cache->filename);
  sprintf(cache->savedirname, "%s-stuff", cache->filename);
  set_default_analysis(cache);
  
  cache->status=CACHE_VALID;

  sprintf(args,"%s_%d",UNTITLED_NAME,untitled_num);
  if(strcmp(args,cache->image_name)==0) {
    untitled_num++;
  }
  return(0);
}

/***************************************************************/
change_image_cache_size(image_mgr)
  Image_Manager *image_mgr;
{
  extern char *args;

  Image_Cache *cache;
  char *progn="change_image_cache_size";
  int malloc_size;
  int old_cache_nframes;

  if((cache=ask_name_get_cache(image_mgr))==NULL) {
/*    printf("\"%d\" curently not in system\n"); */
    return(0);
  }

  old_cache_nframes=cache->nframes;
  printf("Enter number of frames for cache (%d) >> ",old_cache_nframes);
  readline(args,ARGSIZE,stdin);
  sscanf(args,"%d",&cache->nframes);
  cache->nframes=((cache->tdim<cache->nframes)? cache->tdim : cache->nframes);

  if(old_cache_nframes == cache->nframes) {
    printf("Cache is currently %d frames\n",cache->nframes); }
  else {
    printf("Setting cache size to %d frames\n",cache->nframes); }

  if(IS_CACHE_VALID(cache)) {
    cache->status=CACHE_NOT_VALID;

    free_float_seq(cache->images,old_cache_nframes);
    if((cache->images=(float **)malloc_float_seq(cache->nframes,cache->imsize))==NULL) {
      printf("%s: malloc cannot allocate cache images nframes=%d.\n",progn, cache->nframes);
      free_cache_memory(cache); return(-1); }
  }
  else {
    printf("Error: cannot modify an invalid cache\n");
    return(-1);
  }
  
  /* bypass the load_cache now. Let the get_image determine which frames
   * to grab from memory */

  /* start end in cache to invalid frame numbers so that get_image
   * will have to load new set of images. */
  cache->start=(int)(-99); cache->end=(int)(-100);

  cache->status=CACHE_VALID;
  return(0);
}

/*******************************************************/
delete_image(image_mgr)
  Image_Manager *image_mgr;
{
  extern char *args;

  Image_Cache *cache;
  char image_name[256];
  char *progn="delete_image";
  
/*  printf("%s: ...\n",progn); */

  sprintf(image_name,"");
  printf("Enter Delete Image Name >> ");
  readline(args,ARGSIZE,stdin);
  sscanf(args,"%s",image_name);
  if(strlen(image_name)==0) {  /* no input means to quit */
    printf("%s: No image name. No sequences removed.\n",progn); return(0); }

  if((cache=find_cache_name(image_mgr,image_name))!=NULL) {
    init_cache(cache);
    printf("%s: image \"%s\" removed!!\n",progn,image_name); 
  }
  else { printf("%s: image not found\n"); }
  return(0);
}

/*******************************************************/
/*******************************************************/
/*********** Cache interal routines ********************/
/*******************************************************/
Image_Cache *find_cache_name(image_mgr,image_name)
  Image_Manager *image_mgr;
  char *image_name;
{
  int i;
  int len=strlen(image_name);
  char *progn="find_cache_name";
  
/*  printf("%s: ...\n",progn); */

  for(i=0;i<MAXIMAGES;i++) {
    if(strncmp(image_mgr->image_list[i]->image_name,image_name,len)==0) {
      return(image_mgr->image_list[i]);
    }}
  return(NULL);
}
  
/**********************************************************/
init_image_manager(image_mgr)
  Image_Manager *image_mgr;
{
  extern int untitled_num;
  int i;
  char *progn="init_image_manager";

/*  printf("%s: ...\n",progn); */
  for(i=0;i<MAXIMAGES;i++){
    init_cache(image_mgr->image_list[i]);
  }
  untitled_num=0;
}

/**********************************************************/
clear_cache(cache)
  Image_Cache *cache;
{
  char *progn="clear_cache";
  
/*  fprintf(stderr,"%s: ...",progn); */
  cache->images=NULL;
  cache->array=NULL;
  cache->file_array=NULL;
  cache->param=NULL;
}

/**********************************************************/
/* init_cache essentially clears all the cache memory and deletes
**  from the list of image.
*/
init_cache(cache)
  Image_Cache *cache;
{
  char *progn="init_cache";
  
/*  printf("%s: ...",progn); */
  /******* init cache **********/
  if(cache==NULL)  return(-1);
  
  free_cache_memory(cache);
  cache->status=CACHE_NOT_MALLOC;
  cache->nframes=CACHE_NFRAMES;
  cache->start=(int)(-99); cache->end=(int)(-100);
  cache->load_offset=0; /* frame is either start or end when loading */
  cache->prev_frame=0;
  cache->tdim=10000;
  cache->filetype=DATAFILE;
  cache->gamma=0.5;
  cache->headersize=0;
  cache->images=NULL;
  sprintf(cache->image_name,"");
  sprintf(cache->filename,"");
  
  sprintf(cache->datafile_ext,"data");
  cache->ext_start=0;
  
  cache->xdim= 0; cache->ydim= 0;

  cache->file_xdim=0; cache->file_ydim=0;
  cache->xcstart=0; cache->ycstart=0;
  cache->file_imsize=0;
  /*****************************/
}


/*******************************************************/
Image_Cache *ask_name_get_cache(image_mgr)
  Image_Manager *image_mgr;
{
  extern char *args;

  Image_Cache *cache;

  char *image_name=image_mgr->image_name;
  char *progn="ask_name_get_cache";

  printf("Enter Image Name: (%s) >> ",image_name);
  readline(args,ARGSIZE,stdin);
/*  printf("%s:strlen(args)=%d\n",progn,strlen(args)); */
  sscanf(args,"%s",image_name);

  if(strlen(image_name)==0) {
    printf("%s: Load image ignored\n",progn); return(NULL); }
  if((cache=find_cache_name(image_mgr,image_name))==NULL) {
/*    printf("\"%s\" not currently loaded.\n",image_name); */
  }
  return(cache);
}

/*******************************************************/
free_all_memory(image_mgr)
  Image_Manager *image_mgr;
{
  int i;

  for(i=0;i<MAXIMAGES;i++)
    free_cache_memory(image_mgr->image_list[i]);
}


/**************************************************/
malloc_cache(cache)
  Image_Cache *cache;
{
  int malloc_size;
  char *progn="malloc_cache";
  
/*   printf("allocating  memory for cache....");*/
  /**************** malloc cache *************************/
  malloc_size=(unsigned)1*cache->imsize*sizeof(unsigned char);
  if((cache->array=(unsigned char *)malloc((unsigned)malloc_size))==NULL) {
    printf("%s: malloc cannot allocate %d bytes for image cache->\n",
	   progn,malloc_size);
    free_cache_memory(cache); return(-1); }

  malloc_size=(unsigned)1*cache->file_imsize*sizeof(unsigned char);

  if(malloc_size>0) {
    if((cache->file_array=(unsigned char *)malloc((unsigned)malloc_size))
       ==NULL) {
      printf("%s: malloc cannot allocate %d bytes for image cache->\n",
	     progn,malloc_size);
      free_cache_memory(cache); return(-1); }}
  else { cache->file_array=NULL; }
     
  malloc_size=(unsigned)(cache->tdim)*PARAM_WIDTH*sizeof(float);
  if((cache->param=(float *)malloc((unsigned)malloc_size))==NULL) { 
    printf("%s: malloc cannot allocate %d bytes for param array\n",
	   progn,malloc_size);
    free_cache_memory(cache); return(-1); }


  if((cache->images=(float **)malloc_float_seq(cache->nframes,cache->imsize))==NULL) {
    printf("%s: malloc cannot allocate cache images nframes=%d.\n",
	   progn, cache->nframes);
    free_cache_memory(cache); return(-1); }

/*  printf("done!\n");*/
  return(0);
}


/*****************************************************************/
free_cache_memory(cache)
  Image_Cache *cache;
{
  if(cache->images!=NULL) {
    free_float_seq(cache->images,cache->nframes);
    cache->images=NULL;}

  if(cache->array!=NULL) {free((char *)cache->array);cache->array=NULL;}
  if(cache->file_array!=NULL) {free((char *)cache->file_array);cache->file_array=NULL;}
  if(cache->param!=NULL) {free((char *)cache->param);cache->param=NULL;}
  cache->status=CACHE_NOT_MALLOC;
}

/*****************************************************************/
set_image_tdim(cache)
  Image_Cache *cache;
{
  int i;
  FILE *fp;
  char datafilename[256];
  char *progn="set_image_tdim";
  
  switch(cache->filetype) {
  case DATAFILE:
    /********************* start reading datafiles from the start frame **********/
/*    printf("%s: Assuming datafile format->.....\n",progn); */
    for(i=cache->ext_start;;i++){
/*      printf("%d ",i); fflush(stdout); */
      /* make datafile name */
      sprintf(datafilename,"%s/%s%d",cache->filename,cache->datafile_ext,i);
      if((fp=fopen(datafilename,"r"))==NULL) {
/*	printf("%s: fopen cannot read datafile %s\n",progn,datafilename);  */
	if(i > 0){
/*
          printf("i=%d,cache->ext_start=%d\n",i,cache->ext_start);
	  printf("datafilename=%s\n",datafilename);
*/
	  cache->tdim=i;
	  printf("%s: Datafile has %d frames.\n",progn, cache->tdim);
	  fclose(fp); return(0);
	}
	else {  /* should be i==0 here */
	  cache->tdim=0;
	  sprintf(datafilename,"%s/%s",cache->filename,cache->datafile_ext);
	  if((fp=fopen(datafilename,"r"))!=NULL) {
	    cache->tdim=1;
	  }
	  fclose(fp); return(0);
	}
      }
      fclose(fp);
    }
    break;
  case DATASET:
/*    printf("%s: Assuming dataset format......\n",progn);*/
    if(stat(cache->filename,&cache->file_stat)==0) {
      cache->tdim=(int)((float)(cache->file_stat.st_size)/((float)cache->file_imsize));
      printf("%s: Datasetfile has %d frames\n",progn,cache->tdim); return(0); }
    else {
      printf("%s: Invalid raw file %s",progn,cache->filename); return(-1); }
    break;
  default:
    printf("set_image_tdim: Unknown file type \n");
    return(-1);
    break;
  }   /* end switch */
  return(0);
}  


/*******************************************************/
Image_Cache *make_internal_image(image_mgr,xdim,ydim,length)
  Image_Manager *image_mgr;
  int xdim,ydim,length;
{
  extern char *args;

  Image_Cache *cache;
  char *progn="make_internal_image";
  int i;
  char *yn=image_mgr->yn;
  char *image_name=image_mgr->image_name;
  
/*  fprintf(stderr,"In %s: ...\n",progn); */

  cache=NULL;

/*
  printf("make_internal_image: xdim=%d, ydim=%d, length=%d\n...",xdim,ydim,length);
*/

#ifdef VERIFY_DESTROY_OLD
  while((cache=ask_name_get_cache(image_mgr))!=NULL) {
    printf("Image \"%s\" currently exists. \n",image_name);

    sprintf(yn,"n");
    printf("\tDo you want to use current image (y/n/q)? (%s) >> ",yn);
    readline(args,ARGSIZE,stdin);  sscanf(args,"%s",yn);
    if(strncmp(yn,"y",1)==0) { /* reuse the existing image */
      /* if incompatible image dimensions clear cache first then realloc */
      if ((cache->xdim!=xdim)||(cache->ydim!=ydim)||
	  (cache->nframes!=length)) { 
/*	fprintf(stderr,"images have incompatible dimensions...\n"); */
	init_cache(cache); }
/*      else { fprintf(stderr,"using existing cache memory\n");} */
      break; }
    else if(strncmp(yn,"n",1)==0) { /* ask for another image name. */
      continue; }
    else if(strncmp(yn,"q",1)==0) {  /* ignore make command */
      printf("%s: Ignoring make-internal-image command \n",progn);
      return(NULL);
    }
  }
#else      /* don't confirm image distruction of internal images. */
  if((cache=ask_name_get_cache(image_mgr))!=NULL) {
    /* if incompatible image dimensions clear cache first then realloc */
    if ((cache->xdim!=xdim)||(cache->ydim!=ydim)||
	(cache->nframes!=length)) { 
      init_cache(cache); }
  }
#endif

  if(cache==NULL){
    /* new image name */
    printf("Creating  a new image \"%s\"....\n",image_name);

    for(i=0;i<MAXIMAGES;i++) {
      if(IS_CACHE_NOT_MALLOC(image_mgr->image_list[i])) {
	cache=image_mgr->image_list[i];
	image_mgr->n_images++;
/*	printf("%s: using an unused image slot\n",progn); */
	init_cache(cache);
	break;
      }}
  }

  if(cache==NULL) { /* if cache is still NULL. not enough image cache slots */
    printf("%s: Cannot insert image cache into image_mgr.\n",progn);
    printf("%s: Increase MAXIMAGES to allow for more images.\n",progn);
    init_cache(cache);
    return(NULL);
  }

  cache->filetype=INTERNAL_IMAGE;
  sprintf(cache->filename,"*internal_image*");
  sprintf(cache->datafile_ext,"none");
  cache->ext_start=0;
  sprintf(cache->image_name,"%s",image_name);
  cache->xdim=xdim; cache->ydim=ydim;
  cache->imsize=(cache->xdim)*(cache->ydim);
  cache->file_imsize=0;
  cache->tdim=length;
  cache->nframes=cache->tdim;
  cache->start=0;
  cache->end=cache->tdim-1;
  
  if(IS_CACHE_NOT_MALLOC(cache)) {
/*    fprintf(stderr,"allocating cache memory... \n"); */
    if(malloc_cache(cache)== -1) {
      printf("Internal image not created. Cannot allocate memory......\n");
      init_cache(cache);
      return(NULL);}
  }
  else {
    fprintf(stderr,"Using existing image memory....\n"); 
  }

  sprintf(cache->paramfilename, "%s.param", cache->filename);
  sprintf(cache->savedirname, "%s-stuff", cache->filename);

  set_default_analysis(cache);
  
  cache->status=CACHE_VALID;
/*  display_cache(cache); */
  return(cache);
}

/*******************************************************/
check_image_size(cache1,cache2)
  Image_Cache *cache1,*cache2;
{
  if((cache1->xdim==cache2->xdim) &&
     (cache1->ydim==cache2->ydim))
    return(0);
  else
    return(-1);
}
      

/**************************************************/
/** the following is a new version of load_cache() */
/**************************************************/
load_cache(cache,start,end)   /* new version */
  int start,end;
  Image_Cache *cache;
{
  return(load_cache_disk(cache,start,end,NULL));
}

/**************************************************/
/* this is the lower level call to load_cache. when
 * res is not null, we simply want to retrieve a single
 * frame without disturbing the cache.
 */
 
load_cache_disk(cache,start,end,res)  
  int start,end;
  Image_Cache *cache;
  float *res;
{

  char datafilename[256];
  int nbytes;

  char *progn="load_cache_disk";
  
  FILE *fp=NULL;
  int file_imsize=cache->file_imsize;
  int imsize=cache->imsize;

  int i,j,k;
  int min_start;
  int max_end;
  int flag[1000];
  float *im_ptr[1000];
  
/*  fprintf(stderr,"In %s....\n",progn); */

  if(res==NULL) {
    if(cache->filetype==DATAFILE) { min_start=cache->ext_start; }
    else { min_start=0; }

    if(start < min_start) { start=min_start; }
  
    max_end=start+(cache->nframes)-1;
    if(max_end >= (cache->tdim)) { start=(cache->tdim)-(cache->nframes); }

    if(start < min_start) { start=min_start; }
  
    end=start+cache->nframes-1;
    if(end >= cache->tdim) {
      printf("Warning: Cache size is too large.\n");
      printf("Sequence has only %d frames but caching %d frames.\n",
	     cache->tdim,cache->nframes);
      end=cache->tdim-1;
    }
/*    printf("current frames in cache=[%d %d]\n",cache->start,cache->end);  */

#define USED 1
#define NOTUSED 0
    for(i=0;i<1000;i++) { flag[i]=NOTUSED; im_ptr[i]=NULL; }

    /* first pass find the images which will be used */
    for(i=start,j=0;i<=end;i++,j++) {
      im_ptr[j]=cache->images[j];  /* save the pointers to the images */
      if(i>=cache->start && i<=cache->end) {
/*	printf("frame %d currently in cache\n",i); */
	flag[i-cache->start]=(int)USED;    /* tag image used */
      }}

    /* adjust the memory pointers */
    for(i=start,j=0,k=0;i<=end;i++,j++) {
      if(i>=cache->start && i<=cache->end) {
	cache->images[j]=im_ptr[(int)(i-cache->start)];
	im_ptr[(int)(i-cache->start)]=NULL;
      }
      else { /* find the first unused image memory with varible k */
	for(;flag[k]==(int)USED && k<1000;k++);
	cache->images[j]=im_ptr[k]; /* first unused slot */
	flag[k]=(int)USED; im_ptr[k]=NULL;
      }}
    
#define LOADED 1
#define NOTLOADED 0
    /* set the flags */
    for(i=start,j=0;i<=end;i++,j++) {
      if(i>=cache->start && i<=cache->end) { flag[j]=(int)LOADED; }
      else { flag[j]=(int)NOTLOADED; }
    }
  }   /* end of when res is null */
  else  {
    if(start<cache->ext_start || start>=cache->tdim) {
      return(-1); }
/*    printf("loading a single frame #%d from disk into res ... \n",start); */
    end=start;   /* set end to start so that loop only once. */
  }

/*  printf("loading cache with frames %d to %d ....\n",start,end);  */

  /************************************/
  switch(cache->filetype) {
  case DATAFILE:
    /*********** start reading datafiles from the start frame **********/
/*    printf("Using datafile format(file_imsize=%d)......\n",file_imsize);*/
    for(i=start,j=0;i<=end;i++,j++){
/*       fprintf(stderr,"*%d ",i); */
      if(flag[j] == LOADED) {
/*	fprintf(stderr,"frame %d already in memory. skipping read... \n",i); */
	continue; }
      /* make datafile name */
      if(cache->tdim==1) {
	sprintf(datafilename,"%s/%s",cache->filename,cache->datafile_ext); }
      else {
	sprintf(datafilename,
		"%s/%s%d",cache->filename,cache->datafile_ext,i); }

      if((fp=fopen(datafilename,"r"))==NULL) {
	printf("%s: fopen cannot read datafile %s\n",progn,datafilename);
	if(i >= 0){
	  cache->tdim=i;
	  printf("%s: Error .... setting max allow frames to %d.\n",
		 progn,cache->tdim); }
	LOAD_CACHE_EXIT(cache);
      }
/*      printf("%s: open file done\n",progn); */
      /* skip through header info */
      if((fseek(fp,(long)(cache->headersize*sizeof(char)),SEEK_SET))== -1) {
	printf("%s: fseek error in reading header\n",progn);
	LOAD_CACHE_EXIT(cache);
      }
      if((nbytes=fread((char *)cache->file_array,sizeof(unsigned char),
		       file_imsize,fp)) !=file_imsize) {
	printf("%s: fread read %d of %d bytes of frame #%d.\n",
	       progn,nbytes,file_imsize,i);
	LOAD_CACHE_EXIT(cache);
      }
      fclose(fp); fp=NULL;

      convert_subarray_byte_to_float
	(cache->file_array,cache->file_xdim,cache->file_ydim,
	 cache->xcstart,cache->ycstart,
	 (res!=NULL? res: (cache->images)[j]),
	 cache->xdim,cache->ydim);
      if(res!=NULL) { return(0); }  /* get one image and return */
    }
    break;

  case DATASET:
/*    printf("Using dataset format......\n");*/
    /********************* start reading dataset from the start frame **********/
    sprintf(datafilename,"%s",cache->filename);
    if((fp=fopen(datafilename,"r"))==NULL) {
      printf("%s: fopen cannot read %s\n",progn,datafilename);
      LOAD_CACHE_EXIT(cache);
    }
    /* skip through header info */
    if((fseek(fp,(long)(cache->headersize*sizeof(char)),SEEK_SET))== -1) {
      printf("%s: fseek error in reading header\n",progn);
      LOAD_CACHE_EXIT(cache);
    }
    if((fseek(fp,(long)(start*file_imsize*sizeof(char)),SEEK_CUR))== -1) {
      printf("%s: fseek error in searching for frame #%d\n",progn,start);
      LOAD_CACHE_EXIT(cache);
    }
    for(i=start,j=0;i<=end;i++,j++){
/*       fprintf(stderr,"*%d ",i); */
      if(flag[j] == LOADED) {
/*	fprintf(stderr,"frame %d already in memory. skipping read... \n",i); */
	if(fseek(fp,(long)(imsize*sizeof(char)),SEEK_CUR) == -1) {
	  printf("%s: fseek error in skipping over loaded frame %d \n",
		 progn,i);
	}
	else { continue; }
      }

      if((nbytes=fread((char *)cache->file_array,sizeof(unsigned char),
		       file_imsize,fp)) !=file_imsize) {
	printf("%s: fread; read %d of %d bytes of frame #%d.\n",
	       progn,nbytes,file_imsize,i);
	if(i >= 0){
	  cache->tdim=i;
	  printf("%s: Error .... setting maximum allow frames to %d.\n",
		 progn,cache->tdim); }
	LOAD_CACHE_EXIT(cache);
      }
      convert_subarray_byte_to_float
	(cache->file_array,cache->file_xdim,cache->file_ydim,
	 cache->xcstart,cache->ycstart,
	 (res!=NULL? res: (cache->images)[j]),
	 cache->xdim,cache->ydim);
      if(res!=NULL) { fclose(fp); return(0); }  /* get one image and return */
    }
    fclose(fp); fp=NULL;
    break;

  case INTERNAL_IMAGE:
    printf("%s: internal_image \"%s\" cache should not be loaded.\n",
	   progn,cache->image_name);
    break;
  default:
    printf("filetype %d not supported\n",cache->filetype);
    LOAD_CACHE_EXIT(cache);
  }

  if(res!=NULL) {
    fprintf(stderr,"%s: error, should not be here when res is not null\n");
  }

  /**** update cache variables *****/
  cache->start=start;
  cache->end=end;

  SET_CACHE_VALID(cache);
  
/*  printf("done!\n");*/
  return(0);
}  


/*****************************************************************/
float *get_image(cache,frame_number) /* new version */
  Image_Cache *cache;
  int frame_number;
{
  return(get_image_disk(cache,frame_number,NULL));
}

/**************************************************/
/* This is the lower level call to get image. if res is
 * null, then we want to load up the cache. otherwise when
 * res is not null, we just want to load a single frame from 
 * disk into res */

float *get_image_disk(cache,frame_number,res) 
  Image_Cache *cache;
  int frame_number;
  float *res;  /* allow for loading image without changing cache. */
{

  int frame_offset;
  float *im;  

/*  printf("get_image_disk: %s frame %d \n",cache->image_name,frame_number); */

  if((frame_number >= (cache->tdim)) || frame_number < cache->ext_start) {
/*    printf("%s: Error... cannot grab frame %d\n",progn, frame_number); */
    return(NULL);
  }

  frame_offset=frame_number-(cache->start);
  if((frame_number >= (cache->start)) && (frame_number <= (cache->end))) {
    /* currently in cache */
    cache->prev_frame=frame_number;

    im=(cache->images)[frame_offset];
    if(res!=NULL) { copy_im(im,res,cache->imsize); return(res); }
    else { return(im); } /* return pointer to cache memory */
  }

  if(res!=NULL) {
    load_cache_disk(cache,frame_number,frame_number,res);
    return(res);
  }

  /****************************************************************/
  /* current image not in cache. load up cache with the set of images
   * then retrieve the image. */
 
  if((frame_number>=cache->prev_frame)&&
     (frame_number<=(cache->prev_frame+cache->nframes-1))) {
/*
    printf("cache: incr. branch....prev=%d cur=%d prev+nframes-1=%d : loading [%d %d]\n",
	   cache->prev_frame,frame_number,cache->prev_frame+cache->nframes-1,
	   cache->prev_frame,cache->prev_frame+cache->nframes-1);
*/
    load_cache(cache,cache->prev_frame,cache->prev_frame+cache->nframes-1);
  }
  else if ((frame_number <= cache->prev_frame) &&
	   (frame_number >= cache->prev_frame - (cache->nframes)+1)) {
/*
    printf("cache: decreasing?....cur=%d prev=%d : loading [%d %d]\n",
	   frame_number,cache->prev_frame,
	   cache->prev_frame-(cache->nframes)+1,cache->prev_frame);
*/
    load_cache(cache,cache->prev_frame-(cache->nframes)+1,cache->prev_frame);
  }
  else {
    {
      int new_start,new_end;
      int load_offset=cache->load_offset;   /* a new slot in cache */

      if(load_offset>=cache->nframes){ load_offset=cache->nframes-1; }

      if(frame_number > cache->end) { /* increasing */
	new_start=frame_number-load_offset;
	new_end=new_start+(cache->nframes-1);
      }
      else if(frame_number < cache->end) { /* decreasing */
	new_end=frame_number+(cache->nframes-1);
	new_start=new_end-(cache->nframes-1);
      }
      load_cache(cache,new_start,new_end);  /* load the cache */
    }
  }

  /* now get the image after loading the cache with the appropriate images */
  frame_offset=frame_number-(cache->start);
  if((frame_number >= (cache->start)) && (frame_number <= (cache->end))) {
    cache->prev_frame=frame_number;
    return((cache->images)[frame_offset]);
  }
  else {
    printf("get_image: Error in caching cannot get frame %d\n",frame_number);
    return(NULL);
  }
}


/*****************************************************************/
int free_float_seq(images,nframes)
  float **images;
  int nframes;
{

/*  fprintf(stderr,"In free_float_seq...nframes=%d\n",nframes); */
  
  if(images==NULL) { return(-1); }
  for(nframes--;nframes>=0;nframes--) {
    if(images[nframes]!=NULL) { free(images[nframes]); images[nframes]=NULL; }
  }

/*  fprintf(stderr,"freeing images\n"); */
  free(images);
/*  fprintf(stderr,"done free_float_seq\n"); */
  return(0);
}

/*****************************************************************/
/* allocates an array of pointers to float images and also allocates the
 * float image memory. */
float **malloc_float_seq(nframes,imsize)
  int nframes,imsize;
{
  float **ptr;
  int i;

  if((ptr=(float **)malloc(sizeof(float *)*nframes))==NULL) {
    fprintf(stderr,"%s: cannot malloc float sequence. nframes=%d\n",nframes);
    return(NULL); }
  
  for(i=0;i<nframes;i++) {
    if((ptr[i]=(float *)malloc(sizeof(float)*imsize))==NULL) {
      fprintf(stderr,"%s: cannot malloc float image. frame %d. imsize=%d\n",
	      nframes,imsize);
      return(NULL); }
  }
  return((float **)ptr);
}

