/******************************************************************

"Copyright  1994 by the Massachusetts  Institute  of  Technology.  All
rights reserved.

Developed by John Y. A. Wang at the  Media Laboratory, MIT, Cambridge,
Massachusetts, with  support  from Secom Ltd., Goldstar Ltd.,  and the
Television of Tomorrow Program.

For use by authorized users only.

This distribution is approved  by Nicholas Negroponte, Director of the
Media Laboratory, MIT.

Permission to use, copy, or modify this software and its documentation
for educational and research purposes only and without  fee is  hereby
granted, provided that this copyright notice and the original authors'
names  appear  on  all  copies  and   supporting   documentation.   If
individual  files  are  separated  from  this  distribution  directory
structure, this copyright notice must be included.  For any other uses
of  this software, in  original or  modified form, including  but  not
limited to distribution in whole or in part, specific prior permission
must be  obtained  from  MIT.   These  programs  shall  not  be  used,
rewritten,  or  adapted  as  the  basis  of a  commercial software  or
hardware  product without  first  obtaining  appropriate licenses from
MIT.  MIT.  makes no  representations about  the  suitability of  this
software for any purpose.  It is  provided "as is" without  express or
implied warranty.

******************************************************************/


/***************************************************************
*** File: color-index.c
*** Author: John Y. A. Wang
*** Date: 1992.
*** Copyright 1992, Massachusetts Institute of Technology, All Rights Reserved.
***
*** Description: generates a index color RGB image.
***************************************************************/

static  char *copyright="\
\n\n***********************************************\n\
File: color-index.c\n\
Author: John Y. A. Wang.\n\
The MIT Media Lab.\n\
Copyright (c) 1991, All Rights Reserved.\n\
***********************************************\n\n";

#include <stdio.h>
#include "motion.h"

/*
  cc -c -O4 -fsingle color-index.c
*/

#ifdef lint
/*
  lint 
*/
#define register
#endif


#define RED 0
#define GREEN 1
#define BLUE 2
#define COL 27

#define mod(y,b) ((int)((y)-(int)((float)(y)/(float)(b))*(float)(b)))
void internal_color_sort(/*list,n*/);


static int color[COL][3] = {{200,30,10},{10,130,5},{30,30,230},
			      {210,210,5},{5,180,220},{240,5,255},
			      {80,50,50},{170,255,255},{200,150,180},

			      {128,5,5},{2,70,30},{2,2,128},
			      {255,128,0},{10,200,90},{128,10,220},
			      {100,130,100},{255,220,200},{175,200,175},
			    
			      {235,0,130},{130,255,50},{80,230,255},
			      {240,80,30},{10,255,10},{80,40,255},
			      
			      {255,180,90},{90,255,180},{200,110,255}};


/** produces an indexed color RGB images base on the colormap.
 ** Uses defined STATIC color array  if colormap is NULL **/
void internal_index_color_order(image,r_im,g_im,b_im,im_size,z,numcolor,colormap,order)
  register float *image,*r_im,*g_im,*b_im;
  register int im_size;
  int z,numcolor;
  float *colormap;
  int order;
{
  register int j,k,i,imax,y;
  register float ind[1000];
  register float val;
  extern color[COL][3];
  int col;

  if(numcolor > 0)
    col=numcolor;
  else
    col=COL;

/*  printf("internal-index-color-order\n");fflush(stdout); */
  
  for(j=0;(j<im_size && (image[j]<0));j++)
          r_im[j]=g_im[j]=b_im[j]=0.0;
  ind[0]=image[j];
  for(imax=1;(j<im_size) && (imax<1000);j++) {
    val=image[j];
    if(val>=0) {
      for(k=0,y=1;(k<imax)&&y;k++)
	if(ind[k] == val ) { y=0;}
      if(y) {
	ind[imax]=val; imax++;
      }
    }
  }
 
  /* sort from lowest to highest */
  internal_color_sort (ind,imax);
/*
  printf("(");
  for(i=0;i<imax;i++)
    printf("%3.2f ",ind[i]);
  printf(")\n");fflush(stdout);
*/
/*  printf("numcolor=%d,colormap=%d\n",numcolor,colormap);fflush(stdout);*/
  /* set the color images */
  for(j=0;(j<im_size);j++) {
    val=image[j];
    if(val < 0) {
	r_im[j]=0; g_im[j]=0;b_im[j]=0;
      }
    else {
      for(k=0,y=1; (k<imax)&&y; k++) {
	if(ind[k]==val) {
	  if(order==1)
	    i=mod(k+z,col);
	  else
	    i=(int)mod(((int)val)+z,col);

	  if((numcolor==0)||(colormap==NULL)) {
	    r_im[j]=color[i][RED]; g_im[j]=color[i][GREEN]; b_im[j]=color[i][BLUE];
	  }
	  else {
	    i=i*3;
	    r_im[j]=colormap[i+RED]; g_im[j]=colormap[i+GREEN]; b_im[j]=colormap[i+BLUE];
	  }
	}
      }
    }
  }
}


void internal_color_sort(list,n)
int n;
float *list;
{
	int l,j,ir,i;
	float rra;
	float *ra;

	if((n==0) || (n==1))
	  return;
	ra = list - 1;
	l=(n >> 1)+1;
	ir=n;
	for (;;) {
		if (l > 1)
			rra=ra[--l];
		else {
			rra=ra[ir];
			ra[ir]=ra[1];
			if (--ir == 1) {
				ra[1]=rra;
				return;
			}
		}
		i=l;
		j=l << 1;
		while (j <= ir) {
			if (j < ir && ra[j] < ra[j+1]) ++j;
			if (rra < ra[j]) {
				ra[i]=ra[j];
				j += (i=j);
			}
			else j=ir+1;
		}
		ra[i]=rra;
	}
}

