/******************************************************************

"Copyright  1994 by the Massachusetts  Institute  of  Technology.  All
rights reserved.

Developed by John Y. A. Wang at the  Media Laboratory, MIT, Cambridge,
Massachusetts, with  support  from Secom Ltd., Goldstar Ltd.,  and the
Television of Tomorrow Program.

For use by authorized users only.

This distribution is approved  by Nicholas Negroponte, Director of the
Media Laboratory, MIT.

Permission to use, copy, or modify this software and its documentation
for educational and research purposes only and without  fee is  hereby
granted, provided that this copyright notice and the original authors'
names  appear  on  all  copies  and   supporting   documentation.   If
individual  files  are  separated  from  this  distribution  directory
structure, this copyright notice must be included.  For any other uses
of  this software, in  original or  modified form, including  but  not
limited to distribution in whole or in part, specific prior permission
must be  obtained  from  MIT.   These  programs  shall  not  be  used,
rewritten,  or  adapted  as  the  basis  of a  commercial software  or
hardware  product without  first  obtaining  appropriate licenses from
MIT.  MIT.  makes no  representations about  the  suitability of  this
software for any purpose.  It is  provided "as is" without  express or
implied warranty.

******************************************************************/


/***************************************************************
*** File: conversion.c
*** Author: John Y. A. Wang
*** Date: January 15, 1993
*** Copyright 1993, Massachusetts Institute of Technology, All Rights Reserved.
***
*** Description: conversion routines
***************************************************************/

static  char *copyright="\
\n\n***********************************************\n\
File: conversion.c\n\
Author: John Y. A. Wang.\n\
The MIT Media Lab.\n\
Copyright (c) 1993, All Rights Reserved.\n\
***********************************************\n\n";

#include "motion.h"

/****************************************************************************/
/** converts a byte array to a float array  assumes pedestal=0 scale=1. **/
/* destructive modification allowed */
void convert_byte_to_float(array,image,size)
  unsigned char *array;
  float *image;
  int size;
{
  int i;
  for(i=size-1;i>=0;i--) { image[i]=(float)((int)array[i]); }
}

/* converts a byte to a float according to pedestal  and scale */
/* destructive modification allowed */
void convert_byte_to_float2(array,image,size,pedestal,scale)
  unsigned char *array;
  float *image;
  float pedestal,scale;
  int size;
{
  int i;
  for(i=size-1;i>=0;i--) {
    image[i]=(scale*(float)((int)array[i])+pedestal); }
}


/*************************************************************************/
/* converts a float to a byte assumes pedestal=0 scale=1 */
/* destructive modification allowed */
void convert_float_to_byte(image,array,imsize)
  float *image;
  unsigned char *array;
  int imsize;
{
  int i,im;

  for(i=0;i<imsize;i++) {
/*    array[i] = (unsigned char)(image[i]); */
    im=(int)(image[i]+.5);  /* round off */
    if(im>255)
      array[i]=(unsigned char)255;
    else if(im<0)
      array[i]=(unsigned char)0;
    else
      array[i]=(unsigned char)((int)im);
  }
}

/**************************************************************************/
/* converts a sub-array to a float array */
void convert_subarray_byte_to_float(array,axdim,aydim,axstart,aystart,image,xdim,ydim)
  unsigned char *array;
  float *image;
  int axdim,aydim;
  int xdim,ydim;
  int axstart,aystart;
{
  int x,y,ipos,apos;
  int ax,ay;
  
  for(y=0,ay=aystart,ipos=0;y<ydim;y++,ay++) {
    for(x=0,ax=axstart,apos=ay*axdim+ax; x<xdim; x++,ax++,apos++,ipos++) {
      if(inbounds(ay,aydim) && inbounds(ax,axdim))
	image[ipos]=(float)((int)array[apos]);
      else
	image[ipos]=0.0;
    }
  }
}

/********************   ******************************************************/
/** converts a sub-float array to a byte array according to pedestal and scale. **/
void convert_subfloat_to_byte(image,xdim,ydim,xstart,ystart,xend,yend,
			 scale,pedestal,array,axdim,aydim)
  unsigned char *array;
  float *image;
  float pedestal,scale;   
  int axdim,aydim;
  int xdim,ydim;
  int xstart,ystart,xend,yend;
{
  int ix,iy,ipos,apos;
  int ax,ay;
  int im;

  
  if(xstart<0) xstart=0;   if(xend>=xdim) xend=xdim-1;
  if(ystart<0) ystart=0;   if(yend>=ydim) yend=ydim-1;

  for(ay=0,iy=ystart,apos=0;ay<aydim;iy++,ay++) {
    for(ax=0,ix=xstart,ipos=iy*xdim+ix; ax<axdim; ix++ ,ax++,apos++,ipos++) {
      if((iy>=ystart) && (iy<=yend) && (ix>=xstart) && (ix<=xend)) {
	im=(int)(scale*(image[ipos]-pedestal)+0.5); /* round off */

	if(im>255) { array[apos]=(unsigned char)255; }
	else if(im<0) { array[apos]=(unsigned char)0; }
	else { array[apos]=(unsigned char)((int)im); } 
      }
      else { array[apos]=(unsigned char)0;}
    }}
}

