/******************************************************************

"Copyright  1994 by the Massachusetts  Institute  of  Technology.  All
rights reserved.

Developed by John Y. A. Wang at the  Media Laboratory, MIT, Cambridge,
Massachusetts, with  support  from Secom Ltd., Goldstar Ltd.,  and the
Television of Tomorrow Program.

For use by authorized users only.

This distribution is approved  by Nicholas Negroponte, Director of the
Media Laboratory, MIT.

Permission to use, copy, or modify this software and its documentation
for educational and research purposes only and without  fee is  hereby
granted, provided that this copyright notice and the original authors'
names  appear  on  all  copies  and   supporting   documentation.   If
individual  files  are  separated  from  this  distribution  directory
structure, this copyright notice must be included.  For any other uses
of  this software, in  original or  modified form, including  but  not
limited to distribution in whole or in part, specific prior permission
must be  obtained  from  MIT.   These  programs  shall  not  be  used,
rewritten,  or  adapted  as  the  basis  of a  commercial software  or
hardware  product without  first  obtaining  appropriate licenses from
MIT.  MIT.  makes no  representations about  the  suitability of  this
software for any purpose.  It is  provided "as is" without  express or
implied warranty.

******************************************************************/


/**********************************************************************
*** File: display-color.c
*** Author: John Y. A. Wang
*** Date:   June 5, 1992
*** Copyright 1992 Masssachussetts Institute of Technology. All Rights Reserved.
***
*** Display color image and sequences represented as a sequence of 3 images:
*** R, G, and B images using RLE rendering commands.
**********************************************************************/

static  char *copyright="\
\n\n***********************************************\n\
File: display-color.c\n\
Author: John Y. A. Wang.\n\
The MIT Media Lab.\n\
Copyright (c) 1992, All Rights Reserved.\n\
***********************************************\n\n";

#include "motion.h"
#include <unistd.h>

#define RLE_DISPLAY

#ifdef NO_RLE_DISPLAY
#undef RLE_DISPLAY

static char *nodisplay="\n\n***** Image Display facilities currently not available *****\n\
Please modify \"display-color\" module to incorporate display routines at your site.\n\n";

#endif

/************************************************************/
/* displays an RGB image */
extern void internal_display_color_im (r,g,b,ysize,xsize,pedestal,scale,
				gamma,zoom,title,tempim,bufsize)
  float *r,*g,*b;
  int ysize,xsize;
  float pedestal, scale, gamma;
  float zoom;
  char *title;
  unsigned char *tempim;
  int bufsize;
{
#ifdef RLE_DISPLAY

  FILE *pid;    /* file pointers */
  int total_size = xsize*ysize;
  unsigned char *array;
  register int temp;
  int i;
  int allocate=0;

/*
  fprintf(stderr,"internal_display_color_im\n");fflush(stderr);

  fprintf(stderr,"internal_gamma=%f\n", gamma); fflush(stderr);
  fprintf(stderr,"internal_pedestal=%f\n", pedestal);fflush(stderr);
  fprintf(stderr,"internal_scale=%f\n", scale);fflush(stderr);
  fprintf(stderr,"internal_title=%s\n", title);fflush(stderr); 
  fflush(stderr);
*/
  
  if(!(tempim==NULL))
    array = (unsigned char *) tempim;
  else {
    if((array = (unsigned char *)malloc((unsigned)total_size*sizeof(char))) == NULL) {
      fprintf(stderr,"malloc: can't allocate %d bytes.\n", total_size);
      return;
    }
    fprintf(stderr,"allocated memory in C\n");fflush(stderr);
    allocate=1;
  }
  
  if ((pid=OpenRGBScreen(ysize,xsize,title, gamma,zoom,(int)1,(int)1)) == NULL)
    fprintf(stderr,"process cannot be opened\n");
  fprintf(stderr,"opening screen... done\n");
  fflush(pid);

  for (i = 0; i < total_size; i++) {
    temp=(int)((r[i]-pedestal)*scale);
    array[i] = (unsigned char)(temp > 0)?((temp < 255)?temp:255):0;
  }

  DumpBufToScreen(array,ysize,xsize,pid,bufsize); 

  for (i = 0; i < total_size; i++) {
    temp=(int)((g[i]-pedestal)*scale);
    array[i] = (temp > 0)?((temp < 255)?temp:255):0;
  }

  DumpBufToScreen(array,ysize,xsize,pid,bufsize);

  for (i = 0; i < total_size; i++) {
    temp=(int)((b[i]-pedestal)*scale);
    array[i] = (temp > 0)?((temp < 255)?temp:255):0;
  }

  DumpBufToScreen(array,ysize,xsize,pid,bufsize);

  if(allocate==1)
    free(array);

  fflush(pid);
  pclose(pid);
  fflush(stderr);
#endif
}


/******************************************************************/
/* starts a color movie and returns a pid. */
extern double internal_init_color_movie
  (ysize,xsize,zoom,gamma,title,save,filename)
  int ysize,xsize;
  float zoom,gamma;
  char *title;
  int save;
  char *filename;
{

#ifdef RLE_DISPLAY

  FILE  *pid;
  int temp;

  switch (save) {
  case 0:
    if ((pid=OpenRGBMovie(ysize,xsize,title, gamma)) == NULL) {
      fprintf(stderr,"process cannot be opened\n"); fflush(stderr);
      return((double)0);
    }
    fprintf(stderr,"opening screen... done\n");
    break;
  case 1:
  default:
    if ((pid=OpenRGBFile(ysize,xsize,(char *)filename)) == NULL) {
      fprintf(stderr,"process cannot be opened\n"); fflush(stderr);
      return((double) 0);
    }
    fprintf(stderr,"opening file... done\n");
    break;
  }
  temp = (int)pid;
  fprintf(stderr,"internal_init_color_movie(pid=%d,temp=%d,(double)temp=%f)\n"
	 ,pid,temp,(double)temp);
  fflush(stderr);
  return((double) temp);
#else
  return((double) 0);
#endif
}

/***************************************************/
/* clean up movie */
extern void internal_close_color_movie(pid)
  FILE  *pid;
{
#ifdef RLE_DISPLAY
  if(pid==NULL) return;
  fflush(pid);
  fflush(stderr);
  pclose(pid);
  fprintf(stderr,"internal_close_color_movie(pid=%d)\n",pid);
#endif
}


/***************************************************/
/* put up one frame of a movie */
extern void internal_display_color_movie_frame
  (pid,r,g,b,ysize,xsize,pedestal,scale,tempim,bufsize)
  float *r,*g,*b;
  int ysize,xsize;
  float pedestal, scale;
  FILE *pid;    /* file pointers */
  int bufsize;
  unsigned char *tempim;
{
#ifdef RLE_DISPLAY

  int total_size = xsize*ysize;
  unsigned char *array;
  register int temp;
  int i;
  int allocate;
  
/*  fprintf(stderr,"internal_display_color_movie_frame (pid=%d)\n",pid);  fflush(stderr); */
  
  if(pid==NULL) return;

  if(!(tempim==NULL)) {
    array = (unsigned char *) tempim;
    allocate=0;
  }
  else {
    if((array = (unsigned char *)malloc((unsigned)total_size*sizeof(char))) == NULL) {
      fprintf(stderr,"malloc: can't allocate %d bytes.\n", total_size);
      return;
    }
    fprintf(stderr,"allocated memory in C\n");fflush(stderr);
    allocate=1;
  }
  
  for (i = 0; i < total_size; i++) {
    temp=(int)((r[i]-pedestal)*scale);
    array[i] = (temp > 0)?((temp < 255)?temp:255):0;
  }
  DumpBufToScreen(array,ysize,xsize,pid,bufsize);

  for (i = 0; i < total_size; i++) {
    temp=(int)((g[i]-pedestal)*scale);
    array[i] = (temp > 0)?((temp < 255)?temp:255):0;
  }
  DumpBufToScreen(array,ysize,xsize,pid,bufsize);

  for (i = 0; i < total_size; i++) {
    temp=(int)((b[i]-pedestal)*scale);
    array[i] = (temp > 0)?((temp < 255)?temp:255):0;
  }

  DumpBufToScreen(array,ysize,xsize,pid,bufsize);

  if(allocate)
    free(array);

  fflush(stderr);
#endif
}

/******************************************************************************/
/* opens a process for displaying color images. */
extern FILE *OpenRGBScreen(/*height,width,text,gamma,zoom,quant,movie*/);

FILE *OpenRGBScreen(height,width,text,gamma,zoom,quant,movie)
  int height, width;
  char *text;
  float gamma;
  float zoom;
  int quant,movie;
{
#ifdef RLE_DISPLAY

  char ExecRLE[400];
  char zoomstr[20];

/*  FILE *pid; */


  zoomstr[0]='\0';
  if(zoom!=1) sprintf(zoomstr,"| rlezoom %9.2",zoom);

  sprintf(ExecRLE,
	  "rawtorle  -h %ld -w %ld -n 3 -N | rleflip -v %s %s | getx11 -j %s -t \"%s\" -i %f",
	  height, width, zoomstr,
	  ((quant==1)? ("| rlequant -f"):(" ")),
	  ((movie==1)?(" -m "):(" ")),text, gamma);
  fprintf(stderr,"starting process\n");
  fprintf(stderr,"%s\n",ExecRLE);
  return( popen(ExecRLE,"w"));


#else
  extern char *nodisplay;
  fprintf(stderr,"%s",nodisplay);
  return(NULL);
#endif
}

/******************************************************************************/
/* opens a process for displaying gray scale images. */
extern FILE *OpenGrayScreen(height,width,text,gamma,zoom,nframes)
  int height, width;
  char *text;
  float gamma;
  float zoom;
  int nframes;
{
#ifdef RLE_DISPLAY

  char ExecRLE[400];
  char zoomstr[20];

/*
  fprintf(stderr,"height=%d width=%d text=%s gamma=%f zoom=%f nframes=%d\n",
	  height,width,text,gamma,zoom,nframes);
*/
  
  zoomstr[0]='\0';
  if(zoom!=1) sprintf(zoomstr,"| rlezoom %9.2",zoom);

  sprintf(ExecRLE,
	  "rawtorle  -h %ld -w %ld -n 1 | rleflip -v %s | getx11 -j %s -t \"%s\" -i %f",
	  height, width, zoomstr,((nframes>1)? " -m ":""), text, gamma);
  fprintf(stderr,"starting process\n");
  fprintf(stderr,"%s\n",ExecRLE);
  return( popen(ExecRLE,"w"));

#else
  extern char *nodisplay;
  fprintf(stderr,"%s",nodisplay);
  return(NULL);
#endif  
}


/***************************************************/
/* fwrite the data to the displaying process */
extern FILE *DumpBufToScreen(Image,height,width,pid,bufsize)
  unsigned char *Image;
  int height, width;
  FILE *pid;
  int bufsize;
{

#ifdef RLE_DISPLAY

  int num;
  int imsize=height*width;
  int i,n;

/*  fprintf(stderr,"dumptoscreen: bufsize=%d\n",bufsize);fflush(stderr);  */

  if(pid==NULL)
    return(NULL);
  
  fflush(pid);

  if(bufsize<=0) {
    fprintf(stderr,"DumpBufToScreen: invalid bufsize=%d, setting value to 100\n");fflush(stderr);
    bufsize=100;
  }
  n=(int)(imsize/bufsize);
  for(i=0;i<n;i++) {
    num=fwrite((char *)Image,sizeof(char),bufsize,pid);
    if(num!=bufsize) {
      fprintf(stderr,"Warning fwrite: %d of %d bytes written\n",num,bufsize);fflush(stderr);
    }
    Image+=num;
  }
  bufsize=imsize-n*bufsize;
  num=fwrite((char *)Image,sizeof(char),bufsize,pid);
  if(num != bufsize)
     fprintf(stderr,"Warning fwrite: %d of %d bytes written\n",num,bufsize);fflush(stderr);

  fflush(pid);
  return(pid);
#else
  return(NULL);
#endif
} /* end of DumpBufToScreen */


/***************************************************/
extern FILE *OpenRGBMovie(height,width,text,gamma)
  int height, width;
  char *text;
  float gamma;
{
  char ExecRLE[400];

/*  FILE *pid; */

#ifdef RLE_DISPLAY

  sprintf(ExecRLE,
	  "rawtorle  -h %ld -w %ld -n 3 -N | rleflip -v | getx11 -j -m -t \"%s\" -i %f",
	  height, width, text, gamma);
  fprintf(stderr,"starting process\n");
  fprintf(stderr,"%s\n",ExecRLE);
  return( popen(ExecRLE,"w"));

#else
  extern char *nodisplay;
  fprintf(stderr,"%s",nodisplay);
  return(NULL);
#endif
}

/***************************************************/
extern FILE *OpenRGBFile(height,width,filename)
  int height, width;
  char *filename;
{
  char ExecRLE[400];

/*  FILE *pid; */

#ifdef RLE_DISPLAY

  sprintf(ExecRLE,
	  "rawtorle  -h %ld -w %ld -n 3 -N -o %s \n", height, width, filename);
  fprintf(stderr,"starting process\n");
  fprintf(stderr,"%s\n",ExecRLE);
  fprintf(stderr,"\nwriting output to %s\n",filename);
  return( popen(ExecRLE,"w"));

#else
  return(NULL);
#endif
}


/***************************************************/
extern void internal_display_rgb_file(filename,gamma,title)
  float gamma;
  char *filename,*title;
{
  char ExecRLE[400];
  FILE *pid; 
  
#ifdef RLE_DISPLAY

  sprintf(ExecRLE,
	  "rlequant -f -m %s | rleflip -v | getx11 -j -m -t \"%s\" -i %f",filename,title,gamma);
  fprintf(stderr,"starting process\n");
  fprintf(stderr,"%s\n",ExecRLE);
  fprintf(stderr,"\nDisplaying RLE color file %s\n",filename);
  fflush(stderr);
  pid=popen(ExecRLE,"w");
  if(pid==NULL) return;
  pclose(pid);

#endif
}


extern void Close_Screen(pid)
  FILE *pid;
{

/*  fprintf(stderr,"Close_Screen: entering.\n"); */
#ifdef RLE_DISPLAY

  if(pid==NULL) return;

  if(pclose(pid)==-1) {
    fprintf(stderr,"Close_Screen: cannot close process\n");
  }
/*   fprintf(stderr,"Close_Screen: exiting.\n");*/

#endif
}
	 

/*******************************************************************/
extern void display_multichannel_image(seq,nchannels,xdim,ydim,
				       zoom,title,paste)
  float **seq;
  int nchannels,xdim,ydim;
  char *title;
  float zoom;
  int paste;
{
  unsigned char tempim=NULL;
  int bufsize=4096;
  float pedestal=0.0,scale=1.0,gamma=1.0;

  internal_display_multichan_im (seq,ydim,xdim,nchannels,
				 pedestal,scale,gamma,zoom,title,
				 tempim,bufsize,paste);
  return;
}

extern void internal_display_multichan_im
  (seq,ydim,xdim,nchannels,pedestal,scale,
   gamma,zoom,title,tempim,bufsize,paste)
  float **seq;
  int ydim,xdim;
  float pedestal, scale, gamma;
  float zoom;
  char *title;
  unsigned char *tempim;
  int bufsize,nchannels,paste;
{

  FILE *pid;    /* file pointers */
  int total_size = xdim*ydim;
  unsigned char *array;
  register int temp;
  int i,ch;
  int allocate=0;
  float *im;
  
  if(tempim!=NULL) { array = (unsigned char *) tempim; }
  else {
    if((array = (unsigned char *)malloc((unsigned)total_size*sizeof(char)))
       == NULL) {
      fprintf(stderr,"malloc: can't allocate %d bytes.\n", total_size);
      return; }
    allocate=1;
  }
  
  if ((pid=OpenMCScreen(ydim,xdim,title,gamma,zoom,(int)1,(int)1,
			nchannels,paste)) == NULL) {
    fprintf(stderr,"process cannot be opened\n"); }

/*  fprintf(stderr,"opening screen... done\n");*/
  fflush(pid);

  for(ch=0;ch<nchannels;ch++) {
    im=seq[ch];
    for (i = 0; i < total_size; i++) {
      temp=(int)((im[i]-pedestal)*scale);
      array[i] = (unsigned char)(temp > 0)?((temp < 255)?temp:255):0;
    }
    DumpBufToScreen(array,ydim,xdim,pid,bufsize);
  }

  if(allocate==1)
    free(array);

  fflush(pid);  pclose(pid); return;
}


extern FILE *OpenMCScreen(height,width,text,gamma,zoom,quant,
			  movie,nchannels,paste)
  int height, width;
  char *text;
  float gamma;
  float zoom;
  int quant,movie,nchannels,paste;
{
#ifdef RLE_DISPLAY

  char ExecRLE[400];
  char zoomstr[20];

  zoomstr[0]='\0';
  if(zoom!=1) sprintf(zoomstr,"| rlezoom %9.2",zoom);

  if(nchannels==3 && paste==0) {
    sprintf(ExecRLE,
	    "rawtorle  -h %ld -w %ld -n %d -N | rleflip -v %s %s | getx11 -j %s -t \"%s\" -i %f",
	    height, width, nchannels, zoomstr,
	    ((quant==1)? ("| rlequant -f"):(" ")),
	    ((movie==1)?(" -m "):(" ")),text, gamma);
  }
  else { /* display as a paste up */
    sprintf(ExecRLE,
	    "rawtorle  -h %ld -w %ld -n 1 -N | rleflip -v %s | getx11 -j %s -t \"%s\" -i %f",
	    nchannels*height, width, zoomstr,
	    ((movie==1)?(" -m "):(" ")),text, gamma);
  }
    
  fprintf(stderr,"starting process\n");  fprintf(stderr,"%s\n",ExecRLE);
  return( popen(ExecRLE,"w"));
#else
  extern char *nodisplay;
  fprintf(stderr,"%s",nodisplay);
  return(NULL);
#endif
}
