/******************************************************************

"Copyright  1994 by the Massachusetts  Institute  of  Technology.  All
rights reserved.

Developed by John Y. A. Wang at the  Media Laboratory, MIT, Cambridge,
Massachusetts, with  support  from Secom Ltd., Goldstar Ltd.,  and the
Television of Tomorrow Program.

For use by authorized users only.

This distribution is approved  by Nicholas Negroponte, Director of the
Media Laboratory, MIT.

Permission to use, copy, or modify this software and its documentation
for educational and research purposes only and without  fee is  hereby
granted, provided that this copyright notice and the original authors'
names  appear  on  all  copies  and   supporting   documentation.   If
individual  files  are  separated  from  this  distribution  directory
structure, this copyright notice must be included.  For any other uses
of  this software, in  original or  modified form, including  but  not
limited to distribution in whole or in part, specific prior permission
must be  obtained  from  MIT.   These  programs  shall  not  be  used,
rewritten,  or  adapted  as  the  basis  of a  commercial software  or
hardware  product without  first  obtaining  appropriate licenses from
MIT.  MIT.  makes no  representations about  the  suitability of  this
software for any purpose.  It is  provided "as is" without  express or
implied warranty.

******************************************************************/


/***************************************************************
*** File: display-image.c
*** Author: John Y. A. Wang
*** Date: January 15, 1993
*** Copyright 1993, Massachusetts Institute of Technology, All Rights Reserved.
***
*** Description: display images.
***************************************************************/

/*****************************************************************/
#include "dynamo.h"
/*****************************************************************/

/********** IMPORTANT:  put copyright label in all the files ***********/
static  char *copyright="\
\n\n***********************************************\n\
Developed by John Y. A. Wang at the Media Laboratory, MIT, Cambridge,Massachusetts\n\
Copyright  1993 by the Massachusetts  Institute  of  Technology.  All rights reserved.\n\
***********************************************\n\n";

/*
  cc -O4 -fsingle -c display-image.c 
  SUN:  cc -O4 -fsingle -c display-image.c

  lint display-image.c
*/
/*****************************************************************/

/*****************************************************************/

original_display(image_mgr,type)
  Image_Manager *image_mgr;
  int type;
{
  extern char *args;

  float cur_param[6];
  float *res=NULL,*image=NULL,*align_res=NULL;
  unsigned char *rarray=NULL;
  int imsize;
  int ref_frame;
  float factor[6];
  float pre_factor[6];
  int start=0;
  int end=0;
  int ndigits=4;

  FILE *pid=NULL;
  FILE *fp=NULL;
  float zoom=1;
  char title[50];
  int rxdim,rydim,rimsize;
  
  Image_Cache *cache=NULL,*cache_res=NULL;
  Affine_Analysis *an=NULL;
  int i,j;

  char *save_filename=image_mgr->cbuf1;
  char *yn=image_mgr->yn;

  int xs,ys,xe,ye;
  float d_clipmin=0,d_clipmax=255;

  char  *progn="original_display";

  /*********************************/

  fp=NULL;
  for(i=0;i<6;i++) {
    factor[i]=1.0; pre_factor[i]=1.0;
  }
  if((cache=ask_name_get_cache(image_mgr))==NULL)
    return(0);
  
  imsize=cache->imsize;
  an= &cache->affine_analysis;
  
  if(an->start>=0) start=an->start;
  if(an->end>=0) end=an->end;
  
  switch(type) {
  case 0: case 1:
    printf("\tEnter Gamma: (%4.1f) : ", cache->gamma);
    readline(args,ARGSIZE,stdin);
    sscanf(args,"%f",&cache->gamma);
    break;
  }

  printf("\tEnter Start: (%d) : ", start);
  readline(args,ARGSIZE,stdin); sscanf(args,"%d",&start);

  if(start<0 || start>=cache->tdim) {
    printf("Incorrect start specification.\n");return(-1); }
  
  printf("\tEnter End: (%d) : ", end);
  readline(args,ARGSIZE,stdin); sscanf(args,"%d",&end);

  if(end>=(cache->tdim)) { end=(cache->tdim)-1; }
  if(end<0 || end < start) { return(-1); }

  printf("\tEnter Display  Min Max: (%6.2f %6.2f) : ", d_clipmin,d_clipmax);
  readline(args,ARGSIZE,stdin); sscanf(args,"%f %f",&d_clipmin,&d_clipmax);

  /****************************/
  switch(type){
  case 1:   /* align, display, and optional save to file */
  case 2:   /* just align and save to file */
    ref_frame=(int)(start+end)/2;
    printf("\tEnter Reference Frame: (%d) : ", ref_frame);
    readline(args,ARGSIZE,stdin); sscanf(args,"%d",&ref_frame);
    if(ref_frame>an->end) ref_frame=an->end;
    else if(ref_frame<an->start) ref_frame=an->start;

    sprintf(title,"%s Aligned %d-%d-c%d",cache->image_name,start,end,ref_frame);

    printf("\tUpdate Pre-Factor: (%4.2f %4.2f %4.2f %4.2f %4.2f %4.2f) : ",
	   pre_factor[0],pre_factor[1],pre_factor[2],
	   pre_factor[3],pre_factor[4],pre_factor[5]);fflush(stdout);
    readline(args,ARGSIZE,stdin);
    sscanf(args,"%f %f %f %f %f %f",
	   &pre_factor[0],&pre_factor[1],&pre_factor[2],
	   &pre_factor[3],&pre_factor[4],&pre_factor[5]);

    printf("\tUpdate Factor: (%4.2f %4.2f %4.2f %4.2f %4.2f %4.2f) : ",
	   factor[0],factor[1],factor[2],factor[3],factor[4],factor[5]);
    readline(args,ARGSIZE,stdin);
    sscanf(args,"%f %f %f %f %f %f",
	   &factor[0],&factor[1],&factor[2],&factor[3],&factor[4],&factor[5]);
    
    printf("\tReference Frame=%d, factor=(%4.2f %4.2f %4.2f %4.2f %4.2f %4.2f)\n",
	   ref_frame,factor[0],factor[1],factor[2],factor[3],factor[4],factor[5]);
    sprintf(save_filename,"%s-%d-%d-%d.align",
	    cache->image_name,start,end,ref_frame);
    break;

  default:
  case 0:   /* original display */
    sprintf(title,"%s %d-%d",cache->image_name,start,end);
    sprintf(save_filename,"%s-%d-%d.orig",cache->image_name,start,end);
    break;

  case 100: /* test */
    sprintf(title,"Test"); break;
  }
  
  xs=an->xstart;xe=an->xend; ys=an->ystart;ye=an->yend;
  printf("\tCrop Region (xs ys xe ye) : (%d %d %d %d) : ",xs,ys,xe,ye);
  readline(args,ARGSIZE,stdin);
  if(strlen(args)!=0) { sscanf(args,"%d %d %d %d", &xs,&ys,&xe,&ye); }

  switch(type) {   /* result image dimensions */
  case 0:  default:
    rxdim=cache->xdim; rydim=cache->ydim;  rimsize=rydim*rxdim;
    break;
  case 1:  case 2:  
    rxdim=cache->xdim; rydim=cache->ydim;
    printf("\tResult size (rxdim rydim) : (%d %d) : ",rxdim,rydim);
    readline(args,ARGSIZE,stdin);
    if(strlen(args)!=0) { sscanf(args,"%d %d", &rxdim,&rydim); }
    rimsize=rxdim*rydim;
  break;
  }

  switch(type) {
  case 0: case 1: default:  /* include display */
    sprintf(yn,"n"); break;
  case 2:  /* always save but never display */
    sprintf(yn,"y"); break;
  }
    
  printf("\tDo you want to save output (y/n/q)? (%s) : ", yn);
  readline(args,ARGSIZE,stdin);  sscanf(args,"%s",yn);
  if(strncmp(yn,"q",1)==0) { return(0); }  /* quit */

  if(strncmp(yn,"y",1)==0) {  
    if((fp=open_file(save_filename,"w"))!=NULL)
      printf("saving output to %s\n",save_filename);
  }
  

  switch(type) {
  case 0: case 1:   /* start display */
  default:
    /* need to open screen for these cases, but defer until
     * we prompt for zoom value */
    break;
  case 2:  /* always save but never display */
    pid=NULL;
    sprintf(yn,"y");
    printf("\tDo you want to save sequence internally (y/q)? (%s) : ", yn);
    readline(args,ARGSIZE,stdin);  sscanf(args,"%s",yn);
    if(strncmp(yn,"q",1)==0) { return(0); }  /* quit */
    sprintf(image_mgr->image_name,"%s-%d-%d-%d.align",
	    cache->image_name,start,end,ref_frame);

    if((cache_res=make_internal_image(image_mgr,rxdim,rydim,end-start+1))==NULL) {
      printf("Cannot allocate memory for result image\n"); 
      return(-1);
    }
    if(cache_res==cache) {
      printf("%s: Cannot put Result into Input \n",progn); return(-1);
    }
  }

  printf("\tEnter zoom: (%4.2f) : ", zoom);
  readline(args,ARGSIZE,stdin); sscanf(args,"%f",&zoom);

  /* now open the window */
  switch(type) {
  case 0: case 1:   /* start display */
  default:
    pid=OpenGrayScreen(rydim,rxdim,title,cache->gamma,zoom,(end-start+1)); 
    if(pid==NULL) {
      printf("\nImage display capabilities currently not available.\n\n");
      fflush(stdout); close_file(fp); return(-1);
    }
    break;
  case 2:
    /* already done this above, ie create an image sequence. */
    break;
  }

  if(pid==NULL && fp==NULL && cache_res==NULL) { return(0); }
    
  sprintf(title,"%d",cache->tdim); ndigits=strlen(title);
  
  /************ allocate result memory **************/
  if((rarray=(unsigned char *)malloc((unsigned)rimsize*sizeof(float)))==NULL) {
    printf("%s: malloc cannot allocate %d bytes for image rarray \n",
	   progn,rimsize*sizeof(float)); free((char *)rarray); return(-1);
  }
    
  if((res=(float *)malloc((unsigned)rimsize*sizeof(float)))==NULL) {
    printf("%s: malloc cannot allocate %d bytes for image res \n",
	   progn,rimsize*sizeof(float));
    free((char *)res); free((char *) rarray); return(-1);
  }

  for(i=start;i<=end;i++) {
    if((image=get_image(cache,i))==NULL) {
      printf("Align: cannot load frame %d.... \n",i);
      Close_Screen(pid); 
      free((char *) res); free((char *) rarray); close_file(fp); return(-1);
    }

    switch(type) {
    case 0: default:
      copy_im(image,res,rimsize);
      break;
    case 1: case 2:   /* align images */
      printf("warping image....: ");
      fill_im(res,(float)0,rimsize);
      internal_update_affine_params_array(cache->param,PARAM_WIDTH,ref_frame,i,cur_param,factor,pre_factor);
      printf("(%8.2f %8.4f %8.4f %8.2f %8.4f %8.4f) \n",
	     cur_param[0],cur_param[1],cur_param[2],
	     cur_param[3],cur_param[4],cur_param[5]);

      if((cache->xdim!=rxdim) || (cache->ydim!=rydim)) {
	bicubic_affine_warp_larger(image, res,
				  cur_param[3],cur_param[4],cur_param[5],
				  cur_param[0],cur_param[1],cur_param[2],
				  cache->ydim,cache->xdim,rydim,rxdim);
      }
      else {
	bicubic_affine_warp_limit(image, res,
				  cur_param[3],cur_param[4],cur_param[5],
				  cur_param[0],cur_param[1],cur_param[2],
				  cache->ydim,cache->xdim,xs,ys,xe,ye);
      }
      break;
    }

    if(cache_res!=NULL) { /* save internally */
      if((align_res=get_image(cache_res,i))==NULL) {
	printf("Error: in getting align internal image frame %d\n",i);
	return(-1); }
      copy_im(res,align_res,rimsize);
    }

/*    printf("(i=%d)",i); */
    rescale_image(res,res,rimsize,(float)d_clipmin,(float)d_clipmax,
		  (float) 0,(float)255);
    convert_float_to_byte(res,rarray,rimsize);
    display_number_on_image(rarray,rxdim,rydim,
			    (float) i,(int)5,(int)5,(int)255,(int)0,(int)ndigits);
    if(fp!=NULL) {
      fwrite((unsigned char *)(rarray),1,rimsize,fp); }
    if(pid!=NULL) {
      DumpBufToScreen(rarray,rydim,rxdim,pid,rimsize); }
  }
  
  /******** clean up ********/
  if(fp!=NULL) {
    printf("Output saved to %s\n",save_filename);
    close_file(fp);
  }

  free((char *)res); free((char *)rarray);
  Close_Screen(pid);
  return(0);
}

/*****************************************************************/


/**********************************************/
display_frame(cache,frame_number)
  Image_Cache *cache;
  int frame_number;
{
  float *image;

  FILE *pid;
  char title[10];
  float zoom=1;
  float gamma=0.5;
  
  if((image=get_image(cache,frame_number))==NULL) {
    printf("display_frame: cannot display frame %d\n",frame_number);
    return(-1);
  }

  sprintf(title,"Frame %d",frame_number);
  
  pid=OpenGrayScreen(cache->ydim,cache->xdim,title,gamma,zoom,1);
  convert_float_to_byte(image,cache->array,cache->imsize);
  DumpBufToScreen(cache->array,cache->ydim,cache->xdim,pid,cache->imsize);
  Close_Screen(pid);
  return(0);
}


/*****************************************************************/
/**********************************************/
display_float_image(image,xdim,ydim,title)
  float *image;
  int xdim,ydim;
  char *title;
{

  FILE *pid;
  float zoom=1;
  float gamma=1.0;
  unsigned char *array;
  int imsize=xdim*ydim;

  if((array=(unsigned char *) malloc((unsigned)imsize))==NULL) {
    printf("display_float_image; cannot display image %d\n");
    return(-1);
  }

  pid=OpenGrayScreen(ydim,xdim,title,gamma,zoom,1);
  convert_float_to_byte(image,array,imsize);
  DumpBufToScreen(array,ydim,xdim,pid,imsize);
  Close_Screen(pid);
  free((char *)array);
  return(0);
}

/*****************************************************************/
/**********************************************/
display_float_image_opt(image,xdim,ydim,title,gamma,zoom)
  float *image;
  int xdim,ydim;
  char *title;
  float zoom;
  float gamma;
{
  FILE *pid;
  unsigned char *array;
  int imsize=xdim*ydim;

  if((array=(unsigned char *) malloc((unsigned)imsize))==NULL) {
    printf("display_float_image; cannot display image %d\n");
    return(-1);
  }

  pid=OpenGrayScreen(ydim,xdim,title,gamma,zoom,1);
  convert_float_to_byte(image,array,imsize);
  DumpBufToScreen(array,ydim,xdim,pid,imsize);
  Close_Screen(pid);
  free((char *)array);
  return(0);
}

/*****************************************************************/


/*****************************************************************/
/**********************************************/
display_array(array,xdim,ydim,title)
  unsigned char *array;
  int xdim,ydim;
  char *title;
{

  FILE *pid;
  float zoom=1;
  float gamma=1.0;
  int asize=xdim*ydim;

  pid=OpenGrayScreen(ydim,xdim,title,gamma,zoom,1);
  DumpBufToScreen(array,ydim,xdim,pid,asize);
  Close_Screen(pid);
  return(0);
}

/*****************************************************************/

