/******************************************************************

"Copyright  1994 by the Massachusetts  Institute  of  Technology.  All
rights reserved.

Developed by John Y. A. Wang at the  Media Laboratory, MIT, Cambridge,
Massachusetts, with  support  from Secom Ltd., Goldstar Ltd.,  and the
Television of Tomorrow Program.

For use by authorized users only.

This distribution is approved  by Nicholas Negroponte, Director of the
Media Laboratory, MIT.

Permission to use, copy, or modify this software and its documentation
for educational and research purposes only and without  fee is  hereby
granted, provided that this copyright notice and the original authors'
names  appear  on  all  copies  and   supporting   documentation.   If
individual  files  are  separated  from  this  distribution  directory
structure, this copyright notice must be included.  For any other uses
of  this software, in  original or  modified form, including  but  not
limited to distribution in whole or in part, specific prior permission
must be  obtained  from  MIT.   These  programs  shall  not  be  used,
rewritten,  or  adapted  as  the  basis  of a  commercial software  or
hardware  product without  first  obtaining  appropriate licenses from
MIT.  MIT.  makes no  representations about  the  suitability of  this
software for any purpose.  It is  provided "as is" without  express or
implied warranty.

******************************************************************/


/***************************************************************
*** File: drawline.c
*** Author: John Y. A. Wang
*** Date: January 15, 1993
*** Copyright 1993, Massachusetts Institute of Technology, All Rights Reserved.
***
*** Description: drawline routines. including vector fields and arrows...
***************************************************************/

static  char *copyright="\
\n\n***********************************************\n\
File: drawline.c\n\
Author: John Y. A. Wang.\n\
The MIT Media Lab.\n\
Copyright (c) 1993, All Rights Reserved.\n\
***********************************************\n\n";

#include "motion.h"
#include <math.h>

/***************************************************************/
/** draws a line into a byte array specifed by start: xs,ys; end: xe,ye. **/
int drawline(array,axdim,aydim,xs,ys,xe,ye,val)
  unsigned char *array;
  int axdim,aydim,xs,ys,xe,ye;
  int val;
{
  int x,y,t;
  int dx,dy;
  float slope;
  unsigned char cval;
		

/*  printf("(%d %d %d %d)",xs,ys,xe,ye);*/
  
  if(val<0) {cval=(unsigned char)0;}
  else if(val>255) {cval=(unsigned char)255;}
  else {cval=(unsigned char)((unsigned) val);}
  
  dx=xe-xs;
  dy=ye-ys;
  if((abs_val(dx)>abs_val(dy)) && (dx!=0)) {
    slope=(float)(((float)dy)/((float)dx));
    if(dx>0) {
/*      printf("x+ : ");*/
      for(t=0,x=xs;t<=dx;t++,x++) {
	y=(int)(ys+slope*t+.5);
/*	printf("(%d,%d)",x,y);*/
	if((x>=0) && (x<axdim) && (y>=0) && (y<aydim)) {array[y*axdim+x]=cval;}
      }}
    else {
/*      printf("x- : "); */
      for(t=0,x=xs;t>=dx;t--,x--) {
	y=(int)(ys+slope*t+.5);
/*	printf("(%d,%d)",x,y); */
	if((x>=0) && (x<axdim) && (y>=0) && (y<aydim)) {array[y*axdim+x]=cval;}
      }}
  }
  else if(dy!=0) {
    slope=(float)(((float)dx)/((float)dy));
    if(dy>0) {
/*      printf("y+ : ");*/
      for(t=0,y=ys;t<=dy;t++,y++) {
	x=(int)(xs+slope*t+.5);
/*	printf("(%d,%d)",x,y); */
	if((x>=0) && (x<axdim) && (y>=0) && (y<aydim)) {array[y*axdim+x]=cval;}
/*	else {break;} */
      }}
    else {
/*      printf("y- : ");*/
      for(t=0,y=ys;t>=dy;t--,y--) {
	x=(int)(xs+slope*t+.5);
/* 	printf("(%d,%d)",x,y);*/
	if((x>=0) && (x<axdim) && (y>=0) && (y<aydim)) {array[y*axdim+x]=cval;}
/*	else {break;} */
      }}
  }
  return(0);
}

/***************************************************************/
#ifdef NEVER
test_drawline()
{
  unsigned char *array;
  int xs,ys,xe,ye;
  float scale;
  int type,headsize;

  int axdim=100,aydim=100;
  int asize=axdim*aydim;

  if((array=(unsigned char *)malloc((unsigned)asize))==NULL) {
    printf("cannot allocate memory for array\n");
    return(-1);
  }
  
  printf("enter (xs ys xe ye) >> ");
  scanf("%d %d %d %d",&xs,&ys,&xe,&ye);
  
  printf("enter (scale type headsize) >> ");
  scanf("%f %d %d",&scale,&type,&headsize);

  fill_byte(array,(unsigned char)0,asize);
/*  drawline(array,axdim,aydim,xs,ys,xe,ye,(int)255); */
  draw_arrow(array,axdim,aydim,xs,ys,xe,ye,(int)255,scale,type,headsize); 
  display_array(array,axdim,aydim,"test drawline");
  return(0);
}

main()
{
  for(;;) 
    test_drawline();
}
#endif

/***************************************************************/
/*** draws arrow heads allows for 3 different types of arrows ***/
/* arrow head is drawn at (xe,ye) 
 * when scaling:
 *  type  1: arrow head fixed
 *  type  2: arrow tail fixed
 *  type  3: center
 */
void draw_arrow(array,axdim,aydim,xs,ys,xe,ye,val,scale,type,headsize)
  unsigned char *array;
  int xs,ys,xe,ye;
  int axdim,aydim;
  float scale;
  int val,headsize;
{
  int dx,dy,x,y;
  float xm,ym;
  float r,ca,sa;

  dx=xe-xs;dy=ye-ys;
  if(scale!=1) {
    switch(type) {
    case 1:
    default:
/*      printf("fix end pt: ");*/
      xs=(int)(scale*(-dx)+xe+.5); ys=(int)(scale*(-dy)+ye+.5);
      break;
    case 2:
/*       printf("fix start pt: ");*/
      xe=(int)(scale*dx+xs+.5); ye=(int)(scale*dy+ys+.5);
      break;
    case 3:
/*      printf("fix center: ");*/
      xm=(xs+xe)/2; ym=(ys+ye)/2;
      xs=(int)(scale*(-dx)/2+xm+.5); xe=(int)(scale*dx/2+xm+.5);
      ys=(int)(scale*(-dy)/2+ym+.5); ye=(int)(scale*dy/2+ym+.5);
      break;
    }
/*    printf("scale=%f\n",scale); */
  }
  drawline(array,axdim,aydim,xs,ys,xe,ye,val);

  /* arrow heads */
  r=(float)sqrt((double)(dx*dx+dy*dy));
  if(r>=1) {
    r=(float)((float)headsize/r);
    ca=.866; sa=.5; /* 30 degress */
    /*  printf("r=%f\n",r); */

    x=(int)(xe-(ca*(float)dx+sa*(float)dy)*r+.5);
    y=(int)(ye-(-sa*(float)dx+ca*(float)dy)*r+.5);
    drawline(array,axdim,aydim,x,y,xe,ye,val);
  
    x=(int)(xe-(ca*(float)dx-sa*(float)dy)*r+.5);
    y=(int)(ye-(sa*(float)dx+ca*(float)dy)*r+.5);
    drawline(array,axdim,aydim,x,y,xe,ye,val);
  }
}


/***************************************************************/
/* draws vector-fields to byte array */
int display_vector_field(dx,dy,xdim,ydim,array,axdim,aydim,val,xstep,ystep,
		     scale,headsize,type)
  float *dx,*dy;
  int xdim,ydim,axdim,aydim;
  unsigned char *array;
  int xstep,ystep;
  float scale;
  int headsize,type,val;
{
  int x,y,ypos;
  int xs,ys;
  
/*
  printf("val=%d xstep=%d ystep=%d scale=%f headsize=%d type=%d\n",
	 val,xstep,ystep,scale,headsize,type);
*/  
  if((xdim!=axdim) || (ydim!=aydim))
    return(-1);

  drawline(array,axdim,aydim,0,0,0,aydim,val);
  drawline(array,axdim,aydim,0,0,axdim,0,val);
  drawline(array,axdim,aydim,0,aydim-1,axdim,aydim-1,val);
  drawline(array,axdim,aydim,axdim-1,0,axdim-1,aydim,val);

  for(y=0;y<aydim;y+=ystep) {
    for(x=0,ypos=y*xdim;x<axdim;x+=xstep) {
      xs=(int)(x-scale*dx[ypos+x]+.5);
      ys=(int)(y-scale*dy[ypos+x]+.5);
/*      printf("(d:%4.2f %4.2f)",dx[ypos+x],dy[ypos+x]);*/
      draw_arrow(array,axdim,aydim,xs,ys,x,y,val,(float)1,(int)1,headsize); 
}
  }
  return(0);
}
