/******************************************************************

"Copyright  1994 by the Massachusetts  Institute  of  Technology.  All
rights reserved.

Developed by John Y. A. Wang at the  Media Laboratory, MIT, Cambridge,
Massachusetts, with  support  from Secom Ltd., Goldstar Ltd.,  and the
Television of Tomorrow Program.

For use by authorized users only.

This distribution is approved  by Nicholas Negroponte, Director of the
Media Laboratory, MIT.

Permission to use, copy, or modify this software and its documentation
for educational and research purposes only and without  fee is  hereby
granted, provided that this copyright notice and the original authors'
names  appear  on  all  copies  and   supporting   documentation.   If
individual  files  are  separated  from  this  distribution  directory
structure, this copyright notice must be included.  For any other uses
of  this software, in  original or  modified form, including  but  not
limited to distribution in whole or in part, specific prior permission
must be  obtained  from  MIT.   These  programs  shall  not  be  used,
rewritten,  or  adapted  as  the  basis  of a  commercial software  or
hardware  product without  first  obtaining  appropriate licenses from
MIT.  MIT.  makes no  representations about  the  suitability of  this
software for any purpose.  It is  provided "as is" without  express or
implied warranty.

******************************************************************/


/***************************************************************
*** File: dynamo.h
*** Author: John Y. A. Wang
*** Date: January 15, 1993
*** Copyright 1993, Massachusetts Institute of Technology, All Rights Reserved.
***
*** Description: header file for motion analysis package.
***************************************************************/

#include<stdio.h>
#include<sys/types.h>
#include<sys/stat.h>

#ifdef SVR4 
#include <string.h>
#else
#include <strings.h>
#endif

#include <errno.h>

/********** IMPORTANT:  put copyright label in all the files ***********/
static  char *copyright_dynamo="\
\n\n***********************************************\n\
Developed by John Y. A. Wang at the Media Laboratory, MIT, Cambridge,Massachusetts\n\
Copyright  1993 by the Massachusetts  Institute  of  Technology.  All rights reserved.\n\
***********************************************\n\n";

#ifdef lint
#define register
#endif

#define DATAFILE         1
#define DATASET          2
#define INTERNAL_IMAGE   3

#define ARGSIZE 256
			   
/************************/
struct Affine_Analysis {
  int start;                /* start of analysis */
  int end;                  /* end of analysis */
  int tstep;                /* time step */
  int xstart;               /* crop region */
  int xend;                 /* crop region */
  int ystart;               /* crop region */
  int yend;                 /* crop region */
  int level;                /* level of coarse to fine */
  float method[10];         /* iteration method */
  float factor[6];     /* analysis factor mask */
  float param_init[6];      /* analysis factor mask */
};
typedef struct Affine_Analysis Affine_Analysis;

/************************/
struct Optical_Analysis {
  int start;                /* start of analysis */
  int end;                  /* end of analysis */
  int tstep;                /* time step */
  int level;                /* level of coarse to fine */
  int blur;                 /* blur level */
  float method[10];         /* iteration method */
  float noise;              /* noise */
  float xbias;              /* xbias */
  float ybias;              /* ybias */
  float intensity_bias;     /* intensity offset bias */
  float clipmin;
  float clipmax;
  float sum_filt[10];       /* summation filter */
  int sfdim;                /* summation filter dimensions */
  float pyr_filt[10];       /* pyramid  filter */
  int pfdim;                /* pyramid  filter dimensions */
  float exp_filt[10];       /* pyramid upsample interpolation filter */
  int expfdim;              /* pyramid upsample interpolation filter dimensions */
  int stat;                 /* flag to indicate to do statistics */  
};
typedef struct Optical_Analysis Optical_Analysis;

/***********************************/
struct Frame_Rate_Conv {
    int level ;             /* level of coarse to fine in optic flow estimation */
    int start;              /* starting index */
    int end;                /* end index */
    float time_step;          /* conversion ratio */
    float xvar;             /* */
    float yvar;             /* */
    float yscale;             /* */
    float xscale;             /* */
  };
typedef struct Frame_Rate_Conv Frame_Rate_Conv;

/************************/
struct Image_Cache {
  int start;               /* start and end of images in cache */
  int end;
  int nframes;             /* number of frames stored in cache */
  int load_offset;  /* frame relative offset to start or end when loading */
  int status;              /* flag indicating status of cache  */
  int prev_frame;          /* the previous accessed frame */
  char image_name[256];         /* alias image name */
  float gamma;

  int xdim;                /* x dimensions of images in memory */
  int ydim;                /* y dimensions of images in memory */
  int tdim;                /* length of movie */
  int imsize;              /* number of pixels of image in cache */
  int filetype;            /* datafile or dataset type */
  char filename[256];      /* image file name */
  char datafile_ext[20];   /* datafile image file extension name */
  int ext_start;           /* extension start */
  char paramfilename[256]; /* param file name */
  char savedirname[256];   /* save directory file name */
  int headersize;          /* size of image data header */
  struct stat file_stat;   /* file status */

#ifndef CACHE_OLD
  float **images;           /* nframes of float images */
#else
  float *images;           /* nframes of float images */
#endif
  unsigned char *array;             /* byte array of imsize */
  unsigned char *file_array;        /* byte array of file_imsize */

  float *param;             /* affine parameters */
  Affine_Analysis  affine_analysis;   /* affine_analysis */ 
  Optical_Analysis optical_analysis;  /* Optical flow analysis */ 
  Frame_Rate_Conv frame_rate_conv;    /* frame rate conversion */
  int xcstart, ycstart;      /* x and y of start of crop */
  int file_xdim, file_ydim;  /* dimensions of image in the file */
  int file_imsize;            /* dimensions of images on file  */
};
typedef struct Image_Cache Image_Cache;

#define MAXIMAGES 10
struct Image_Manager {
  int n_images;
  Image_Cache *image_list[MAXIMAGES];
  char args[ARGSIZE];
  char image_name[256];
  char cbuf1[256];
  char cbuf2[256];
  char yn[10];
  int dummy;
};
typedef struct Image_Manager Image_Manager;
/*******************************************/
/******** MENU DEFINITIONS **********/
#define MAIN_IMAGE_MANAGEMENT_MENU_OPT  1
#define MAIN_AFFINE_ANALYSIS_OPT        2
#define MAIN_OPTICAL_FLOW_ANALYSIS_OPT  3
#define MAIN_DISPLAY_IMAGES_OPT         4
#define MAIN_ENHANCE_IMAGES_OPT         5
#define MAIN_BINARY_OPERATION_OPT       6
#define MAIN_UNARY_OPERATION_OPT        7
#define MAIN_QUIT_OPT                   9

#define IMAGE_LOAD_OPT                1 
#define IMAGE_CHANGE_CACHE_SIZE_OPT   2
#define IMAGE_STATUS_OPT              3
#define IMAGE_DELETE_OPT              4
#define IMAGE_SAVE_OPT                5
#define IMAGE_RESET_MEMORY_OPT        6

#define AFFINE_DISPLAY_ANALYSIS_OPT  1
#define AFFINE_RUN_ANALYSIS_OPT      2
#define AFFINE_DISPLAY_PARAMS_OPT    3
#define AFFINE_SAVE_PARAMS_OPT       4
#define AFFINE_LOAD_PARAMS_OPT       5
#define AFFINE_ALIGN_IMAGES_OPT       6

#define OPTICAL_FLOW_RUN_ANALYSIS_OPT   1
#define OPTICAL_FLOW_DISPLAY_OPT        2
#define OPTICAL_FLOW_DISPLAY_FIELD_OPT  3

#define DISPLAY_ORIGINAL_OPT        1
#define DISPLAY_AFFINE_ALIGN_OPT    2

#define ENHANCE_MEDIAN_ALIGN_OPT 1
#define ENHANCE_AVERAGE_ALIGN_OPT 2

#define BIN_ADD_OPT   1
#define BIN_SUB_OPT   2
#define BIN_MUL_OPT   3
#define BIN_DIV_OPT   4

#define UNI_MEDIAN_OPT   1
#define UNI_AVERAGE_OPT  2

#define MAIN_FRAME_RATE_CONV_OPT 8
/******* GLOBALS ***************/
#define PARAM_WIDTH 6
#define TEMPIMAGES 5
#define PROGN "DYNAMO"
/***********************************/
#define inbounds(x,dim)  ( ((x)>=0) && ((x)<(dim)) )

#define READ 0
#define WRITE 1

/*** comment parser ****/
#define COMMENT_CHAR ';'

FILE *OpenGrayScreen(/*height,width,text,gamma,zoom,nframes*/);
FILE *DumpBufToScreen(/*Image,height,width,pid,bufsize*/);
void Close_Screen(/*pid*/);

/**** fileio.c *****/
FILE *open_file(/*filename,type*/);
int save_dir(/*savedirname*/);
int save_descriptor(/*filename,xsize,ysize,datasets,datafiles,channels,
		    scale,pedestal,comment*/);
int save_image_to_file(/*image_mgr,cache,image,xdim,ydim,file_extension,comment*/);
int close_file(/*fp*/);
char *readline(/*_ args,size,fp _*/);

void free_images(/*malloc_images,nimages*/);
void close_files(/*opened_files,nfiles*/);


float *get_image(/*cache,frame_number*/);
void display_number_on_image(/*array,xdim,ydim,number,xoffset,yoffset,fg_val,bg_val,ndigits*/);

/**** cache.c ***/
Image_Cache *find_cache_name(/*image_mgr,image_name*/);
Image_Cache *ask_name_get_cache(/*image_mgr*/);
Image_Cache *make_internal_image(/*image_mgr,xdim,ydim,length*/);
float *get_image_disk(/*cache,frame_number,res*/);
int load_cache(/*cache,start,end*/);
int load_cache_disk(/*cache,start,end,res*/);

float internal_interpolate_mask(/*im, x, y, xdim, ydim, mask, match_val, eq, print*/);
  
float combinatorial(/*k,n*/);

float **malloc_float_seq(/*nframes,imsize*/);
int free_float_seq(/*images,nframes*/);


#include "motion.h"
