/******************************************************************

"Copyright  1994 by the Massachusetts  Institute  of  Technology.  All
rights reserved.

Developed by John Y. A. Wang at the  Media Laboratory, MIT, Cambridge,
Massachusetts, with  support  from Secom Ltd., Goldstar Ltd.,  and the
Television of Tomorrow Program.

For use by authorized users only.

This distribution is approved  by Nicholas Negroponte, Director of the
Media Laboratory, MIT.

Permission to use, copy, or modify this software and its documentation
for educational and research purposes only and without  fee is  hereby
granted, provided that this copyright notice and the original authors'
names  appear  on  all  copies  and   supporting   documentation.   If
individual  files  are  separated  from  this  distribution  directory
structure, this copyright notice must be included.  For any other uses
of  this software, in  original or  modified form, including  but  not
limited to distribution in whole or in part, specific prior permission
must be  obtained  from  MIT.   These  programs  shall  not  be  used,
rewritten,  or  adapted  as  the  basis  of a  commercial software  or
hardware  product without  first  obtaining  appropriate licenses from
MIT.  MIT.  makes no  representations about  the  suitability of  this
software for any purpose.  It is  provided "as is" without  express or
implied warranty.

******************************************************************/


/***************************************************************
*** File: enhance.c
*** Author: John Y. A. Wang
*** Date: January 15, 1993
*** Copyright 1993, Massachusetts Institute of Technology, All Rights Reserved.
***
*** Description: enhance images using affine motion
***************************************************************/

/*****************************************************************/
#include "dynamo.h"
/*****************************************************************/

/********** IMPORTANT:    **********************************/
/********** put copyright label in all the files ***********/

static  char *copyright="\
\n\n***********************************************\n\
Developed by John Y. A. Wang at the Media Laboratory, MIT, Cambridge,Massachusetts\n\
Copyright  1993 by the Massachusetts  Institute  of  Technology.  All rights reserved.\n\
***********************************************\n\n";

/*
  cc -O4 -fsingle -c enhance.c 
  SUN:  cc -O4 -fsingle -c enhance.c

  lint enhance.c
*/
/*****************************************************************/

void bicubic_affine_warp_limit2_weight(/*im, res, ax, bx, cx, ay, by, cy,
				       ydim, xdim,xstart,ystart,xend,yend,
				       rydim,rxdim,rxstart,rystart,rxend,ryend,
				       weight*/);

void bicubic_affine_warp_weight_abs(/*im, res, ax, bx, cx, ay, by, cy,
				       ydim, xdim,xstart,ystart,xend,yend,
				       rydim,rxdim,rxstart,rystart,rxend,ryend,
				       weight*/);


/*******************************************************/
enhance_median_images(image_mgr)
  Image_Manager *image_mgr;
{
  extern char *args;

  float *im_addr[100];
  int i;
  char *name=image_mgr->image_name;
  Image_Cache *cache=NULL,*cache_res=NULL;
  int start,end,nframes,rxdim,rydim,rimsize;

  /* these are malloc-ed here */
  float *res=NULL,*res_save=NULL,*count=NULL,*comp_param=NULL; 

  float equal_val=1.0;
  float setval=0.0;

  char *yn=image_mgr->yn;
  float xzoom=1,yzoom=1;
  int ref_frame=0;
  float factor[6],pre_factor[6],p_res1[6],p_res2[6];
  float xc,yc;
  Affine_Analysis *an=NULL;
  float *cur_param=NULL;
  
  char *progn="median_enhance_images";

  if((cache=ask_name_get_cache(image_mgr))==NULL)
    return(0);

  sprintf(yn,"q");
  printf("Do you want to use current results of Affine Motion Analysis(y/n/q)? (%s) >> ",yn);
  readline(args,ARGSIZE,stdin); sscanf(args,"%s",yn);
  if(strncmp(yn,"q",1)==0) {return(0);}
  else if(strncmp(yn,"n",1)==0) {
    run_global_affine_motion_analysis(image_mgr);
    an= &cache->affine_analysis; start=an->start; end=an->end;
  }
  else if(strncmp(yn,"y",1)==0) {
    cache=find_cache_name(image_mgr,name);
    an= &cache->affine_analysis; start=an->start; end=an->end+1;

    printf("Currently available parameters in memory ... start=%d end=%d\n",
	   start,end);

    printf("Enter Start Frame: (%d) >> ",start);
    readline(args,ARGSIZE,stdin); sscanf(args,"%d",&start);

    printf("Enter End Frame: (%d) >> ",end);
    readline(args,ARGSIZE,stdin); sscanf(args,"%d",&end);

    if (start < an->start) start=an->start;
    if (end > (an->end+1)) end=an->end+1;

    printf("%s: Using current parameters in memory\n",progn);
  }

  printf("%s: Enhancing with frames %d to %d\n",progn,start,end);

  ref_frame=(int)((end+start)/2);
  printf("Enter Reference Frame: (%d) >> ",ref_frame);
  readline(args,ARGSIZE,stdin); sscanf(args,"%d",&ref_frame);

  cache=find_cache_name(image_mgr,name);
  an= &cache->affine_analysis;
  
  nframes=end-start+1;

#define IS_IN_CACHE(frame,cache) (((frame) >= (cache)->start) && ((x) >= (frame)->end))
  
  if((nframes > cache->nframes) && (nframes <= 100)) {
    printf("%s:cache not large enough for operation\n",progn);
    return(-1);
    }

 
  xzoom=1;yzoom=1;
  printf("Enter Amount of XZoom  and YZoom Enhance: (%6.2f %6.2f) >> ",xzoom,yzoom);
  readline(args,ARGSIZE,stdin); sscanf(args,"%f %f",&xzoom,&yzoom);
  for(i=0;i<6;i++) {
    factor[i]=1.0;pre_factor[i]=1.0;
  }

  rxdim=cache->xdim;
  printf("Enter result xdim: (%d) >> ",rxdim);
  readline(args,ARGSIZE,stdin); sscanf(args,"%d",&rxdim);

  rydim=cache->ydim;
  printf("Enter result ydim: (%d) >> ",rydim);
  readline(args,ARGSIZE,stdin); sscanf(args,"%d",&rydim);

  rimsize=rxdim*rydim;
  
  xc=(an->xstart + an->xend)/2;
  yc=(an->ystart + an->yend)/2;

  printf("\tCenter Region: (%6.2f %6.2f) : ",xc,yc);
  readline(args,ARGSIZE,stdin);
  if(strlen(args)!=0) { sscanf(args,"%f %f",&xc,&yc); }

  /*************** Allocate memory ****************************/
  if((res=(float *)malloc((unsigned)rimsize*sizeof(float)))==NULL) {
    printf("cannot allocate memory for analysis (im_flow)\n");
    free((char *)res); 
    return(-1);
  }
  if((count=(float *)malloc((unsigned)rimsize*sizeof(float)))==NULL) {
    printf("cannot allocate memory for analysis (im_flow)\n");
    free((char *)res); free((char *)count);
    return(-1);
  }
  if((comp_param=(float *)malloc((unsigned)(nframes*PARAM_WIDTH)*sizeof(float)))==NULL) {
    printf("cannot allocate memory for analysis (im_flow)\n");
    free((char *)res); free((char *)count); free((char *)comp_param);
    return(-1);
  }

  /***************************************************/

/*  printf("ref_frame=%d,start=%d,end=%d\n",ref_frame,start,end); */
  for(i=0;i<6;i++) { p_res1[i]=0; }

  p_res1[0]=(cache->ydim)/2 - yc;
  p_res1[3]=(cache->xdim)/2 - xc;

  zoom_param(p_res2,xzoom,yzoom);
/*
  printf("p_res1: (%6.2f %6.4f  %6.4f %6.2f %6.4f  %6.4f)\n",
	 p_res1[0],p_res1[1],p_res1[2],p_res1[3],p_res1[4],p_res1[5]);
  printf("p_res2: (%6.2f %6.4f  %6.4f %6.2f %6.4f  %6.4f)\n",
	 p_res2[0],p_res2[1],p_res2[2],p_res2[3],p_res2[4],p_res2[5]);
*/
  for(i=start;i<=end;i++) {
    im_addr[(i-start)]=get_image(cache,i);
    cur_param= &comp_param[PARAM_WIDTH*(i-start)];
    internal_update_affine_params_array(cache->param,PARAM_WIDTH,ref_frame,i,
				       cur_param,factor,pre_factor);

    internal_update_affine_params_from(cur_param,p_res1,cur_param,factor);
    internal_update_affine_params_from(cur_param,p_res2,cur_param,factor);
  }

#ifdef NEVER
  /***** additional refinement ********/
  {
    float method[1];
    float *ttemp=NULL,*w_images=NULL;
    int ttempsize=0,j;
    
    int imsize;
    imsize=cache->imsize;
    ttempsize=cache->imsize*5;
    
    if((ttemp=(float *)malloc((unsigned)ttempsize*sizeof(float)))==NULL) {
      printf("cannot allocate memory for analysis (ttemp)\n");
      free((char *)res); free((char *)count); free((char *)comp_param);
      free((char *)ttemp);
      return(-1);
    }
    
    if((w_images=(float *)malloc((unsigned)imsize*2*sizeof(float)))==NULL) {
      printf("cannot allocate memory for analysis (ttemp)\n");
      free((char *)res); free((char *)count); free((char *)comp_param);
      free((char *)ttemp); free((char *) w_images);
      return(-1);
    }

    method[0]=1;
    for(i=0;i<6;i++) {factor[i]=1.0;}

    im1=w_images; im2=w_images+imsize;

    cur_param=comp_param[PARAM_WIDTH*(ref_frame-start)];
    bicubic_affine_warp(im_addr[(ref_frame-start)], im2,
			cur_param[3],cur_param[4],cur_param[5],
			cur_param[0],cur_param[1],cur_param[2],
			cache->ydim,cache->xdim);

    for(i=start,j=0;i<=end;i++,j++) {
      cur_param= &comp_param[PARAM_WIDTH*(i-start)];
      bicubic_affine_warp(im_addr[(i-start)], im1,
			  cur_param[3],cur_param[4],cur_param[5],
			  cur_param[0],cur_param[1],cur_param[2],
			  cache->ydim,cache->xdim);
      internal_coarse_fine_affine_estimator(im1,im2,cur_param,cur_param,cache->xdim,
					    cache->ydim,
					    0,method,
					    0,0,cache->xdim,cache->ydim,
					    ttemp,ttempsize,
					    factor,NULL,NULL);
    }
    free((char *)ttemp); free((char *) w_images);
  }
#endif
  
  fill_im(res,(float)0,rimsize);

  printf("\n\nPerforming temporal median filtering. Please wait .....\n\n");

  internal_warp_median_seq(im_addr,NULL,cache->xdim,cache->ydim,0,nframes-1,
			   res,count,rxdim,rydim,
			   equal_val,setval,
			   comp_param,(int)PARAM_WIDTH,nframes,0);

  free((char *)count);free((char *)comp_param);
  /****** all done with enhancement *********/

  sprintf(args,"Median Enhance Image: %d-%d-c%d", start,end,ref_frame);
/*  display_float_image(res,rxdim,rydim,args); */
  display_float_image_opt(res,rxdim,rydim,args,cache->gamma,(float)1.0);

  /* save results internally */
  sprintf(yn,"n"); printf("\tDo you want to save result internally (y/n)? (n) : ");
  readline(args,ARGSIZE,stdin);  sscanf(args,"%s",yn);
  if(strncmp(yn,"y",1)==0) {
    sprintf(image_mgr->image_name,"%s-med-enhance",cache->image_name);
    if((cache_res=make_internal_image(image_mgr,rxdim,rydim,(int)1))==NULL) {
      printf("Cannot allocate memory for result image\n"); 
    }
    if((res_save=get_image(cache_res,0))==NULL) {
      printf("%s: Progn error. Cannot get cache_res float array.\n",progn);
      return(-1); }
    copy_im(res,res_save,(int)(rxdim*rydim));
    printf("Results saved to %s....\n",cache_res->image_name);
  }

  /* save to file */
  sprintf(yn,"n");
  printf("\tDo you want to save output to disk (y/n)? (n) : ");
  readline(args,ARGSIZE,stdin); sscanf(args,"%s",yn);
  if(strncmp(yn,"y",1)==0) {
    sprintf(args, "MCNR median ehancement of %s\nstart=%d,end=%d\n",
	    cache->filename,start,end);
    save_image_to_file(image_mgr,cache,res,rxdim,rydim,"enhance",args);
    }
  free((char *)res);
  return(0);
}


/*******************************************************/
#define near_val(x,c,e) (((x)<=((c)+(e)))&&((x)>=((c)-(e))))

internal_warp_median_seq (im_addr,mask_addr, xdim, ydim, start_frame, end_frame,
			  res, count, rxdim, rydim,
			  equal_val, setval,
			  aff_array, axdim ,aydim,wdim)
  float *res,*count;
  register float *aff_array;
  register float **im_addr; /* array of pointers to images */
  register float **mask_addr; /* array of pointers to mask images */
  register int xdim,ydim;
  int rxdim,rydim;
  int axdim, aydim;
  float equal_val;
  float setval;
  int start_frame, end_frame;
  int wdim;
{
  int i,j,f,n;
  float arr[400];
  int index;
  register  float *aff,*image,*mask;
  float x,y;
  float x_off, y_off;
  float rx_off, ry_off;
  float x_diff, y_diff;
  register float dx,dy;
  register int pos;
    
  x_off = (xdim - 1) / 2.0;
  y_off = (ydim - 1) / 2.0;
  rx_off = (rxdim - 1) / 2.0;
  ry_off = (rydim - 1) / 2.0;

  x_diff = rx_off - x_off;
  y_diff = ry_off - y_off;

/*
  printf("equal-val=%4.2f; xdim=%d; ydim=%d\n",equal_val,xdim,ydim);fflush(stdout);

  printf("start_frame=%d; end_frame=%d; axdim=%d; aydim=%d\n",
	 start_frame,end_frame,axdim,aydim);fflush(stdout);
*/
  for(j=0,pos=0; j<rydim; j++)  {      /* j is the y var */
/*    printf("*%d* ",j);fflush(stdout);  */
    for(i=0; i<rxdim; i++,pos++) {    /* i is the x var */
      dx = ((float)i - rx_off); dy=((float)j - ry_off);
      for(f=start_frame, n=0; f<=end_frame; f++) {
	if(mask_addr !=NULL)
	  mask=mask_addr[f];
	else
	  mask=NULL;
	aff= &aff_array[f*axdim];

	y = j - y_diff - (aff[0] + aff[1] * dx + aff[2] * dy);
	x = i - x_diff - (aff[3] + aff[4] * dx + aff[5] * dy);

	if (inbounds(x,xdim) && inbounds(y,ydim)
	    && (mask==NULL || near_val(mask[((int)y)*xdim + (int)x],equal_val,0.1))) {
/*	  printf("(%d,%d)",(int)y,(int)x);fflush(stdout); */
	  image=(im_addr[f]);
	  arr[n]=internal_interpolate_mask(image,x,y,xdim,ydim,mask,equal_val,(int)1,(int)0);
/*	  printf("mask=%4.2f\n",mask[((int)y)*xdim + (int)x]);fflush(stdout); */

	  /* load the array */
	  if(arr[n]!=0)
	    n++;
	}
      }

      count[pos]=(float) n; 
      switch (n) {
      case 0:
	res[pos] = setval;
	break;
      default:
	internal_sort(arr,n);       /* sort list */
	index = (int) (n/2);        /* determine index */
	if(1&n)                     /* odd */
	  res[pos] = arr[index]; 
	else
	  res[pos]= (arr[index] + arr[index-1])/2.0;
      }
    }}
  fflush(stdout);
}


/************************************************************/
/* use bicubic if 4 points are available, linear if 2 are available.
** interpolation is describe by the mask function within the result image
** dimensions
*/

float internal_interpolate_mask(im, x, y, xdim, ydim, mask, match_val, eq, print)
  float *im,*mask; 
  float x,y;
  int xdim, ydim;
  int print;
  float match_val;
  int eq;
{
  register float frac;	/* accessed 20x in loop */
  register float frac2, frac3;  /* accessed 10x in loop */
  register float x_coeff_1, x_coeff0, x_coeff1, x_coeff2; /* 5x in loop */
  register float res_val;
  register float xbase,ybase;

  int pos,ypos;
  int ix,xb,xp,xbin,xs;
  int iy,yb,yp,ybin,ys;
  float x_arr[4];
  float y_arr[4];

  /**********************/
  xbase = (int)(x);
  ybase = (int)(y);

  if(print==1) {
    printf("interp: x=%f y=%f :: xbase=%f ybase=%f print=%d\n",x,y,xbase,ybase,print);
    fflush(stdout);
  }

  if((((xbase) >= 0) && ((xbase + 1) < xdim)) &&    
     (((ybase) >= 0) && ((ybase + 1) < ydim))) {  /* quick check */

    res_val = 0.0;
      
    /* Compute X fractional position and base index */
    frac = x - xbase;
    frac2 = frac*frac;
    frac3 = frac*frac2;
	    
    x_coeff_1 = (frac2/2.0 - frac/3.0 - frac3/6.0);  /* c_1 = 0 -1/3  1/2 -1/6 */
    x_coeff0  = (1.0 - frac2  + (frac3 - frac)/2.0); /* c0  = 1 -1/2  -1   1/2 */
    x_coeff1 = (frac + (frac2 - frac3)/2.0);         /* c1  = 0  1    1/2 -1/2 */
    x_coeff2 = ((frac3 - frac)/6.0);                  /* c2  = 0 -1/6  0    1/6 */
      
    /* Compute Y fractional position and base index */
      
    y_arr[0] = 0.0; y_arr[1] = 0.0; y_arr[2] = 0.0; y_arr[3] = 0.0;

    xs=xbase-1; ys=ybase-1;
    if(print==1) printf("x=%6.2f y=%6.2f\n",x,y);
    for(iy=0,yb=0,ybin=1,yp=ys,ypos=yp*xdim;iy<4;iy++,ybin*=2,yp++,ypos+=xdim){
      if(print==1 && iy==0) printf("\n%d: %6d  %6d  %6d  %6d ", 0, xs, xs+1, xs+2, xs+3);

      if(inbounds(yp,ydim)) {
	for(ix=0,xb=0,xbin=1,xp=xs,pos=ypos+xp;ix<4;ix++,xbin*=2,xp++,pos++) {
	  if(inbounds(xs,xdim) &&
	     (mask==NULL || (eq && near_val(mask[pos],match_val,0.1)) || (!eq && mask[pos]>match_val))) {
	    x_arr[ix]=im[pos]; xb+=xbin; }
	  else 	  x_arr[ix]=0;
	}
	if(print==1) {
	  printf("\n%d: %6.2f  %6.2f  %6.2f  %6.2f ",
	       yp, x_arr[0], x_arr[1], x_arr[2], x_arr[3]);
	printf("  xb=%d : ",xb);
	}
	switch (xb) {
	case 15:   /* cubic */
	  y_arr[iy]= x_coeff_1 * x_arr[0] + x_coeff0 * x_arr[1] +
	            x_coeff1  * x_arr[2] + x_coeff2 * x_arr[3];
	  if(print==1) printf("cubic  result = %6.2f ",y_arr[iy]);
	  yb+=ybin; break;
	case 6: case 7: case 14:  /* linear */
/*	  frac = x - xbase; */
	  y_arr[iy] = (1-frac)*x_arr[1] + frac*x_arr[2]; yb+=ybin;
	  if(print==1) printf("linear result = %6.2f ",y_arr[iy]);
	  break;
	default: /* other combo ignore*/
	  y_arr[iy]=0;
	  if(print==1) printf("none   result = %6.2f ",y_arr[iy]);
	  break;
	  }
	if(print==1) printf("yb=%d,ybin=%d",yb,ybin);
      }
      else y_arr[iy]=0;
    }
    if(print==1) {
      printf("\ny_arr: %6.2f  %6.2f  %6.2f  %6.2f ",y_arr[0], y_arr[1], y_arr[2], y_arr[3]);
      printf("  yb=%d : ",yb);
    }
    switch (yb) {
    case 15:   /* cubic in y */
      frac = y - ybase; frac2 = frac*frac; frac3 = frac*frac2;
      res_val = (frac2/2.0 - frac/3.0 - frac3/6.0)  * y_arr[0] +
	(1.0 - frac2  + (frac3 - frac)/2.0) * y_arr[1] +
	  (frac + (frac2 - frac3)/2.0)        * y_arr[2] +
	    ((frac3 - frac)/6.0)                * y_arr[3];
      if(print==1) printf("  yb=%d : cubic  result = %6.2f ",yb,res_val);
      break;
    case 6: case 7: case 14:   /* linear in y*/
      frac = y - ybase;
      res_val= (1-frac)*y_arr[1] + frac*y_arr[2];
      if(print==1) printf("  yb=%d : linear result = %6.2f ",yb,res_val);
      break;
    default: /* other combo ignore*/
      res_val=0;
      if(print==1) printf("  yb=%d : none   result = %6.2f ",yb,res_val);
      break;
    }
  }
  else {
    if(print==1) {
      printf("not inbounds setting values to 0.0\n");fflush(stdout);
      res_val=0;
    }
  }

  if(print==1){
    printf("\n********** value=%6.2f\n",res_val);fflush(stdout);
  }
  
  if(print==1) fflush(stdout);
  return((float)res_val);
}



/*******************************************************/
enhance_average_images(image_mgr)
  Image_Manager *image_mgr;
{
  extern char *args;

  float *im=NULL, *res_save=NULL;
  int i,j;

  char *yn=image_mgr->yn;
  char *name=image_mgr->image_name;

  Image_Cache *cache=NULL,*cache_res=NULL;
  int start,end,nframes,rxdim,rydim,rimsize;
  float *res=NULL,*count=NULL,*comp_param=NULL; /* these are malloc-ed here */

  float xzoom,yzoom;
  int ref_frame;
  float factor[6],pre_factor[6],p_res1[6],p_res2[6];
  float xc,yc;
  Affine_Analysis *an=NULL;
  float *cur_param=NULL;
  
  int imsize;
  
  char *progn="enhance_average_images";

  if((cache=ask_name_get_cache(image_mgr))==NULL)
    return(0);

  imsize=cache->imsize;

  sprintf(yn,"q");
  printf("Do you want to use current results of Affine Motion Analysis(y/n/q)? (%s) >> ",yn);
  readline(args,ARGSIZE,stdin); sscanf(args,"%s",yn);
  if(strncmp(yn,"q",1)==0) {return(0);}
  else if(strncmp(args,"n",1)==0) {
    run_global_affine_motion_analysis(image_mgr);
    an= &cache->affine_analysis; start=an->start; end=an->end+1;
  }
  else if(strncmp(args,"y",1)==0) {
    cache=find_cache_name(image_mgr,name);
    an= &cache->affine_analysis; start=an->start; end=an->end+1;

    printf("current affine motion estimation results for frame %d to %d\n",
	   an->start,an->end);

    printf("Enter Start Frame: (%d) >> ",start);
    readline(args,ARGSIZE,stdin); sscanf(args,"%d",&start);

    printf("Enter End Frame: (%d) >> ",end);
    readline(args,ARGSIZE,stdin); sscanf(args,"%d",&end);

    if (start < an->start) start=an->start;
    if (end > (an->end+1)) end=an->end+1;

    printf("%s: Using current parameters in memory\n",progn);
  }

  printf("%s: Enhancing with frames %d to %d\n",progn,start,end);

  ref_frame=(int)((end+start)/2);
  printf("Enter Reference Frame: (%d) >> ",ref_frame);
  readline(args,ARGSIZE,stdin); sscanf(args,"%d",&ref_frame);

  cache=find_cache_name(image_mgr,name);
  an= &cache->affine_analysis;
  
  nframes=end-start+1;
 
  xzoom=1;yzoom=1;
  printf("Enter Amount of XZoom  and YZoom Enhance: (%6.2f %6.2f) >> ",xzoom,yzoom);
  readline(args,ARGSIZE,stdin); sscanf(args,"%f %f",&xzoom,&yzoom);
  for(i=0;i<6;i++) {
    factor[i]=1.0;pre_factor[i]=1.0;
  }

  rxdim=cache->xdim;
  printf("Enter result xdim: (%d) >> ",rxdim);
  readline(args,ARGSIZE,stdin); sscanf(args,"%d",&rxdim);

  rydim=cache->ydim;
  printf("Enter result ydim: (%d) >> ",rydim);
  readline(args,ARGSIZE,stdin); sscanf(args,"%d",&rydim);

  rimsize=rxdim*rydim;
  
  printf("ref_frame=%d,start=%d,end=%d\n",ref_frame,start,end);

  xc=(an->xstart + an->xend)/2;  /* x center of analysis region */
  yc=(an->ystart + an->yend)/2;  /* y center of analysis region */

  printf("\tCenter Region: (%6.2f %6.2f) : ",xc,yc);
  readline(args,ARGSIZE,stdin);
  if(strlen(args)!=0) { sscanf(args,"%f %f",&xc,&yc); }

  /******************* allocate memory ********************/

  if((res=(float *)malloc((unsigned)rimsize*sizeof(float)))==NULL) {
    printf("cannot allocate memory for analysis (im_flow)\n");
    free((char *)res); 
    return(-1);
  }
  if((count=(float *)malloc((unsigned)rimsize*sizeof(float)))==NULL) {
    printf("cannot allocate memory for analysis (im_flow)\n");
    free((char *)res); free((char *)count);
    return(-1);
  }
  if((comp_param=(float *)malloc((unsigned)(nframes*PARAM_WIDTH)*sizeof(float)))==NULL) {
    printf("cannot allocate memory for analysis (im_flow)\n");
    free((char *)res); free((char *)count); free((char *)comp_param);
    return(-1);
  }

  /***************************************************/

  for(i=0;i<6;i++) {  p_res1[i]=0;}
  p_res1[0]=(cache->ydim)/2 - yc; p_res1[3]=(cache->xdim)/2 - xc;
  zoom_param(p_res2,xzoom,yzoom);

/*
  printf("p_res1: (%6.2f %6.4f  %6.4f %6.2f %6.4f  %6.4f)\n",
	 p_res1[0],p_res1[1],p_res1[2],p_res1[3],p_res1[4],p_res1[5]);
  printf("p_res2: (%6.2f %6.4f  %6.4f %6.2f %6.4f  %6.4f)\n",
	 p_res2[0],p_res2[1],p_res2[2],p_res2[3],p_res2[4],p_res2[5]);
*/

  printf("\n\nPerforming temporal average filtering. Please wait .....\n\n");

  fill_im(res,(float)0,rimsize);
  fill_im(count,(float)0,rimsize);
 
  for(i=start;i<=end;i++) {
    while((im=get_image(cache,i))==NULL) {
      fprintf(stderr,"%s: cannot find image %d... continuing....\n",progn,i);
    }
    cur_param= &comp_param[PARAM_WIDTH*(i-start)];
    internal_update_affine_params_array(cache->param,PARAM_WIDTH,ref_frame,i,
				       cur_param,factor,pre_factor);

    internal_update_affine_params_from(cur_param,p_res1,cur_param,factor); /* shift center */
    internal_update_affine_params_from(cur_param,p_res2,cur_param,factor); /* apply zoom */

    bicubic_affine_warp_weight_abs
      (im, res,
       cur_param[3],cur_param[4],cur_param[5],
       cur_param[0],cur_param[1],cur_param[2],
       cache->ydim,cache->xdim,
/*       an->xstart,an->ystart,an->xend,an->yend, */
       (int)5,(int)5,cache->xdim-5,cache->ydim-5,
       rydim,rxdim,(int)0,(int)0,rxdim,rydim,
       count);

  }

/*  fill_im(count,(float)1.0,rimsize); */

  div_im(res,count,res,rimsize);

  sprintf(args,"Average Enhance Image: %d-%d-c%d", start,end,ref_frame);
  display_float_image_opt(res,rxdim,rydim,args,cache->gamma,(float)1.0);

  /* save results internally */
  sprintf(yn,"n"); printf("\tDo you want to save result internally (y/n)? (n) : ");
  readline(args,ARGSIZE,stdin);  sscanf(args,"%s",yn);
  if(strncmp(yn,"y",1)==0) {
    sprintf(image_mgr->image_name,"%s-avg-enhance",cache->image_name);
    if((cache_res=make_internal_image(image_mgr,rxdim,rydim,(int)1))==NULL) {
      printf("Cannot allocate memory for result image\n"); 
    }
    if((res_save=get_image(cache_res,0))==NULL) {
      printf("%s: Progn error. Cannot get cache_res float array.\n",progn);
      return(-1); }
    copy_im(res,res_save,(int)(rxdim*rydim));
    printf("Results saved to %s....\n",cache_res->image_name);
  }

  sprintf(yn,"n");
  printf("\tDo you want to save output to disk (y/n)? (n) : ");
  readline(args,ARGSIZE,stdin); sscanf(args,"%s",yn);
  if(strncmp(yn,"y",1)==0) {
    sprintf(args, "MCNR average ehancement of %s\nstart=%d,end=%d\n",
	    cache->filename,start,end);
    save_image_to_file(image_mgr,cache,res,rxdim,rydim,"enhance",args);
    }

  free((char *)res);free((char *)count);free((char *)comp_param);
  return(0);
}




/* affine warp only regions within: rxstart,rystart,rxend,ryend (not include end).
 * also check for input crop region. 
 * code is same as bicubic_affine_warp except for the checking in the result loop */
/* returns a weighted square distance to the nearest pixel */
/* adds to resulting image */
#define abs_val(x) (((x) < 0) ? -(x) : (x))
#define roundoff(x) (((x)>0)?((int)((x)+0.5)):((int)((x)-0.5)))

void bicubic_affine_warp_weight_abs(im, res, ax, bx, cx, ay, by, cy,
				       ydim, xdim,xstart,ystart,xend,yend,
				       rydim,rxdim,rxstart,rystart,rxend,ryend,
				       weight)
  float *im, *res;
  float ax, bx, cx, ay, by, cy;
  int xdim, ydim;
  int rxdim,rydim;
  int rxstart,rystart,rxend,ryend;
  int xstart,ystart,xend,yend;
  float *weight;

{
  register float *im_ptr;	/* accessed 38x in loop */
  register int base;		/* accessed 30x in loop */
  register float frac, temp;	/* accessed 20x in loop */
  register float frac2, frac3;  /* accessed 10x in loop */
  int x_inbounds_1, x_inbounds0, x_inbounds1, x_inbounds2; /* 6x in loop */
  register float x_coeff_1, x_coeff0, x_coeff1, x_coeff2; /* 5x in loop */
  register float res_val;
  register int i,j;
  register float xfrac,xbase,yfrac,ybase;
  float x_off, y_off;
  float rx_off,ry_off,x_diff,y_diff;
  int pos;
  float xx,yy,weight_val;

  x_off = (xdim - 1) / 2.0; rx_off = (rxdim - 1) / 2.0;
  y_off = (ydim - 1) / 2.0; ry_off = (rydim - 1) / 2.0;
  x_diff = rx_off - x_off; y_diff = ry_off - y_off;

  /*
  **  wx = ax + bx * x + cx * y;
  **  wy = ay + by * x + cy * y;
  */

/*
  printf ("\nbicubic_affine_warp_limit2_weight: (%5f %5f %5f %5f %5f %5f)", ay, by, cy, ax, bx, cx);
  printf( "\n(%5f %5f: %d %d)", y_off, x_off, ydim, xdim);
*/
  
/*  memset(res, 0, xdim*ydim*sizeof(float)); */

  if(rxstart<0) rxstart=0;  if(rxend>rxdim) rxend=rxdim-1;
  if(rystart<0) rystart=0;  if(ryend>rydim) ryend=rydim-1;

  if(xstart<0) xstart=0;  if(xend>xdim) xend=xdim-1;
  if(ystart<0) ystart=0;  if(yend>ydim) yend=ydim-1;

/*
  printf("bicubic_affine_warp_limit2_weight: xdim=%d, rydim=%d: xs=%d ys=%d xe=%d ye=%d\n", xdim,ydim,xstart,ystart,xend,yend);
  printf("bicubic_affine_warp_limit2_weight: xdim=%d, rydim=%d: rxs=%d rys=%d txe=%d rye=%d\n", rxdim,rydim,rxstart,rystart,rxend,ryend);
*/

  for(j=rystart; j<rydim; j++)  /* j is the y var */
    for(i=rxstart,pos=j*rxdim+i; i<rxdim; i++,pos++) /* i is the x var */
      {
	xfrac = i - x_diff - (ax + bx * (i - rx_off) + cx * (j - ry_off));
	xbase = (int)(xfrac);
	yfrac = j - y_diff - (ay + by * (i - rx_off) + cy * (j - ry_off));
	ybase = (int)(yfrac);

	if((((xbase-1) >= xstart) || ((xbase + 3) >= xstart)) &&
	   (((xbase-1) < xend) || ((xbase + 3) < xend)) &&
	   (((ybase-1) >= ystart) || ((ybase + 3) >= ystart)) &&
	   (((ybase-1) < yend) || ((ybase + 3) < yend))
	   )
	  {
	    xx = xfrac-roundoff(xfrac); xx=abs_val(xx);
	    yy = yfrac-roundoff(yfrac); yy=abs_val(yy);
	    weight_val =  (float)(1/((abs_val(xx) + abs_val(yy)) + 1.0));

	    res_val = 0.0;
	    /* Compute X fractional position and base index */
	    frac = xfrac - xbase;
	    frac2 = frac*frac;
	    frac3 = frac*frac2;
	    base  = xbase - 1;
	    im_ptr = im + base;
	    
	    x_inbounds_1 = ((base >= 0) && (base < xdim))?1:0;
	    if (x_inbounds_1) x_coeff_1 = (frac2/2.0 - frac/3.0 - frac3/6.0);
	    base++;
	    x_inbounds0  = ((base >= 0) && (base < xdim))?1:0;
	    if (x_inbounds0)  x_coeff0  = (1.0 - frac/2.0 - frac2 + frac3/2.0);
	    base++;
	    x_inbounds1  = ((base >= 0) && (base < xdim))?1:0;
	    if (x_inbounds1)  x_coeff1 = (frac + frac2/2.0 - frac3/2.0);
	    base++;
	    x_inbounds2  = ((base >= 0) && (base < xdim))?1:0;
	    if (x_inbounds2)  x_coeff2 = (frac3/6.0 - frac/6.0);
	    
	    /* Compute Y fractional position and base index */
	    frac = yfrac - ybase;
	    frac2 = frac*frac;
	    frac3 = frac*frac2;
	    base = ybase - 1 ;
	    im_ptr += base*xdim;
	    
	    if ((base >= 0) && (base < ydim))  /* y_inbounds_1 */
	      {
		temp = 0.0;
		if (x_inbounds_1) temp += x_coeff_1 * *im_ptr;
		im_ptr++;
		if (x_inbounds0)  temp += x_coeff0  * *im_ptr;
		im_ptr++;
		if (x_inbounds1)  temp += x_coeff1  * *im_ptr;
		im_ptr++;
		if (x_inbounds2)  temp += x_coeff2  * *im_ptr;
		im_ptr -= 3;
		res_val += temp * (frac2/2.0 - frac/3.0 - frac3/6.0);
	      }
	    base++;  im_ptr += xdim;
	    if ((base >= 0) && (base < ydim))  /* y_inbounds0 */
	      {
		temp = 0.0;
		if (x_inbounds_1) temp += x_coeff_1 * *im_ptr;
		im_ptr++;
		if (x_inbounds0)  temp += x_coeff0  * *im_ptr;
		im_ptr++;
		if (x_inbounds1)  temp += x_coeff1  * *im_ptr;
		im_ptr++;
		if (x_inbounds2)  temp += x_coeff2  * *im_ptr;
		im_ptr -= 3;	    
		res_val += temp * (1 - frac/2.0 - frac2 + frac3/2.0);
	      }
	    base++;  im_ptr += xdim; 
	    if ((base >= 0) && (base < ydim))  /* y_inbounds1 */
	      {
		temp = 0.0;
		if (x_inbounds_1) temp += x_coeff_1 * *im_ptr;
		im_ptr++;
		if (x_inbounds0)  temp += x_coeff0  * *im_ptr;
		im_ptr++;
		if (x_inbounds1)  temp += x_coeff1  * *im_ptr;
		im_ptr++;
		if (x_inbounds2)  temp += x_coeff2  * *im_ptr;
		im_ptr -= 3;	    
		res_val += temp * (frac + frac2/2.0 - frac3/2.0);
	      }
	    base++;  im_ptr += xdim;
	    if ((base >= 0) && (base < ydim))  /* y_inbounds2 */
	      {
		temp = 0.0;
		if (x_inbounds_1) temp += x_coeff_1 * *im_ptr;
		im_ptr++;
		if (x_inbounds0)  temp += x_coeff0  * *im_ptr;
		im_ptr++;
		if (x_inbounds1)  temp += x_coeff1  * *im_ptr;
		im_ptr++;
		if (x_inbounds2)  temp += x_coeff2  * *im_ptr;
		im_ptr -= 3;	    
		res_val += temp * (frac3/6.0 - frac/6.0);
	      }
	    res[pos] += (float)(weight_val*res_val);
	    weight[pos] += (float)weight_val;
	  }
      }
  fflush(stdout);
}
