/******************************************************************

"Copyright  1994 by the Massachusetts  Institute  of  Technology.  All
rights reserved.

Developed by John Y. A. Wang at the  Media Laboratory, MIT, Cambridge,
Massachusetts, with  support  from Secom Ltd., Goldstar Ltd.,  and the
Television of Tomorrow Program.

For use by authorized users only.

This distribution is approved  by Nicholas Negroponte, Director of the
Media Laboratory, MIT.

Permission to use, copy, or modify this software and its documentation
for educational and research purposes only and without  fee is  hereby
granted, provided that this copyright notice and the original authors'
names  appear  on  all  copies  and   supporting   documentation.   If
individual  files  are  separated  from  this  distribution  directory
structure, this copyright notice must be included.  For any other uses
of  this software, in  original or  modified form, including  but  not
limited to distribution in whole or in part, specific prior permission
must be  obtained  from  MIT.   These  programs  shall  not  be  used,
rewritten,  or  adapted  as  the  basis  of a  commercial software  or
hardware  product without  first  obtaining  appropriate licenses from
MIT.  MIT.  makes no  representations about  the  suitability of  this
software for any purpose.  It is  provided "as is" without  express or
implied warranty.

******************************************************************/


/***************************************************************
*** File: fileio.c
*** Author: John Y. A. Wang
*** Date: January 15, 1993
*** Copyright 1993, Massachusetts Institute of Technology, All Rights Reserved.
***
*** Description: fileio routines
***************************************************************/

/*****************************************************************/
#include "dynamo.h"
/*****************************************************************/

/********** IMPORTANT:    **********************************/
/********** put copyright label in all the files ***********/

static  char *copyright="\
\n\n***********************************************\n\
Developed by John Y. A. Wang at the Media Laboratory, MIT, Cambridge,Massachusetts\n\
Copyright  1993 by the Massachusetts  Institute  of  Technology.  All rights reserved.\n\
***********************************************\n\n";


/*
  cc -O4 -fsingle -c fileio.c 
  SUN:  cc -O4 -fsingle -c fileio.c

  lint fileio.c
*/
/*****************************************************************/

/*****************************************************************/
FILE *open_file(filename,type)
  char *filename;
  char *type;
{
  extern char *args;

  int tt;
  int checkfile=0;
  char yn[20];
  FILE *fp;
  char *progn="open_file";
  
  if(strncmp(type,"w",1)==0)
    tt=WRITE;
  else if(strncmp(type,"r",1)==0)
    tt=READ;
  
  while(checkfile==0) {
    sprintf(yn,"q");
    printf("%s %s ok (y/n/q) ? (q) >> ",((tt==READ)? "Input from":
				((tt==WRITE)?"Output to":"Unknown I/O on")),filename);
    readline(args,ARGSIZE,stdin);
    sscanf(args,"%s", yn);

    if(strncmp(yn,"y",1)==0) {
      printf("opening file %s\n",filename);
      if((fp=fopen(filename,type))==NULL) {
	printf("%s: cannot open %s for I/O (%s)\n",progn,filename,type);
      }
      else {
	checkfile=1;
      }
    }
    else if(strncmp(yn,"n",1)==0) {
      printf("Enter new filename >> ");
      readline(args,ARGSIZE,stdin);
      sscanf(args,"%s",filename);
    }
    else
      return(NULL);
  }
  return(fp);
}
/*****************************************************************/

/*****************************************************************/
/*********************************/
int save_dir(savedirname)
  char *savedirname;
{
  extern char *args;

  mode_t dir_mask;
  char yn[5];

  dir_mask = 0755;  /* drwxr-xr-x */

  while(1) {
    printf("\n\tEnter output directory: (%s) :\n\t\t>> ",savedirname);
    readline(args,ARGSIZE,stdin);
    if(strlen(args) != 0)
      sscanf(args,"%s", savedirname);
  
    sprintf(yn,"y");
    printf("Save images to directory %s?(y/n/q) : (%s) >> ", savedirname,yn);
    readline(args,ARGSIZE,stdin);
    sscanf(args,"%s", yn);
    if(strncmp(args,"q",1)==0)
      return(-1);

    printf("\n\tCreating Directory %s ... \n", savedirname);fflush(stdout);
    if(mkdir(savedirname,(mode_t)dir_mask)!=0) {
      if(errno == EEXIST) {
	sprintf(yn,"n");
	printf("\n\t ****** directory exits... Overwrite (y/n)? (%s) >> ",yn);
	readline(args,ARGSIZE,stdin);
	sscanf(args,"%s", yn);
	if(strncmp(yn,"y",1)==0) {return(0);}
      }
      else {
	printf("\t   .....Cannot create %s\n",savedirname);
      }
    }
    else {
/*      printf("done\n"); */
      return(0);
    }
  }
}

/*****************************************************************/
/*********************************/
int save_descriptor(filename,xsize,ysize,datasets,datafiles,channels,
		    scale,pedestal,comment)
  char *filename;
  int ysize,xsize,channels;
  float scale, pedestal;
  char *comment;
{
  FILE *fp;

/*  printf("\nWriting Descriptor File %s:\n",filename); */
  if ((fp=fopen(filename,"w")) == NULL) {
    printf("******* cannot open descriptor file: %s ********\n", filename);
    return(-1);
  }
  fprintf(fp,"(_data_sets %d)\n", datasets); 
  if(datafiles>1)
    fprintf(fp,"(_data_files %d)\n",datafiles);
  fprintf(fp,"(_channels %d)\n", channels);
  fprintf(fp,"(_dimensions %d %d)\n",ysize, xsize);
/*  fprintf(fp,"(_data_type \"unsigned_1\")\n"); */
  fprintf(fp,"(_data_type unsigned_1)\n"); 
  fprintf(fp,"(scale %f)\n",(float) (1/scale));
  fprintf(fp,"(pedestal %f)\n", pedestal);
  fprintf(fp,"(_comment \"DYNAMO: %s \")",comment);
  close_file(fp);
  return(0);
}
/*******************************************************/
/*******************************************************/
int save_image_to_file(image_mgr,cache,image,xdim,ydim,file_extension,comment)
  Image_Manager *image_mgr;
  Image_Cache *cache;
  float *image;
  int xdim,ydim;
  char *file_extension;
  char *comment;
{  

  char *dirname=image_mgr->cbuf1;
  char *save_filename=image_mgr->cbuf2;
/*  char *yn=image_mgr->yn; */
  FILE *fp;

  unsigned char *array=NULL;   /* this is malloc-ed here */
  int imsize=xdim*ydim;
  int num;
  char *progn="save_image_to_file";
  

  sprintf(dirname,"%s-%s",cache->savedirname,file_extension);
  if(save_dir(dirname)==0) {
    sprintf(save_filename,"%s/data",dirname);
    fp=NULL;
    if((fp=open_file(save_filename,"w"))==NULL) {
      printf("%s: cannot open file %s. Not saving \"%s\" results to %s.\n",
	     save_filename,file_extension);
      return(-1);
    }

    if((array=(unsigned char *)malloc((unsigned)imsize))==NULL) {
      printf("%s: Cannot allocate memory to convert image\n");
      return(-1);
    }
      
    convert_float_to_byte(image,array,imsize);

    if((num=fwrite(array,sizeof(unsigned char),imsize,fp))!=imsize) {
      printf("%s: incomplete write. %d of %d bytes written to %s\n",progn,num,imsize,save_filename);
    }
    if(fp!=NULL) {
      sprintf(save_filename,"%s/descriptor",dirname);
      save_descriptor(save_filename,xdim,ydim,(int)1,(int) 1,(int)1,
		      (float)1.0,(float)0.0,comment);
    }
  }

  free((char *) array);
  close_file(fp);
  return(0);
}


/* protected fclose. Check for NULL fp before closing */
int close_file(fp)
  FILE *fp;
{
  if(fp==NULL) { return(0); }
  else { return(fclose(fp));}
}
	   


char *readline(/*_ args,size,fp _*/);

/* reads a line and ingores lines beginning with semicolon */
char *readline(buf,size,fp)
  char *buf;
  int size;
  FILE* fp;
{
  char temp=0;
  
  fgets(buf,size,fp); 
  while(buf[0]==';') {   /* ; on the first char on the line is comment */
    fgets(buf,size,fp);
  } 

  sscanf(buf,"%1s",&temp); 
  if(temp==COMMENT_CHAR) {  /* whitespace followed by ; is null args */
    sprintf(buf,"\n");
  }
  
/*  printf("buf=%s\n",buf); */
  return(buf);
}


/*************************************************************/
/** free all the malloced images.  The idea here is to keep all the
 ** pointers of malloced memory in an array to ease in cleaning up.
 **/
void free_images(malloc_images,nimages)
  char **malloc_images;
  int nimages;
{
  int i;
  if(malloc_images==NULL) return;
  for(i=0;i<nimages;i++) {
    if(malloc_images[i] != NULL) {
      free(malloc_images[i]); malloc_images[i]=NULL;
    }
  }
}

/*************************************************************/
/** close all the opened files.  The idea here is to keep all the
 ** file pointers of opened files in an array to ease in cleaning up
 **/
void close_files(opened_files,nfiles)
  FILE **opened_files;
  int nfiles;
{
  int i;
  if(opened_files==NULL) return;
  for(i=0;i<nfiles;i++) {
    if(opened_files[i] != NULL) {
      if(fclose(opened_files[i])==0) { opened_files[i]=NULL; } 
      else { fprintf(stderr,"error closing file #%d \n",i);}
    }
  }
}

