/******************************************************************

"Copyright  1994 by the Massachusetts  Institute  of  Technology.  All
rights reserved.

Developed by John Y. A. Wang at the  Media Laboratory, MIT, Cambridge,
Massachusetts, with  support  from Secom Ltd., Goldstar Ltd.,  and the
Television of Tomorrow Program.

For use by authorized users only.

This distribution is approved  by Nicholas Negroponte, Director of the
Media Laboratory, MIT.

Permission to use, copy, or modify this software and its documentation
for educational and research purposes only and without  fee is  hereby
granted, provided that this copyright notice and the original authors'
names  appear  on  all  copies  and   supporting   documentation.   If
individual  files  are  separated  from  this  distribution  directory
structure, this copyright notice must be included.  For any other uses
of  this software, in  original or  modified form, including  but  not
limited to distribution in whole or in part, specific prior permission
must be  obtained  from  MIT.   These  programs  shall  not  be  used,
rewritten,  or  adapted  as  the  basis  of a  commercial software  or
hardware  product without  first  obtaining  appropriate licenses from
MIT.  MIT.  makes no  representations about  the  suitability of  this
software for any purpose.  It is  provided "as is" without  express or
implied warranty.

******************************************************************/


/***************************************************************
*** File: flow-src.c
*** Author: John Y. A. Wang
*** Date: January 15, 1993
*** Copyright 1993, Massachusetts Institute of Technology, All Rights Reserved.
***
*** Description: optic flow using gradient method
***************************************************************/

/*****************************************************************/
#include "dynamo.h"
/*****************************************************************/

/********** IMPORTANT:    **********************************/
/********** put copyright label in all the files ***********/

static  char *copyright="\
\n\n***********************************************\n\
Developed by John Y. A. Wang at the Media Laboratory, MIT, Cambridge,Massachusetts\n\
Copyright  1993 by the Massachusetts  Institute  of  Technology.  All rights reserved.\n\
***********************************************\n\n";

/*
  cc -O4 -fsingle -c flow-src.c 
  SUN:  cc -O4 -fsingle -c flow-src.c

  lint flow-src.c
*/
/*****************************************************************/


/*****************************************************************/
run_optical_flow_analysis(image_mgr)
  Image_Manager *image_mgr;
{
  extern char *args;

  float *im2=NULL,*temp=NULL;   /* these are not malloc-ed here*/
  float *im1=NULL,*xflow=NULL,*yflow=NULL, *im_offset=NULL,*ttemp=NULL; /* these are malloc-ed here */
  int ttempsize=0;

  Optical_Analysis *an;
  Image_Cache *cache;

  int i;
  char *title=image_mgr->cbuf1;
  char *yn=image_mgr->yn;
  FILE *xpid,*ypid,*flowpid,*xflow_fp,*yflow_fp,*vflow_fp;
  int ndigits;
  int imsize;
  float d_clipmin= -10,d_clipmax=10;
  
  float zoom=1.0,pedestal,scale;
  
  int analysis_type=1;
  int valid=0;
  char *save_filename=image_mgr->cbuf1;
  int num;

  int xstep,ystep;
  float flow_scale;
  int headsize;

  int overlay;
  int save_float=0;

  Image_Cache *cache_xflow, *cache_yflow;
  int length;

  /**************** Analysis *******************************/
  printf("Running Optic Flow analysis with the following options:\n");

  if((cache=ask_name_get_cache(image_mgr))==NULL)
    return(0);

  imsize=cache->imsize;
  ttempsize=10*imsize;
  an= &(cache->optical_analysis);
  
  change_optical_flow_options(cache);
  
  valid=0;
  while(!valid) {
    printf("\n\tThere are two optical flow analysis types:\n");
    printf("\t  1) 2 Parameter Analyis.\n");
    printf("\t  2) 3 Parameter Analyis (allows for changes in intensity).\n");
    printf("\n\tEnter Optical Flow Type: (%d) : ",analysis_type);
    readline(args,ARGSIZE,stdin);
    sscanf(args,"%d",&analysis_type);

    
    if(strncmp(args,"q",1)==0)
      return(0);
      
    switch(analysis_type) {
    case 1:
      printf("Motion Analysis using 2 parameter model.\n");
      ttempsize=10*imsize; valid=1; break;
    case 2:
      printf("Motion Analysis using 3 parameter model.\n");
      ttempsize=12*imsize;
      if((im_offset=(float *)malloc((unsigned)imsize*sizeof(float)))==NULL) {
	printf("cannot allocate memory for analysis (im_flow)\n");
	free((char *)im_offset);
	return(-1);
      }
      valid=1; break;
    default:
      printf("%d Option not currently support.\n",analysis_type);
      valid=0; break;
    }
  }

  printf("\tEnter Display Flow Min Max: (%6.2f %6.2f) : ", d_clipmin,d_clipmax);
  readline(args,ARGSIZE,stdin);
  sscanf(args,"%f %f",&d_clipmin,&d_clipmax);

  pedestal=d_clipmin;
  scale=(float)(255/(d_clipmax-d_clipmin));


  xstep=10; ystep=10;flow_scale=1.0; headsize=2;
  printf("\tEnter Display Flow Field :  \n\t  (xstep,ystep,scale,headsize): (%d %d %6.2f %d): "
	 ,xstep,ystep,flow_scale,headsize);
  readline(args,ARGSIZE,stdin);
  sscanf(args,"%d %d %f %d",&xstep,&ystep,&flow_scale,&headsize);


  sprintf(yn,"n");overlay=0;
  printf("\tOverlay flow field on image(y/n)? (%s): ",yn);
  readline(args,ARGSIZE,stdin);
  if((strlen(args)!=0) && (strncmp(args,"y",1)==0))
    overlay=1;

  /******* save results ********/
  sprintf(yn,"n");
  xflow_fp=NULL;yflow_fp=NULL;vflow_fp=NULL;
  printf("\tDo you want to save output optical flow to files(y/n)? (%s): ",yn);
  readline(args,ARGSIZE,stdin);
  if((strlen(args)!=0) && (strcmp(args,"\n")!=0))
    sscanf(args,"%s",yn);
  if(strncmp(yn,"y",1)==0) {

    save_float=0; sprintf(yn,"n");
    printf("\tDo you want to save output as float(y/n)? (%s): ",yn);
    readline(args,ARGSIZE,stdin);
    if((strlen(args)!=0) && (strcmp(args,"\n")!=0))
      sscanf(args,"%s",yn);
    if(strncmp(yn,"y",1)==0)
      save_float=1;

    if(save_dir(cache->savedirname)==0) {

      xflow_fp=NULL;yflow_fp=NULL;

      printf("\nSave XFLOW: \n");fflush(stdout);
      sprintf(save_filename,"%s/xflow",cache->savedirname);
      if((xflow_fp=open_file(save_filename,"w"))==NULL) {
	printf("Not saving xflow results to %s.\n",save_filename);
      }

      printf("\nSave YFLOW: \n");fflush(stdout);
      sprintf(save_filename,"%s/yflow",cache->savedirname);
      if((yflow_fp=open_file(save_filename,"w"))==NULL) {
	printf("Not saving yflow results to %s.\n",save_filename);
      }


      printf("\nSave FLOWFIELD: \n");fflush(stdout);
      sprintf(save_filename,"%s/vflow",cache->savedirname);
      if((vflow_fp=open_file(save_filename,"w"))==NULL) {
	printf("Not saving vector field results to %s.\n",save_filename);
      }
      if(xflow_fp!=NULL || yflow_fp!=NULL || vflow_fp!=NULL) {
	sprintf(save_filename,"%s/descriptor",cache->savedirname);
	printf("saving descriptor... scale=%f,pedestal=%f\n",scale,pedestal);
	sprintf(args,"clipmin=%f, clipmax=%f : directory containing xflow, yflow, vector-field images \n Note images saved as %s.",
		d_clipmin, d_clipmax, ((save_float)? "float": "unsigned byte"));
	save_descriptor(save_filename,cache->xdim,cache->ydim,
			((an->end)-(an->start)+1),(int)1,(int)1,
			(float)scale,(float)pedestal,args);
      }
    }
  }

  /***************************************************************/
  sprintf(yn,"n");
  printf("\tDo you want to save optic flow results internally (y/n)? (%s): ",yn);
  readline(args,ARGSIZE,stdin);
  sscanf(args,"%s",yn);
  if(strncmp(yn,"y",1)==0) {
    length=an->end-an->start+1;
    sprintf(image_mgr->image_name,"%s-xflow",cache->image_name);
    if((cache_xflow=make_internal_image(image_mgr,cache->xdim,cache->ydim,length))
       ==NULL) {
      printf("Not saving xflow results internally.\n"); 
/*      init_cache(cache_xflow); cache_xflow=NULL; */
    }
/*    printf("x: okay\n"); */
    sprintf(image_mgr->image_name,"%s-yflow",cache->image_name);
    if((cache_yflow=make_internal_image(image_mgr,cache->xdim,cache->ydim,length))
       ==NULL) {
      printf("Not saving yflow results internally.\n"); 
/*      init_cache(cache_yflow); cache_yflow=NULL; */
    }
  }
  else {
    cache_xflow=NULL; cache_yflow=NULL;
  }

  /***************************************************************/
  
  if((ttemp=(float *)malloc((unsigned)ttempsize*sizeof(float)))==NULL) {
    printf("cannot allocate memory for analysis\n");
    free((char *)im_offset);free((char *)ttemp);
    return(-1);
  }

  if((cache_xflow==NULL) &&
     ((xflow=(float *)malloc((unsigned)imsize*sizeof(float)))==NULL)) {
    printf("cannot allocate memory for analysis (xflow)\n");
    free((char *)im_offset);free((char *)ttemp);free((char *)xflow);
    return(-1);
  }
  if((cache_yflow==NULL) &&
     ((yflow=(float *)malloc((unsigned)imsize*sizeof(float)))==NULL)) {
    printf("cannot allocate memory for analysis (yflow)\n");
    free((char *)im_offset);free((char *)ttemp);free((char *)xflow);
    free((char *)yflow);
    return(-1);
  }

  if((im1=(float *)malloc((unsigned)imsize*sizeof(float)))==NULL) {
    printf("cannot allocate memory for analysis (im_flow)\n");
    free((char *)im_offset);free((char *)ttemp);free((char *)xflow);free((char *)yflow);
    free((char *)im1);
    return(-1);
  }

  /*****************************/

  sprintf(title,"xflow %d-%d-of%d",an->start,an->end,analysis_type);
  xpid=OpenGrayScreen(cache->ydim,cache->xdim,title,cache->gamma,zoom,2);

  sprintf(title,"yflow %d-%d-of%d",an->start,an->end,analysis_type);
  ypid=OpenGrayScreen(cache->ydim,cache->xdim,title,cache->gamma,zoom,2);
  
  sprintf(title,"flow %d-%d-of%d",an->start,an->end,analysis_type);
  flowpid=OpenGrayScreen(cache->ydim,cache->xdim,title,cache->gamma,zoom,2);

/*  printf("xpid=%d, ypid=%d, flowpid=%d\n",xpid,ypid,flowpid);fflush(stdout);*/

  sprintf(title,"%d",cache->tdim); ndigits=strlen(title);

  /*************************************************/

  printf("Running Optical Flow Analysis ........\n");
  for(i=an->start;i<=an->end;i++) {
    printf("Analysis on frames %d and %d .... \n",i,i+an->tstep);

    if((temp=get_image(cache,i))==NULL) {
      printf("Analysis: cannot get frame %d\n",i);
      break;
    }
    /* save im1 into static space in case cache changes while getting im2 */
    copy_im(temp,im1,cache->imsize);  
    if((im2=get_image(cache,(i+an->tstep)))==NULL){
      printf("Analysis: cannot get frame %d\n",i);
      break;
    }
    
    if(cache_xflow!=NULL) {
/*      printf("Using internal images\n");*/
      if((xflow=get_image(cache_xflow, i-an->start ))==NULL){
	printf("Analysis: cannot get frame %d of xflow result.\n",i);
	break;
      }}

    if(cache_yflow!=NULL) {
/*      printf("Using internal images\n");*/
      if((yflow=get_image(cache_yflow, i-an->start ))==NULL){
	printf("Analysis: cannot get frame %d of yflow result.\n",i);
	break;
      }}
    
    fill_im(xflow,(float)0,imsize);
    fill_im(yflow,(float)0,imsize);
    fill_im(ttemp,(float)0,ttempsize);
/*     printf("im1=%d,im2=%d,xflow=%d,yflow=%d,ttemp=%d\n",im1,im2,xflow,yflow,ttemp); */

    switch(analysis_type) {
    case 1:
    default:
      /* normal 2 parameter optical flow */
      internal_coarse_fine_gradient_flow
	(im1,im2,xflow,yflow,cache->xdim,cache->ydim,
	 an->level,an->method,an->blur,an->noise,
	 ttemp,ttempsize,an->xbias,an->ybias,
	 an->clipmin, an->clipmax, NULL,NULL,
	 an->sum_filt,an->sfdim,an->pyr_filt,an->pfdim,an->exp_filt,an->expfdim,
	 (float)0,an->stat);
      break;
    case 2:
      /* 3 parameter optical flow with extra intersity change parameter */
      fill_im(im_offset,(float)0,imsize);
      internal_coarse_fine_gradient_flow_offset
	(im1,im2,xflow,yflow,im_offset,cache->xdim,cache->ydim,
	 an->level,an->method,an->blur,an->noise,
	 ttemp,ttempsize,an->xbias,an->ybias,
	 an->clipmin, an->clipmax, NULL,NULL,
	 an->sum_filt,an->sfdim,an->pyr_filt,an->pfdim,an->exp_filt,an->expfdim,
	 (float)0,an->stat);
      break;
    }
      
    /******* display xflow *********/
    convert_subfloat_to_byte(xflow,cache->xdim,cache->ydim,
			     0,0,cache->xdim-1,cache->ydim-1,
			     scale,pedestal,
			     cache->array,cache->xdim,cache->ydim);
/*
    rescale_image(xflow,xflow,imsize,(float)d_clipmin,(float)d_clipmax,(float) 0,(float)255);
    convert_float_to_byte(xflow,cache->array,imsize);
*/
    if(xflow_fp!=NULL) {
      printf("saving xflow frame %d... ",i);
      if(save_float==0) {
	if((num=fwrite((char *)(cache->array),sizeof(char),imsize,xflow_fp))!=imsize) {
	  printf("Warning: saved only %d of %d bytes of xflow frame %d\n",num,imsize,i);
	  break;
	}
      }
      else {
	printf("saved as float... "); 
	if((num=fwrite((char *)xflow,sizeof(float),imsize,xflow_fp))!=imsize) {
	  printf("Warning: saved only %d of %d bytes of yflow frame %d\n",num,imsize,i);
	  break;
	}
      }
      printf("num=%d",num);
      printf("\n");
    }
    
    display_number_on_image(cache->array,cache->xdim,cache->ydim,
			    (float) i,(int)5,(int)5,(int)255,(int)0,(int)ndigits);
    DumpBufToScreen(cache->array,cache->ydim,cache->xdim,xpid,cache->imsize);

    /******* display yflow **********/
    convert_subfloat_to_byte(yflow,cache->xdim,cache->ydim,
			     0,0,cache->xdim-1,cache->ydim-1,
			     scale,pedestal,
			     cache->array,cache->xdim,cache->ydim);
    if(yflow_fp!=NULL) {
      printf("saving yflow frame %d... ",i); 
      if(save_float==0) {
	if((num=fwrite((char *)(cache->array),sizeof(char),imsize,yflow_fp))!=imsize) {
	  printf("Warning: saved only %d of %d bytes of yflow frame %d\n",num,imsize,i);
	  break;
	}
      }
      else {
	printf("saved as float... ");
	if((num=fwrite((char *)yflow,sizeof(float),imsize,yflow_fp))!=imsize) {
	  printf("Warning: saved only %d of %d bytes of yflow frame %d\n",num,imsize,i);
	  break;
	}
      }
      printf("num=%d\n",num); 
      printf("\n");
    }
    display_number_on_image(cache->array,cache->xdim,cache->ydim,
			    (float) i,(int)5,(int)5,(int)255,(int)0,(int)ndigits);

    DumpBufToScreen(cache->array,cache->ydim,cache->xdim,ypid,cache->imsize);

    /******** display  vector field *****/
    if(overlay) {
      convert_subfloat_to_byte(im2,cache->xdim,cache->ydim,
			       0,0,cache->xdim-1,cache->ydim-1,
			       (float)1,(float)0,
			       cache->array,cache->xdim,cache->ydim);
    }
    else {
      fill_byte(cache->array,(unsigned char)0,cache->imsize);
    }
/*
    printf("analysis: xstep=%d,ystep=%d,scale=%f,headsize=%d\n",
	   xstep,ystep,flow_scale,headsize);
*/
    display_vector_field(xflow,yflow,cache->xdim,cache->ydim,
			 cache->array,cache->xdim,cache->ydim,(int)255,
			 (int)xstep,(int)ystep,(float)flow_scale,(int)headsize,(int)1);
    if(vflow_fp!=NULL) {
/*      printf("saving xflow frame %d...\n",i);*/
      if((num=fwrite((char *)(cache->array),sizeof(char),imsize,vflow_fp))!=imsize) {
	printf("Warning: saved only %d of %d bytes of vflow frame %d\n",num,imsize,i);
	break;
      }
    }

    display_number_on_image(cache->array,cache->xdim,cache->ydim,
			    (float) i,(int)5,(int)5,(int)255,(int)0,(int)ndigits);
    DumpBufToScreen(cache->array,cache->ydim,cache->xdim,flowpid,cache->imsize);
  }
  fclose(xflow_fp);fclose(yflow_fp);

/*  printf("closing displays....\n");fflush(stdout); */
  Close_Screen(xpid);
/*  printf("closed xflow display\n");fflush(stdout); */

  Close_Screen(ypid);
/*  printf("closed yflow display\n");fflush(stdout); */

  Close_Screen(flowpid);
/*  printf("closed vector flow display\n");fflush(stdout); */

  free((char *)ttemp);
  if (cache_xflow == NULL)
    free((char *)xflow);
  if (cache_yflow == NULL)
  free((char *)yflow);
 
 free((char *)im_offset);
  free((char *)im1);
  return(0);
}

/*********************************/
change_optical_flow_options(cache)
  Image_Cache *cache;
{
  extern char *args;

  Optical_Analysis *an=&(cache->optical_analysis);

  char yn[5];
  int i;

  printf("\tLevel: (%d) : ",an->level);
  readline(args,ARGSIZE,stdin);
  if(strlen(args)!=0)
    sscanf(args,"%d",&an->level);

  printf("\tMethod: (");
  for(i=0;i<=an->level;i++){
    printf("%d ",(int)an->method[i]);
  }
  printf(") : ");
  readline(args,ARGSIZE,stdin);
  if(strlen(args)!=0)
    sscanf(args,"%f %f %f %f %f %f %f %f %f %f",
	   &an->method[0],&an->method[1],&an->method[2],&an->method[3],&an->method[4],
	   &an->method[5],&an->method[6],&an->method[7],&an->method[8],&an->method[9]);


  printf("\tStart frame: (%d) : ",an->start);
  readline(args,ARGSIZE,stdin);
  if(strlen(args)!=0)
    sscanf(args,"%d",&an->start);

  printf("\tEnd frame: (%d) : ",an->end);
  readline(args,ARGSIZE,stdin);
  if(strlen(args)!=0)
    sscanf(args,"%d",&an->end);


  printf("\tTime step between analysis: (%d) : ",an->tstep);
  readline(args,ARGSIZE,stdin);
  if(strlen(args)!=0)
    sscanf(args,"%d",&an->tstep);


  if(an->end>(cache->ext_start + cache->tdim - 1 - an->tstep)) {
    an->end = cache->ext_start + cache->tdim - 1 - an->tstep; }
/*
  an->end=(((an->end)<=(tdim-(an->tstep)-1)) ?
	   (an->end) : (tdim-(an->tstep)-1));
*/

  sprintf(yn,"n");
  printf("\tChange other options? (%s) : ",yn);
  readline(args,ARGSIZE,stdin);
  sscanf(args,"%s",yn);
  if(strncmp(yn,"n",1)==0)
    return(0);

  printf("\tEnter xbias ybias : (%f %f) : ",an->xbias,an->ybias);
  readline(args,ARGSIZE,stdin);
  if(strlen(args)!=0)
    sscanf(args,"%f %f",&an->xbias,&an->ybias);

  printf("\tEnter intensity bias : (%f) : ",an->intensity_bias);
  readline(args,ARGSIZE,stdin);
  if(strlen(args)!=0)
    sscanf(args,"%f",&an->intensity_bias);

  printf("\tEnter noise bias : (%f) : ",an->noise);
  readline(args,ARGSIZE,stdin);
  if(strlen(args)!=0)
    sscanf(args,"%f",&an->noise);
  
  return(0);
}

/*****************************************************************/
/*********************************/
display_optical_flow_options(image_mgr)
  Image_Manager *image_mgr;
{
  int i;
  Image_Cache *cache;
  Optical_Analysis *an;
  
  if((cache=ask_name_get_cache(image_mgr))==NULL)
    return(0);
  an= &cache->optical_analysis;

  printf("\tLevel: %d\n",an->level);
  printf("\tMethod: (");
  for(i=0;i<=an->level;i++){
    printf("%d ",(int)an->method[i]);
  }
  printf(")\n");
  printf("\tStart frame: %d\n",an->start);
  printf("\tEnd frame: %d\n",an->end);
  printf("\tTime step between analysis: %d\n",an->tstep);
  printf("\tClip Max and  Min: %6.2f %6.2f\n",an->clipmax,an->clipmin);
  printf("\tBlurlevel : %d\n",an->blur);
  printf("\tXbias Ybias : %f %f\n",an->xbias,an->ybias);
  printf("\tNoise bias : %f\n",an->noise);
  printf("\tintensity bias : %f\n",an->intensity_bias);

  printf("\tSummation filter : (");
  for(i=0;i<an->sfdim;i++){
    printf("%8.3f ",an->sum_filt[i]);
  }
  printf(")\n");
  
  printf("\tPyramid Subsample filter : (");
  for(i=0;i<an->pfdim;i++){
    printf("%8.3f ",an->pyr_filt[i]);
  }
  printf(")\n");
  
  printf("\tPyramid Upsample filter : (");
  for(i=0;i<an->expfdim;i++){
    printf("%8.3f ",an->exp_filt[i]);
  }
  printf(")\n");
  return(0);
}


/****************************************************************/
display_vector_field_driver(image_mgr)
  Image_Manager *image_mgr;
{
  extern char *args;

  Image_Cache *cache_xflow,*cache_yflow,*cache_overlay,*cache_result;
  int start,end;
  float *xflow,*yflow,*image,*res;
  int xstep,ystep, headsize,type;
  float flow_scale=1.0;
  int val=255;
  int i;
  FILE *flowpid;

  char *title=image_mgr->cbuf1;
  char *yn=image_mgr->yn;
  float gamma;
  int ndigits;
  int t_offset;

  cache_result=NULL;

  printf("Enter X-flow \n");
  if((cache_xflow=ask_name_get_cache(image_mgr))==NULL)
    return(-1);
  
  printf("Enter Y-flow \n");
  if((cache_yflow=ask_name_get_cache(image_mgr))==NULL)
    return(-1);

  gamma=1.0;
  sprintf(image_mgr->image_name,"*none*");
  cache_overlay=NULL;
  printf("Enter Image to Overlay \n");
  if((cache_overlay=ask_name_get_cache(image_mgr))!=NULL) {
    gamma=cache_overlay->gamma;
    printf("Enter Gamma: (%f) : ",gamma);
    readline(args,ARGSIZE,stdin);
    sscanf(args,"%f",&gamma);
    cache_overlay->gamma=gamma;
    t_offset=0;
    printf("Enter time offset: (%d) : ",t_offset);
    readline(args,ARGSIZE,stdin);
    sscanf(args,"%d",&t_offset);
  }
  else {
    printf(".... not overlaying any images \n");fflush(stdout);
  }

  start=cache_xflow->start;
  end=cache_xflow->end;
  printf("Enter Start and End: (%d %d) : ",start,end);
  readline(args,ARGSIZE,stdin);
  sscanf(args,"%d %d",&start,&end);

  if(start<cache_xflow->start) start=cache_xflow->start;
  if(start<cache_yflow->start) start=cache_yflow->start;
  if(end>cache_xflow->end) end=cache_xflow->end;
  if(end>cache_yflow->end) end=cache_yflow->end;

  xstep=10;ystep=10;flow_scale=1.0; headsize=2;
  printf("\tEnter Vector Display Options(xstep,ystep,scale,headsize): \n\t\t(%d %d %6.2f %d) : "
	 ,xstep,ystep,flow_scale,headsize);
  readline(args,ARGSIZE,stdin);
  sscanf(args,"%d %d %f %d",&xstep,&ystep,&flow_scale,&headsize);
  
  type=0;
  
  
  sprintf(title,"%s:%s",cache_xflow->image_name,cache_yflow->image_name);
  flowpid=OpenGrayScreen(cache_xflow->ydim,cache_xflow->xdim,title,gamma,1.0,(end-start+1));

/*  printf("flowpid=%d\n",flowpid);fflush(stdout); */
  
  sprintf(title,"%d",cache_xflow->tdim); ndigits=strlen(title);

  sprintf(yn,"n");
  printf("\tDo you want to save  flow fields results internally (y/n)? (%s): ",yn);
  readline(args,ARGSIZE,stdin);
  sscanf(args,"%s",yn);
  if(strncmp(yn,"y",1)==0) {
    if((cache_result=make_internal_image(image_mgr,cache_xflow->xdim,cache_xflow->ydim,end-start+1))
       ==NULL) {
      printf("Not saving flow field results internally.\n"); 
/*      init_cache(cache_result); cache_result=NULL; */
    }
  }

  for(i=start;i<=end;i++) {
    printf("%d ",i);fflush(stdout);
    if((xflow=get_image(cache_xflow,i))==NULL){
      printf("vector field display: cannot get frame %d of xflow result.\n",i);
      break;
    }
    if((yflow=get_image(cache_yflow,i))==NULL){
      printf("vector field display: cannot get frame %d of yflow result.\n",i);
      break;
    }

    if(cache_overlay==NULL)
      fill_byte(cache_xflow->array,(unsigned char)0,cache_xflow->imsize);
    else {
      if((image=get_image(cache_overlay,i+t_offset+1))==NULL){
	printf("vector field display: cannot get frame %d of overlay image.\n",i);
	break;
      }
      convert_float_to_byte(image,cache_xflow->array,cache_xflow->imsize);
    }
    display_vector_field(xflow,yflow,cache_xflow->xdim,cache_xflow->ydim,
			 cache_xflow->array,cache_xflow->xdim,cache_xflow->ydim,
			 val,xstep,ystep,
			 flow_scale,headsize,type);

    if(cache_result!=NULL) {
      if((res=get_image(cache_result,i))==NULL){
	printf("vector field display: cannot get frame %d of result image.\n",i);
	break;
      }
      convert_byte_to_float(cache_xflow->array,res,cache_xflow->imsize);
    }

    display_number_on_image(cache_xflow->array,cache_xflow->xdim,cache_xflow->ydim,
			    (float) i,(int)5,(int)5,(int)255,(int)0,(int)ndigits);
    DumpBufToScreen(cache_xflow->array,cache_xflow->ydim,cache_xflow->xdim,flowpid,cache_xflow->imsize);
  }

  Close_Screen(flowpid);
  return(0);
}

