/******************************************************************

"Copyright  1994 by the Massachusetts  Institute  of  Technology.  All
rights reserved.

Developed by John Y. A. Wang at the  Media Laboratory, MIT, Cambridge,
Massachusetts, with  support  from Secom Ltd., Goldstar Ltd.,  and the
Television of Tomorrow Program.

For use by authorized users only.

This distribution is approved  by Nicholas Negroponte, Director of the
Media Laboratory, MIT.

Permission to use, copy, or modify this software and its documentation
for educational and research purposes only and without  fee is  hereby
granted, provided that this copyright notice and the original authors'
names  appear  on  all  copies  and   supporting   documentation.   If
individual  files  are  separated  from  this  distribution  directory
structure, this copyright notice must be included.  For any other uses
of  this software, in  original or  modified form, including  but  not
limited to distribution in whole or in part, specific prior permission
must be  obtained  from  MIT.   These  programs  shall  not  be  used,
rewritten,  or  adapted  as  the  basis  of a  commercial software  or
hardware  product without  first  obtaining  appropriate licenses from
MIT.  MIT.  makes no  representations about  the  suitability of  this
software for any purpose.  It is  provided "as is" without  express or
implied warranty.

******************************************************************/


/*
*	m a t r i x - u t i l s . c		dead@media-lab	87/03/02
*/

#ifndef lint
static char copyright[] =
"Copyright (c) 1987 by David T. Chen.  All rights reserved.";
#endif

#include <stdio.h>
#include <math.h>
#include "matrix.h"

typedef	float		WorldType;

/*
==========================================================================
MatrixType_Set_N_and_M( )
VectorType_Set_N( )
IndexType_Set_N( )

==========================================================================
*/
MatrixType_Set_N_and_M( matrix, n, m )
MatrixType *matrix;
int n, m;
{
	if (n < 0 || m < 0 || n >= MAX_RANK || m >= MAX_RANK)
	{	fprintf(stderr, "MatrixType_Set_N_and_M: paramters out of bounds\n");
		return(-1);
	}
	matrix->n = n;
	matrix->m = m;
	return(0);
}

VectorType_Set_N( vector, n )
VectorType *vector;
int n;
{
	if (n < 0 || n >= MAX_RANK)
	    {	fprintf(stderr, "VectorType_Set_N: `n' out of bounds: %d\n",n);
		return(-1);
	}
	vector->n = n;
	return(0);
}

IndexType_Set_N( index, n )
IndexType *index;
int n;
{
	if (n < 0 || n >= MAX_RANK)
	{	fprintf(stderr, "IndexType_Set_N: `n' out of bounds\n");
		return(-1);
	}
	index->n = n;
	return(0);
}



/*
==========================================================================
MatrixType_Fprintf()

==========================================================================
*/
MatrixType_Fprintf( stream, A )
register FILE *stream;
MatrixType *A;
{
register i, j;
int n = MatrixType_Get_N(A);
int m = MatrixType_Get_M(A);
register MatrixData A_data = MatrixType_Get_Data(A);
	fprintf(stream, "(\n");
	for (i=0; i<n; i++)
	{
		fprintf(stream, "(");
		for (j=0; j<m-1; j++)
			fprintf(stream, "%g ", A_data[i][j]);

		fprintf(stream, "%g)\n", A_data[i][j]);
	}
	fprintf(stream, ")\n");
}



/*
==========================================================================
VectorType_Fprintf()

==========================================================================
*/
VectorType_Fprintf( stream, A )
register FILE *stream;
VectorType *A;
{
register i;
int n = VectorType_Get_N(A);
register VectorData A_data = VectorType_Get_Data(A);

	fprintf(stream, "(");

	for (i=0; i<n-1; i++)
		fprintf(stream, "%g ", A_data[i]);

	fprintf(stream, "%g)\n", A_data[i]);
}



/*
==========================================================================
IndexType_Fprintf()

==========================================================================
*/
IndexType_Fprintf( stream, index )
register FILE *stream;
IndexType *index;
{
register i;
int n = IndexType_Get_N(index);
register IndexData index_data = IndexType_Get_Data(index);

	fprintf(stream, "( ");
	for (i=0; i<n; i++)
		fprintf(stream, "%d ", index_data[i]);

	fprintf(stream, ")\n");
}


/*
==========================================================================
MatrixType_Copy( ) -- Copy m2 to m1.

==========================================================================
*/
MatrixType_Copy( m1, m2 )
MatrixType *m1, *m2;
{
register i, j;
int n = MatrixType_Get_N(m2);
int m = MatrixType_Get_M(m2);
register MatrixData M1 = MatrixType_Get_Data(m1);
register MatrixData M2 = MatrixType_Get_Data(m2);

	MatrixType_Set_N_and_M( m1, n, m );

	for (i=0; i<n; i++)
		for (j=0; j<m; j++)
			M1[i][j] = M2[i][j];
}


/*
==========================================================================
MatrixType_Transpose( ) -- Transpose m1, answer in m2.

==========================================================================
*/
MatrixType_Transpose( m1, m2 )
MatrixType *m1, *m2;
{
register i, j;
int n = MatrixType_Get_N(m1);
int m = MatrixType_Get_M(m1);
register MatrixData M1 = MatrixType_Get_Data(m1);
register MatrixData M2 = MatrixType_Get_Data(m2);

	MatrixType_Set_N_and_M( m2, m, n );

	for (i=0; i<n; i++)
		for (j=0; j<m; j++)
			M2[j][i] = M1[i][j];
}


/*
==========================================================================
MatrixType_Multiply( ) -- m1 * m2 = m3

==========================================================================
*/
MatrixType_Multiply( m1, m2, m3 )
MatrixType *m1, *m2, *m3;
{
register i, j, k;
int n_m1 = MatrixType_Get_N(m1);
int m_m1 = MatrixType_Get_M(m1);
int m_m2 = MatrixType_Get_M(m2);
register MatrixData M1 = MatrixType_Get_Data(m1);
register MatrixData M2 = MatrixType_Get_Data(m2);
register MatrixData M3 = MatrixType_Get_Data(m3);

	if ( m_m1 != MatrixType_Get_N(m2) )
	{	fprintf(stderr, "MatrixType_Multiply: rank mis-match.\n");
		return(-1);
	}
	MatrixType_Set_N_and_M( m3, n_m1, m_m2 );

	for(k=0; k<n_m1; k++)
		for(j=0; j<m_m2; j++)
		{
			M3[k][j] = 0.0;
			for(i=0; i<m_m1; i++)
				M3[k][j] += M1[k][i]*M2[i][j];
		}

	return(0);
}



/*
==========================================================================
VectorType_Add( ) -- v1 + v2 = v3

==========================================================================
*/
VectorType_Add( v1, v2, v3 )
VectorType *v1, *v2, *v3;
{
register i;
register n = VectorType_Get_N(v1);
register VectorData V1 = VectorType_Get_Data(v1);
register VectorData V2 = VectorType_Get_Data(v2);
register VectorData V3 = VectorType_Get_Data(v3);

	if (n != VectorType_Get_N(v2))
	{	fprintf(stderr, "VectorType_Add: rank mis-match.\n");
		return(-1);
	}
	VectorType_Set_N( v3, n );

	for (i=0; i<n; i++)
		V3[i] = V1[i] + V2[i];

	return(0);
}


/*
==========================================================================
VectorType_Scale( ) -- v1 * r = v2

==========================================================================
*/
VectorType_Scale( v1, r, v2 )
VectorType *v1, *v2;
register WorldType r;
{
register i;
register n = VectorType_Get_N(v1);
register VectorData V1 = VectorType_Get_Data(v1);
register VectorData V2 = VectorType_Get_Data(v2);

	VectorType_Set_N( v2, n );

	for (i=0; i<n; i++)
		V2[i] = V1[i] * r;
}


/*
==========================================================================
VectorType_Copy( ) -- copy v2 to v1

==========================================================================
*/
VectorType_Copy( v1, v2 )
VectorType *v1, *v2;
{
register i;
register n = VectorType_Get_N(v2);
register VectorData V1 = VectorType_Get_Data(v1);
register VectorData V2 = VectorType_Get_Data(v2);

	VectorType_Set_N( v1, n );

	for (i=0; i<n; i++)
		V1[i] = V2[i];
}







