/******************************************************************

"Copyright  1994 by the Massachusetts  Institute  of  Technology.  All
rights reserved.

Developed by John Y. A. Wang at the  Media Laboratory, MIT, Cambridge,
Massachusetts, with  support  from Secom Ltd., Goldstar Ltd.,  and the
Television of Tomorrow Program.

For use by authorized users only.

This distribution is approved  by Nicholas Negroponte, Director of the
Media Laboratory, MIT.

Permission to use, copy, or modify this software and its documentation
for educational and research purposes only and without  fee is  hereby
granted, provided that this copyright notice and the original authors'
names  appear  on  all  copies  and   supporting   documentation.   If
individual  files  are  separated  from  this  distribution  directory
structure, this copyright notice must be included.  For any other uses
of  this software, in  original or  modified form, including  but  not
limited to distribution in whole or in part, specific prior permission
must be  obtained  from  MIT.   These  programs  shall  not  be  used,
rewritten,  or  adapted  as  the  basis  of a  commercial software  or
hardware  product without  first  obtaining  appropriate licenses from
MIT.  MIT.  makes no  representations about  the  suitability of  this
software for any purpose.  It is  provided "as is" without  express or
implied warranty.

******************************************************************/


/***************************************************************
*** File: median-convolve.c
*** Author: John Y. A. Wang
*** Date:    March 9, 1992
*** Copyright 1992, Massachusetts Institute of Technology, All Rights Reserved.
***
*** Description:
   Median convolution with filter specified by
   filt. User can specify one of two types of median
   convolutions:
   type=1: convolution with equal weighting on all non-zero
           taps of the filter.
   type=2: weighted convolution with weights specified by
           the filter taps.
   
   xfilt_space, and yfilt_space, is the distance between
   the filter taps. 
   
   grid_start, and grid_step are not currently used. These are
   used by the OBVIUS convolution routine.

***************************************************************/

static  char *copyright="\
\n\n***********************************************\n\
File: median-convolve.c\n\
Author: John Y. A. Wang.\n\
The MIT Media Lab.\n\
Copyright (c) 1992, All Rights Reserved.\n\
***********************************************\n\n";

#include <stdio.h>
#include <math.h>
#include "motion.h"


/* convolution with variable filter tap spacing.
** edge-handler reflects back at edges : -1 -> 1 and  dim=256; 256 -> 254
** Warning: will add to the result image 
*/

/*
  cc -c -fsingle -O4 median-convolve.c
*/

double internal_median();
double internal_weighted_median();

int internal_median_filter_reflect
  (image, x_dim, y_dim, filt, temp,x_fdim, y_fdim,result,
   xfilt_space, yfilt_space,type)
  register float *image, *temp;
  float *filt; 
  register int x_fdim, x_dim;
  register float *result;
  int y_dim, y_fdim;
  int xfilt_space, yfilt_space;
  char *type;

{
  register float fval;
  register float arr[1000], warr[1000];
  register int xfp,yfp,xp,yp, x_pos, y_pos, yf, x, y, pos;
  register int y_fmid = (y_fdim+(y_fdim-1)*(yfilt_space-1))/2;
  register int x_fmid = (x_fdim+(x_fdim-1)*(xfilt_space-1))/2;

  register int arrn;
#define MED_UNIFORM 0
#define MED_WEIGHTED 1
  int type_n=MED_UNIFORM;

/*
  fprintf(stderr,"\n internal_median_filter_reflect filter_space(%d, %d) \n", 
	 xfilt_space, yfilt_space); 
*/
  if(strcmp(type,"uniform")==0) {
/*    fprintf(stderr,"performing equally weighted median\n"); */
    type_n=MED_UNIFORM;}
  else if(strcmp(type,"weighted")==0) {
/*    fprintf(stderr,"performing filter weighted median\n"); */
    type_n=MED_WEIGHTED;}
  else {
    fprintf(stderr,"Error, incorrect median type operator %s\n",type);
    return(-1);
  }
  
  for(y=0,pos=0; y < y_dim; y++) {
    for(x=0; x < x_dim; x++,pos++) {
      arrn=0;               /* start from beginning of array */
      for(yf=0,y_pos=y-y_fmid,yfp=0; yf<y_fdim; yf++,y_pos+=yfilt_space,yfp+=x_fdim) {
	if(y_pos < 0)
	  yp = -x_dim*y_pos;
	else
	  if(y_pos >= y_dim)
	    yp = x_dim*(y_dim - (y_pos - y_dim + 2));
	  else
	    yp = x_dim*y_pos;
	
	for(xfp=0,x_pos=x-x_fmid; xfp < x_fdim; xfp++,x_pos+=xfilt_space) {
	  if(x_pos < 0)
	    xp = - x_pos;
	  else
	    if(x_pos >= x_dim)
	     xp = x_dim - (x_pos - x_dim + 2);
	    else
	      xp=x_pos;

	   if((fval=filt[yfp + xfp]) != 0.0) {
	     arr[arrn] = image[yp + xp];
	     warr[arrn] = fval;
	     arrn++;
	   }
	 }  /* end xf */
      }     /* end yf */
      if(type_n == MED_UNIFORM) {
	temp[pos] = (float)internal_median(arr,arrn); }
      else {
	temp[pos] = (float)internal_weighted_median(arr,warr,arrn); }
    }       /* end x */
  }           /* end y */
  
  copy_im(temp,result,x_dim*y_dim);
  return(0);
}

#ifdef OLD
void internal_median_filter_reflect(image, x_dim, y_dim, filt, temp, x_fdim, y_fdim,
		xgrid_start,xgrid_step,ygrid_start,ygrid_step,result,edges,
		 xfilt_space, yfilt_space,type)
  register float *image, *temp;
  register int x_fdim, x_dim;
  register int xgrid_step;
  register float *result;
  int ygrid_step;
  int xgrid_start, ygrid_start;
  float *filt; 
  int y_dim, y_fdim;
  char *edges;
  int xfilt_space, yfilt_space;
  int type;

{
  register float fval;
  register float arr[1000], warr[1000];
  register int xfp,yfp,xp,yp, x_pos, y_pos, xf, yf, x, y, pos;
  register int y_fmid = (y_fdim+(y_fdim-1)*(yfilt_space-1))/2;
  register int x_fmid = (x_fdim+(x_fdim-1)*(xfilt_space-1))/2;

  register int arrn;
  int i;

/*
  printf("\n internal_median_filter_reflect filter_space(%d, %d) \n", 
	 xfilt_space, yfilt_space); 
*/
  if(type == 1)
    printf("\n performing equally weighted median");
  else
    printf("\n performing filter weighted median");

  for(yf=0,arrn=0; yf < y_fdim; yf++) {
    printf("\n(");
    for(xf=0; xf < x_fdim; xf++) 
      printf("%6.4f ", filt[yf*x_fdim + xf]);
    printf(")");
  }
  fflush(stdout);

  for(y=0,pos=0; y < y_dim; y++) {
    for(x=0; x < x_dim; x++,pos++) {
      arrn=0;               /* start from beginning of array */
      for(yf=0,y_pos=y-y_fmid,yfp=0; yf<y_fdim; yf++,y_pos+=yfilt_space,yfp+=x_fdim) {
	if(y_pos < 0)
	  yp = -x_dim*y_pos;
	else
	  if(y_pos >= y_dim)
	    yp = x_dim*(y_dim - (y_pos - y_dim + 2));
	  else
	    yp = x_dim*y_pos;
	
	for(xfp=0,x_pos=x-x_fmid; xfp < x_fdim; xfp++,x_pos+=xfilt_space) {
	  if(x_pos < 0)
	    xp = - x_pos;
	  else
	    if(x_pos >= x_dim)
	     xp = x_dim - (x_pos - x_dim + 2);
	    else
	      xp=x_pos;

	   if((fval=filt[yfp + xfp]) != 0.0) {
	     arr[arrn] = image[yp + xp];
	     warr[arrn] = fval;
	     arrn++;
	   }
	 }  /* end xf */
      }     /* end yf */
      if(type == 1)
	temp[pos] = (float)internal_median(arr,arrn);
      else
	temp[pos] = (float)internal_weighted_median(arr,warr,arrn);
    }       /* end x */
  }           /* end y */
  
  copy_im(temp,result,x_dim*y_dim);
  fflush(stdout);
}
#endif



