/******************************************************************

"Copyright  1994 by the Massachusetts  Institute  of  Technology.  All
rights reserved.

Developed by John Y. A. Wang at the  Media Laboratory, MIT, Cambridge,
Massachusetts, with  support  from Secom Ltd., Goldstar Ltd.,  and the
Television of Tomorrow Program.

For use by authorized users only.

This distribution is approved  by Nicholas Negroponte, Director of the
Media Laboratory, MIT.

Permission to use, copy, or modify this software and its documentation
for educational and research purposes only and without  fee is  hereby
granted, provided that this copyright notice and the original authors'
names  appear  on  all  copies  and   supporting   documentation.   If
individual  files  are  separated  from  this  distribution  directory
structure, this copyright notice must be included.  For any other uses
of  this software, in  original or  modified form, including  but  not
limited to distribution in whole or in part, specific prior permission
must be  obtained  from  MIT.   These  programs  shall  not  be  used,
rewritten,  or  adapted  as  the  basis  of a  commercial software  or
hardware  product without  first  obtaining  appropriate licenses from
MIT.  MIT.  makes no  representations about  the  suitability of  this
software for any purpose.  It is  provided "as is" without  express or
implied warranty.

******************************************************************/


/******************************************************************************
***  File: median.c
***  Author: John Y. A. Wang
***  Date:   March 9, 1992.

*** Copyright 1992 Masssachussetts Institute of Technology. All Rights Reserved.

***  Description:
***       obtain median on a array by first sorting  and then
***       returning the value at midpoint of sort.
***       Sort routine from Numerical Recipes in C.
******************************************************************************/

static  char *copyright="\
\n\n***********************************************\n\
File: median.c\n\
Author: John Y. A. Wang.\n\
The MIT Media Lab.\n\
Copyright (c) 1992, All Rights Reserved.\n\
***********************************************\n\n";

#include <stdio.h>
#include <memory.h>
#include <math.h>
#include "motion.h"

/*
  cc -c -O4 -fsingle median.c
*/

#ifdef lint
#define register
#endif

double internal_median(/*list, n*/);
/* median on the array but ignore values that are equal to
 * checkval */   
double internal_median_checkval(/*im, size, checkval*/);

double internal_median_window(/*image,size,thresh,ww*/);

/* sort "list" of length "n" and returns "list" that is sorted. */
void internal_sort(/*list,n*/);

double internal_weighted_median(/*list, weight, n*/);

/* sort on list and order list2 using list as well. */
void internal_sort2(/*list,list2,n*/);

/**************************************************************************/

double internal_median(list, n)
register float *list;
register int n;
{
  register int index;
  
  if(n==0)
    return(0.0);
  if(n==1)
    return((double)list[0]);
  if(n==2)
    return((double)((list[0]+list[1])/2.0));

  internal_sort(list,n);       /* sort list */
  index = (int) (n/2); /* determine index */
  if(1&n)              /* odd */
    return((double)list[index]);
  else
    return((double) (list[index] + list[index-1])/2);
}

 
/* sort "list" of length "n" and returns "list" that is sorted. */
void internal_sort(list,n)
int n;
float *list;
{
	int l,j,ir,i;
	float rra;
	float *ra;

	if((n==0) || (n==1))
	  return;
	ra = list - 1;
	l=(n >> 1)+1;
	ir=n;
	for (;;) {
		if (l > 1)
			rra=ra[--l];
		else {
			rra=ra[ir];
			ra[ir]=ra[1];
			if (--ir == 1) {
				ra[1]=rra;
				return;
			}
		}
		i=l;
		j=l << 1;
		while (j <= ir) {
			if (j < ir && ra[j] < ra[j+1]) ++j;
			if (rra < ra[j]) {
				ra[i]=ra[j];
				j += (i=j);
			}
			else j=ir+1;
		}
		ra[i]=rra;
	}
}


/** this obtains the weighted median, stochastically, this  is the ML
  estimate. If we let the normalized weights be the probabililties then
  ML is where the CMF cumulative  mass function is 0.5. This algorithm
  calculates the CMF by sorting according to input values with corresponding
  weights, then scans the sorted weight list to find the point where 
  CMF is equal to the halfway point.
  **/
double internal_weighted_median(list, weight, n)
  register float *list, *weight;
  register int n;
{
  register int index;
  register float cdf, midpt;

  if(n==0)
    return(0.0);
  if(n==1)
    return((double)list[0]);
  if(n==2)
    return((double)((list[0]+list[1])/2.0));

  internal_sort2(list,weight,n);       /* sort (value,weight) list */

  for(index=0,midpt=0; index<n; index++)  {  /* find total mass */
    midpt += weight[index];
  }
  midpt = midpt/2;    /* find the halfway mass */
  for(index=0,cdf=0; index<n; index++) {  /* do CMF until the halfway pt */
    if((cdf += weight[index]) >= midpt)
      break;
  }
  return((double)list[index]);
}

/* sort on list and order list2 using list as well. */
void internal_sort2(list,list2,n)
  int n;
  float *list,*list2;
{
  int l,j,ir,i;
  float rra,rrw;
  float *ra,*rw;
  
  if((n==0) || (n==1))
    return;
  ra = list - 1;
  rw = list2 - 1;
  l=(n >> 1)+1;
  ir=n;
  for (;;) {
    if (l > 1) {
      rra=ra[--l];
      rrw=rw[l];
    }
    else {
      rra=ra[ir];
      rrw=rw[ir];
      ra[ir]=ra[1]; rw[ir]=rw[1];
      if (--ir == 1) {
	ra[1]=rra; rw[1]=rrw;
	return;
      }
    }
    i=l;
    j=l << 1;
    while (j <= ir) {
      if (j < ir && ra[j] < ra[j+1]) ++j;
      if (rra < ra[j]) {
	ra[i]=ra[j];
	rw[i]=rw[j];
	j += (i=j);
      }
      else j=ir+1;
    }
    ra[i]=rra;
    rw[i]=rrw;
  }
}



/* median on the array but ignore values that are equal to
 * checkval */   
double internal_median_checkval (im, size, checkval)
  register float *im, checkval;
  register int size;
{
  register int n,i;
  register float *arr;
  register double res;
  
  printf("internal_median-checkval\n");
  fflush(stdout);
  arr = (float *)malloc(size * sizeof(float)); 
  if (arr == 0) {
    printf("\nerror: internal_median-checkval could not allocate enough memory\n");
    fflush(stdout);
    return((float) -999999);
  }

  for (i=0, n=0; i< size; i++)
    if (im[i] !=  checkval) {
      arr[n] = im[i];
      n++;
    }
  
  switch (n) {
  case 0:
    free((char *) arr);
    return ((double) checkval);
    break;
  case 1:
    free(arr);
    return ((double) arr[0]);
    break;
  default:
    res = internal_median (arr, n);
    free(arr);
    return (res);
    }
}


double internal_median_window(image,size,thresh,ww)
  float *image;
  int ww,size;
  float thresh;
{
  register int ll,uu;
  register float  mu,ml;
  register float medianval;
  int w2,i;

  w2 = 2*ww+5;  /* window limit */
  medianval = (float) internal_median(image,size); /* sort and get median */
/*  printf("medianval(before)=%f,ww=%d",medianval,ww); */

  if((thresh>0) && (size>w2)) {  /* if window is big enough then do the neat stuff */
    i=(int)(size/2);             /* start at the median position */

/*    printf("\nsize=%d,index=%d,image[%d]=%f",size,i,i,image[i]); */
    mu = image[i + ww] + thresh;   /* upper value */
    ml = image[i - ww] - thresh;   /* lower value */

    /* search for upper and lower index  */
    for(uu=(int)(i+ww); ((mu > image[uu]) && (uu < size)) ; uu++);
    for(ll=(int)(i-ww); ((ml < image[ll]) && (ll >= 0)) ; ll--);

/*    printf("\nmedianval=%f,ll=%d, uu=%d ",medianval,ll,uu); */
    if(1&(i=(uu+ll)))  { /* find the midpt. described by uu and ll */
      medianval=image[(int)(i/2)];
/*      printf("\nodd: index=%d,medianval=%f",(int)(i/2), medianval); */
    }
    else {
      medianval= (image[(int)(i/2)] + image[(int)(i/2) + 1])/2.0;
/*      printf("\neven: index=%d,medianval=%f",(int)(i/2), medianval);  */
    }
  } 
/*  printf("\nmedianval=%f\n",medianval);*/
  fflush(stdout);
  return((double) medianval);
}

/*
#define YY 86
#define XX 54
*/

double internal_get_addr(im)
  float *im;
{
  int addr;

  addr = (int)(&im[0]);
  return((double)addr);
}

int internal_median_seq (im_addr, n_frames, res, size, check_val, setval)
  register float *res;
  register float **im_addr; /* array of pointers to images */
  register float check_val;
  register float setval;
  register int size, n_frames;
{
  register int i,j,n;
  register float arr[400];
  int index;
  
/* printf("internal_median_seq\n");  */

  if(n_frames==0) {return(0);}
  if(n_frames>400) {return(-1);}

  for(i=0; i<size; i++) {
    for(j=0, n=0; j<n_frames; j++) {
      if ( ((im_addr[j])[i]) != check_val) {
	arr[n]=(im_addr[j])[i];  /* load the array */
	n++;
      }}

    switch (n) {
    case 0: res[i] = setval;  break;
    case 1: res[i] = arr[0];  break;
    default:
      internal_sort(arr,n);       /* sort list */
      index = (int) (n/2); /* determine index */
      if(1&n) { res[i] = arr[index]; } /* odd number of frames */
      else { res[i]= (arr[index] + arr[index-1])/2.0; }
    }
  }
  return(0);
}


int internal_weighted_median_seq (im_addr, weight, n_frames, res, size,
				   check_val, setval)
  register float *res,*weight;
  register float **im_addr; /* array of pointers to images */
  register float check_val;
  register float setval;
  register int size, n_frames;
{
  register int i,j,n;
  register float arr[400], warr[400];
  int index;
  
/* printf("internal_weighted_median_seq\n");  */

  if(n_frames==0) {return(0);}
  if(n_frames>400) {return(-1);}

  for(i=0; i<size; i++) {
    for(j=0, n=0; j<n_frames; j++) {
      if ( ((im_addr[j])[i]) != check_val) {
	arr[n]=(im_addr[j])[i];  /* load the array */
	warr[n]=weight[j];       /* load the weights */
	n++;
      }}
    switch(n) {
    case 0: res[i] = setval; break;
    case 1: res[i] = arr[0]; break;
    default:
      res[i]=(float)internal_weighted_median(arr,warr,n);
    }
  }
  return(0);
}

/* find the variance of a sequences */
int internal_variance_seq (im_addr, n_frames, mean_im, res, size, check_val)
  register float *res, *mean_im;
  register float **im_addr; /* array of pointers to images */
  register float check_val;
  register int size, n_frames;
{
  register int i,j,n;
  register float error, tmp, mean;

  for(i=0; i<size; i++) {
    mean = mean_im[i];
    error = 0;
    for(j=0, n=0; j<n_frames; j++) {
      if ( ((im_addr[j])[i]) != check_val) {
	tmp = ((im_addr[j])[i] - mean);   /* error */
	error += tmp*tmp;
	n++;
      }
    }
    res[i] = error/(float)n;
  }
}

/* subtract two sequences frame by frame */
int internal_sub_seq (im_addr, n_frames, im, out_addr, size)
  register float  *im;
  register float **im_addr; /* array of pointers to images */
  register float **out_addr; /* array of pointers to output images */
  register int size;
{
  register int i,n;
  register float *out_frame, *in_frame;

    for( n=0; n < n_frames; n++) {
      in_frame = im_addr[n];
      out_frame = out_addr[n];
	for(i=0; i<size; i++) {
	  out_frame[i] = in_frame[i] - im[i];
	}
    }      
}



#ifdef NEVER
main()
{
  int n,i;
  float arr[10];
  float med;

  printf("enter n >> ");
  scanf("%d", &n);
  for(i=0; i<n; i++)
    scanf("%f", &arr[i]);
  for(i= 0; i< n;i++)
    printf("%4.2f ",arr[i]);
  printf("\n");

  med = internal_median(arr,n);
  for(i= 0; i< n;i++)
    printf("%4.2f ",arr[i]);
  printf("\nmed = %4.2f\n",med);
}

#endif


