/******************************************************************

"Copyright  1994 by the Massachusetts  Institute  of  Technology.  All
rights reserved.

Developed by John Y. A. Wang at the  Media Laboratory, MIT, Cambridge,
Massachusetts, with  support  from Secom Ltd., Goldstar Ltd.,  and the
Television of Tomorrow Program.

For use by authorized users only.

This distribution is approved  by Nicholas Negroponte, Director of the
Media Laboratory, MIT.

Permission to use, copy, or modify this software and its documentation
for educational and research purposes only and without  fee is  hereby
granted, provided that this copyright notice and the original authors'
names  appear  on  all  copies  and   supporting   documentation.   If
individual  files  are  separated  from  this  distribution  directory
structure, this copyright notice must be included.  For any other uses
of  this software, in  original or  modified form, including  but  not
limited to distribution in whole or in part, specific prior permission
must be  obtained  from  MIT.   These  programs  shall  not  be  used,
rewritten,  or  adapted  as  the  basis  of a  commercial software  or
hardware  product without  first  obtaining  appropriate licenses from
MIT.  MIT.  makes no  representations about  the  suitability of  this
software for any purpose.  It is  provided "as is" without  express or
implied warranty.

******************************************************************/



/***************************************************************
*** File: motion-imops.c
*** Author: John Y. A. Wang
*** Date: January 28, 1993.
*** Copyright 1993, Massachusetts Institute of Technology, All Rights Reserved.
***
*** Description: Common image utilities used in motion algorithms.
***************************************************************/

static  char *copyright="\
\n\n***********************************************\n\
File: motion-imops.c\n\
Author: John Y. A. Wang.\n\
The MIT Media Lab.\n\
Copyright (c) 1993, All Rights Reserved.\n\
***********************************************\n\n";

#include "motion.h"

/*********** affine-motion-cf.c ***************/
/* fill  a float array with val */
void fill_im(a,val,size)
  register float *a;
  register float val;
  register int size;
{
  register int i;
  for(i=0;i<size;i++)
    a[i]=val;
}

/************ gradient-flow-cf.c gradient-flow-offset-cf.c *************/

/*
**  update the flow estimate by composing the
**  current flow map with the new flow map
*/

void internal_update_flow(xflow,yflow,xnewflow,ynewflow,xdim,ydim,clipmin,clipmax)
  register float *xflow,*yflow,*xnewflow,*ynewflow;
  int xdim,ydim;
  register float clipmax, clipmin;
{
  register int i;
  register float val;
  register int im_size = xdim*ydim;

/*  printf("(%e,%e)",clipmin,clipmax);fflush(stdout); */
  for(i=0;i<im_size;i++) {
    if((val=ynewflow[i]) > clipmax) 
      yflow[i] += (val - clipmax)/(1 + val - clipmax);
    else
      if(val < clipmin)
	yflow[i] += (clipmin - val)/(-1 + val - clipmin); 
      else
	yflow[i] += val;

    if((val=xnewflow[i]) > clipmax) 
      xflow[i] += (val - clipmax)/(1 + val - clipmax); 
    else
      if(val < clipmin)
	xflow[i] += (clipmin - val)/(-1 + val - clipmin);
      else
	xflow[i] += val;
  }
}

/*
** expand the flow to the next level in the pyramid
** 1) upsample the flow and 2) apply filter to smooth image
** temp_in is assumed to be 2x the size of the image
*/

void internal_expand_flow(xflow,yflow,xres,yres,temp,xdim,ydim,rxdim,rydim,exp_filt,fdim,
		     thresh)
  float *xflow,*yflow,*xres,*yres,*temp;
  int xdim,ydim;
  float *exp_filt;
  int fdim;
  float thresh;
{
/*  static float g_filt[5]={0.125, 0.5, 0.75, 0.5, 0.125}; *
/*  static float g_filt2[5]={0.25, 1.0, 1.5, 1.0, 0.25}; */
  register float exp_filt2[20];
  register int i;
  register float sum;
  
  for(i=0,sum=0;i<fdim;i++)
    sum+=exp_filt[i];
  sum=8.0/(sum*sum);
  for(i=0;i<fdim;i++)
    exp_filt2[i]=exp_filt[i]*sum;

/*  printf("\nexpand-flow: (%d,%d) (%d,%d) (%d,%d) (%d,%d)\n"
	 ,xdim,ydim,rxdim,rydim,xflow,xres,yflow,yres);fflush(stdout);
*/
  expand_sep_filter_reflect(xflow,xres,temp,xdim, ydim, rxdim, rydim,
			    exp_filt,exp_filt2,fdim,fdim,2,2);
  expand_sep_filter_reflect(yflow,yres,temp,xdim, ydim, rxdim, rydim,
			    exp_filt,exp_filt2,fdim,fdim,2,2);
  if(thresh!=0) {  /* thresh reduces the effect of small motion */
    fprintf(stderr,"\ninternal_expand_flow: thresh!=0 no longer supported.\n");
/*    remap_image(xflow,rxdim,rydim,thresh); */
/*    remap_image(yflow,rxdim,rydim,thresh); */
  }
}

/* expands the image by space parameter
** requires temp_im space 2x the size of the result image
*/
void expand_sep_filter_reflect(image, result, temp_im, xdim, ydim, rxdim, rydim,
			       xfilt, yfilt, xfdim, yfdim,xfilt_space, yfilt_space)
  float *image, *result, *temp_im;
  int xfdim,yfdim;
  int xdim, ydim;
  float *xfilt, *yfilt; 
  int xfilt_space, yfilt_space;
  int rydim, rxdim;
{
  register int x,y,outskip;
  register float *in_pos, *out_pos;
  register int im_size;

/*  
printf("exapnd_sep_filter_reflect(%d,%d)(%d,%d) (%d,%d) (%d,%d)",
	 xdim,ydim,rxdim,rydim,xfdim,yfdim,xfilt_space,yfilt_space); fflush(stdout);
*/
  if((rydim < ydim*yfilt_space) && (rydim >= ((ydim+1)*yfilt_space)) &&
     (rxdim <= xdim*xfilt_space) && (rxdim >= ((xdim+1)*xfilt_space))) {
    printf("\n expand_sep_filter_reflect: result sizes do not match:\n");
    printf("input (%d,%d): output (%d,%d)\n",ydim,xdim,rydim,rxdim);
    fflush(stdout);
    return;
  }
  im_size = rxdim*rydim;

  for(x=0; x<im_size; x++)
    temp_im[x] = 0.0;

  outskip=yfilt_space*rxdim;
  in_pos=image;
  for(y=0; y < ydim; y++) {
    out_pos = temp_im+y*outskip;
    for(x=0; x < xdim; x++) {
      *out_pos += *in_pos;
      out_pos += xfilt_space;
      in_pos++;
    }
  }
  apply_sep_filter_reflect(temp_im,result,temp_im+(rxdim*rydim),
			   rxdim,rydim,xfilt,yfilt,xfdim,yfdim,1,1); 
}

/* clip values of a float array with clipmin and clipmanx */
void clip_im(array, clipmin, clipmax, size)
  register float *array, clipmin, clipmax;
  register int size;
{
  register int i;
  register float val;

  for(i=0;i<size;i++) {
    if((val=array[i]) > clipmax)
      array[i] += clipmax;
    else
      if(val < clipmin)
	array[i] += clipmin;
      else
	array[i] += val;
  }}

/* remaps the values of an float image such that resulting values is:
**  res = val*abs_val(val)/thresh  if abs_val(val)< threshold */
void remap_image(im,xdim,ydim,thresh)
  register float *im,thresh;
  int xdim,ydim;
{
  register int im_size=xdim*ydim;
  register int i;
  register float val;

  if(thresh<=0) return;

  for(i=0;i<im_size;i++) {
    val=im[i];
    if(abs_val(val) < thresh)
      im[i]=val*abs_val(val)/thresh;
  }
}

/*************** common to all *****************/

/***** multiply images *****/

void mul_im(a1, a2, res, size)
  register float *a1, *a2, *res;  register int size;
{  register int i;  for(i=0;i<size;i++) {  res[i] = a1[i] * a2[i];}}

void mul_im_sc(a1, val, res, size)
  register float *a1, val, *res;  register int size;
{  register int i;  for(i=0;i<size;i++) {  res[i] = a1[i] * val;}}

/***** divide images *****/
void div_im(a1, a2, res, size)
  register float *a1, *a2, *res;  register int size;
{
  register int i;
   for(i=0;i<size;i++) { if (a2[i] == 0) { res[i]=0; } else { res[i] = a1[i] / a2[i]; }}
}

void div_im_sc(a1, val, res, size)
  register float *a1, val, *res;  register int size;
{
  register int i;
  if(val==0.0) { zero_im(a1,size); return; }
  for(i=0;i<size;i++) { res[i] = (float)(a1[i] / val); }}

/***** subtract images *****/
void sub_im(a1, a2, res, size)
  register float *a1, *a2, *res;  register int size;
{  register int i;  for(i=0;i<size;i++) { res[i] = a1[i] - a2[i];}}

/***** add images *****/
void add_im(a1, a2, res, size)
  register float *a1, *a2, *res;  register int size;
{  register int i;  for(i=0;i<size;i++) { res[i] = a1[i] + a2[i]; }}

void add_im_sc(a1,val,res,size)
  register float *a1,val, *res;  register int size;  
{  register int i;  for(i=0;i<size;i++) { res[i] = a1[i] + val; }}

/***** copy images *****/
void copy_im(a1, res, size)
  register float *a1, *res;  register int size;
{  register int i;  for(i=0;i<size;i++) { res[i] = a1[i]; }}

/****** zero image ********/
void zero_im(image, size)
  register float *image;  register int size;
{  int i;  for(i=0;i<size;i++) { image[i] = 0.0; }}

/**** abs image *****/
void abs_im(a1, res, size)
  register float *a1, *res;  register int size;
{  register int i;  for(i=0;i<size;i++) { res[i] = abs_val(a1[i]); }}

/******************************************/
/** image compare **/

void internal_greater_than_im(im,res,val,size)
  register float *im,*res;
  register float val;
  register int size;
{
  register int i;
  for(i=0;i<size;i++) {
    if(im[i]>val)
      res[i] = 1.0;
    else
      res[i] = 0.0;
  }
}

void internal_less_than_im(im,res,val,size)
  register float *im,*res;
  register float val;
  register int size;
{
  register int i;
  for(i=0;i<size;i++) {
    if(im[i]<val)
      res[i] = 1.0;
    else
      res[i] = 0.0;
  }
}

void internal_equal_im(im,res,val,size)
  register float *im,*res;
  register float val;

  register int size;
{
  register int i;
  for(i=0;i<size;i++) {
    if(im[i]==val)
      res[i] = 1.0;
    else
      res[i] = 0.0;
  }
}

/*************************************************************/
/** finds the maximum of the image **/
float max_image(im,imsize) 
  float *im;
  int imsize;
{
  int i;
  float max;
  
  for(i=0,max=im[0];i<imsize;i++)
    if(max<im[i]) max=im[i];
  
  return((float) max);
}


/******************************************/

void blur_image(image,result,temp,xsize,ysize,level,g_filt,fdim)
  float *image,*result,*temp,*g_filt;
  int xsize,ysize,level;
{
  int i,space;
/*  static float g_filt[5] = {0.0625,0.25,0.375,0.25,0.0625}; */
/*
  printf("blur_image: level=%d, xsize=%d, ysize=%d\n",level,xsize,ysize);fflush(stdout);
  printf("\timage=%d, result=%d, temp=%d\n",image,result,temp);fflush(stdout);
*/
  if(level == 0) 
    if(image == result)
      return;
    else {
      copy_im(image,result,xsize*ysize);
      return;
    }

  if(level >= 1) {
    apply_sep_filter_reflect(image,result,temp,xsize,ysize,g_filt,g_filt,fdim,fdim,1,1); 
  }

  for(i=1,space=2; i<level ;i++,space*=2){
    apply_sep_filter_reflect(result,result,temp,xsize,ysize,
				   g_filt,g_filt,fdim,fdim,space,space);
  }
}

/**************************************************/
void apply_sep_filter_reflect(image, result, temp_im, xdim, ydim,
			       xfilt, yfilt, xfdim, yfdim,xfilt_space, yfilt_space)
  float *image, *result, *temp_im;
  register int xfdim;
  int xdim;
  register float *xfilt, *yfilt; 
  int ydim;
  register int yfdim;
  int xfilt_space, yfilt_space;

{
  register double sum;
  int y_fmid = (yfdim+(yfdim-1)*(yfilt_space-1))/2;
  int x_fmid = (xfdim+(xfdim-1)*(xfilt_space-1))/2;
  int x,y;
  register int x_pos, y_pos, xf, yf;
  register int yoff, xoff;
  register float *ximage, *yimage, *xtemp, *ytemp; 

/*  printf("\n internal_separable_filter_conv \n"); fflush(stdout); */
  /* yfilter */

  if(xfdim==0 && yfdim == 0) /* don't convolve at all */
    return;
  else
    if(yfdim == 0)  {  /* convolve x only */
      ximage = image;
      xtemp = result;
    }
    else
      if(xfdim==0) {   /* convolve y only */
	yimage = image;
	ytemp = result;
      }
      else {
	yimage = image;
	ytemp = temp_im;     /* convolve both x and y */
	ximage = temp_im;
	xtemp = result;
      }

  /******** convolve in the y direction with the yfilter ********/
  if(yfdim != 0)  /* convolve y */
    for(y=0; y < ydim; y++) {
      yoff = y - y_fmid;
      x_pos = y*xdim;
	for(x=0; x < xdim; x++) {
	  sum=0.0;
	  for(yf=0; yf < yfdim; yf++) {
	    y_pos = yoff + yf*yfilt_space;
	    
	    if(y_pos < 0)
	      y_pos = -y_pos;
	    else
	      if(y_pos >= ydim)
		y_pos = ydim - (y_pos - ydim + 2);
	    
	    sum+= yimage[y_pos*xdim + x]*yfilt[yf];
	  }     /* end yf */
	  ytemp[x_pos + x] = sum;
	}       /* end x */
    }         /* end y */

  /******** convolve in the x direction with the xfilter ********/
  if(xfdim != 0)  /* convolve x */
    for(y=0; y < ydim; y++) {
      y_pos =y*xdim;
      for(x=0; x < xdim; x++) {
	xoff = x - x_fmid;
	sum=0.0;
	for(xf=0; xf < xfdim; xf++) {
	  x_pos = xoff + xf*xfilt_space;
	  if(x_pos < 0)
	    x_pos = -x_pos;
	  else
	    if(x_pos >= xdim)
	      x_pos = xdim - (x_pos - xdim + 2);
	  sum+= ximage[y_pos + x_pos]*xfilt[xf];
	}     /* end xf */
	xtemp[y_pos + x] = sum;
      }       /* end x */
    }         /* end y */
}


/**************************************************************/
/* destructive modification not allowed */
void apply_nonsep_filter_reflect(im,res,xdim,ydim,filt,xfdim,yfdim)
  float *im,*res,*filt;
  int xdim,ydim,xfdim,yfdim;
{

  int x,y,pos;
  int xfmid=(int)xfdim/2;
  int yfmid=(int)yfdim/2;
  int yf,yp,fp,ypos;
  int xf,xp;
  float sum=0;

  if(im == res) {
    fprintf(stderr,"\napply_nonsep_filter_reflect: destructive modification not allowed\n\n");
    return;
  }

  for(y=0,pos=0;y<ydim;y++) {
    for(x=0;x<xdim;x++,pos++) {

      sum=0;
      for(yf=0,yp=y-yfmid,fp=0;yf<yfdim;yf++,yp++) {
	if(yp < 0) { yp = -yp; }
	else if(yp >= ydim) { yp = 2*ydim-yp-2; }
	ypos=yp*xdim;

	for(xf=0,xp=x-xfmid;xf<xfdim;xf++,xp++,fp++) {
	  if(xp < 0) { xp = -xp; }
	  else if(xp >= xdim) { xp = 2*xdim-xp-2; }
	  sum+=filt[fp]*im[ypos+xp];
	}}
      res[pos]=sum;
    }}
}



/**************************************************************/
/** makes a binomial distributed  filter of fdim; **/
void binomial_filter(filt,fdim)
  float *filt;
  int fdim;
{
  int i;
  double sum;
  
/*  fprintf(stderr,"filt=%d\n",filt); */
  for(i=0,sum=0;i<fdim;i++) {
    filt[i]=combinatorial(i,fdim-1); 
/*  fprintf(stderr,"filt[%d]=%e ",i,filt[i]); */
    sum+=(double)filt[i];
  }
/*  fprintf(stderr,"sum=%f\n",sum); */
  for(i=0;i<fdim;i++) {
    filt[i]=(float)(((double)filt[i])/((double)sum));
/*     fprintf(stderr,"filt[%d]=%e\n",i,filt[i]);  */
  }
}


/**************************************************************/
/**  returns the number of combination of n choose k ************/
float combinatorial(k,n)
  int k,n;
{
  int i;
  double numer,denom;
  
  if(k<0) k=0;
  if(k>n) k=n;

  if(k>(int)(n/2)) { k=(n-k); }  /* for stability use reflection */

  for(i=n,numer=1;i>(n-k);numer*=i,i--);
  for(i=k,denom=1;i>1;denom*=i,i--);
/*  fprintf(stderr,"numer=%e, denom=%e\n",numer,denom);  */
  return((float)(numer/denom));
}

/*********************************/
/** rescales an image from clipmin,clipmax to min,max so that
 ** clipmin is mapped to min, and clipmax is mapped to max. */
void rescale_image(im,res,imsize,clipmin,clipmax,min,max)
  float *im,*res;
  int imsize;
  float clipmin;
  float clipmax;
  float min,max;
{
  float scale=(max-min)/(clipmax-clipmin);
  int i;
  float im_val;
    
/*
  fprintf(stderr,"(min=%6.2f,max=%6.2f),(clipmin=%6.2f, clipmax=%6.2f) scale=%6.2f\n"
	 ,min,max,clipmin,clipmax,scale);fflush(stderr);
*/  
  for(i=0;i<imsize;i++) {
    im_val=scale*(im[i]-clipmin)+min;
    if(im_val<min)
      res[i]=min;
    else if(im_val>max)
      res[i]=max;
    else
      res[i]=im_val;
  }
}


/*********** crop an image ******************/
void crop_im(im,xdim,ydim,xstart,ystart,res,rxdim,rydim)
  float *im, *res;
  int xstart,ystart,xdim,ydim,rxdim,rydim;
{
  int x,y;
  int rxend=rxdim,ryend=rydim; /* not inclusive */
  int rxstart=0, rystart=0;
  int rypos,iypos;
  int cpysize;

  if(xstart<0) { rxstart= -xstart; xstart=0;}
  if((xstart+rxdim)>xdim) {rxend=xdim-xstart;} 
  if(ystart<0) { rystart= -ystart; ystart=0;}
  if((ystart+rydim)>ydim) {ryend=ydim-ystart;} 
  
  for(x=0,y=rxdim*rydim;x<y;res[x]=(float)0.0,x++);

  cpysize=sizeof(float)*(rxend-rxstart);

  for(y=rystart,iypos=ystart*xdim+xstart,rypos=rystart*rxdim+rxstart; y<ryend ;
      y++,iypos+=xdim,rypos+=rxdim) {
    memcpy(res+rypos,im+iypos,cpysize);
  }
}


/******** paste a sub image into res **********/
void paste_im(subim,xdim,ydim,res,rxstart,rystart,rxdim,rydim)
  float *subim,*res;
  int xdim,ydim,rxdim,rydim;
  int rxstart,rystart;
{
  int x,y;
  int rxend=rxstart+xdim, ryend=rystart+ydim;
  int xstart=0, ystart=0;
  int rypos,iypos;
  int cpysize;

  if(rxstart<0) { xstart=-rxstart; rxstart=0;}
  if(rxend>rxdim) {rxend=rxdim;} 
  if(rystart<0) { ystart=-rystart; rystart=0;}
  if(ryend>rydim) {ryend=rydim;} 

  cpysize=sizeof(float)*(rxend-rxstart);
  if(cpysize==0) return;
/*
  fprintf(stderr,"xstart=%d ystart=%d; rxstart=%d rystart=%d; rxend=%d ryend=%d cpysize=%d\n",
	       xstart,ystart,rxstart,rystart,rxend,ryend,cpysize);fflush(stderr);
*/
  for(y=rystart,iypos=ystart*xdim+xstart,rypos=rystart*rxdim+rxstart; y<ryend ;
      y++,iypos+=xdim,rypos+=rxdim) {
    memcpy(res+rypos,subim+iypos,cpysize);
  }
}


