/******************************************************************

"Copyright  1994 by the Massachusetts  Institute  of  Technology.  All
rights reserved.

Developed by John Y. A. Wang at the  Media Laboratory, MIT, Cambridge,
Massachusetts, with  support  from Secom Ltd., Goldstar Ltd.,  and the
Television of Tomorrow Program.

For use by authorized users only.

This distribution is approved  by Nicholas Negroponte, Director of the
Media Laboratory, MIT.

Permission to use, copy, or modify this software and its documentation
for educational and research purposes only and without  fee is  hereby
granted, provided that this copyright notice and the original authors'
names  appear  on  all  copies  and   supporting   documentation.   If
individual  files  are  separated  from  this  distribution  directory
structure, this copyright notice must be included.  For any other uses
of  this software, in  original or  modified form, including  but  not
limited to distribution in whole or in part, specific prior permission
must be  obtained  from  MIT.   These  programs  shall  not  be  used,
rewritten,  or  adapted  as  the  basis  of a  commercial software  or
hardware  product without  first  obtaining  appropriate licenses from
MIT.  MIT.  makes no  representations about  the  suitability of  this
software for any purpose.  It is  provided "as is" without  express or
implied warranty.

******************************************************************/


/***************************************************************
*** File: motion-main.c
*** Author: John Y. A. Wang
*** Date: January 15, 1993
*** Copyright 1993, Massachusetts Institute of Technology, All Rights Reserved.
***
*** Description: main motion analysis. The calls the sub-menus.
***************************************************************/

/*****************************************************************/
#include "dynamo.h"
#include "version.h"
/*****************************************************************/

/********** IMPORTANT:  put copyright label in all the files ***********/
static  char *copyright="\
\n\n***********************************************\n\
Developed by John Y. A. Wang at the Media Laboratory, MIT, Cambridge,Massachusetts\n\
Copyright  1993 by the Massachusetts  Institute  of  Technology.  All rights reserved.\n\
***********************************************\n\n";



/********* externals variables ***************/
char *args=NULL;
int untitled_num=0;


/*********************************************/


main()
{
  int i;
  int option;
  extern char *args;
  extern int untitled_num;

  Image_Manager image_mgr_struct;
  Image_Manager *image_mgr = &image_mgr_struct;

  Image_Cache cache_array[MAXIMAGES];
  
  /*************************************************************************/

  args=image_mgr->args;   /* set up extern args pointer */
  
  option = -1;

  printf("%s",COPYRIGHT_MSG); 

  if(option==-1) {
    printf("press RETURN key to continue....\n");
    fgets(args,ARGSIZE,stdin); sscanf(args,"%d",&option); }

  printf("%s",COPYRIGHT_EXT); 

  if(option == -1 ) {
    printf("press RETURN key to continue....\n");
    fgets(args,ARGSIZE,stdin); sscanf(args,"%d",&option); }

  /**************** initialize  cache *********************/
  for(i=0;i<MAXIMAGES;i++) {
    clear_cache(&cache_array[i]);
    image_mgr->image_list[i] = &cache_array[i];
  }
  
  init_image_manager(image_mgr);
  
  /*************** main driver ***********************************************/
  for(;;option=-1) {
    display_main_menu();
    printf("Please select option: ");
    while(option== -1) {
      readline(args,ARGSIZE,stdin); sscanf(args,"%d",&option);
    }

    switch(option){

    case MAIN_IMAGE_MANAGEMENT_MENU_OPT:
      image_management_menu(image_mgr); break;
    case MAIN_AFFINE_ANALYSIS_OPT:
      affine_analysis_menu(image_mgr); break;
    case MAIN_OPTICAL_FLOW_ANALYSIS_OPT:
      optical_flow_analysis_menu(image_mgr); break;
    case MAIN_DISPLAY_IMAGES_OPT:
      display_images_menu(image_mgr); break;
    case MAIN_ENHANCE_IMAGES_OPT:
      enhance_images_menu(image_mgr); break;
    case MAIN_BINARY_OPERATION_OPT:
      binary_operation_menu(image_mgr); break;
    case MAIN_UNARY_OPERATION_OPT:
      unary_operation_menu(image_mgr); break;
    case MAIN_FRAME_RATE_CONV_OPT:
      frame_rate_conv(image_mgr); break;
    case 999:
      test_cache(image_mgr); break;

    case MAIN_QUIT_OPT:
      free_all_memory(image_mgr);
      printf("%s\n",COPYRIGHT_MSG); 
      exit(0);
      break;
    default:
      printf("Option %d current not supported.\n", option);
      printf("\n\n");  break;
    } /* end switch */
  }
}

/*****************************************************************/
/*****************************************************************/
image_management_menu(image_mgr)
  Image_Manager *image_mgr;
{
  int option;
  extern char *args;

  for(;;) {
    display_image_management_menu();
    option = -1;
    printf("Please select option: ");
    readline(args,ARGSIZE,stdin);
   if((strncmp(args,"q",1)==0) || (strncmp(args,"\n",1)==0))
      return;
    sscanf(args,"%d",&option);
    
    switch(option) {
    case IMAGE_LOAD_OPT:
      load_image(image_mgr); break;
    case IMAGE_CHANGE_CACHE_SIZE_OPT:
      change_image_cache_size(image_mgr); break;
    case IMAGE_STATUS_OPT:
      display_image_list(image_mgr); break;
    case IMAGE_DELETE_OPT:
      delete_image(image_mgr); break;
    case IMAGE_SAVE_OPT:
      save_image(image_mgr); break;
    case IMAGE_RESET_MEMORY_OPT:
      clear_all_memory(image_mgr); break;
    default:
      printf("Option %d current not supported.\n", option); break;
    }
  }
}
/*****************************************************************/
/*****************************************************************/
affine_analysis_menu(image_mgr)
  Image_Manager *image_mgr;
{
  int option;
  extern char *args;

  for(;;){
    display_affine_analysis_menu();
    option = -1;
    printf("Please select option: ");
    readline(args,ARGSIZE,stdin);
    if((strncmp(args,"q",1)==0) || (strncmp(args,"\n",1)==0))
      return;
    sscanf(args,"%d",&option);
    
    switch(option) {
    case AFFINE_RUN_ANALYSIS_OPT:
      run_global_affine_motion_analysis(image_mgr); break;
    case AFFINE_DISPLAY_ANALYSIS_OPT:
      display_affine_motion_options(image_mgr); break;
    case AFFINE_SAVE_PARAMS_OPT:
      save_param_array(image_mgr); break;
    case AFFINE_DISPLAY_PARAMS_OPT:
      display_affine_params(image_mgr); break;
    case AFFINE_LOAD_PARAMS_OPT:
      load_affine_params(image_mgr); break;
    case AFFINE_ALIGN_IMAGES_OPT:
      original_display(image_mgr,2);  break;
    default:
      printf("Option %d current not supported.\n", option); break;
    }
  }
}

/*****************************************************************/
/*****************************************************************/
optical_flow_analysis_menu(image_mgr)
  Image_Manager *image_mgr;
{
  int option;
  extern char *args;

  for(;;) {
    display_optical_flow_analysis_menu();
    option = -1;
    printf("Please select option: ");
    readline(args,ARGSIZE,stdin);
    if((strncmp(args,"q",1)==0) || (strncmp(args,"\n",1)==0))
      return;
    sscanf(args,"%d",&option);
    
    switch(option) {
    case OPTICAL_FLOW_RUN_ANALYSIS_OPT:
      run_optical_flow_analysis(image_mgr); break;
    case OPTICAL_FLOW_DISPLAY_OPT:
      display_optical_flow_options(image_mgr); break;
    case OPTICAL_FLOW_DISPLAY_FIELD_OPT:
      display_vector_field_driver(image_mgr); break;
    default:
      printf("Option %d current not supported.\n", option); break;
    }
  }
}
/*****************************************************************/
/*****************************************************************/
display_images_menu(image_mgr)
  Image_Manager *image_mgr;
{
  int option;
  extern char *args;
  
  for(;;) {
    display_display_images_menu();
    option = -1;
    printf("Please select option: ");
    readline(args,ARGSIZE,stdin);
    if((strncmp(args,"q",1)==0) || (strncmp(args,"\n",1)==0))
      return;
    sscanf(args,"%d",&option);
    
    switch(option) {
    case DISPLAY_ORIGINAL_OPT:
      original_display(image_mgr,0);  break;
    case DISPLAY_AFFINE_ALIGN_OPT:
      original_display(image_mgr,1);  break;
    default:
      printf("Option %d current not supported.\n", option);
      break;
    }
  }
}  

/*****************************************************************/
/*****************************************************************/
enhance_images_menu(image_mgr)
  Image_Manager *image_mgr;
{
  int option;
  extern char *args;
  
  for(;;) {
    display_enhance_images_menu();
    option = -1;
    printf("Please select option: ");
    readline(args,ARGSIZE,stdin);
    if((strncmp(args,"q",1)==0) || (strncmp(args,"\n",1)==0))
      return;
    sscanf(args,"%d",&option);
    
    switch(option) {
    case ENHANCE_MEDIAN_ALIGN_OPT:
      enhance_median_images(image_mgr); break; 
    case ENHANCE_AVERAGE_ALIGN_OPT:
      enhance_average_images(image_mgr);  break;
    default:
      printf("Option %d current not supported.\n", option);
      break;
    }
  }
}  


/*****************************************************************/
/*****************************************************************/
binary_operation_menu(image_mgr)
  Image_Manager *image_mgr;
{
  int option;
  extern char *args;
  
  for(;;) {
    display_binary_ops_menu();
    option = -1;
    printf("Please select option: ");
    readline(args,ARGSIZE,stdin);
    if((strncmp(args,"q",1)==0) || (strncmp(args,"\n",1)==0))
      return;
    sscanf(args,"%d",&option);
    
    switch(option) {
    case BIN_ADD_OPT:
      binary_operate_images(image_mgr,option);  break;
    case BIN_SUB_OPT:
      binary_operate_images(image_mgr,option);  break;
    case BIN_MUL_OPT:
      binary_operate_images(image_mgr,option);  break;
    case BIN_DIV_OPT:
      binary_operate_images(image_mgr,option);  break;
    default:
      printf("Option %d current not supported.\n", option);
      break;
    }
  }
}


/*****************************************************************/
/*****************************************************************/
unary_operation_menu(image_mgr)
  Image_Manager *image_mgr;
{
  int option;
  extern char *args;
  
  for(;;) {
    display_unary_ops_menu();
    option = -1;
    printf("Please select option: ");
    readline(args,ARGSIZE,stdin);
    if((strncmp(args,"q",1)==0) || (strncmp(args,"\n",1)==0))
      return;
    sscanf(args,"%d",&option);
    
    switch(option) {
    case UNI_MEDIAN_OPT:
      temporal_filter_sequence(image_mgr,"median");  break;
    case UNI_AVERAGE_OPT:
      temporal_filter_sequence(image_mgr,"average");  break;
    default:
      printf("Option %d current not supported.\n", option);
      break;
    }
  }
}




test_cache(image_mgr)
  Image_Manager *image_mgr;
{

  extern char *args;
  Image_Cache *cache;
  int frame=0;
  float *res=NULL;
  char *progn="test_cache";

  while(1) {
    if((cache=ask_name_get_cache(image_mgr))==NULL)
      return(0);
    printf("\n");
    printf("cache: start=%d end=%d prev=%d\n",
	   cache->start,cache->end,cache->prev_frame);
    printf("Enter Frame Number (%d) >> ", frame);
    readline(args,ARGSIZE,stdin);
    sscanf(args,"%d",&frame);

    if(frame == -999) { continue; } 
    else if(frame < 0) {
      {
	float *temp=NULL;
	frame = -frame;
	if((temp=(float *)malloc(cache->imsize*sizeof(float)))==NULL) {
	  fprintf(stderr,"cannot get enough memory...\n"); continue; }
	if(get_image_disk(cache,frame,temp)!=temp) {
	  fprintf(stderr,"cannot read from disk frame %d\n",frame);
	}
	free(temp);
      }
    }
    else if((frame > 0) && (res=get_image(cache,frame))==NULL) {
      printf("%s: Progn error. Cannot get cache_res float array.\n",progn);
    }
  }
}

