/******************************************************************

"Copyright  1994 by the Massachusetts  Institute  of  Technology.  All
rights reserved.

Developed by John Y. A. Wang at the  Media Laboratory, MIT, Cambridge,
Massachusetts, with  support  from Secom Ltd., Goldstar Ltd.,  and the
Television of Tomorrow Program.

For use by authorized users only.

This distribution is approved  by Nicholas Negroponte, Director of the
Media Laboratory, MIT.

Permission to use, copy, or modify this software and its documentation
for educational and research purposes only and without  fee is  hereby
granted, provided that this copyright notice and the original authors'
names  appear  on  all  copies  and   supporting   documentation.   If
individual  files  are  separated  from  this  distribution  directory
structure, this copyright notice must be included.  For any other uses
of  this software, in  original or  modified form, including  but  not
limited to distribution in whole or in part, specific prior permission
must be  obtained  from  MIT.   These  programs  shall  not  be  used,
rewritten,  or  adapted  as  the  basis  of a  commercial software  or
hardware  product without  first  obtaining  appropriate licenses from
MIT.  MIT.  makes no  representations about  the  suitability of  this
software for any purpose.  It is  provided "as is" without  express or
implied warranty.

******************************************************************/


/***************************************************************
*** File: motion.h
*** Author: John Y. A. Wang
*** Date: May 26, 1993.
*** Copyright 1993, Massachusetts Institute of Technology, All Rights Reserved.
***
*** Description: header for motion analysis source
***      These routines are part of the DYNAMO software.
***************************************************************/

#ifndef DYNAMO_MOTION_INCLUDED
#define DYNAMO_MOTION_INCLUDED

static  char *copyright_m="\
\n\n***********************************************\n\
Author: John Y. A. Wang.\n\
DYNAMO Software\n\
The MIT Media Lab.\n\
Copyright (c) 1993 by Massachusetts Institute of Technology\n\
All Rights Reserved.\n\
***********************************************\n\n";

#include <stdio.h>
#include <malloc.h>
#include <stdlib.h>

#ifdef lint
/*
  lint 
*/
#define register
#endif

#define inbounds(x,dim)  ( ((x)>=0) && ((x)<(dim)) )
#define abs_val(x) (((x) < 0) ? -(x) : (x))

/************************** affine-mot-cf.c ********************/
extern void internal_coarse_fine_affine_estimator
  (/*_ float *im1, float *im2, float *ap, float *init_ap,
     int xdim, int ydim, int level, float *method,
     int xstart,int ystart,int xend,int yend,
     float *ttemp, int tempsize,
     float *f_array, float *mask, float *masktemp,
     int limit _*/);

/************************** affine-motion.c ********************/
extern void internal_affine_estimator_global
  (/* float *im1, float *im2, float *ap, int xdim, int ydim,
      int xstart, int ystart, int xend, int yend,
      float * ttemp,int tempsize, float * mask */);

/************************** affine-util.c ********************/

/* finds composite affine param  of  transformation with p1 followed by p2 */
extern void internal_update_affine_params_from
  (/* float *p1, float *p2, float *pr, float *factor */);

/* inverts the affine param p1 and  puts it in pr */
extern void internal_invert_affine_params_from
  (/*float *p1, float *pr */);

/** finds the composite aff param between from->to and puts it in pr. **/
extern void internal_update_affine_params_array
  (/* float *aram_array, int array_width, int to, int from, int pr,
      float *factor_in, float *pre_factor_in */);

/** takes an array of aff params and finds the composite params
 * to center_frame. returns -1 on error and 0 when complete. */
extern int internal_make_composite_aff_array
  (/* float *aff_array, float *res_array, int arraywidth,
      int center_frame, int nframes, int invert */);

/**** updates an affine parameter array with the following result:
  ***       pre_param -> param_array -> pos_param
  *** destructive modification allowed **/
extern void internal_update_aff_array
  (/* float *aff_array, float *res_array, int arraywidth,
      float *pre_param,float *post_param, int nframes */);

/** zoom the param with factors. result is xfrm with param
 ** followed by xy-zoom. returns results in param.
 ** destructive modification allowed. */
extern void zoom_affine_param
  (/* float *param,float factor1, float factor2 */);

/************************** display-color.c ********************/

extern void internal_display_color_im
  (/* float *r, float *g, float *b, int ysize, int xsize,
      float pedestal,float scale,float gamma, float zoom,
      char *title,unsigned char *tempim, int bufsize */);

extern double internal_init_color_movie
  (/* int ysize, int xsize, float zoom, float gamma,
      char *title, int save, char *filename */);

extern void internal_close_color_movie(/* FILE *pid */);

extern FILE *OpenRGBScreen
  (/* int height, int width, char *text, float *gamma,float *zoom,
      int quant, int movie */);

extern FILE *OpenGrayScreen
  (/* int height, int width, char *text, float *gamma, float *zoom,
      int nframes */);

extern FILE *DumpBufToScreen
  (/* float *Image, int height, int width, FILE *pid, int bufsize */);

extern FILE *OpenRGBMovie
  (/* int height, int width, char *text, float gamma */);

extern FILE *OpenRGBFile(/* int height, int width, char *filename */);

extern void internal_display_rgb_file
  (/* char *filename, float gamma, char *title*/);

extern void Close_Screen(/* FILE *pid */);

extern void internal_display_color_movie_frame
  (/* FILE *pid, float *r, float *g, float *b, int ysize, int xsize,
      float pedestal, float scale, float *tempim, int bufsize */);

extern void display_multichannel_image
  (/* float **seq, int nchannels, int xdim, int ydim,
      float zoom, char *title, int paste */);

extern void internal_display_multichan_im
  (/* float **seq, int ydim, int xdim, int nchannels,
      float pedestal, float scale, float gamma, float zoom,
      char *title, unsigned char *tempim, int bufsize, int paste*/);

extern FILE *OpenMCScreen
  (/* int height, int width, char *text, float gamma, float zoom, int quant,
      int movie, int nchannels,paste*/);

/************************** display-color.c ********************/

/** produces an indexed color RGB images base on the colormap.
 ** Uses defined STATIC color array  if colormap is NULL **/
void internal_index_color_order(/*image,r_im,g_im,b_im,im_size,z,numcolor,colormap,order*/);

/************************** display-digit.c ********************/

/* display the number as digits on an image.  array is of IMAGE_TYPE */
void display_number_on_image(/*array,xdim,ydim,number,xoffset,yoffset,fg_val,bg_val,ndigits*/);

/*************** display-number.c *******************/
/* displays numbers on images of various datatypes.
 * To replace display-digit.c. See display-number.h for prototypes */
#include "display-number.h"

/************************** grad-flow-cf.c  ********************/

/** coarse fine gradient flow by gaussian pyramids
 ** see file for more details **/
void internal_coarse_fine_gradient_flow(/*im1, im2, xflow, yflow, xdim, ydim,level,method,
				   flowlevel,f,ttemp,ttempsize,xbias,ybias,
				   clipmin,clipmax,xinit,yinit,
				   sum_filt,fdim,pyr_filt,pfdim,exp_filt,expfdim,
				   thresh,stat
				   */);


/************************** grad-off-cf.c  ********************/

/** coarse fine gradient flow by gaussian pyramids using 3 param model.
 ** see file for more details **/
void internal_coarse_fine_gradient_flow_offset
  (/*im1,im2,xflow,yflow,offset,
     xdim, ydim,level,method,
     flowlevel,f,ttemp,ttempsize,xbias,ybias,
     clipmin,clipmax,xinit,yinit,
     sum_filt,fdim,pyr_filt,pfdim,exp_filt,expfdim,
     thresh,stat
     */);

/************************** gradient-flow.c  ********************/

/** gradient flow using simple 2 param model.
 ** see file for more details **/
void gradient_flow_c(/*im1,im2,xflow,yflow,xdim,ydim,flowlevel,
		       sum_filt,fdim,f,
		       ttemp,tempsize,xbias,ybias,stat*/);


/************************** grad-off.c  ********************/

/** gradient flow using 3 param model.
 ** see file for more details **/
void gradient_flow_offset_c(/*im1,im2,xflow,yflow,offset,xdim,ydim,flowlevel,
			      sum_filt,fdim,f,
			      ttemp,tempsize,xbias,ybias,stat*/);


/************************** grad-flow-cf3.c  ********************/

/* multichannel implementation of coarse-fine gradient flow.
 * calls either the 2 or 3 parameter motion estimation depending
 *    on the nparams argument.
 * imseq1 and imseq2 are arrays of pointers to float
 *    arrays with nchannels. 
 * caller can specify the channels for analysis via select_channel.
 * try_nchannels specifies the number of select items in select_channel.
 */

void cf_grad_flow3 (/*imseq1, imseq2, xflow, yflow,offset,
		    xdim, ydim,level,method,
		    flowlevel,f,ttemp,ttempsize,xbias,ybias,
		    clipmin,clipmax,xinit,yinit,
		    nsum_filt,fdim,pyr_filt,pfdim,exp_filt,expfdim,
		    thresh,stat,nchannels,
		    ptemp,ptempsize,
		    nparams,
		    select_channel,try_nchannels
		    */);

/************************** gradient-flow3.c  ********************/
/* multichannel implementation of 2 parameter gradient flow
 * algorithm. imseq1 and imseq2 are arrays of pointers to float
 * arrays. nchannels specifies the number channels in imseq.
 */

void gradient_flow3_c(/*imseq1,imseq2,xflow,yflow,xdim,ydim,
		      flowlevel,sum_filt,fdim,f,
		      ttemp,tempsize,xbias,ybias,stat,nchannels*/);

/************************** grad-off3.c  ********************/
/* multichannel implementation of 3 parameter gradient flow
 * algorithm. imseq1 and imseq2 are arrays of pointers to float
 * arrays. nchannels specifies the number channels in imseq.
 */
void gradient_flow_offset3_c(/*imseq1,imseq2,xflow,yflow,offset,
			     xdim,ydim,flowlevel,sum_filt,fdim,f,
			     ttemp,tempsize,xbias,ybias,stat,nchannels*/);
void accum_mul_im_scale(/*scale, a1, a2, res, size*/);
void accum_im_scale(/*scale, a1, res,size*/);

/************************** median.c  ********************/

double internal_median(/*list, n*/);

/* sort "list" of length "n" and returns "list" that is sorted. */
void internal_sort(/*list,n*/);


/** this obtains the weighted median, stochastically, this  is the ML
  estimate. If we let the normalized weights be the probabililties then
  ML is where the CMF cumulative  mass function is 0.5. This algorithm
  calculates the CMF by sorting according to input values with corresponding
  weights, then scans the sorted weight list to find the point where 
  CMF is equal to the halfway point.
  **/
double internal_weighted_median(/*list, weight, n*/);

/* sort on list and order list2 using list as well. */
void internal_sort2(/*list,list2,n*/);

/* median on the array but ignore values that are equal to
 * checkval */   
double internal_median_checkval(/*im, size, checkval*/);

double internal_median_window(/*image,size,thresh,ww*/);

int internal_median_seq (/*im_addr, n_frames, res, size, check_val, setval*/);

int internal_weighted_median_seq (/* im_addr, weight, n_frames, res, size,
				     check_val, setval*/);

/* find the variance of a sequences */
int internal_variance_seq(/*im_addr, n_frames, mean_im, res, size, check_val*/);

/* subtract two sequences frame by frame */
int internal_sub_seq (/*im_addr, n_frames, im, out_addr, size*/);

/*********************** median-convolve.c ********************************/
int internal_median_filter_reflect
  (/*image, x_dim, y_dim, filt, temp,x_fdim, y_fdim,result,
   xfilt_space, yfilt_space,type*/);

/**************************************************************************/
/************************** motion-imops.c  ********************/

/* set a float array to val */
void fill_im(/*a,val,size*/);
#define fill_array(a,val,size) fill_im(a,val,size)

/** update the flow estimate by composing the current flow map with the new flow map
 ** used by grad-flow-cf.c grad-offset-cf.c **/
void internal_update_flow(/*xflow,yflow,xnewflow,ynewflow,xdim,ydim,clipmin,clipmax*/);

/* expand the flow to the next level in the pyramid
** 1) upsample the flow and 2) apply filter to smooth image
** temp_in is assumed to be 2x the size of the image
** used by grad-flow-cf.c grad-offset-cf.c **/
void internal_expand_flow(/*xflow,yflow,xres,yres,temp,xdim,ydim,rxdim,rydim,exp_filt,fdim,
		     thresh*/);

/* expands the image by space parameter
** requires temp_im space 2x the size of the result image
** used by grad-flow-cf.c grad-offset-cf.c **/
void expand_sep_filter_reflect(/*image, result, temp_im, xdim, ydim, rxdim, rydim,
			       xfilt, yfilt, xfdim, yfdim,xfilt_space, yfilt_space*/);

/* clip values of a float array with clipmin and clipmanx */
void clip_array (/*array, clipmin, clipmax, size*/);

/* remaps the values of an float image such that resulting values is:
**  res = val*abs_val(val)/thresh  if abs_val(val)< threshold */
void remap_image(/*im,xdim,ydim,thresh*/);

/***** the following in motion-imops.c are common to all motion source *******/

/***** multiply images *****/
void mul_im(/*a1, a2, res, size*/);
#define mul_array(a1, a2, res, size) mul_im(a1, a2, res, size)

void mul_im_sc(/*a1, val, res, size*/);

/***** divide images *****/
void div_im(/*a1, a2, res, size*/);
#define div_array(a1, a2, res, size) div_im(a1, a2, res, size)

#ifdef NEVER
void div_im_sc(/*a1, val, res, size*/);
#endif

/***** subtract images *****/
void sub_im (/*a1, a2, res, size*/);
#define sub_array(a1, a2, res, size)  sub_im(a1, a2, res, size)

/***** add images *****/
void add_im(/*a1, a2, res, size*/);
#define add_array(a1, a2, res, size)  add_im(a1, a2, res, size)

void add_im_sc(/*a1,val,res,size*/);

/***** copy images *****/
void copy_im(/*a1, res, size*/);
#define copy_array(a1, res, size)  copy_im(a1, res, size)

/****** zero image ********/
void zero_im(/*image, size*/);
#define zero_array(a1, res, size)  zero_im(a1, res, size)

/**** abs image *****/
void abs_im(/*a1, res, size*/);
#define abs_array(a1, res, size)  abs_im(a1, res, size)

/****** finds the maximum of the image *****/
float max_image(/*im,imsize*/); 

/** image compare **/
void internal_greater_than_im (/*im,res,val,size*/);
void internal_less_than_im (/*im,res,val,size*/);
void internal_equal_im(/*im,res,val,size*/);
void blur_image(/*image,result,temp,xsize,ysize,level,g_filt,fdim*/);
void apply_sep_filter_reflect(/*image, result, temp_im, xdim, ydim,
			       xfilt, yfilt, xfdim, yfdim,xfilt_space, yfilt_space*/);
void apply_nonsep_filter_reflect(/*im,res,xdim,ydim,filt,xfdim,yfdim*/);

/************* additional stuff in motion-imops.c **************/

/** makes a binomial distributed  filter of fdim; **/
void binomial_filter(/*filt,fdim*/);

/**  returns the number of combination of n choose k ************/
float combinatorial(/*k,n*/);

/** rescales an image from clipmin,clipmax to min,max so that
 ** clipmin is mapped to min, and clipmax is mapped to max. */
void rescale_image(/*im,res,imsize,clipmin,clipmax,min,max*/);

/*********** crop an image ******************/
/* Note: when res==im, funny things will happen */
void crop_im(/*im,xdim,ydim,xstart,ystart,res,rxdim,rydim*/);

/******** paste a sub image into res **********/
void paste_im(/*sub_im,xdim,ydim,res,rxstart,rystart,rxdim,rydim*/);

/**************************************************************************/
/**************************************************************************/


/************************** new-warper.c  ********************/

/** affine warp using bicubic interpolation **/
void bicubic_affine_warp(/*im, res, ax, bx, cx, ay, by, cy, ydim, xdim*/);

/** affine warp using bicubic interpolation. Allows for warping into larger res **/
void bicubic_affine_warp_larger(/*im, res, ax, bx, cx, ay, by, cy, ydim, xdim,
			   rydim, rxdim*/);

/* affine warp only regions within: rxstart,rystart,rxend,ryend (not including *end).
** code is same as bicubic_affine_warp except for the checking in the result loop */
void bicubic_affine_warp_limit(/*im, res, ax, bx, cx, ay, by, cy, ydim, xdim,
			  rxstart,rystart,rxend,ryend*/);

/* affine warp only regions within: rxstart,rystart,rxend,ryend (not including *end).
** code is same as bicubic_affine_warp except for the checking in the result loop */
/* returns a weighted square distance to the nearest pixel */
void bicubic_affine_warp_limit_weight(/*im, res, ax, bx, cx, ay, by, cy, ydim, xdim,
			  rydim,rxdim,rxstart,rystart,rxend,ryend,weight*/);


/* affine warp only regions within: rxstart,rystart,rxend,ryend (not include end).
 * also check for input crop region. 
 * code is same as bicubic_affine_warp except for the checking in the result loop */
/* returns a weighted square distance to the nearest pixel */
/* adds to resulting image */
void bicubic_affine_warp_limit2_weight(/*im, res, ax, bx, cx, ay, by, cy,
				       ydim, xdim,xstart,ystart,xend,yend,
				       rydim,rxdim,rxstart,rystart,rxend,ryend,
				       weight*/);

/* warps an image with warp maps wx and wy using bicubic interpolation */
void bicubic_warp(/*im, res, wy, wx, ydim, xdim*/);



/***************************  pyramid .c *******************/

/* base image points to the input image. Other levels of the pyramid are
** assigned to segments in the pyrmemory.
** pyrmemory and temp should be at least the size of im.
** Actually does the filtering if make==1, otherwise just set pointers. */
void make_pyramid(/*im, level, xdim, ydim, pyr, pyrmemory, temp, make, filter, fdim*/);

/* convolves and subsamples the image */
void apply_sep_filter_subsample(/*image, result, temp_im, xdim, ydim,
			       xfilt, yfilt, xfdim, yfdim,xfilt_space, yfilt_space*/);


/********************** conversion.c  ******************************/

/** converts a byte array to a float array **/
void convert_byte_to_float(/*array,image,size*/);

/* converts a byte to a float according to pedestal  and scale */
void convert_byte_to_float2(/*array,image,size,pedestal,scale*/);

/* converts a float to a byte assumes pedestal=0 scale=1 */
void convert_float_to_byte(/*image,array,imsize*/);

/* converts a sub-array to a float array */
void convert_subarray_byte_to_float(/*array,axdim,aydim,axstart,aystart,image,xdim,ydim*/);

/** converts a sub-float array to a byte array according to pedestal and scale. **/
void convert_subfloat_to_byte(/*image,xdim,ydim,xstart,ystart,xend,yend,
			 scale,pedestal,array,axdim,aydim*/);


/********************* byteops.c ********************************/

/* fill a byte array with val */
void fill_byte(/*array,val,asize*/);

/* crop a byte 2D array. Note: when res==im, funny things will happen */
void crop_byte(/*im,xdim,ydim,xstart,ystart,res,rxdim,rydim*/);

/********************* drawline.c ********************************/

/* draws vector-fields to byte array */
int display_vector_field(/*dx,dy,xdim,ydim,array,axdim,aydim,val,xstep,ystep,
		     scale,headsize,type*/);

/*** draws arrow heads allows for 3 different types of arrows ***/
/* arrow head is drawn at (xe,ye)  */
void draw_arrow(/*array,axdim,aydim,xs,ys,xe,ye,val,scale,type,headsize*/);

/** draws a line into a byte array specifed by start: xs,ys; end: xe,ye. **/
int drawline(/*array,axdim,aydim,xs,ys,xe,ye,val*/);


/*************************** fill-image. c ************************/
/** NOTE: In this file, image type will be (unsigend char)
 ** if compiled with DBYTE_8. Otherwise image type is (float).
 **/

/* iterative fill-in image.
** Warning input images  will be modified upon returning.
** Caller should take care in preserving the original. Returns 0
** when fill-in was complete, otherwise 1.
**/
int fill_image_n(/*im,temp_im,xdim,ydim,mask,temp_mask,f_val,niter,type*/);


#endif
