/******************************************************************

"Copyright  1994 by the Massachusetts  Institute  of  Technology.  All
rights reserved.

Developed by John Y. A. Wang at the  Media Laboratory, MIT, Cambridge,
Massachusetts, with  support  from Secom Ltd., Goldstar Ltd.,  and the
Television of Tomorrow Program.

For use by authorized users only.

This distribution is approved  by Nicholas Negroponte, Director of the
Media Laboratory, MIT.

Permission to use, copy, or modify this software and its documentation
for educational and research purposes only and without  fee is  hereby
granted, provided that this copyright notice and the original authors'
names  appear  on  all  copies  and   supporting   documentation.   If
individual  files  are  separated  from  this  distribution  directory
structure, this copyright notice must be included.  For any other uses
of  this software, in  original or  modified form, including  but  not
limited to distribution in whole or in part, specific prior permission
must be  obtained  from  MIT.   These  programs  shall  not  be  used,
rewritten,  or  adapted  as  the  basis  of a  commercial software  or
hardware  product without  first  obtaining  appropriate licenses from
MIT.  MIT.  makes no  representations about  the  suitability of  this
software for any purpose.  It is  provided "as is" without  express or
implied warranty.

******************************************************************/


/***********************************************************
  File: new-warper.c
  Author: John Y. A. Wang
  Date:   July 23, 1992

  Discription: Performs from_warp using bicubic interpolation.

 Copyright 1992, Masschusetts Institute of Technology,
 All Rights Reserved

***********************************************************/

static  char *copyright="\
\n\n***********************************************\n\
File: new-warper.c\n\
Author: John Y. A. Wang.\n\
The MIT Media Lab.\n\
Copyright (c) 1992, All Rights Reserved.\n\
***********************************************\n\n";

#include <stdio.h>
#include <math.h>
#include "motion.h"

/***************************************************
Bicubic interpolation initially implemented by E. Simoncelli
for Object-Based Vision and Image Understanding System (OBVIUS)
Copyright 1988 by MIT Media Laboratory, Massachusetts Institute of
Technology. 
***************************************************/
 
#define bound(x, n)	((x)<0 ? 0 : ((x)>(n) ? (n) : (x)))
#define mod(x, n) ((x)>0?((x)%(n)) : ((n)-(-(x)%(n))%(n)))
#define abs_val(x) (((x) < 0) ? -(x) : (x))
#define roundoff(x) (((x)>0)?((int)((x)+0.5)):((int)((x)-0.5)))

/** affine warp using bicubic interpolation **/
/* adds to resulting image */
void bicubic_affine_warp(im, res, ax, bx, cx, ay, by, cy, ydim, xdim)
  float *im, *res; 
  float ax, bx, cx, ay, by, cy;
  int xdim, ydim;

{
  register float *im_ptr;	/* accessed 38x in loop */
  register int base;		/* accessed 30x in loop */
  register float frac, temp;	/* accessed 20x in loop */
  register float frac2, frac3;  /* accessed 10x in loop */
  int x_inbounds_1, x_inbounds0, x_inbounds1, x_inbounds2; /* 6x in loop */
  register float x_coeff_1, x_coeff0, x_coeff1, x_coeff2; /* 5x in loop */
  register float res_val;
  register int i,j;
  register float xfrac,xbase,yfrac,ybase;
  float x_off, y_off;
    
  x_off = (float)((xdim - 1) / 2.0);
  y_off = (float)((ydim - 1) / 2.0);

  /*
  **  wx = ax + bx * x + cx * y;
  **  wy = ay + by * x + cy * y;
  */
/*
  printf ("\nbicubic_affine_warp: (%5f %5f %5f %5f %5f %5f)", ay, by, cy, ax, bx, cx);
  printf( "\n(%5f %5f: %d %d)", y_off, x_off, ydim, xdim);
*/
  
/*  memset(res, 0, xdim*ydim*sizeof(float)); */
  for(j=0; j<ydim; j++)  /* j is the y var */
    for(i=0; i<xdim; i++)  /* i is the x var */
	{

	xfrac = i - (ax + bx * (i - x_off) + cx * (j - y_off));
	xbase = (int)(xfrac);
	yfrac = j - ( ay + by * (i - x_off) + cy * (j - y_off));
	ybase = (int)(yfrac);

	if((((xbase) >= 0) && ((xbase) < xdim)) &&
	   (((ybase) >= 0) && ((ybase) < ydim)))
	  {
	    res_val = 0.0;
	    /* Compute X fractional position and base index */
	    frac = xfrac - xbase;
	    frac2 = frac*frac;
	    frac3 = frac*frac2;
	    base  = xbase - 1;
	    im_ptr = im + base;
	    
	    x_inbounds_1 = ((base >= 0) && (base < xdim))?1:0;
	    if (x_inbounds_1) x_coeff_1 = (frac2/2.0 - frac/3.0 - frac3/6.0);
	    base++;
	    x_inbounds0  = ((base >= 0) && (base < xdim))?1:0;
	    if (x_inbounds0)  x_coeff0  = (1.0 - frac/2.0 - frac2 + frac3/2.0);
	    base++;
	    x_inbounds1  = ((base >= 0) && (base < xdim))?1:0;
	    if (x_inbounds1)  x_coeff1 = (frac + frac2/2.0 - frac3/2.0);
	    base++;
	    x_inbounds2  = ((base >= 0) && (base < xdim))?1:0;
	    if (x_inbounds2)  x_coeff2 = (frac3/6.0 - frac/6.0);
	    
	    /* Compute Y fractional position and base index */
	    frac = yfrac - ybase;
	    frac2 = frac*frac;
	    frac3 = frac*frac2;
	    base = ybase - 1 ;
	    im_ptr += base*xdim;
	    
	    if ((base >= 0) && (base < ydim)) { /* y_inbounds_1 */
	      temp = 0.0;
	      if (x_inbounds_1) temp += x_coeff_1 * *im_ptr; im_ptr++;
	      if (x_inbounds0)  temp += x_coeff0  * *im_ptr; im_ptr++;
	      if (x_inbounds1)  temp += x_coeff1  * *im_ptr; im_ptr++;
	      if (x_inbounds2)  temp += x_coeff2  * *im_ptr; im_ptr -= 3;
	      res_val += temp * (frac2/2.0 - frac/3.0 - frac3/6.0);
	    }
	    base++;  im_ptr += xdim;
	    if ((base >= 0) && (base < ydim)) {  /* y_inbounds0 */
	      temp = 0.0;
	      if (x_inbounds_1) temp += x_coeff_1 * *im_ptr; im_ptr++;
	      if (x_inbounds0)  temp += x_coeff0  * *im_ptr; im_ptr++;
	      if (x_inbounds1)  temp += x_coeff1  * *im_ptr; im_ptr++;
	      if (x_inbounds2)  temp += x_coeff2  * *im_ptr; im_ptr -= 3;
	      res_val += temp * (1 - frac/2.0 - frac2 + frac3/2.0);
	      }
	    base++;  im_ptr += xdim; 
	    if ((base >= 0) && (base < ydim)) { /* y_inbounds1 */
	      temp = 0.0;
	      if (x_inbounds_1) temp += x_coeff_1 * *im_ptr; im_ptr++;
	      if (x_inbounds0)  temp += x_coeff0  * *im_ptr; im_ptr++;
	      if (x_inbounds1)  temp += x_coeff1  * *im_ptr; im_ptr++;
	      if (x_inbounds2)  temp += x_coeff2  * *im_ptr; im_ptr -= 3;
	      res_val += temp * (frac + frac2/2.0 - frac3/2.0);
	    }
	    base++;  im_ptr += xdim;
	    if ((base >= 0) && (base < ydim)) { /* y_inbounds2 */
	      temp = 0.0;
	      if (x_inbounds_1) temp += x_coeff_1 * *im_ptr; im_ptr++;
	      if (x_inbounds0)  temp += x_coeff0  * *im_ptr; im_ptr++;
	      if (x_inbounds1)  temp += x_coeff1  * *im_ptr; im_ptr++;
	      if (x_inbounds2)  temp += x_coeff2  * *im_ptr; im_ptr -= 3;
	      res_val += temp * (frac3/6.0 - frac/6.0);
	    }
	    *res += (float)res_val;
	  }
	res++;
      }
  fflush(stdout);
}


/** affine warp using bicubic interpolation. Allows for warping into larger res **/
/* adds to resulting image */
void bicubic_affine_warp_larger(im, res, ax, bx, cx, ay, by, cy, ydim, xdim,
			   rydim, rxdim)
  float *im, *res;
  float ax, bx, cx, ay, by, cy;
  int xdim, ydim;
  int rxdim, rydim;

{
  register float *im_ptr;	/* accessed 38x in loop */
  register int base;		/* accessed 30x in loop */
  register float frac, temp;	/* accessed 20x in loop */
  register float frac2, frac3;  /* accessed 10x in loop */
  int x_inbounds_1, x_inbounds0, x_inbounds1, x_inbounds2; /* 6x in loop */
  register float res_val;
  float x_coeff_1, x_coeff0, x_coeff1, x_coeff2; /* 5x in loop */
  register int i,j;
  register float x_off, y_off;
  register float rx_off, ry_off;
  register float x_diff, y_diff;
  float xfrac,xbase,yfrac,ybase;
  
  x_off = (xdim - 1) / 2.0;
  y_off = (ydim - 1) / 2.0;
  rx_off = (rxdim - 1) / 2.0;
  ry_off = (rydim - 1) / 2.0;

  x_diff = rx_off - x_off;
  y_diff = ry_off - y_off;

  /*
  **  wx = ax + bx * x + cx * y;
  **  wy = ay + by * x + cy * y;
  */

/*
  printf ("\n%5f %5f %5f %5f %5f %5f", ay, by, cy, ax, bx, cx);
  printf( "\n%5f %5f: %d %d", y_off, x_off, ydim, xdim);
*/
  
/*  memset(res, 0, xdim*ydim*sizeof(float)); */
  for(j=0; j<rydim; j++)  /* j is the y var */
    for(i=0; i<rxdim; i++) /* i is the x var */
      {
	xfrac = i - x_diff - (ax + bx * (i - rx_off) + cx * (j - ry_off));
	xbase = (int)(xfrac);
	yfrac = j - y_diff - (ay + by * (i - rx_off) + cy * (j - ry_off));
	ybase = (int)(yfrac);

	if((((xbase-1) >= 0) || ((xbase + 3) >= 0)) &&
	   (((xbase-1) < xdim) || ((xbase + 3) < xdim)) &&
	   (((ybase-1) >= 0) || ((ybase + 3) >= 0)) &&
	   (((ybase-1) < ydim) || ((ybase + 3) < ydim))
	   )
	  {
	    res_val = 0.0;
	    /* Compute X fractional position and base index */
	    frac = xfrac - xbase;
	    frac2 = frac*frac;
	    frac3 = frac*frac2;
	    base  = xbase - 1;
	    im_ptr = im + base;
	    
	    x_inbounds_1 = ((base >= 0) && (base < xdim))?1:0;
	    if (x_inbounds_1) x_coeff_1 = (frac2/2.0 - frac/3.0 - frac3/6.0);
	    base++;
	    x_inbounds0  = ((base >= 0) && (base < xdim))?1:0;
	    if (x_inbounds0)  x_coeff0  = (1.0 - frac/2.0 - frac2 + frac3/2.0);
	    base++;
	    x_inbounds1  = ((base >= 0) && (base < xdim))?1:0;
	    if (x_inbounds1)  x_coeff1 = (frac + frac2/2.0 - frac3/2.0);
	    base++;
	    x_inbounds2  = ((base >= 0) && (base < xdim))?1:0;
	    if (x_inbounds2)  x_coeff2 = (frac3/6.0 - frac/6.0);
	    
	    /* Compute Y fractional position and base index */
	    frac = yfrac - ybase;
	    frac2 = frac*frac;
	    frac3 = frac*frac2;
	    base = ybase - 1 ;
	    im_ptr += base*xdim;
	    
	    if ((base >= 0) && (base < ydim)) { /* y_inbounds_1 */
	      temp = 0.0;
	      if (x_inbounds_1) temp += x_coeff_1 * *im_ptr; im_ptr++;
	      if (x_inbounds0)  temp += x_coeff0  * *im_ptr; im_ptr++;
	      if (x_inbounds1)  temp += x_coeff1  * *im_ptr; im_ptr++;
	      if (x_inbounds2)  temp += x_coeff2  * *im_ptr; im_ptr -= 3;
	      res_val += temp * (frac2/2.0 - frac/3.0 - frac3/6.0);
	    }
	    base++;  im_ptr += xdim;
	    if ((base >= 0) && (base < ydim)) { /* y_inbounds0 */
	      temp = 0.0;
	      if (x_inbounds_1) temp += x_coeff_1 * *im_ptr; im_ptr++;
	      if (x_inbounds0)  temp += x_coeff0  * *im_ptr; im_ptr++;
	      if (x_inbounds1)  temp += x_coeff1  * *im_ptr; im_ptr++;
	      if (x_inbounds2)  temp += x_coeff2  * *im_ptr; im_ptr -= 3;
	      res_val += temp * (1 - frac/2.0 - frac2 + frac3/2.0);
	    }
	    base++;  im_ptr += xdim; 
	    if ((base >= 0) && (base < ydim)) { /* y_inbounds1 */
	      temp = 0.0;
	      if (x_inbounds_1) temp += x_coeff_1 * *im_ptr; im_ptr++;
	      if (x_inbounds0)  temp += x_coeff0  * *im_ptr; im_ptr++;
	      if (x_inbounds1)  temp += x_coeff1  * *im_ptr; im_ptr++;
	      if (x_inbounds2)  temp += x_coeff2  * *im_ptr; im_ptr -= 3;
	      res_val += temp * (frac + frac2/2.0 - frac3/2.0);
	    }
	    base++;  im_ptr += xdim;
	    if ((base >= 0) && (base < ydim)) { /* y_inbounds2 */
	      temp = 0.0;
	      if (x_inbounds_1) temp += x_coeff_1 * *im_ptr; im_ptr++;
	      if (x_inbounds0)  temp += x_coeff0  * *im_ptr; im_ptr++;
	      if (x_inbounds1)  temp += x_coeff1  * *im_ptr; im_ptr++;
	      if (x_inbounds2)  temp += x_coeff2  * *im_ptr; im_ptr -= 3;
	      res_val += temp * (frac3/6.0 - frac/6.0);
	    }
	    *res += (float)res_val;
	  }
	res++;
      }
  fflush(stdout);
}


/* affine warp only regions within: rxstart,rystart,rxend,ryend (not including *end).
** code is same as bicubic_affine_warp except for the checking in the result loop */
/* adds to resulting image */
void bicubic_affine_warp_limit(im, res, ax, bx, cx, ay, by, cy, ydim, xdim,
			  rxstart,rystart,rxend,ryend)
  float *im, *res;
  float ax, bx, cx, ay, by, cy;
  int xdim, ydim;
  int rxstart,rystart,rxend,ryend;

{
  register float *im_ptr;	/* accessed 38x in loop */
  register int base;		/* accessed 30x in loop */
  register float frac, temp;	/* accessed 20x in loop */
  register float frac2, frac3;  /* accessed 10x in loop */
  int x_inbounds_1, x_inbounds0, x_inbounds1, x_inbounds2; /* 6x in loop */
  register float x_coeff_1, x_coeff0, x_coeff1, x_coeff2; /* 5x in loop */
  register float res_val;
  register int i,j;
  register float xfrac,xbase,yfrac,ybase;
  float x_off, y_off;
  int pos;
    
  x_off = (float)((xdim - 1) / 2.0);
  y_off = (float)((ydim - 1) / 2.0);

  /*
  **  wx = ax + bx * x + cx * y;
  **  wy = ay + by * x + cy * y;
  */
/*
  printf ("\nbicubic_affine_warp_limit: (%5f %5f %5f %5f %5f %5f)", ay, by, cy, ax, bx, cx);
  printf( "\n(%5f %5f: %d %d)", y_off, x_off, ydim, xdim);
*/
  
/*  memset(res, 0, xdim*ydim*sizeof(float)); */

  if(rxstart<0) rxstart=0;  if(rxend>xdim) rxend=xdim-1;
  if(rystart<0) rystart=0;  if(ryend>ydim) ryend=ydim-1;

/*
  printf("bicubic_affine_warp_limit: rxstart=%d,rystart=%d,rxend=%d,ryend=%d\n",
	 rxstart,rystart,rxend,ryend);
*/

  for(j=rystart; j<ryend; j++)  /* j is the y var */
    for(i=rxstart,pos=j*xdim+i; i<rxend; i++,pos++)  /* i is the x var */
	{

	xfrac = i - (ax + bx * (i - x_off) + cx * (j - y_off));
	xbase = (int)(xfrac);
	yfrac = j - ( ay + by * (i - x_off) + cy * (j - y_off));
	ybase = (int)(yfrac);

	if((((xbase) >= 0) && ((xbase) < xdim)) &&
	   (((ybase) >= 0) && ((ybase) < ydim)))
	  {
	    res_val = 0.0;
	    /* Compute X fractional position and base index */
	    frac = xfrac - xbase;
	    frac2 = frac*frac;
	    frac3 = frac*frac2;
	    base  = xbase - 1;
	    im_ptr = im + base;
	    
	    x_inbounds_1 = ((base >= 0) && (base < xdim))?1:0;
	    if (x_inbounds_1) x_coeff_1 = (frac2/2.0 - frac/3.0 - frac3/6.0);
	    base++;
	    x_inbounds0  = ((base >= 0) && (base < xdim))?1:0;
	    if (x_inbounds0)  x_coeff0  = (1.0 - frac/2.0 - frac2 + frac3/2.0);
	    base++;
	    x_inbounds1  = ((base >= 0) && (base < xdim))?1:0;
	    if (x_inbounds1)  x_coeff1 = (frac + frac2/2.0 - frac3/2.0);
	    base++;
	    x_inbounds2  = ((base >= 0) && (base < xdim))?1:0;
	    if (x_inbounds2)  x_coeff2 = (frac3/6.0 - frac/6.0);
	    
	    /* Compute Y fractional position and base index */
	    frac = yfrac - ybase;
	    frac2 = frac*frac;
	    frac3 = frac*frac2;
	    base = ybase - 1 ;
	    im_ptr += base*xdim;
	    
	    if ((base >= 0) && (base < ydim)) { /* y_inbounds_1 */
	      temp = 0.0;
	      if (x_inbounds_1) temp += x_coeff_1 * *im_ptr; im_ptr++;
	      if (x_inbounds0)  temp += x_coeff0  * *im_ptr; im_ptr++;
	      if (x_inbounds1)  temp += x_coeff1  * *im_ptr; im_ptr++;
	      if (x_inbounds2)  temp += x_coeff2  * *im_ptr; im_ptr -= 3;
	      res_val += temp * (frac2/2.0 - frac/3.0 - frac3/6.0);
	    }
	    base++;  im_ptr += xdim;
	    if ((base >= 0) && (base < ydim)) { /* y_inbounds0 */
	      temp = 0.0;
	      if (x_inbounds_1) temp += x_coeff_1 * *im_ptr; im_ptr++;
	      if (x_inbounds0)  temp += x_coeff0  * *im_ptr; im_ptr++;
	      if (x_inbounds1)  temp += x_coeff1  * *im_ptr; im_ptr++;
	      if (x_inbounds2)  temp += x_coeff2  * *im_ptr; im_ptr -= 3;
	      res_val += temp * (1 - frac/2.0 - frac2 + frac3/2.0);
	    }
	    base++;  im_ptr += xdim; 
	    if ((base >= 0) && (base < ydim)) { /* y_inbounds1 */
	      temp = 0.0;
	      if (x_inbounds_1) temp += x_coeff_1 * *im_ptr; im_ptr++;
	      if (x_inbounds0)  temp += x_coeff0  * *im_ptr; im_ptr++;
	      if (x_inbounds1)  temp += x_coeff1  * *im_ptr; im_ptr++;
	      if (x_inbounds2)  temp += x_coeff2  * *im_ptr; im_ptr -= 3;
	      res_val += temp * (frac + frac2/2.0 - frac3/2.0);
	    }
	    base++;  im_ptr += xdim;
	    if ((base >= 0) && (base < ydim)) { /* y_inbounds2 */
	      temp = 0.0;
	      if (x_inbounds_1) temp += x_coeff_1 * *im_ptr; im_ptr++;
	      if (x_inbounds0)  temp += x_coeff0  * *im_ptr; im_ptr++;
	      if (x_inbounds1)  temp += x_coeff1  * *im_ptr; im_ptr++;
	      if (x_inbounds2)  temp += x_coeff2  * *im_ptr; im_ptr -= 3;
	      res_val += temp * (frac3/6.0 - frac/6.0);
	    }
	    res[pos] += (float)res_val;
	  }
      }
  fflush(stdout);
}


/* affine warp only regions within: rxstart,rystart,rxend,ryend (not including *end).
** code is same as bicubic_affine_warp except for the checking in the result loop */
/* returns a weighted square distance to the nearest pixel */
/* adds to resulting image */
void bicubic_affine_warp_limit_weight(im, res, ax, bx, cx, ay, by, cy, ydim, xdim,
			  rydim,rxdim,rxstart,rystart,rxend,ryend,weight)
  float *im, *res;
  float ax, bx, cx, ay, by, cy;
  int xdim, ydim;
  int rxdim,rydim;
  int rxstart,rystart,rxend,ryend;
  float *weight;

{
  register float *im_ptr;	/* accessed 38x in loop */
  register int base;		/* accessed 30x in loop */
  register float frac, temp;	/* accessed 20x in loop */
  register float frac2, frac3;  /* accessed 10x in loop */
  int x_inbounds_1, x_inbounds0, x_inbounds1, x_inbounds2; /* 6x in loop */
  register float x_coeff_1, x_coeff0, x_coeff1, x_coeff2; /* 5x in loop */
  register float res_val;
  register int i,j;
  register float xfrac,xbase,yfrac,ybase;
  float x_off, y_off;
  float rx_off,ry_off,x_diff,y_diff;
  int pos;
  float xx,yy,weight_val;

  x_off = (xdim - 1) / 2.0;
  y_off = (ydim - 1) / 2.0;
  rx_off = (rxdim - 1) / 2.0;
  ry_off = (rydim - 1) / 2.0;

  x_diff = rx_off - x_off;
  y_diff = ry_off - y_off;
  

  /*
  **  wx = ax + bx * x + cx * y;
  **  wy = ay + by * x + cy * y;
  */
/*
  printf ("\nbicubic_affine_warp_limit: (%5f %5f %5f %5f %5f %5f)", ay, by, cy, ax, bx, cx);
  printf( "\n(%5f %5f: %d %d)", y_off, x_off, ydim, xdim);
*/
  
/*  memset(res, 0, xdim*ydim*sizeof(float)); */

  if(rxstart<0) rxstart=0;  if(rxend>rxdim) rxend=rxdim-1;
  if(rystart<0) rystart=0;  if(ryend>rydim) ryend=rydim-1;

  printf("bicubic_affine_warp_limit_weight: xdim=%d, rydim=%d: rxstart=%d,rystart=%d,rxend=%d,ryend=%d\n",
	 rxdim,rydim,rxstart,rystart,rxend,ryend);

  for(j=rystart; j<=ryend; j++)  /* j is the y var */
    for(i=rxstart,pos=j*rxdim+i; i<=rxend; i++,pos++)  /* i is the x var */
	{

	xfrac = i - x_diff - (ax + bx * (i - rx_off) + cx * (j - ry_off));
	xbase = (int)(xfrac);
	yfrac = j - y_diff - ( ay + by * (i - rx_off) + cy * (j - ry_off));
	ybase = (int)(yfrac);

	if((((xbase) >= 0) && ((xbase) < xdim)) &&
	   (((ybase) >= 0) && ((ybase) < ydim)))
	  {
	    xx = abs_val(xfrac-roundoff(xfrac));
	    yy = abs_val(yfrac-roundoff(yfrac));
	    weight_val =  (float)1/((xx*xx + yy*yy) + 0.1);

	    res_val = 0.0;
	    /* Compute X fractional position and base index */
	    frac = xfrac - xbase;
	    frac2 = frac*frac;
	    frac3 = frac*frac2;
	    base  = xbase - 1;
	    im_ptr = im + base;
	    
	    x_inbounds_1 = ((base >= 0) && (base < xdim))?1:0;
	    if (x_inbounds_1) x_coeff_1 = (frac2/2.0 - frac/3.0 - frac3/6.0);
	    base++;
	    x_inbounds0  = ((base >= 0) && (base < xdim))?1:0;
	    if (x_inbounds0)  x_coeff0  = (1.0 - frac/2.0 - frac2 + frac3/2.0);
	    base++;
	    x_inbounds1  = ((base >= 0) && (base < xdim))?1:0;
	    if (x_inbounds1)  x_coeff1 = (frac + frac2/2.0 - frac3/2.0);
	    base++;
	    x_inbounds2  = ((base >= 0) && (base < xdim))?1:0;
	    if (x_inbounds2)  x_coeff2 = (frac3/6.0 - frac/6.0);
	    
	    /* Compute Y fractional position and base index */
	    frac = yfrac - ybase;
	    frac2 = frac*frac;
	    frac3 = frac*frac2;
	    base = ybase - 1 ;
	    im_ptr += base*xdim;
	    
	    if ((base >= 0) && (base < ydim)) { /* y_inbounds_1 */
	      temp = 0.0;
	      if (x_inbounds_1) temp += x_coeff_1 * *im_ptr; im_ptr++;
	      if (x_inbounds0)  temp += x_coeff0  * *im_ptr; im_ptr++;
	      if (x_inbounds1)  temp += x_coeff1  * *im_ptr; im_ptr++;
	      if (x_inbounds2)  temp += x_coeff2  * *im_ptr; im_ptr -= 3;
	      res_val += temp * (frac2/2.0 - frac/3.0 - frac3/6.0);
	    }
	    base++;  im_ptr += xdim;
	    if ((base >= 0) && (base < ydim)) { /* y_inbounds0 */
	      temp = 0.0;
	      if (x_inbounds_1) temp += x_coeff_1 * *im_ptr; im_ptr++;
	      if (x_inbounds0)  temp += x_coeff0  * *im_ptr; im_ptr++;
	      if (x_inbounds1)  temp += x_coeff1  * *im_ptr; im_ptr++;
	      if (x_inbounds2)  temp += x_coeff2  * *im_ptr; im_ptr -= 3;
	      res_val += temp * (1 - frac/2.0 - frac2 + frac3/2.0);
	      }
	    base++;  im_ptr += xdim; 
	    if ((base >= 0) && (base < ydim)) { /* y_inbounds1 */
	      temp = 0.0;
	      if (x_inbounds_1) temp += x_coeff_1 * *im_ptr; im_ptr++;
	      if (x_inbounds0)  temp += x_coeff0  * *im_ptr; im_ptr++;
	      if (x_inbounds1)  temp += x_coeff1  * *im_ptr; im_ptr++;
	      if (x_inbounds2)  temp += x_coeff2  * *im_ptr; im_ptr -= 3;
	      res_val += temp * (frac + frac2/2.0 - frac3/2.0);
	    }
	    base++;  im_ptr += xdim;
	    if ((base >= 0) && (base < ydim)) { /* y_inbounds2 */
	      temp = 0.0;
	      if (x_inbounds_1) temp += x_coeff_1 * *im_ptr; im_ptr++;
	      if (x_inbounds0)  temp += x_coeff0  * *im_ptr; im_ptr++;
	      if (x_inbounds1)  temp += x_coeff1  * *im_ptr; im_ptr++;
	      if (x_inbounds2)  temp += x_coeff2  * *im_ptr; im_ptr -= 3;
	      res_val += temp * (frac3/6.0 - frac/6.0);
	      }
	    res[pos] += weight_val*(float)res_val;
	    weight[pos] += weight_val;
	  }
      }
  fflush(stdout);
}



/* affine warp only regions within: rxstart,rystart,rxend,ryend (not include end).
 * also check for input crop region. 
 * code is same as bicubic_affine_warp except for the checking in the result loop */
/* returns a weighted square distance to the nearest pixel */
/* adds to resulting image */
void bicubic_affine_warp_limit2_weight(im, res, ax, bx, cx, ay, by, cy,
				       ydim, xdim,xstart,ystart,xend,yend,
				       rydim,rxdim,rxstart,rystart,rxend,ryend,
				       weight)
  float *im, *res;
  float ax, bx, cx, ay, by, cy;
  int xdim, ydim;
  int rxdim,rydim;
  int rxstart,rystart,rxend,ryend;
  int xstart,ystart,xend,yend;
  float *weight;

{
  register float *im_ptr;	/* accessed 38x in loop */
  register int base;		/* accessed 30x in loop */
  register float frac, temp;	/* accessed 20x in loop */
  register float frac2, frac3;  /* accessed 10x in loop */
  int x_inbounds_1, x_inbounds0, x_inbounds1, x_inbounds2; /* 6x in loop */
  register float x_coeff_1, x_coeff0, x_coeff1, x_coeff2; /* 5x in loop */
  register float res_val;
  register int i,j;
  register float xfrac,xbase,yfrac,ybase;
  float x_off, y_off;
  float rx_off,ry_off,x_diff,y_diff;
  int pos;
  float xx,yy,weight_val;

  x_off = (xdim - 1) / 2.0; rx_off = (rxdim - 1) / 2.0;
  y_off = (ydim - 1) / 2.0; ry_off = (rydim - 1) / 2.0;
  x_diff = rx_off - x_off; y_diff = ry_off - y_off;

  /*
  **  wx = ax + bx * x + cx * y;
  **  wy = ay + by * x + cy * y;
  */

  printf ("\nbicubic_affine_warp_limit2_weight: (%5f %5f %5f %5f %5f %5f)", ay, by, cy, ax, bx, cx);
  printf( "\n(%5f %5f: %d %d)", y_off, x_off, ydim, xdim);

  
/*  memset(res, 0, xdim*ydim*sizeof(float)); */

  if(rxstart<0) rxstart=0;  if(rxend>rxdim) rxend=rxdim-1;
  if(rystart<0) rystart=0;  if(ryend>rydim) ryend=rydim-1;

  if(xstart<0) xstart=0;  if(xend>xdim) xend=xdim-1;
  if(ystart<0) ystart=0;  if(yend>ydim) yend=ydim-1;

  printf("bicubic_affine_warp_limit2_weight: xdim=%d, rydim=%d: xs=%d ys=%d xe=%d ye=%d\n", xdim,ydim,xstart,ystart,xend,yend);
  printf("bicubic_affine_warp_limit2_weight: xdim=%d, rydim=%d: rxs=%d rys=%d txe=%d rye=%d\n", rxdim,rydim,rxstart,rystart,rxend,ryend);


  for(j=rystart; j<rydim; j++)  /* j is the y var */
    for(i=rxstart,pos=j*rxdim+i; i<rxdim; i++,pos++) /* i is the x var */
      {
	xfrac = i - x_diff - (ax + bx * (i - rx_off) + cx * (j - ry_off));
	xbase = (int)(xfrac);
	yfrac = j - y_diff - (ay + by * (i - rx_off) + cy * (j - ry_off));
	ybase = (int)(yfrac);

	if((((xbase-1) >= xstart) || ((xbase + 3) >= xstart)) &&
	   (((xbase-1) < xend) || ((xbase + 3) < xend)) &&
	   (((ybase-1) >= ystart) || ((ybase + 3) >= ystart)) &&
	   (((ybase-1) < yend) || ((ybase + 3) < yend))
	   )
	  {
	    xx = xfrac-roundoff(xfrac); xx=abs_val(xx);
	    yy = yfrac-roundoff(yfrac); yy=abs_val(yy);
	    weight_val =  (float)1/((xx*xx + yy*yy) + 0.1);

	    res_val = 0.0;
	    /* Compute X fractional position and base index */
	    frac = xfrac - xbase;
	    frac2 = frac*frac;
	    frac3 = frac*frac2;
	    base  = xbase - 1;
	    im_ptr = im + base;
	    
	    x_inbounds_1 = ((base >= 0) && (base < xdim))?1:0;
	    if (x_inbounds_1) x_coeff_1 = (frac2/2.0 - frac/3.0 - frac3/6.0);
	    base++;
	    x_inbounds0  = ((base >= 0) && (base < xdim))?1:0;
	    if (x_inbounds0)  x_coeff0  = (1.0 - frac/2.0 - frac2 + frac3/2.0);
	    base++;
	    x_inbounds1  = ((base >= 0) && (base < xdim))?1:0;
	    if (x_inbounds1)  x_coeff1 = (frac + frac2/2.0 - frac3/2.0);
	    base++;
	    x_inbounds2  = ((base >= 0) && (base < xdim))?1:0;
	    if (x_inbounds2)  x_coeff2 = (frac3/6.0 - frac/6.0);
	    
	    /* Compute Y fractional position and base index */
	    frac = yfrac - ybase;
	    frac2 = frac*frac;
	    frac3 = frac*frac2;
	    base = ybase - 1 ;
	    im_ptr += base*xdim;
	    
	    if ((base >= 0) && (base < ydim)) { /* y_inbounds_1 */
	      temp = 0.0;
	      if (x_inbounds_1) temp += x_coeff_1 * *im_ptr; im_ptr++;
	      if (x_inbounds0)  temp += x_coeff0  * *im_ptr; im_ptr++;
	      if (x_inbounds1)  temp += x_coeff1  * *im_ptr; im_ptr++;
	      if (x_inbounds2)  temp += x_coeff2  * *im_ptr; im_ptr -= 3;
	      res_val += temp * (frac2/2.0 - frac/3.0 - frac3/6.0);
	    }
	    base++;  im_ptr += xdim;
	    if ((base >= 0) && (base < ydim)) { /* y_inbounds0 */
	      temp = 0.0;
	      if (x_inbounds_1) temp += x_coeff_1 * *im_ptr; im_ptr++;
	      if (x_inbounds0)  temp += x_coeff0  * *im_ptr; im_ptr++;
	      if (x_inbounds1)  temp += x_coeff1  * *im_ptr; im_ptr++;
	      if (x_inbounds2)  temp += x_coeff2  * *im_ptr; im_ptr -= 3;
	      res_val += temp * (1 - frac/2.0 - frac2 + frac3/2.0);
	    }
	    base++;  im_ptr += xdim; 
	    if ((base >= 0) && (base < ydim)) { /* y_inbounds1 */
	      temp = 0.0;
	      if (x_inbounds_1) temp += x_coeff_1 * *im_ptr; im_ptr++;
	      if (x_inbounds0)  temp += x_coeff0  * *im_ptr; im_ptr++;
	      if (x_inbounds1)  temp += x_coeff1  * *im_ptr; im_ptr++;
	      if (x_inbounds2)  temp += x_coeff2  * *im_ptr; im_ptr -= 3;
	      res_val += temp * (frac + frac2/2.0 - frac3/2.0);
	    }
	    base++;  im_ptr += xdim;
	    if ((base >= 0) && (base < ydim)) { /* y_inbounds2 */
	      temp = 0.0;
	      if (x_inbounds_1) temp += x_coeff_1 * *im_ptr; im_ptr++;
	      if (x_inbounds0)  temp += x_coeff0  * *im_ptr; im_ptr++;
	      if (x_inbounds1)  temp += x_coeff1  * *im_ptr; im_ptr++;
	      if (x_inbounds2)  temp += x_coeff2  * *im_ptr; im_ptr -= 3;
	      res_val += temp * (frac3/6.0 - frac/6.0);
	    }
	    res[pos] += (float)(weight_val*res_val);
	    weight[pos] += (float)weight_val;
	  }
      }
  fflush(stdout);
}

/*****************************************************************/
/* 
FROM_WARP warps array A into B according to inverse warp map [wx,wy].
The warp equation is

  B[i][j]  <-  A[i - wx[i][j]][j - wy[i][j]]

where we use linear interpolation to determine between-pixel values;
should be augmented to allow a user-selected filter mask instead of
linear interpolation...  References to pixels outside of A are
considered to be zero.
*/
/* warps an image with warp maps wx and wy using bicubic interpolation */
void bicubic_warp(im, res, wy, wx, ydim, xdim)
  float *im, *res;
  float *wy, *wx;
  int xdim, ydim;

{
  register float *im_ptr;	/* accessed 38x in loop */
  register int base;		/* accessed 30x in loop */
  register float frac, temp;	/* accessed 20x in loop */
  register float frac2, frac3;  /* accessed 10x in loop */
  int x_inbounds_1, x_inbounds0, x_inbounds1, x_inbounds2; /* 6x in loop */
  register float x_coeff_1, x_coeff0, x_coeff1, x_coeff2; /* 5x in loop */
  register float res_val;
  register int i,j;
  register float xfrac,xbase,yfrac,ybase;

/*  printf("bicubic_warp: (%d %d)\n",ydim,xdim);fflush(stdout); */
/*  memset(res, 0, xdim*ydim*sizeof(float)); */
  for(j=0; j<ydim; j++)  /* j is the y var */
    for(i=0; i<xdim; i++)  /* i is the x var */
	{
	xfrac = i - *wx++;
	xbase = (int)(xfrac);
	yfrac = j - *wy++;
	ybase = (int)(yfrac);

	if((((xbase) >= 0) && ((xbase + 0) < xdim)) &&
	   (((ybase) >= 0) && ((ybase + 0) < ydim)))
	  {
	    res_val = 0.0;

	    /* Compute X fractional position and base index */
	    frac = xfrac - xbase;
	    frac2 = frac*frac;
	    frac3 = frac*frac2;
	    base  = xbase - 1;
	    im_ptr = im + base;
	    
	    x_inbounds_1 = ((base >= 0) && (base < xdim))?1:0;
	    if (x_inbounds_1) x_coeff_1 = (frac2/2.0 - frac/3.0 - frac3/6.0);
	    base++;
	    x_inbounds0  = ((base >= 0) && (base < xdim))?1:0;
	    if (x_inbounds0)  x_coeff0  = (1.0 - frac/2.0 - frac2 + frac3/2.0);
	    base++;
	    x_inbounds1  = ((base >= 0) && (base < xdim))?1:0;
	    if (x_inbounds1)  x_coeff1 = (frac + frac2/2.0 - frac3/2.0);
	    base++;
	    x_inbounds2  = ((base >= 0) && (base < xdim))?1:0;
	    if (x_inbounds2)  x_coeff2 = (frac3/6.0 - frac/6.0);
	    
	    /* Compute Y fractional position and base index */
	    frac = yfrac - ybase;
	    frac2 = frac*frac;
	    frac3 = frac*frac2;
	    base = ybase - 1 ;
	    im_ptr += base*xdim;
	    
	    if ((base >= 0) && (base < ydim)) { /* y_inbounds_1 */
	      temp = 0.0;
	      if (x_inbounds_1) temp += x_coeff_1 * *im_ptr; im_ptr++;
	      if (x_inbounds0)  temp += x_coeff0  * *im_ptr; im_ptr++;
	      if (x_inbounds1)  temp += x_coeff1  * *im_ptr; im_ptr++;
	      if (x_inbounds2)  temp += x_coeff2  * *im_ptr; im_ptr -= 3;
	      res_val += temp * (frac2/2.0 - frac/3.0 - frac3/6.0);
	    }
	    base++;  im_ptr += xdim;
	    if ((base >= 0) && (base < ydim)) { /* y_inbounds0 */
	      temp = 0.0;
	      if (x_inbounds_1) temp += x_coeff_1 * *im_ptr; im_ptr++;
	      if (x_inbounds0)  temp += x_coeff0  * *im_ptr; im_ptr++;
	      if (x_inbounds1)  temp += x_coeff1  * *im_ptr; im_ptr++;
	      if (x_inbounds2)  temp += x_coeff2  * *im_ptr; im_ptr -= 3;
	      res_val += temp * (1 - frac/2.0 - frac2 + frac3/2.0);
	    }
	    base++;  im_ptr += xdim; 
	    if ((base >= 0) && (base < ydim)) { /* y_inbounds1 */
	      temp = 0.0;
	      if (x_inbounds_1) temp += x_coeff_1 * *im_ptr; im_ptr++;
	      if (x_inbounds0)  temp += x_coeff0  * *im_ptr; im_ptr++;
	      if (x_inbounds1)  temp += x_coeff1  * *im_ptr; im_ptr++;
	      if (x_inbounds2)  temp += x_coeff2  * *im_ptr; im_ptr -= 3;
	      res_val += temp * (frac + frac2/2.0 - frac3/2.0);
	    }
	    base++;  im_ptr += xdim;
	    if ((base >= 0) && (base < ydim)) { /* y_inbounds2 */
	      temp = 0.0;
	      if (x_inbounds_1) temp += x_coeff_1 * *im_ptr; im_ptr++;
	      if (x_inbounds0)  temp += x_coeff0  * *im_ptr; im_ptr++;
	      if (x_inbounds1)  temp += x_coeff1  * *im_ptr; im_ptr++;
	      if (x_inbounds2)  temp += x_coeff2  * *im_ptr; im_ptr -= 3;
	      res_val += temp * (frac3/6.0 - frac/6.0);
	    }
	    *res = (float)res_val;
	  } 
	else{
	  *res = 0.0; 
/*	  *res = im[j*xdim + i];  */
	}
	res++;
      }
  fflush(stdout);
}
