/******************************************************************

"Copyright  1994 by the Massachusetts  Institute  of  Technology.  All
rights reserved.

Developed by John Y. A. Wang at the  Media Laboratory, MIT, Cambridge,
Massachusetts, with  support  from Secom Ltd., Goldstar Ltd.,  and the
Television of Tomorrow Program.

For use by authorized users only.

This distribution is approved  by Nicholas Negroponte, Director of the
Media Laboratory, MIT.

Permission to use, copy, or modify this software and its documentation
for educational and research purposes only and without  fee is  hereby
granted, provided that this copyright notice and the original authors'
names  appear  on  all  copies  and   supporting   documentation.   If
individual  files  are  separated  from  this  distribution  directory
structure, this copyright notice must be included.  For any other uses
of  this software, in  original or  modified form, including  but  not
limited to distribution in whole or in part, specific prior permission
must be  obtained  from  MIT.   These  programs  shall  not  be  used,
rewritten,  or  adapted  as  the  basis  of a  commercial software  or
hardware  product without  first  obtaining  appropriate licenses from
MIT.  MIT.  makes no  representations about  the  suitability of  this
software for any purpose.  It is  provided "as is" without  express or
implied warranty.

******************************************************************/


/* 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;  File:  pinverse.c
;;;  Author: Simoncelli
;;;  Description: Matrix inversion
;;;  Creation Date:
;;;  ----------------------------------------------------------------
;;;    Object-Based Vision and Image Understanding System (OBVIUS),
;;;      Copyright 1988, Vision Science Group,  Media Laboratory,  
;;;              Massachusetts Institute of Technology.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
*/

/* -- Uses routines by dead@media-lab in pseudo.c, lud.c, matrix-utils.c -- */

#include <stdio.h>
#include <math.h>
#include "matrix.h"

/* 
stuff for testing this code:

cc -o test -f68881 pinverse.c pseudo.c lud.c matrix-utils.c 

static float mat1[] =     
  {1.0, 2.0, 3.0,
  4.0, 5.0, 6.0,
     7.0, 8.0, 9.0};

static float mat2[] =     
  {1.0, 0.0, 0.0,
  0.0, 1.0, 0.0,
     0.0, 0.0, 1.0};
  
static float mat3[] =     
  {1.0, 0.0, 0.0,
  0.0, 1.0, 0.0,
     0.0, 0.0, 1.0};

main(argc, argv)
  int argc;
  char *argv[];
  {
  print_matrix (mat3, 3, 3);
  internal_pinverse (mat2, mat3, 3, 3);
  print_matrix (mat3, 3, 3);
  internal_pinverse (mat1, mat3, 3, 3);
  print_matrix (mat3, 3, 3);
  }
  
print_matrix(mat,n,m)
  float mat[];
  int n,m;
  {
  int row, col, row_lin;

  for (row=0;  row<=n;  row++)
      {
      printf("\n");
      for (col=0; col<m; col++)
	printf("%6.1f  ", (float) mat[row*n+col]);
      }
  }
*/

/*-- Finds the pseudo-inverse of mat and puts it into imat --*/
/* -- n is number of rows, m is number of columns in mat -- */
internal_pinverse(mat,imat,n,m)
  float *mat, *imat;
  int n,m;
  {
  int row, col, row_lin;
  MatrixType smat;   
  MatrixType  ismat;

  if (n>MAX_RANK||m>MAX_RANK)
      {
      printf("Error: matrix too big to invert!");
      return(0);
      }

  MatrixType_Set_N_and_M( &smat, n, m );
  MatrixType_Set_N_and_M( &ismat, m, n ); 
  
  /* copy mat to smat */
  for (row=row_lin=0;  row<n;  row++,row_lin+=m)
    for (col=0; col<m; col++)
      (smat.data)[row][col] = (RobotType) mat[row_lin+col];

  PseudoInverse(&smat, &ismat); 

  /* copy ismat to imat */
  for (row=row_lin=0; row<m; row++,row_lin+=n)
    for (col=0; col<n; col++)
      imat[row_lin+col] = (float) (ismat.data)[row][col]; 
}


/* Local Variables: */
/* buffer-read-only: t */
/* End: */
