/******************************************************************

"Copyright  1994 by the Massachusetts  Institute  of  Technology.  All
rights reserved.

Developed by John Y. A. Wang at the  Media Laboratory, MIT, Cambridge,
Massachusetts, with  support  from Secom Ltd., Goldstar Ltd.,  and the
Television of Tomorrow Program.

For use by authorized users only.

This distribution is approved  by Nicholas Negroponte, Director of the
Media Laboratory, MIT.

Permission to use, copy, or modify this software and its documentation
for educational and research purposes only and without  fee is  hereby
granted, provided that this copyright notice and the original authors'
names  appear  on  all  copies  and   supporting   documentation.   If
individual  files  are  separated  from  this  distribution  directory
structure, this copyright notice must be included.  For any other uses
of  this software, in  original or  modified form, including  but  not
limited to distribution in whole or in part, specific prior permission
must be  obtained  from  MIT.   These  programs  shall  not  be  used,
rewritten,  or  adapted  as  the  basis  of a  commercial software  or
hardware  product without  first  obtaining  appropriate licenses from
MIT.  MIT.  makes no  representations about  the  suitability of  this
software for any purpose.  It is  provided "as is" without  express or
implied warranty.

******************************************************************/


/*
*	p s e u d o . c			dead@media-lab		87/03/04
*
*		General purpose pseudo inverse functions.
*/

#ifndef lint
static char copyright[] =
"Copyright (c) 1987 by David T. Chen.  All rights reserved.";
#endif

#include <stdio.h>
#include <math.h>
#include "matrix.h"

/*
==========================================================================
PseudoInverse() -- Returns PsuedoInverse of a matrix.

==========================================================================
*/
PseudoInverse( J, Jinv )
MatrixType *J, *Jinv;
{
int no_rows = MatrixType_Get_N(J);
int no_cols = MatrixType_Get_M(J);
int flag;
MatrixType Jtranspose;
MatrixType JJ;
MatrixType JJinv;

	if (no_rows == no_cols)			/* Square Matrix */
	{	flag = Invert( J, Jinv );
	}
	else if (no_rows > no_cols)		/* Overspecified */
	{
		JtransposeTimesJ( J, &JJ );

		if ((flag = Invert(&JJ, &JJinv)) != SINGULAR)
		{
			MatrixType_Transpose(J, &Jtranspose);
			MatrixType_Multiply( &JJinv, &Jtranspose, Jinv );
		}
	}
	else							/* Underspecified */
	{	JtimesJtranspose( J, &JJ );

		if ((flag = Invert(&JJ, &JJinv)) != SINGULAR)
		{
			MatrixType_Transpose(J, &Jtranspose);
			MatrixType_Multiply( &Jtranspose, &JJinv, Jinv );
		}
	}
	return( flag );
}
/* PseudoInverse */



/*
==========================================================================
PseudoInverseNoSecondaryGoals() -- Singularity status is returned

	Finds `result' such that:

	J*result - Xdot  has the minimum norm.

==========================================================================
*/
PseudoInverseNoSecondaryGoals( Xdot, J, result)

VectorType *Xdot;
MatrixType *J;
VectorType *result;
{
int no_rows = MatrixType_Get_N(J);
int no_cols = MatrixType_Get_M(J);
int flag;
MatrixType JJ;
VectorType B;

	if (no_rows == no_cols)
	{
		/* Square Matrix */
		flag = LUdecomposition( Xdot, J, result );
	}
	else if (no_rows > no_cols)
	{
		/* Overspecified */
		JtransposeTimesJ( J, &JJ );

		if (JtransposeTimesVector( J, Xdot, &B ) < 0)
			return( SINGULAR );

		flag = LUdecomposition( &B, &JJ, result );
	}
	else
	{	/* Underspecified */
		JtimesJtranspose( J, &JJ );

		if ((flag = LUdecomposition( Xdot, &JJ, &B )) != SINGULAR)
			if (JtransposeTimesVector( J, &B, result ) < 0)
				return( SINGULAR );
	}
	return( flag );
}
/* PseudoInverseNoSecondaryGoals */



/*
==========================================================================
PseudoInverseSecondaryGoals() -- Singularity status is returned

==========================================================================
*/
PseudoInverseSecondaryGoals( Xdot, J, z, result )

VectorType *Xdot;
MatrixType *J;
VectorType *z;
VectorType *result;
{
int no_rows = MatrixType_Get_N(J);
int no_cols = MatrixType_Get_M(J);
int flag;
MatrixType JJ;
VectorType XJZ;
VectorType B;
VectorType C;

	if (no_rows == no_cols)
	{
		/* Square Matrix */
		fprintf(stderr, "PseudoInverseSecondaryGoals: `J' is square, ignoring secondary goal.\n");

		flag = LUdecomposition( Xdot, &JJ, result );
	}
	else if (no_rows > no_cols)
	{
		/* Overspecified */
		JtransposeTimesJ( J, &JJ );

		if (XdotMinusJtimesZ( Xdot, J, z, &XJZ ) < 0)
			return( SINGULAR );

		if (JtransposeTimesVector( J, &XJZ, &B ) < 0)
			return( SINGULAR );

		if ((flag = LUdecomposition( &B, &JJ, &C )) != SINGULAR)
			if (VectorType_Add( &C, z, result ) < 0)
				return( SINGULAR );
	}
	else
	{	/* Underspecified */
		JtimesJtranspose( J, &JJ );

		if (XdotMinusJtimesZ( Xdot, J, z, &XJZ ) < 0)
			return( SINGULAR );

		if ((flag = LUdecomposition( &XJZ, &JJ, &B )) != SINGULAR)
		{
			if (JtransposeTimesVector( J, &B, &C ) < 0)
				return( SINGULAR );

			if (VectorType_Add( &C, z, result ) < 0)
				return( SINGULAR );
		}
	}
	return( flag );
}
/* PseudoInverseSecondaryGoals */


/*
==========================================================================
JtimesJtranspose()

==========================================================================
*/
static JtimesJtranspose( J_matrix, JJ_matrix )
MatrixType *J_matrix;
MatrixType *JJ_matrix;
{
register i, j, k;
RobotType tmp;
register MatrixData J  = MatrixType_Get_Data(J_matrix);
register MatrixData JJ = MatrixType_Get_Data(JJ_matrix);
int no_rows = MatrixType_Get_N(J_matrix);
int no_cols = MatrixType_Get_M(J_matrix);

	MatrixType_Set_N_and_M( JJ_matrix, no_rows, no_rows );

	for (i=0; i<no_rows; i++)
		for (j=0; j<=i; j++)
		{
			tmp = 0.0;
			for (k=0; k<no_cols; k++)
				tmp += J[i][k] * J[j][k];
			JJ[i][j] = tmp;
		}

	for (i=0; i<no_rows; i++)
		for (j=i+1; j<no_rows; j++)
			JJ[i][j] = JJ[j][i];
}
/* JtimesJtranspose */


/*
==========================================================================
JtransposeTimesJ()

==========================================================================
*/
static JtransposeTimesJ( J_matrix, JJ_matrix )
MatrixType *J_matrix;
MatrixType *JJ_matrix;
{
register i, j, k;
RobotType tmp;
register MatrixData J  = MatrixType_Get_Data(J_matrix);
register MatrixData JJ = MatrixType_Get_Data(JJ_matrix);
int no_rows = MatrixType_Get_N(J_matrix);
int no_cols = MatrixType_Get_M(J_matrix);

	MatrixType_Set_N_and_M( JJ_matrix, no_cols, no_cols );

	for (i=0; i<no_cols; i++)
		for (j=0; j<=i; j++)
		{
			tmp = 0.0;
			for (k=0; k<no_rows; k++)
				tmp += J[k][i] * J[k][j];
			JJ[i][j] = tmp;
		}

	for (i=0; i<no_cols; i++)
		for (j=i+1; j<no_cols; j++)
			JJ[i][j] = JJ[j][i];
}
/* JtransposeTimesJ */


/*
==========================================================================
XdotMinusJtimesZ()

==========================================================================
*/
XdotMinusJtimesZ( Xdot_vector, J_matrix, Z_vector, XJZ_vector )

VectorType *Xdot_vector;
MatrixType *J_matrix;
VectorType *Z_vector;
VectorType *XJZ_vector;
{
int i, j;
int no_rows = MatrixType_Get_N(J_matrix);
int no_cols = MatrixType_Get_M(J_matrix);

register VectorData Xdot = VectorType_Get_Data(Xdot_vector);
register MatrixData J = MatrixType_Get_Data(J_matrix);
register VectorData z = VectorType_Get_Data(Z_vector);
register VectorData XJZ = VectorType_Get_Data(XJZ_vector);
RobotType tmp;

	if (no_cols != VectorType_Get_N(Z_vector))
	{	fprintf(stderr, "XdotMinusJtimesZ: rank mis-match with `z'.\n");
		return(-1);
	}
	if (no_cols != VectorType_Get_N(Xdot_vector))
	{	fprintf(stderr, "XdotMinusJtimesZ: rank mis-match with `Xdot'.\n");
		return(-1);
	}
	VectorType_Set_N(XJZ_vector, no_cols);

	for (i=0; i<no_rows; i++)
	{
		tmp = Xdot[i];

		for (j=0; j<no_cols; j++)
			tmp -= J[i][j] * z[j];

		XJZ[i] = tmp;
	}
	return(0);
}
/* XdotMinusJtimesZ */



/*
==========================================================================
JtransposeTimesVector()

==========================================================================
*/
static JtransposeTimesVector( J_matrix, B_vector, result_vector )
MatrixType *J_matrix;
VectorType *B_vector;
VectorType *result_vector;
{
register i, j;
RobotType tmp;
register MatrixData J = MatrixType_Get_Data(J_matrix);
register VectorData B = VectorType_Get_Data(B_vector);
register VectorData result = VectorType_Get_Data(result_vector);
int no_rows = MatrixType_Get_N(J_matrix);
int no_cols = MatrixType_Get_M(J_matrix);

	if (no_rows != VectorType_Get_N(B_vector))
	{	fprintf(stderr, "JtransposeTimesVector: rank mis-match.\n");
		return(-1);
	}
	VectorType_Set_N( result_vector, no_cols );

	for (i=0; i<no_cols; i++)
	{
		tmp = 0.0;
		for (j=0; j<no_rows; j++)
			tmp += J[j][i] * B[j];
		result[i] = tmp;
	}
	return(0);
}
/* JtransposeTimesVector */
