/******************************************************************

"Copyright  1994 by the Massachusetts  Institute  of  Technology.  All
rights reserved.

Developed by John Y. A. Wang at the  Media Laboratory, MIT, Cambridge,
Massachusetts, with  support  from Secom Ltd., Goldstar Ltd.,  and the
Television of Tomorrow Program.

For use by authorized users only.

This distribution is approved  by Nicholas Negroponte, Director of the
Media Laboratory, MIT.

Permission to use, copy, or modify this software and its documentation
for educational and research purposes only and without  fee is  hereby
granted, provided that this copyright notice and the original authors'
names  appear  on  all  copies  and   supporting   documentation.   If
individual  files  are  separated  from  this  distribution  directory
structure, this copyright notice must be included.  For any other uses
of  this software, in  original or  modified form, including  but  not
limited to distribution in whole or in part, specific prior permission
must be  obtained  from  MIT.   These  programs  shall  not  be  used,
rewritten,  or  adapted  as  the  basis  of a  commercial software  or
hardware  product without  first  obtaining  appropriate licenses from
MIT.  MIT.  makes no  representations about  the  suitability of  this
software for any purpose.  It is  provided "as is" without  express or
implied warranty.

******************************************************************/


/***************************************************************
*** File: pyramid.c
*** Author: John Y. A. Wang
*** Date: November 18, 1992.
*** Copyright 1992, Institute of Technology, All Rights Reserved.
***
*** Description: pyramid code in C.
***************************************************************/

static  char *copyright="\
\n\n***********************************************\n\
File: pyramid.c\n\
Author: John Y. A. Wang.\n\
The MIT Media Lab.\n\
Copyright (c) 1992, All Rights Reserved.\n\
***********************************************\n\n";

#include <stdio.h>
#include "pyramid.h"
#include "motion.h"

void apply_sep_filter_subsample(/*image, result, temp_im, xdim, ydim,
			       xfilt, yfilt, xfdim, yfdim,xfilt_space, yfilt_space*/);
void make_pyramid(/*im, level, xdim, ydim, pyr, pyrmemory, temp, make, filter, fdim*/);


/**********************************************************************/
/* base image points to the input image. Other levels of the pyramid are
** assigned to segments in the pyrmemory.
** pyrmemory and temp should be at least the size of im.
** Actually does the filtering if make==1, otherwise just set pointers. */
void make_pyramid(im, level, xdim, ydim, pyr, pyrmemory, temp, make, filter, fdim)
  float *im, *filter, *temp;
  int level, xdim, ydim, fdim;
  Pyramid *pyr;
  float *pyrmemory;
  int make;    /* 0: just setup, 1: setup and filter */
{
  
  int xsize, ysize,i;
  float *ptr;

  if(level>PYRMAXLEVEL) {
    printf("make_pyramid: pyramid level %d currently not supported. max. level=%d\n",
	   level, PYRMAXLEVEL);fflush(stdout);
    return;
  }
  /* initialize the pyramid structure */
  pyr->level[0].xdim = xdim;
  pyr->level[0].ydim = ydim;
  pyr->level[0].array = im;

  for(ptr=pyrmemory, i=1; i<=level; i++) {
    pyr->level[i].xdim = xsize = (int) (((float)pyr->level[i-1].xdim / 2) +.5);
    pyr->level[i].ydim = ysize = (int) (((float)pyr->level[i-1].ydim / 2) +.5);
    pyr->level[i].array = ptr;
    if(make==1)
      apply_sep_filter_subsample(pyr->level[i-1].array, pyr->level[i].array, temp,
				 pyr->level[i-1].xdim, pyr->level[i-1].ydim,
				 filter, filter,fdim,fdim,2,2);
    ptr=ptr+ysize*xsize;
  }
}

/**********************************************************************/
/* convolves and subsamples the image */

void apply_sep_filter_subsample(image, result, temp_im, xdim, ydim,
			       xfilt, yfilt, xfdim, yfdim,xfilt_space, yfilt_space)
  float *image, *result, *temp_im;
  register int xfdim;
  int xdim;
  register float *xfilt, *yfilt; 
  int ydim;
  register int yfdim;
  int xfilt_space, yfilt_space;

{
  register double sum;
  int y_fmid = yfdim/2;
  int x_fmid = xfdim/2;
  int x,y,x2,y2,y_pos2;
  int xdim2 = (int) ((float)(xdim)/xfilt_space+.5);
  int ydim2 = (int) ((float)(ydim)/yfilt_space+.5);
  register int x_pos, y_pos, xf, yf;
  register int yoff, xoff;
  register float *ximage, *yimage, *xtemp, *ytemp; 

  if(xfdim==0 && yfdim == 0) /* don't convolve at all */
    return;
  else
    if(yfdim == 0)  {  /* convolve x only */
      ximage = image;
      xtemp = result;
    }
    else
      if(xfdim==0) {   /* convolve y only */
	yimage = image;
	ytemp = result;
      }
      else {
	yimage = image;
	ytemp = temp_im;     /* convolve both x and y */
	ximage = temp_im;
	xtemp = result;
      }

/*  printf("\nxdim=%d,ydim=%d, xdim2=%d, ydim2=%d",xdim,ydim,xdim2,ydim2);fflush(stdout); */

  /******** convolve in the y direction with the yfilter ********/
  if(yfdim != 0)  /* convolve y */
    for(y=0, y2=0; y < ydim; y+=yfilt_space,y2++) {
      yoff = y - y_fmid;
      y_pos2 = y2*xdim;
	for(x=0; x < xdim; x++) {
	  sum=0.0;
	  for(yf=0; yf < yfdim; yf++) {
	    y_pos = yoff + yf;
	    
	    if(y_pos < 0)
	      y_pos = -y_pos;
/*	      y_pos += ydim; */
	    else
	      if(y_pos >= ydim)
		y_pos = ydim - (y_pos - ydim + 2); 
/*		y_pos = y_pos - ydim; */
	    
	    sum+= yimage[y_pos*xdim + x]*yfilt[yf];
	  }     /* end yf */
	  ytemp[y_pos2 + x] = sum;
	}       /* end x */
    }         /* end y */

  /******** convolve in the x direction with the xfilter ********/
  if(xfdim != 0)  /* convolve x */
    for(y=0; y < ydim2; y++) {
      y_pos = y*xdim;
      y_pos2 = y*xdim2;
      for(x=0,x2=0; x < xdim; x+=xfilt_space, x2++) {
	xoff = x - x_fmid;
	sum=0.0;
	for(xf=0; xf < xfdim; xf++) {
	  x_pos = xoff + xf;
	  if(x_pos < 0)
	    x_pos = -x_pos;
/*	    x_pos += xdim; */
	  else
	    if(x_pos >= xdim)
	      x_pos = xdim - (x_pos - xdim + 2); 
/*	      x_pos = x_pos - xdim; */
	  sum+= ximage[y_pos + x_pos]*xfilt[xf];
	}     /* end xf */
	xtemp[y_pos2 + x2] = sum;
      }       /* end x */
    }         /* end y */
}



/**************************************************************/
show_pyr(im,xdim,ydim, pyr0, pyr1, pyr2, pyr3, temp, temp1,level)
  float *im,*pyr0,*pyr1,*pyr2,*pyr3,*temp,*temp1;
  int xdim,ydim;
  int level;
{
  Pyramid pyr;
  int xsize, ysize,i;
  float *images[5];
  static float g_filt[5]={0.0625, 0.25, 0.375, 0.25, 0.0625};
  
  images[0]=pyr0; images[1]=pyr1;  images[2]=pyr2;  images[3]=pyr3;
  
  make_pyramid(im,level,xdim,ydim,&pyr,temp,temp1,1,g_filt,5);

  for(i=0; i<=3; i++) {
    xsize = pyr.level[i].xdim;
    ysize = pyr.level[i].ydim;
    printf("\nmake-pyr level %d: xsize=%d ysize=%d",i,xsize, ysize);fflush(stdout);
    memcpy((char *)images[i],(char *)pyr.level[i].array,xsize*ysize*sizeof(float)); 
  }
}
  
