/******************************************************************

"Copyright  1994 by the Massachusetts  Institute  of  Technology.  All
rights reserved.

Developed by John Y. A. Wang at the  Media Laboratory, MIT, Cambridge,
Massachusetts, with  support  from Secom Ltd., Goldstar Ltd.,  and the
Television of Tomorrow Program.

For use by authorized users only.

This distribution is approved  by Nicholas Negroponte, Director of the
Media Laboratory, MIT.

Permission to use, copy, or modify this software and its documentation
for educational and research purposes only and without  fee is  hereby
granted, provided that this copyright notice and the original authors'
names  appear  on  all  copies  and   supporting   documentation.   If
individual  files  are  separated  from  this  distribution  directory
structure, this copyright notice must be included.  For any other uses
of  this software, in  original or  modified form, including  but  not
limited to distribution in whole or in part, specific prior permission
must be  obtained  from  MIT.   These  programs  shall  not  be  used,
rewritten,  or  adapted  as  the  basis  of a  commercial software  or
hardware  product without  first  obtaining  appropriate licenses from
MIT.  MIT.  makes no  representations about  the  suitability of  this
software for any purpose.  It is  provided "as is" without  express or
implied warranty.

******************************************************************/


/***************************************************************
*** File: uniops.c
*** Author: John Y. A. Wang
*** Date: January 15, 1993
*** Copyright 1993, Massachusetts Institute of Technology, All Rights Reserved.
***
*** Description: uniary operations on images.
***************************************************************/

/*****************************************************************/
#include "dynamo.h"
/*****************************************************************/

/********** IMPORTANT:    **********************************/
/********** put copyright label in all the files ***********/

static  char *copyright="\
\n\n***********************************************\n\
Developed by John Y. A. Wang at the Media Laboratory, MIT, Cambridge,Massachusetts\n\
Copyright  1993 by the Massachusetts  Institute  of  Technology.  All rights reserved.\n\
***********************************************\n\n";

/*
  cc -O4 -fsingle -c uniops.c 
  SUN:  cc -O4 -fsingle -c uniops.c

  lint uniops.c
*/
/*****************************************************************/

int temporal_filter_sequence(image_mgr,type_str)
  Image_Manager *image_mgr;
  char *type_str;
{
  extern char *args;

  char *progn="temporal_filter_sequence";
  int start,end;
  int wsize;
  Image_Cache *cache=NULL,*cache_res=NULL;
  int length;
  
  char *image_name=image_mgr->image_name;
  char *yn=image_mgr->yn;
  char *save_filename=image_mgr->cbuf1;

#define MED 1
#define AVG 2  
  int type;
  int rc;
  
  FILE *fp=NULL;

  /*********************************/

  if(strncmp(type_str,"median",6)==0) {
    type=MED; }
  else if(strncmp(type_str,"average",7)==0) {
    type=AVG; }
  else {
    printf("%s: type=\"%s\" current not supported.",progn,type_str);
    return(-1);
  }

  printf("\nPerforming window-centered temporal %s filter on sequence...\n",
	  type_str);

  if((cache=ask_name_get_cache(image_mgr))==NULL) { return(0); }

  start=cache->ext_start; end=start;
  printf("Enter Start and End Frame : (%d %d) >> ", start,end);
  readline(args,ARGSIZE,stdin); sscanf(args,"%d %d",&start,&end);
  if(start>cache->tdim) start=cache->tdim;
  if(end>=cache->ext_start+cache->tdim) end=cache->ext_start+cache->tdim-1;

  length=end-start+1;

  wsize=1; 
  printf("Enter %s window size : (%d) >> ",type_str, wsize);
  readline(args,ARGSIZE,stdin);  sscanf(args,"%d",&wsize);
  
  if(wsize>cache->tdim) wsize=cache->tdim;
  
  sprintf(yn,"n");
  printf("\tDo you want to save results internally (y/n)? (%s): ",yn);
  readline(args,ARGSIZE,stdin);
  sscanf(args,"%s",yn);
  if(strncmp(yn,"y",1)==0) {
    sprintf(image_name,"%s-%s-w%d",cache->image_name,type_str,wsize);
    printf("Result Image name :\n\t");
    if((cache_res=make_internal_image(image_mgr,cache->xdim,cache->ydim,length))==NULL) {
      printf("Not saving results internally.\n");
/*      init_cache(cache_res); cache_res=NULL; */
    }
    cache_res->gamma=cache->gamma;
  }
 
  if(cache_res==cache) {
    printf("%s: Cannot put Result into Input \n",progn);  return(-1); }

  sprintf(yn,"n");
  printf("\tDo you want to save results to disk (y/n)? (%s): ",yn);
  readline(args,ARGSIZE,stdin); sscanf(args,"%s",yn);
  if(strncmp(yn,"y",1)==0) {
    printf("\n");
    sprintf(cache->savedirname,"t-%s-w%d",type_str,wsize);
    if(save_dir(cache->savedirname)==0) {
      sprintf(save_filename,"%s/data",cache->savedirname);
      if((fp=open_file(save_filename,"w"))==NULL) {
	printf("Not saving saving results to %s.\n",save_filename);
      }
    }
    sprintf(args,"Temporal %s filter with window width = %d. \n\t Window is centered to the current frame.\n\t start=%d end=%d \0",type_str,wsize,start,end);
    sprintf(save_filename,"%s/descriptor",cache->savedirname);
    save_descriptor(save_filename,cache->xdim,cache->ydim,
		    (int)(end-start+1),(int)1,(int)1,
		    (float)1.0,(float)0.0,args);
  }

  printf("Applying temporal %s with a window size of %d. \n\n",
	 type_str,wsize);

  switch(type) {
  case MED:
    rc=median_sequence(cache,cache_res,start,end,wsize,fp);
    close_file(fp); return(rc);
  case AVG:
    rc=average_sequence(cache,cache_res,start,end,wsize,fp);
    close_file(fp); return(rc);
  default:
    close_file(fp);
    return(-1);
  }
  return(0);
}




#define TMED_FILT_RET(rc) \
 return(clean_up_ret_func(malloc_images,nmalloc,NULL,0,display_pid,rc));


median_sequence(cache,cache_res,start,end,wsize,fp)
  Image_Cache *cache,*cache_res;
  int start,end,wsize;
  FILE *fp;
{
  extern char *args;

  float **im_addr;
  char *progn="median_sequence";

  int start_med,end_med;
  int nframes,i,frame,wsize2;
  float *res=NULL,*temp_res=NULL;
  int imsize=cache->imsize;

  float checkval=0, setval=0;
  
  int cache_start,cache_end;

  FILE *display_pid=NULL;
  char title[20];
  char *malloc_images[5];
  int nmalloc=0,num;

  /************* allocate memory ************************/

  if(wsize>cache->nframes) {
    printf("%s: Current cache size of %d is smaller than window size of %d\n",
	   progn, cache->nframes,wsize);
    printf("\tPlease increase cache size to %d before trying again.\n",wsize);
    return(-1);
  }

  if(cache_res==NULL) {
    if((temp_res=(float *)malloc(cache->imsize*sizeof(float)))==NULL) {
      fprintf(stderr,"Cannot allocate memory for temporal median filtering.\n");
      TMED_FILT_RET(-1);
    }
    malloc_images[nmalloc]=(char *)temp_res; nmalloc++;
  }
  
  if((im_addr=(float **)malloc(wsize*sizeof(float)))==NULL) {
    printf("%s: cannot allocate memory for im_addr\n",progn);
    TMED_FILT_RET(-1); }
  malloc_images[nmalloc]=(char *)im_addr; nmalloc++;
  
  wsize2=(int)(wsize/2);  /* 1->0; 2->1; 3->1; 4->2 */
    
  sprintf(title,"T-Med w=%d ",wsize);
  display_pid=OpenGrayScreen(cache->ydim,cache->xdim,title,
			     cache->gamma,(float)1.0,(int)(end-start+1));

  /************** begin temporal filtering ********************/
  for(frame=start;frame<=end;frame++,printf("\n")){
    printf("%d: ",frame);

    start_med=frame-wsize2;
    end_med=start_med+wsize-1;
    
    if(start_med<cache->ext_start) { start_med=cache->ext_start;}
    if(end_med>(cache->ext_start+cache->tdim-1)) {
      end_med = cache->ext_start+cache->tdim-1; }

/*    printf("current-frame = %d: ",i);*/

    /* set up the cache. load up cache with start to end frame. */
    get_image(cache,start_med);
    get_image(cache,end_med);

    /* check to insure constant cache while loading image. */
    cache_start=cache->start; cache_end=cache->end;  

    for(i=start_med,nframes=0; i<=end_med; i++) {
      if((i >= cache->ext_start) &&
	 (i < cache->ext_start+cache->tdim)) {
	printf("%d ",i); 
	
	if((cache_start!=cache->start)||(cache_end!=cache->end)) {
	  printf("%s: error, caching while doing median on sequence.\n");
	  TMED_FILT_RET(-1); }

	if(nframes>=wsize)
	  printf("fatal error: nframes=%d >= wsize=%d\n",nframes,wsize);

	if((im_addr[nframes]=get_image(cache,i))==NULL) {
	  printf("%s: cannot access image number %d\n",progn,i);
	  TMED_FILT_RET(-1); }
	nframes++;
      }
    }

    if(cache_res==NULL) { res=temp_res; }
    else if((res=get_image(cache_res,frame-start))==NULL) {
      printf("%s: cannot access result frame number %d\n",progn,frame);
	  TMED_FILT_RET(-1); }

    internal_median_seq(im_addr,nframes,res,imsize,checkval,setval);
/*    printf("\n"); */

    if(display_pid!=NULL || fp!=NULL) {
      convert_float_to_byte(res,cache->array,cache->imsize);

      if(fp!=NULL) {
	if((num=fwrite(cache->array,sizeof(char),imsize,fp))!=imsize) {
	  fprintf(stderr,"Warning: incomplete fwrite. %d of %d bytes written\n",
		  num, imsize);
	  TMED_FILT_RET(-1); }}

      

      display_number_on_image(cache->array,cache->xdim,cache->ydim,
			      (float)frame,(int)5,(int)5,(int)255,(int)0,
			      (int)ndigits(cache->tdim));

      DumpBufToScreen(cache->array,cache->ydim,cache->xdim,
		      display_pid,cache->imsize);
    }

  }
  TMED_FILT_RET(0);
}

int ndigits(n)
  int n;
{
  char numstr[10];
  sprintf(numstr,"%d",n);
  return((int)strlen(numstr));
}

#define TAVG_FILT_RET(rc) \
 return(clean_up_ret_func(malloc_images,nmalloc,NULL,0,display_pid,rc));

/****************************************************************/
/* average of a sequence */

average_sequence(cache,cache_res,start,end,wsize,fp)
  Image_Cache *cache,*cache_res;
  int start,end,wsize;
  FILE *fp;
{
  extern char *args;

  float *im;
  char *progn="average_sequence";
  int n,k;
  int wsize2,frame;
  double *temp_im;

  float *res=NULL,*temp_res=NULL;
  int imsize=cache->imsize;
  
  int cur_res,start_add,end_add,aframe,allocated=0,cache_nframes;
  float *cache_temp_im=NULL;

  FILE *display_pid=NULL;
  char title[20];
  char *malloc_images[5];
  int nmalloc=0,num;


  /************ allocate memory ********************/
  if(cache_res==NULL) {
    if((temp_res=(float *)malloc(cache->imsize*sizeof(float)))==NULL) {
      fprintf(stderr,"Cannot allocate memory for temporal median filtering.\n");
      TAVG_FILT_RET(-1);
    }
/*    fprintf(stderr,"no internal_image. using temp res...\n"); */
    malloc_images[nmalloc]=(char *)temp_res; nmalloc++;
  }

  if((temp_im=(double *)malloc(imsize*sizeof(double)))==NULL) {
    printf("%s: cannot allocate memory for temp_im\n",progn);
    TAVG_FILT_RET(-1); }
  malloc_images[nmalloc]=(char *)temp_im; nmalloc++;

  /**************************************************/
  
/*  fprintf(stderr,"done mallocing images\n"); */

  wsize2=(int)(wsize/2);  /* 1->0; 2->1; 3->1; 4->2 */

  sprintf(title,"T-Avg w=%d ",wsize);
  display_pid=OpenGrayScreen(cache->ydim,cache->xdim,title,
			     cache->gamma,(float)1.0,(int)(end-start+1));

  /************** begin temporal filtering ********************/
  for(frame=start,cur_res=0,n=0;frame<=end;frame++,cur_res++) {
    fprintf(stderr,"%d ",frame);
    start_add=frame-wsize2;
    end_add=start_add+wsize-1;

    if(cache_res==NULL) { res=temp_res; }
    else if((res=get_image(cache_res,cur_res))==NULL) {
      printf("%s: Progn error. Cannot get cache_res float array.\n",progn);
      TAVG_FILT_RET(-1); }

    if(frame==start) { /* first average */
      if(start_add<cache->ext_start) { start_add=cache->ext_start; }
      if(end_add >= cache->ext_start+cache->tdim) {
	end_add=cache->ext_start+cache->tdim-1; }

      accum_op(temp_im,NULL,imsize,"0");  /* initialize accumulator */

      for(k=start_add,n=0;k<=end_add;k++) {
/*	fprintf(stderr,"%d ",k); */
	if((im=get_image(cache,k))!=NULL) {
	  accum_op(temp_im,im,imsize,"+"); n++; }
      }
/*
      fprintf(stderr,"\nafter first accum: cache->start=%d cache->end=%d\n", 
	      cache->start, cache->end);
*/
    }   /* end of condition for first average */
 
    else {     /* frame!=start accumulation for subsequent frames */

      /* cache too small to span the window */
      if(cache->nframes < wsize) {
/*	printf("cache size smaller than window\n"); */
	if(cache_temp_im==NULL) {
	  if((cache_temp_im=(float *)malloc(imsize*sizeof(float)))==NULL) {
	    fprintf(stderr,"cannot allocate memory for average \n");
	    cache_temp_im=NULL; }
	  else {
	    fprintf(stderr,"allocated a temp image for running avg\n");
	    allocated=1; }
	}
	if(cache_temp_im==NULL) { /* worst case situation, borrow from cache */
	  fprintf(stderr,"\n\n************ Warning ***********\n");
	  fprintf(stderr,"short on memory: this implementation not tested.\n\n");
	  cache_temp_im=get_image(cache,cache->end); allocated=0;
	  cache_nframes=cache->nframes;
	  if(cache_nframes>1) cache->nframes=cache->nframes-1;
	}
      }
      else { cache_temp_im=NULL; }
      
      /* subtract off previous start */
      aframe=start_add-1;
      if((im=get_image_disk(cache,(int)aframe,cache_temp_im))!=NULL) {
	accum_op(temp_im,im,imsize,"-"); n--; }
/*
  fprintf(stderr,"after subtracting: cache->start=%d cache->end=%d\n",
	      cache->start, cache->end);
*/      
      /* add on current end */
      aframe=end_add;
      if((im=get_image(cache,(int)aframe))!=NULL) {
	accum_op(temp_im,im,imsize,"+"); n++; }
/*
      fprintf(stderr,"after adding: cache->start=%d cache->end=%d\n",
	      cache->start, cache->end);
*/
    } /* end condition subsequent frames */
    
    /* normalize by number of frames in accumulation */
    { int j; 
      for(j=0;j<imsize;j++) {res[j]=(float)((double)temp_im[j]/(double)n);}
    }

    if(display_pid!=NULL || fp!=NULL) {
      convert_float_to_byte(res,cache->array,cache->imsize);

      if(fp!=NULL) {
	if((num=fwrite(cache->array,sizeof(char),imsize,fp))!=imsize) {
	  fprintf(stderr,"Warning: incomplete fwrite. %d of %d bytes written\n",
		  num, imsize);
	  TAVG_FILT_RET(-1); }}

      display_number_on_image(cache->array,cache->xdim,cache->ydim,
			      (float)frame,(int)5,(int)5,(int)255,(int)0,
			      (int)ndigits(cache->tdim));
      DumpBufToScreen(cache->array,cache->ydim,cache->xdim,
		      display_pid,cache->imsize);
    }
    
  } /* end of moving window loop */

  if(allocated==1 && cache_temp_im!=NULL) {
    fprintf(stderr,"freeing temp image for running avg.\n");
    free(cache_temp_im); }
  if(allocated==0 && cache_temp_im!=NULL) {
    fprintf(stderr,"\n******* Warning ***********\n");
    fprintf(stderr,"Was short on memory for running avg. Cache may be corrupted.\n\n");
    cache->nframes=cache_nframes; }
  
      TAVG_FILT_RET(0); 
}


/* accumulate into a double buffer. specify operator */
int accum_op(accum,im,imsize,op)
  double *accum;
  float *im;
  int imsize;
  char *op;
{
  int i;
  
/*  fprintf(stderr,"accum_op: imsize=%d op=%s accum=%d\n",imsize,op,accum); */

  if(strncmp(op,"+",1)==0) {
/*    fprintf(stderr,"adding..."); */
    for(i=0;i<imsize;i++) { accum[i] += (double)im[i];} }
  else if(strncmp(op,"-",1)==0) {
/*    fprintf(stderr,"subtracting..."); */
    for(i=0;i<imsize;i++) { accum[i] -= (double)im[i];} }
  else if(strncmp(op,"0",1)==0) {
/*    fprintf(stderr,"initializing..."); */
    for(i=0;i<imsize;i++) { accum[i] = (double)0.0;} }
  else {
    fprintf(stderr,"accum_op: unknown operator \"%s\"",op);
    return(-1);
  }
/*  fprintf(stderr,"done\n"); */
  return(0);
}


/***** free all malloced images and close open files and procs ****/
int clean_up_ret_func(images,nimages,files,nfiles,pid,rc)
  char **images;
  int nimages,nfiles,rc;
  FILE **files,*pid;
{
  if(images!=NULL) { free_images(images,nimages); }
  if(files!=NULL) { close_files(files,nfiles); }
  if(pid!=NULL) {pclose(pid);}
  return(rc);
}
